
/******************************************************************************
* MODULE     : algebraic.hpp
* DESCRIPTION: Algebraic extensions based on a primitive element
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_ALGEBRAIC_HPP
#define __MMX_ALGEBRAIC_HPP
#include "dense_polynomial.hpp"
#include "symbol.hpp"
namespace mmx {
#define TMPL_DEF template<typename C>
#define TMPL template<typename C>
#define Polynomial polynomial<C>
#define Primitive_element primitive_element<C>
#define Algebraic algebraic<C>
TMPL class algebraic;
TMPL class primitive_element;
TMPL inline bool compatible (const Algebraic& x, const Algebraic& y);
TMPL inline Polynomial value (const Algebraic& x);
TMPL inline Polynomial modulus (const Algebraic& x);
TMPL inline Primitive_element prim (const Algebraic& x);

/******************************************************************************
* Data for characterizing algebraic extensions
******************************************************************************/

TMPL_DEF
class primitive_element {
MMX_ALLOCATORS;
public:
  Polynomial mp; // minimal polynomial
  inline primitive_element (const Polynomial& mp2): mp (mp2) {}
};

TMPL inline bool
hard_eq (const Primitive_element& d1, const Primitive_element& d2) {
  return hard_eq (d1.mp, d2.mp); }

/******************************************************************************
* The algebraic type
******************************************************************************/

TMPL_DEF
class algebraic {
MMX_ALLOCATORS;
private:
  Polynomial p;
  Primitive_element q;
public:
  static Primitive_element trivial;
public:
  inline algebraic (): p (0), q (trivial) {}
  template<typename T> inline algebraic (const T& c):
    p (c), q (trivial) {}
  inline algebraic (const Polynomial& P, const Polynomial& Q):
    p (P), q (Q) {}
  inline algebraic (const Polynomial& P, const Primitive_element& Q):
    p (P), q (Q) {}
  inline algebraic (const Polynomial& P, const Polynomial& Q, bool F):
    p (rem (P, Q)), q (Q) { assert (F); }
  inline algebraic (const Polynomial& P, const Primitive_element& Q, bool F):
    p (rem (P, Q.mp)), q (Q) { assert (F); }
  friend bool compatible LESSGTR (const Algebraic& x, const Algebraic& y);
  friend Polynomial value LESSGTR (const Algebraic& x);
  friend Polynomial modulus LESSGTR (const Algebraic& x);
  friend Primitive_element prim LESSGTR (const Algebraic& x);
};

TMPL Primitive_element
Algebraic::trivial = Primitive_element (Polynomial (seq (C(-1), C(1))));

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline Polynomial value (const Algebraic& x) { return x.p; }
TMPL inline Polynomial modulus (const Algebraic& x) { return x.q.mp; }
TMPL inline Primitive_element prim (const Algebraic& x) { return x.q; }

/* FIXME: why does this not work? A compiler bug in gcc 2.95.3?
template<typename Op, typename C> inline nat
unary_hash (const Algebraic& x) {
  nat h= Op::op (value (x));
  return (h<<1) ^ (h<<5) ^ (h>>29) ^ Op::op (modulus (x));
}

template<typename Op, typename C> bool
binary_test (const Algebraic& x1, const Algebraic& x2) {
  return Op::op (value (x1), value (x2)) &&
         Op::op (modulus (x1), modulus (x2));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Algebraic)
SOFT_IDENTITY_OP_SUGAR(TMPL,Algebraic)
HARD_IDENTITY_OP_SUGAR(TMPL,Algebraic)
*/

TMPL inline nat hash (const Algebraic& x) {
  nat h= hash (value (x));
  return (h<<1) ^ (h<<5) ^ (h>>29) ^ hash (modulus (x)); }
TMPL inline nat soft_hash (const Algebraic& x) {
  nat h= soft_hash (value (x));
  return (h<<1) ^ (h<<5) ^ (h>>29) ^ soft_hash (modulus (x)); }
TMPL inline nat hard_hash (const Algebraic& x) {
  nat h= hard_hash (value (x));
  return (h<<1) ^ (h<<5) ^ (h>>29) ^ hard_hash (modulus (x)); }
TMPL inline bool operator == (const Algebraic& x1, const Algebraic& x2) {
  return value (x1) == value (x2) &&
         modulus (x1) == modulus (x2); }
TMPL inline bool operator != (const Algebraic& x1, const Algebraic& x2) {
  return !(x1 == x2); }
TMPL inline bool eq (const Algebraic& x1, const Algebraic& x2) {
  return eq (value (x1), value (x2)) &&
         eq (modulus (x1), modulus (x2)); }
TMPL inline bool neq (const Algebraic& x1, const Algebraic& x2) {
  return !eq (x1, x2); }
TMPL inline bool hard_eq (const Algebraic& x1, const Algebraic& x2) {
  return hard_eq (value (x1), value (x2)) &&
         hard_eq (modulus (x1), modulus (x2)); }
TMPL inline bool hard_neq (const Algebraic& x1, const Algebraic& x2) {
  return !hard_eq (x1, x2); }

EQUAL_INT_SUGAR(TMPL,Algebraic,Algebraic);

TMPL generic
flatten (const Algebraic& x) {
  return gen ("algebraic", flatten (value (x)), flatten (modulus (x)));
}

PRINT_SUGAR(TMPL,Algebraic)

/******************************************************************************
* Joining algebraic extensions
******************************************************************************/

TMPL Primitive_element
join (const Primitive_element& d1, const Primitive_element& d2,
      Polynomial& z1, Polynomial& z2)
{
  // Return smallest common extension dd of d1 and d2
  // On exit, z1 and z2 contains the primitive elements of d1 and d2 inside dd
  if (hard_eq (d1, Algebraic::trivial)) {
    z1= Polynomial (C(1));
    z2= Polynomial (C(1), (nat) 1);
    return d2;
  }
  else if (hard_eq (d2, Algebraic::trivial)) {
    z1= Polynomial (C(1), (nat) 1);
    z2= Polynomial (C(1));
    return d1;
  }
  else if (hard_eq (d1, d2) || d1.mp == d2.mp) {
    z1= Polynomial (C(1), (nat) 1);
    z2= Polynomial (C(1), (nat) 1);
    return d1;
  }
  else ASSERT (false, "joining extensions not yet implemented");
}

TMPL Algebraic
upgrade (const Algebraic& x, const Primitive_element& d, const Polynomial& z) {
  if (hard_eq (prim (x), Algebraic::trivial))
    return Algebraic (value (x), d);
  else if (hard_eq (prim (x), d) || prim (x).mp == d.mp)
    return x;
  else ASSERT (false, "joining extensions not yet implemented");
}

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL Algebraic
operator + (const Algebraic& x1, const Algebraic& x2) {
  if (hard_eq (prim (x1), prim (x2)))
    return Algebraic (value (x1) + value (x2), prim (x1));
  else {
    Polynomial z1, z2;
    Primitive_element c= join (prim (x1), prim (x2), z1, z2);
    return upgrade (x1, c, z1) + upgrade (x2, c, z2);
  }
}

TMPL inline Algebraic
operator + (const Algebraic& x1, const C& x2) {
  return Algebraic (value (x1) + x2, prim (x1));
}

TMPL inline Algebraic
operator + (const C& x1, const Algebraic& x2) {
  return Algebraic (x1 + value (x2), prim (x2));
}

TMPL inline Algebraic
operator - (const Algebraic& x) {
  return Algebraic (-value (x), prim (x));
}

TMPL Algebraic
operator - (const Algebraic& x1, const Algebraic& x2) {
  if (hard_eq (prim (x1), prim (x2)))
    return Algebraic (value (x1) - value (x2), prim (x1));
  else {
    Polynomial z1, z2;
    Primitive_element c= join (prim (x1), prim (x2), z1, z2);
    return upgrade (x1, c, z1) - upgrade (x2, c, z2);
  }
}

TMPL inline Algebraic
operator - (const Algebraic& x1, const C& x2) {
  return Algebraic (value (x1) - x2, prim (x1));
}

TMPL inline Algebraic
operator - (const C& x1, const Algebraic& x2) {
  return Algebraic (x1 - value (x2), prim (x2));
}

TMPL Algebraic
operator * (const Algebraic& x1, const Algebraic& x2) {
  if (hard_eq (prim (x1), prim (x2)))
    return Algebraic (value (x1) * value (x2), prim (x1), true);
  else {
    Polynomial z1, z2;
    Primitive_element c= join (prim (x1), prim (x2), z1, z2);
    return upgrade (x1, c, z1) * upgrade (x2, c, z2);
  }
}

TMPL inline Algebraic
operator * (const Algebraic& x1, const C& x2) {
  return Algebraic (value (x1) * x2, prim (x1));
}

TMPL inline Algebraic
operator * (const C& x1, const Algebraic& x2) {
  return Algebraic (x1 * value (x2), prim (x2));
}

TMPL Algebraic
operator / (const C& x1, const Algebraic& x2) {
  xgcd_matrix<Polynomial > m= xgcd (value (x2), modulus (x2));
  return Algebraic (x1 * m.a, prim (x2));
}

TMPL inline Algebraic
operator / (const Algebraic& x1, const C& x2) {
  return Algebraic (value (x1) / x2, prim (x1));
}

TMPL inline Algebraic
operator / (const Algebraic& x1, const Algebraic& x2) {
  return x1 * (C(1) / x2);
}

ARITH_INT_SUGAR(TMPL,Algebraic,C);

#undef TMPL_DEF
#undef TMPL
#undef Primitive_element
#undef Algebraic
} // namespace mmx
#endif // __MMX_ALGEBRAIC_HPP
