
/******************************************************************************
* MODULE     : fft_bench.cpp
* DESCRIPTION: Bench FFT
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "tft.hpp"
#include "basix_utils.hpp"
using namespace mmx;

#define K modular<FFTP>

/******************************************************************************
* Univariate case
******************************************************************************/

nat
up_2 (nat i) {
  return 1 << (log_2 (i-1) + 1);
}

template<typename C> tft_tree<C>
make (nat n) {
  C* a= mmx_new<C> (n);
  for (nat i=0; i<n; i++)
    a[i]= i*i;
  tft_tree<C> t= tft_leaf<C> (log_2 (up_2 (n)));
  insert (t, 0, a, n);
  mmx_delete<C> (a, n);
  return t;
}

template<typename C> C*
make_segment (nat n) {
  C* a= mmx_new<C> (up_2 (n));
  for (nat i=0; i<n; i++)
    a[i]= i*i;
  return a;
}

void
bench_tft (nat n, int type) {
  K* a= make_segment<K> (n);
  tft_tree<K> t= make<K> (n);
  nat count= 0;
  nat start= mmx_time ();
  while (mmx_time () < start + 200) {
    switch (type) {
    case 0: tft (t); break;
    case 1: itft (t); break;
    case 2: tft (a, n, log_2 (n-1) + 1); break;
    case 3: itft (a, n, log_2 (n-1) + 1); break;
    }
    count++;
  }
  double total= mmx_time () - start;
  cout.width (10); cout << n << " : ";
  cout.width (18); cout << (total/(count*n)) << " ms ";
  cout.width (18); cout << (total/count) << " ms\n";
  mmx_delete<K> (a, up_2 (n));
}

/******************************************************************************
* Multivariate case
******************************************************************************/

nat
encode (nat i, nat base) {
  if (i == 0) return 0;
  return base * encode (i>>1, base) + (i&1);
}

template<typename C> void
construct (tft_tree<C>& t, nat offset, nat n, nat dim, nat N, nat base) {
  if (dim == 0) {
    C a[1];
    a[0]= 1;
    insert (t, offset, a, 1);
  }
  else for (nat i=0; i<n; i++)
    construct (t, offset + (encode (i, base) << (dim-1)), n-i, dim-1, N, base);
}

template<typename C> tft_tree<C>
make (nat n, nat dim) {
  tft_tree<C> t= tft_leaf<C> (dim * log_2 (up_2 (n)));
  construct (t, 0, n, dim, up_2 (n), 1<<dim);
  simplify (t);
  return t;
}

void
bench_tft (nat n, nat dim, int type) {
  tft_tree<K> t= make<K> (n, dim);
  // bench_tft (size (t)); return;
  nat sz   = size (t);
  nat count= 0;
  nat start= mmx_time ();
  while (mmx_time () < start + 200) {
    switch (type) {
    case 0: tft (t); break; 
    case 1: itft (t); break;
    case 2: pyramidal_tft (t, dim); break;
    case 3: pyramidal_itft (t, dim); break;
    }
    count++;
  }
  double total= mmx_time () - start;
  cout.width (10); cout << sz << " : ";
  cout.width (18); cout << (total/(count*sz)) << " ms ";
  cout.width (18); cout << (total/count) << " ms\n";
}

#ifdef FFT_COUNT
void
count_tft (nat n, nat dim, int type) {
  tft_tree<K> t= make<K> (n, dim);
  nat sz       = size (t);
  cross_count  = 0;
  icross_count = 0;
  flip_count   = 0;
  switch (type) {
  case 0: tft (t); break;
  case 1: itft (t); break;
  case 2: pyramidal_tft (t, dim); break;
  case 3: pyramidal_itft (t, dim); break;
  }
  nat    sum  = cross_count + icross_count + flip_count;
  double total= sum;
  cout.width (10); cout << sz << " : ";
  cout.width (18); cout << (total/sz) << " average ";
  cout.width (18); cout << sum << " total\n";
}
#endif

/******************************************************************************
* Main routine
******************************************************************************/

int
main () {
  /* NOTE: necessitates bivariate fft_transformers for tft, itft, etc.
  tft_tree<K> t= make<K> (7, 2);
  print (t);
  tft (t);
  print (t);
  itft (t);
  print (t);
  pyramidal_tft (t, 2);
  print (t);
  pyramidal_itft (t, 2);
  print (t);
  */

  /*
#ifdef FFT_COUNT
  cross_count = 0;
  icross_count= 0;
  flip_count  = 0;
  tft_tree<K> t= make<K> (19, 2);
  print (t);
  cout << "-----------------------------------\n";
  tft (t);
  cout << "-----------------------------------\n";
  print (t);
  cout << "-----------------------------------\n";
  cout << "count  = " << cross_count << "\n";
  cout << "size   = " << size (t) << "\n";
  cout << "average= " << ((double) cross_count) / size (t) << "\n";
#endif
  */

  /*
#ifdef FFT_COUNT
  cross_count = 0;
  icross_count= 0;
  flip_count  = 0;
  tft_tree<K> t= make<K> (153, 1);
  print (t);
  cout << "-----------------------------------\n";
  tft (t);
  cout << "-----------------------------------\n";
  print (t);
  cout << "-----------------------------------\n";
  cout << "count  = " << cross_count << "\n";
  cout << "size   = " << size (t) << "\n";
  cout << "average= " << ((double) cross_count) / size (t) << "\n";
#endif
  */

  /*
  for (double i=16.0000001; i<1048576.1; i *= 1.0717734625362931)
    bench_tft ((nat) i, 0);
  for (double i=8.0000001; i<1024.1; i *= 1.0717734625362931)
    bench_tft ((nat) i, 2, 0);
  for (double i=166666; i<1200000.1; i *= 1.16)
    bench_tft ((nat) i, 0);
  for (double i=100.0; i<200.0; i *= 1.0617734625362931)
    bench_tft ((nat) i, 3, 0);
  for (double i=1000000.0; i<3000000.1; i *= 1.0717734625362931)
    bench_tft ((nat) i, 0);
  for (double i=1410; i<2500.1; i *= 1.0717734625362931)
    bench_tft ((nat) i, 2, 0);
  */

  //for (double i=16.0000001; i<1048576.1; i *= 1.0717734625362931) {
  for (double i=16.0000001; i<2100000.0; i *= 1.41421356237309505) {
    bench_tft ((nat) i, 0);
    bench_tft ((nat) i, 2);
    bench_tft ((nat) i, 1);
    bench_tft ((nat) i, 3);
  }

  /*
#ifdef FFT_COUNT
  for (double i=1000000.0; i<3000000.1; i *= 1.0717734625362931)
    count_tft ((nat) i, 1, 0);
  for (double i=1410; i<2500.1; i *= 1.0717734625362931)
    count_tft ((nat) i, 2, 0);
  for (double i=166666; i<1200000.1; i *= 1.16)
    count_tft ((nat) i, 1, 0);
  for (double i=100.0; i<200.0; i *= 1.0617734625362931)
    count_tft ((nat) i, 3, 0);
#endif
  */

  /*
  for (nat i=16; i<=(1<<21); i*=2) {
    cout << i << "\t";
    bench_tft (i, 1, 0);
    cout << i << "\t";
    bench_tft (i, 1, 1);
    cout << i+1 << "\t";
    bench_tft (i+1, 1, 0);
    cout << i+1 << "\t";
    bench_tft (i+1, 1, 1);
  }
  */

#ifdef FFT_COUNT
  for (nat i=16; i<=(1<<21); i*=2) {
    cout << i << "\t";
    count_tft (i, 1, 0);
    cout << i << "\t";
    count_tft (i, 1, 1);
    cout << i+1 << "\t";
    count_tft (i+1, 1, 0);
    cout << i+1 << "\t";
    count_tft (i+1, 1, 1);
  }
#endif

  cout << "-----------------------------------\n";
  for (nat i=8; i<=2048; i*=2) {
    cout << i << "\t";
    bench_tft (i, 2, 0);
    cout << i << "\t";
    bench_tft (i, 2, 1);
    cout << i << "\t";
    bench_tft (i, 2, 2);
    cout << i << "\t";
    bench_tft (i, 2, 3);
    cout << i+1 << "\t";
    bench_tft (i+1, 2, 0);
    cout << i+1 << "\t";
    bench_tft (i+1, 2, 1);
    cout << i+1 << "\t";
    bench_tft (i+1, 2, 2);
    cout << i+1 << "\t";
    bench_tft (i+1, 2, 3);
  }

#ifdef FFT_COUNT
  for (nat i=8; i<=2048; i*=2) {
    cout << i << "\t";
    count_tft (i, 2, 0);
    cout << i << "\t";
    count_tft (i, 2, 1);
    cout << i << "\t";
    count_tft (i, 2, 2);
    cout << i+1 << "\t";
    count_tft (i+1, 2, 0);
    cout << i+1 << "\t";
    count_tft (i+1, 2, 1);
    cout << i+1 << "\t";
    count_tft (i+1, 2, 2);
  }
#endif

  cout << "-----------------------------------\n";
  for (nat i=8; i<=256; i*=2) {
    cout << i << "\t";
    bench_tft (i, 3, 0);
    cout << i << "\t";
    bench_tft (i, 3, 1);
    cout << i << "\t";
    bench_tft (i, 3, 2);
    cout << i << "\t";
    bench_tft (i, 3, 3);
    cout << i+1 << "\t";
    bench_tft (i+1, 3, 0);
    cout << i+1 << "\t";
    bench_tft (i+1, 3, 1);
    cout << i+1 << "\t";
    bench_tft (i+1, 3, 2);
    cout << i+1 << "\t";
    bench_tft (i+1, 3, 3);
  }

  cout << "-----------------------------------\n";
  for (nat i=4; i<=64; i*=2) {
    cout << i << "\t";
    bench_tft (i, 4, 0);
    cout << i << "\t";
    bench_tft (i, 4, 1);
    cout << i << "\t";
    bench_tft (i, 4, 2);
    cout << i << "\t";
    bench_tft (i, 4, 3);
    cout << i+1 << "\t";
    bench_tft (i+1, 4, 0);
    cout << i+1 << "\t";
    bench_tft (i+1, 4, 1);
    cout << i+1 << "\t";
    bench_tft (i+1, 4, 2);
    cout << i+1 << "\t";
    bench_tft (i+1, 4, 3);
  }

  cout << "-----------------------------------\n";
  for (nat i=2; i<=32; i++) {
    cout << i << "\t";
    bench_tft (i, 5, 0);
  }
  
  cout << "-----------------------------------\n";
  for (nat i=2; i<=16; i++) {
    cout << i << "\t";
    bench_tft (i, 6, 0);
  }

  cout << "-----------------------------------\n";
  for (nat i=2; i<=16; i++) {
    cout << i << "\t";
    bench_tft (i, 7, 0);
  }

  cout << "-----------------------------------\n";
  for (nat i=2; i<=8; i++) {
    cout << i << "\t";
    bench_tft (i, 8, 0);
  }

  return 0;
}
