
/******************************************************************************
* MODULE     : dense_matrix.hpp
* DESCRIPTION: Dense matrices
* COPYRIGHT  : (C) 2004  Olivier Ruatta
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_DENSE_MATRIX_HPP
#define __MMX_DENSE_MATRIX_HPP
#include "matrix.hpp"
#include "block.hpp"
namespace mmx {
#define TMPL template<typename C, typename V>
#define Variant dense_matrix<V>
#define Matrix matrix<C,Variant >
#define Matrix_rep matrix_rep<C,Variant >
#define Abs_matrix matrix<Abs_type(C),Variant >
TMPL inline nat cols (const Matrix& m);
TMPL inline nat rows (const Matrix& m);
TMPL inline bool is_scalar    (const Matrix& m);
TMPL inline bool is_matricial (const Matrix& m);
TMPL inline C* tab (Matrix& m);
TMPL inline const C* tab (const Matrix& m);

/******************************************************************************
* Dense matrix variants
******************************************************************************/

template<typename V>
class dense_matrix {
public:
  static const nat def_rows = 0;
  static const nat def_cols = 0;
  static const nat init_rows= 1;
  static const nat init_cols= 1;
};

template<typename V, typename RS, typename CS> class rectangular {};
template<typename V, typename RS, typename CS>
class dense_matrix<rectangular<V,RS,CS> > {
public:
  static const nat def_rows = RS::val;
  static const nat def_cols = CS::val;
  static const nat init_rows= RS::val;
  static const nat init_cols= CS::val;
};

/******************************************************************************
* Dense matrices of arbitrary sizes
******************************************************************************/

TMPL
class Matrix_rep REP_STRUCT {
  C*   a;
  nat  nr;
  nat  nc;
  bool scalar_flag;
public:
  inline matrix_rep (C* a2, nat nr2, nat nc2, bool flag):
    a (a2), nr (nr2), nc (nc2), scalar_flag (flag) {}
  inline ~matrix_rep () { mmx_delete<C> (a, nr*nc); }
  friend class Matrix;
  friend nat cols LESSGTR (const Matrix& m);
  friend nat rows LESSGTR (const Matrix& m);
  friend C* tab LESSGTR (Matrix& m);
  friend const C* tab LESSGTR (const Matrix& m);
  friend bool is_scalar    LESSGTR (const Matrix& m);
  friend bool is_matricial LESSGTR (const Matrix& m);
};

TMPL
class Matrix {
INDIRECT_PROTO_2 (matrix, matrix_rep, C, Variant)
private:
  static C zero;
public:
  matrix () {
    nat nc = Variant::def_rows;
    nat nr = Variant::def_cols;
    C* tab= mmx_new<C> (nr*nc);
    rep= new Matrix_rep (tab, nr, nc, false);
  }

  template<typename T> matrix (const T& c) {
    nat nr= Variant::init_rows;
    nat nc= Variant::init_cols;
    C* tab= mmx_new<C> (nr*nc);
    rep= new Matrix_rep (tab, nr, nc, true);
    block_set (rep->a, C(c), nr, nc);
  }

  template<typename T, typename W> matrix (const matrix<T,W>& m) {
    C* t= mmx_new<C> (rows(m)*cols(m));
    rep= new Matrix_rep (t, rows(m), cols(m), is_scalar (m));
    segment_cast (rep->a, tab(m), rows(m) * cols(m));
  }

  matrix (const C& x) {
    nat nr= Variant::init_rows;
    nat nc= Variant::init_cols;
    C* tab= mmx_new<C> (nr*nc);
    rep= new Matrix_rep (tab, nr, nc, true);
    block_set (rep->a, x, nr, nc);
  }

  matrix (const C& x, nat nr, nat nc) {
    C* tab= mmx_new<C> (nr*nc);
    rep= new Matrix_rep (tab, nr, nc, false);
    block_set (rep->a, x, nr, nc);
  }

  matrix (C* tab, nat nr=Variant::def_rows, nat nc=Variant::def_cols) {
    rep= new Matrix_rep (tab, nr, nc, false);
  }

  matrix (const iterator<C>& it) {
    nat i, j, nr= Variant::def_rows, nc= Variant::def_cols;
    if (nr==0) nr=1;
    C* tab= mmx_new<C> (nr*nc);
    for (i=0; i<nr; i++, ++it)
      for (j=0; j<nc; j++, ++it)
	tab[i*nc+j]= *it;
    rep= new Matrix_rep (tab, nc, nr, false);
  }

  inline const C& operator () (nat i, nat j) const {
    if (is_scalar (*this)) return i==j? rep->a[0]: zero;
    return rep->a[i*cols(*this)+j]; }
  inline C& operator () (nat i, nat j) {
    ASSERT (is_matricial (*this), "non-scalar matrix expected");
    return rep->a[i*cols(*this)+j]; }
  inline C& scalar_ref () {
    ASSERT (is_scalar (*this), "scalar matrix expected");
    return rep->a[0]; }
};
INDIRECT_IMPL_2 (matrix, matrix_rep, typename C, C, typename V, Variant)

TMPL C Matrix::zero= C(0);
TMPL inline nat rows (const Matrix& m) { return m->nr; }
TMPL inline nat cols (const Matrix& m) { return m->nc; }
TMPL inline bool is_scalar (const Matrix& m) { return m->scalar_flag; }
TMPL inline bool is_matricial (const Matrix& m) { return !m->scalar_flag; }
TMPL inline const C& read (const Matrix& m, nat i, nat j) { return m(i,j); }
TMPL inline const C* tab (const Matrix& m) { return m->a; }
TMPL inline C* tab (Matrix& m) {
  ASSERT (is_matricial (m), "non-scalar matrix expected");
  m.secure(); return m->a; }
TMPL inline Matrix extend (const Matrix& m, const Matrix& n) {
  ASSERT (is_scalar (m), "scalar matrix expected");
  ASSERT (is_matricial (n), "non-scalar matrix expected");
  return Matrix (m (0, 0), rows (n), cols (n)); }

STYPE_TO_TYPE(TMPL,scalar_type,Matrix,C);
STYPE_TO_TYPE(TMPL,abs_type,Matrix,Abs_matrix);

template<typename T, typename F, typename TV, typename FV>
struct cast_helper<matrix<T,dense_matrix<TV> >,
		   matrix<F,dense_matrix<FV> > >
{
  static inline matrix<T,dense_matrix<TV> >
  cv (const matrix<F,dense_matrix<FV> >& m) {
    nat n= rows(m) * cols(m);
    const F* sm= tab (m);
    T* c= mmx_new<T> (n);
    for (nat i=0; i<n; i++) c[i]= as<T> (sm[i]);
    return matrix<T,dense_matrix<TV> > (c, rows(m), cols(m)); }
};

/******************************************************************************
* Fixed size dense matrices
******************************************************************************/

#define RMatrix matrix<C,dense_matrix<rectangular<V,RS,CS> > >
#define RMatrix_rep matrix_rep<C,dense_matrix<rectangular<V,RS,CS> > >

template<typename C, typename V, typename RS, typename CS>
class RMatrix_rep REP_STRUCT {
  C* a;
public:
  inline matrix_rep (C* a2, nat dummy1, nat dummy2, bool dummy3):
    a (a2) { (void) dummy1; (void) dummy2; (void) dummy3; }
  inline virtual ~matrix_rep () { mmx_delete<C> (a, RS::val * CS::val); }
  friend class RMatrix;
  friend nat cols LESSGTR (const RMatrix& m);
  friend nat rows LESSGTR (const RMatrix& m);
  friend C*  tab  LESSGTR (const RMatrix& m);
  friend bool is_scalar    LESSGTR (const Matrix& m);
  friend bool is_matricial LESSGTR (const Matrix& m);
};

template<typename C, typename V, typename RS, typename CS> inline nat
cols (const RMatrix& m) { return CS::val; }
template<typename C, typename V, typename RS, typename CS> inline nat
rows (const RMatrix& m) { return RS::val; }
template<typename C, typename V, typename RS, typename CS> inline bool
is_scalar (const RMatrix& m) { (void) m; return false; }
template<typename C, typename V, typename RS, typename CS> inline bool
is_matricial (const RMatrix& m) { (void) m; return true; }

#undef RMatrix
#undef RMatrix_rep

/******************************************************************************
* Abstract routines on dense matrices
******************************************************************************/

template<typename Op, typename C, typename V> nat
unary_hash (const Matrix& m) {
  register nat i, j, h= 214365, nr= rows (m), nc= cols (m);
  for (i=0; i<nr; i++)
    for (j=0; j<nc; j++)
      h= (h<<1) ^ (h<<5) ^ (h>>27) ^ Op::op (m (i, j));
  return h;
}

template<typename Op, typename C, typename V> inline Matrix
unary_map (const Matrix& m) {
  if (is_scalar (m)) return Matrix (Op::op (m (0, 0)));
  nat nrows= rows (m);
  nat ncols= cols (m);
  C* r= mmx_new<C> (nrows*ncols);
  segment_unary<Op> (r, tab (m), nrows*ncols);
  return Matrix (r, nrows, ncols);
}

template<typename Op, typename C, typename V> inline Matrix
binary_map (const Matrix& m, const Matrix& n) {
  if (is_scalar (m) || is_scalar (n)) {
    if (is_matricial (m)) return binary_map<Op> (m, extend (n, m));
    if (is_matricial (n)) return binary_map<Op> (extend (m, n), n);
    return Matrix (Op::op (m (0, 0), n (0, 0)));
  }
  nat nrows= rows (m);
  nat ncols= cols (m);
  ASSERT (rows (n) == nrows, "unequal number of rows");
  ASSERT (cols (n) == ncols, "unequal number of columns");
  C* r= mmx_new<C> (nrows*ncols);
  segment_binary<Op> (r, tab (m), tab (n), nrows*ncols);
  return Matrix (r, nrows, ncols);
}

template<typename Op, typename C, typename V, typename X> inline Matrix
extrary_map (const Matrix& m, const X& x) {
  if (is_scalar (m)) return Matrix (Op::op (m (0, 0), x));
  nat nrows= rows (m);
  nat ncols= cols (m);
  C* r= mmx_new<C> (nrows*ncols);
  segment_extrary<Op> (r, tab (m), x, nrows*ncols);
  return Matrix (r, nrows, ncols);
}

template<typename Op, typename C, typename V> inline Matrix&
in_place_binary_map (Matrix& m, const Matrix& n) {
  if (is_scalar (m) || is_scalar (n)) {
    if (is_matricial (m))
      return in_place_binary_map<Op> (m, extend (n, m));
    if (is_matricial (n))
      m= extend (m, n);
    else {
      Op::in_place (m.scalar_ref (), n (0, 0));
      return m;
    }
  }
  nat nrows= rows (m);
  nat ncols= cols (m);
  ASSERT (rows (n) == nrows, "unequal number of rows");
  ASSERT (cols (n) == ncols, "unequal number of columns");
  segment_in_place_binary<Op> (tab (m), tab (n), nrows*ncols);
  return m;
}

template<typename Op, typename C, typename V, typename X> inline Matrix&
in_place_extrary_map (Matrix& m, const X& x) {
  if (is_scalar (m)) {
    Op::in_place (m.scalar_ref (), x);
    return m;
  }
  nat nrows= rows (m);
  nat ncols= cols (m);
  segment_in_place_extrary<Op> (tab (m), x, nrows*ncols);
  return m;
}

template<typename Op, typename C, typename V> inline bool
binary_test (const Matrix& m, const Matrix& n) {
  if (is_scalar (m) || is_scalar (n)) {
    if (is_matricial (m)) return binary_test<Op> (m, extend (n, m));
    if (is_matricial (n)) return binary_test<Op> (extend (m, n), n);
    return Op::op (m (0, 0), n (0, 0));
  }
  nat nrows= rows (m);
  nat ncols= cols (m);
  if (rows (n) != nrows || cols (n) != ncols) return false;
  return segment_test<Op> (tab (m), tab (n), nrows*ncols);
}

/******************************************************************************
* Instantiations of abstract routines on matrices
******************************************************************************/

TMPL Matrix copy (const Matrix& m) {
  return unary_map<copy_op> (m); }
TMPL Matrix operator - (const Matrix& m) {
  return unary_map<neg_op> (m); }
TMPL Matrix operator + (const Matrix& m, const Matrix& n) {
  return binary_map<add_op> (m, n); }
TMPL Matrix operator + (const Matrix& m, const C& n) {
  return binary_map<add_op> (m, Matrix (n)); }
TMPL Matrix operator + (const C& m, const Matrix& n) {
  return binary_map<add_op> (Matrix (m), n); }
TMPL Matrix operator - (const Matrix& m, const Matrix& n) {
  return binary_map<sub_op> (m, n); }
TMPL Matrix operator - (const Matrix& m, const C& n) {
  return binary_map<sub_op> (m, Matrix (n)); }
TMPL Matrix operator - (const C& m, const Matrix& n) {
  return binary_map<sub_op> (Matrix (m), n); }
TMPL Matrix operator * (const Matrix& m, const C& c) {
  return extrary_map<rmul_op> (m, c); }
TMPL Matrix operator * (const C& c, const Matrix& m) {
  return extrary_map<lmul_op> (m, c); }
TMPL Matrix operator / (const Matrix& m, const C& c) {
  return extrary_map<rdiv_op> (m, c); }
TMPL Matrix quo (const Matrix& m, const C& c) {
  return extrary_map<rquo_op> (m, c); }
TMPL Matrix rem (const Matrix& m, const C& c) {
  return extrary_map<rrem_op> (m, c); }
TMPL Matrix& operator += (Matrix& m, const Matrix& n) {
  return in_place_binary_map<add_op> (m, n); }
TMPL Matrix& operator -= (Matrix& m, const Matrix& n) {
  return in_place_binary_map<sub_op> (m, n); }
TMPL Matrix& operator *= (Matrix& m, const C& x) {
  return in_place_extrary_map<rmul_op> (m, x); }
TMPL Matrix& operator /= (Matrix& m, const C& x) {
  return in_place_extrary_map<rdiv_op> (m, x); }
TMPL bool operator <= (const Matrix& m, const Matrix& n) {
  return binary_test<lesseq_op> (m, n); }
TMPL bool operator >= (const Matrix& m, const Matrix& n) {
  return binary_test<gtreq_op> (m, n); }
template<typename C, typename V, typename K> bool
operator == (const Matrix& m, const K& c) {
  return m == Matrix (c, rows (m), cols (m)); }
template<typename C, typename V, typename K> bool
operator <= (const Matrix& m, const C& c) {
  return m <= Matrix (c, rows (m), cols (m)); }
template<typename C, typename V, typename K> bool
operator >= (const Matrix& m, const C& c) {
  return m >= Matrix (c, rows (m), cols (m)); }

TRUE_IDENTITY_OP_SUGAR(TMPL,Matrix)
SOFT_IDENTITY_OP_SUGAR(TMPL,Matrix)
ARITH_INT_SUGAR(TMPL,Matrix,C);
STRICT_COMPARE_SUGAR(TMPL,Matrix)

TMPL Matrix derive (const Matrix& m) {
  return unary_map<derive_op> (m); }
TMPL Matrix integrate (const Matrix& m) {
  return unary_map<integrate_op> (m); }

/******************************************************************************
* Operations with cast
******************************************************************************/

template<typename Op, typename R, typename C, typename V>
inline matrix<R,Variant >
cast_unary_map (const Matrix& m) {
  if (is_scalar (m))
    return matrix<R,Variant >
      (Op::template cast_helper<R>::cast_op (m (0, 0)));
  nat nr= rows (m), nc= cols (m);
  R* r= mmx_new<R> (nr*nc);
  segment_cast_unary<Op> (r, tab (m), nr*nc);
  return matrix<R,Variant > (r, nr*nc);
}

template<typename Op, typename R, typename S, typename V>
struct cast_unary_helper<Op,matrix<R,Variant >,matrix<S,Variant > > {
  static inline matrix<R,Variant >
  op (const matrix<S,Variant>& x) {
    return cast_unary_map<Op,R> (x); }
};

template<typename Op, typename R, typename C, typename V, typename X>
inline matrix<R,Variant >
cast_extrary_map (const Matrix& m, const X& x) {
  if (is_scalar (m))
    return matrix<R,Variant >
      (Op::template cast_helper<R>::cast_op (m[0], x));
  nat nr= rows (m), nc= cols (m);
  R* r= mmx_new<R> (nr*nc);
  segment_cast_extrary<Op> (r, tab (m), x, nr*nc);
  return matrix<R,Variant > (r, nr*nc);
}

template<typename Op, typename R, typename S1, typename S2, typename V>
struct cast_binary_helper<Op,matrix<R,Variant >,matrix<S1,Variant >,S2> {
  static inline matrix<R,Variant >
  op (const matrix<S1,Variant>& x, const S2& y) {
    return cast_extrary_map<typename Op::sc_right,R> (x, y); }
};

template<typename Op, typename R, typename S1, typename S2, typename V>
struct cast_binary_helper<Op,matrix<R,Variant>,S1,matrix<S2,Variant> > {
  static inline matrix<R,Variant>
  op (const S2& x, const matrix<S1,Variant>& y) {
    return cast_extrary_map<typename Op::sc_left,R> (y, x); }
};

TMPL Abs_matrix abs (const Matrix& m) {
  return cast_abs<Abs_matrix > (m); }

/******************************************************************************
* Further routines on matrices
******************************************************************************/

TMPL Matrix
transpose (const Matrix & m) {
  if (is_scalar (m)) return m;
  nat nrows= rows (m), ncols= cols (m);
  C* r = mmx_new<C> (nrows*ncols);
  block_transpose (r, tab (m), nrows, ncols);
  return Matrix (r, ncols, nrows);
}

TMPL Matrix
submatrix (const Matrix& m, nat rdeb, nat rfin, nat cdeb, nat cfin) {
  if (is_scalar (m)) return Matrix (m (0, 0), rfin-rdeb, cfin-cdeb);
  nat nrows= rows (m), ncols= cols (m);
  C* r = mmx_new<C> ((rfin-rdeb)*(cfin-cdeb));
  block_extraction (r, tab (m), nrows, rdeb, rfin, ncols, cdeb, cfin);
  return Matrix (r, rfin-rdeb, cfin-cdeb);
}

TMPL Matrix
operator * (const Matrix& m, const Matrix& n) {
  if (is_scalar (m) || is_scalar (n)) {
    if (is_matricial (m)) return m * n (0, 0);
    if (is_matricial (n)) return m (0, 0) * n;
    return Matrix (m (0, 0) * n (0, 0));
  }
  nat mrows= rows (m), mcols= cols (m), nrows= rows(n), ncols= cols(n);
  ASSERT (nrows == mcols, "numbers of rows and columns don't match");
  C* r = mmx_new<C> (mrows*ncols);
  block_naive_mul (r, tab (m), tab (n), mrows, mcols, ncols);
  return Matrix (r, mrows, ncols);
}

/******************************************************************************
* Operations on rows or columns
******************************************************************************/

TMPL void
in_place_swap_row (Matrix& m, nat i, nat j) {
  ASSERT (is_matricial (m), "non-scalar matrix expected");
  nat mrows= rows (m), mcols= cols (m);
  ASSERT (i < mrows && j < mrows, "out of range");
  row_in_place_swap (tab (m), i, j, mrows, mcols);
}

TMPL void
in_place_swap_col (Matrix& m, nat i, nat j) {
  ASSERT (is_matricial (m), "non-scalar matrix expected");
  nat mrows= rows (m), mcols= cols (m);
  ASSERT (i < mrows && j < mrows, "out of range");
  col_in_place_swap (tab (m), i, j, mrows, mcols);
}

#define INLINE_SCAL_LINE_OP(name1,name2)			\
TMPL void name1 (Matrix& m, C c, nat i) {			\
  ASSERT (is_matricial (m), "non-scalar matrix expected");	\
  nat mrows=rows(m), mcols=cols(m);				\
  name2 (tab(m), c, i, mrows, mcols);				\
}

#define INLINE_LINES_OP(name1, name2)				\
TMPL void name1 (Matrix& m, nat i, C ci, nat j, C cj) {		\
  ASSERT (is_matricial (m), "non-scalar matrix expected");	\
  nat mrows=rows(m), mcols=cols(m);				\
  name2 (tab(m), i, ci, j, cj, mrows, mcols);			\
}

INLINE_SCAL_LINE_OP(in_place_row_mul, row_in_place_mul);
INLINE_SCAL_LINE_OP(in_place_row_div, row_in_place_div);
INLINE_SCAL_LINE_OP(in_place_col_mul, col_in_place_mul);
INLINE_SCAL_LINE_OP(in_place_col_div, col_in_place_div);
INLINE_LINES_OP(in_place_rows_linadd,row_combine_add);
INLINE_LINES_OP(in_place_cols_linadd,col_combine_add);
INLINE_LINES_OP(in_place_rows_linsub,row_combine_sub);
INLINE_LINES_OP(in_place_cols_linsub,col_combine_sub);

#undef TMPL
#undef Variant
#undef Matrix
#undef Matrix_rep
#undef Abs_matrix
} // namespace mmx
#endif // __MMX_DENSE_MATRIX_HPP
