
/******************************************************************************
* MODULE     : dense_pivot.hpp
* DESCRIPTION: Pivoting for dense matrices
* COPYRIGHT  : (C) 2005  Olivier Ruatta and Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_DENSE_PIVOT_HPP
#define __MMX_DENSE_PIVOT_HPP
#include "dense_matrix.hpp"

namespace mmx {
#define TMPL template<typename C, typename V>
#define Matrix matrix<C,dense_matrix<V> >
#define Matrix_rep matrix_rep<C,dense_matrix<V> >

/******************************************************************************
* Pivoting
******************************************************************************/

// NOTE for Olivier:
// please add or replace by corrected versions of the previous algorithms

TMPL void
naive_pivoting (Matrix& m,
		const nat& ri, const nat& ci,
		const nat& rf, const nat& cf)
{
  ASSERT ((ri < rows(m)) && (ci < cols(m)) &&
	  (rf <= rows(m)) && (cf <= cols(m)),
	  "out of range");

  // search pivot
  nat i= ci, j= ri;
  for (j= ci; j<cf; j++) {
    for (i= ri; i<rf; i++)
      if (m (i, j) != 0) break;
    if (i<rf) break;
  }

  if (i<rf && j<cf) {
    // put pivot on the top and the left
    if (i != ri) in_place_swap_row (m, ri, i);
    if (j != ci) in_place_swap_col (m, ci, j);
    // operations are made on the rows
    for (nat index = 1; index < rf-ri; index++) 
      in_place_rows_linsub (m, ri + index, m (ri, ci),
			       ri, m (ri+index, ci));
    // call the pivoting on the non triangular sub-matrix
    if (ri+1 < rf && ci+1 < cf) 
      naive_pivoting (m, ri+1, ci+1, rf, cf);
  }
}

TMPL C
naive_extended_pivoting (Matrix& m, Matrix& k,
			 nat ri, nat ci,
			 nat rf, nat cf)
{
  C d= 1;
  ASSERT ((ri < rows(m)) && (ci < cols(m)) &&
	  (rf <= rows(m)) && (cf <= cols(m)),
	  "out of range");

  // search pivot
  nat i= ci, j= ri;
  for (j= ci; j<cf; j++) {
    for (i= ri; i<rf; i++)
      if (m (i, j) != 0) break;
    if (i<rf) break;
  }

  if (i<rf && j<cf) {
    // put pivot on the top and the left
    if (i != ri) {
      in_place_swap_row (k, ri, i);
      in_place_swap_row (m, ri, i);
      d= -d;
    }
    ci= j;
    // operations are made on the rows
    d *= m (ri, ci);
    for (nat index = 1; index < rf-ri; index++) {
      d /= m (ri, ci);
      in_place_rows_linsub (k, ri + index, m (ri, ci),
			       ri, m (ri+index, ci));
      in_place_rows_linsub (m, ri + index, m (ri, ci),
			       ri, m (ri+index, ci));
    }
    // call the pivoting on the non triangular sub-matrix
    if (ri+1 < rf && ci+1 < cf)
      return d * naive_extended_pivoting (m, k, ri+1, ci+1, rf, cf);
    return d;
  }
  else return 0;
}

/******************************************************************************
* Applications
******************************************************************************/

TMPL nat
rank (const Matrix& m) {
  ASSERT (is_matricial (m), "non-scalar matrix expected");
  Matrix c (copy (m));
  naive_pivoting (c, 0, 0, rows(c), cols(c));
  nat i;
  for (i=0; i < min (rows (c), cols (c)) && c (i, i) != 0; i++);
  return i;
}

TMPL Matrix
triangulate (const Matrix& m) {
  ASSERT (is_matricial (m), "non-scalar matrix expected");
  Matrix c (copy (m));
  Matrix k (C(1), rows(m), rows(m));
  (void) naive_extended_pivoting (c, k, 0, 0, rows(c), cols(c));
  return c;
}

TMPL Matrix
kernel (const Matrix& m) {
  ASSERT (is_matricial (m), "non-scalar matrix expected");
  Matrix c (copy (m));
  Matrix k (C(1), rows(m), rows(m));
  (void) naive_extended_pivoting (c, k, 0, 0, rows(c), cols(c));
  nat i, j;
  for (i=0; i<rows(c); i++)
    if (segment_equal (tab(c)+i*cols(c), C(0), cols(c))) break;
  k= submatrix (k, i, rows(m), 0, rows(m));
  for (i=0; i<rows(k); i++)
    for (j=0; j<cols(k); j++)
      if (k (i, j) != 0) return k / k (i, j);
  return k;
}

TMPL Matrix
image (const Matrix& m) {
  ASSERT (is_matricial (m), "non-scalar matrix expected");
  Matrix c (copy (m));
  Matrix k (C(1), rows(m), rows(m));
  (void) naive_extended_pivoting (c, k, 0, 0, rows(c), cols(c));
  nat i, j;
  for (i=0; i<rows(c); i++)
    if (segment_equal (tab(c)+i*cols(c), C(0), cols(c))) break;
  return submatrix (c, 0, i, 0, cols(m));
}

TMPL Matrix
invert (const Matrix& m) {
  ASSERT (is_matricial (m), "non-scalar matrix expected");
  ASSERT (cols(m) == rows(m), "square matrix expected");
  nat n= rows(m);
  Matrix t (copy (m));
  Matrix a (C(1), n, n);
  (void) naive_extended_pivoting (t, a, 0, 0, n, n);
  VERIFY (a * m == t, "internal error, pivoting failed");
  // inv(m) = inv(t) a
  Matrix u= transpose (t);
  Matrix b (C(1), n, n);
  (void) naive_extended_pivoting (u, b, 0, 0, n, n);
  VERIFY (b * transpose(t) == u, "internal error, pivoting failed");
  // inv(t) = tr(b) inv(u)
  b= transpose (b);
  VERIFY (t * b == u, "internal error, transposition failed");
  // inv(t) = b inv(u)
  for (nat i=0; i<n; i++) {
    ASSERT (u (i, i) != 0, "non-invertible matrix");
    C inv= 1 / u (i, i);
    for (nat j=0; j<n; j++)
      b (j, i) = inv * b (j, i);
  }
  return b * a;
}

TMPL C
det (const Matrix& m) {
  ASSERT (is_matricial (m), "non-scalar matrix expected");
  ASSERT (cols(m) == rows(m), "square matrix expected");
  Matrix c (copy (m));
  Matrix k (C(1), rows(m), rows(m));
  return naive_extended_pivoting (c, k, 0, 0, rows(c), cols(c));
}

TMPL Matrix
horizontal_join (const Matrix& m1, const Matrix& m2) {
  ASSERT (is_matricial (m1) || is_matricial (m2),
	  "non-scalar matrix expected");
  if (!is_matricial (m1))
    return horizontal_join (Matrix (m1(0,0), rows (m2), rows (m2)), m2);
  if (!is_matricial (m2))
    return horizontal_join (m1, Matrix (m2(0,0), rows (m1), rows (m1)));
  ASSERT (rows (m1) == rows (m2), "unequal number of rows");
  Matrix r (C(0), rows (m1), cols (m1) + cols (m2));
  for (nat i=0; i<rows(m1); i++) {
    for (nat j=0; j<cols(m1); j++)
      r(i,j)= m1(i,j);
    for (nat j=0; j<cols(m2); j++)
      r(i,j+cols(m1))= m2(i,j);
  }
  return r;
}

TMPL Matrix
vertical_join (const Matrix& m1, const Matrix& m2) {
  ASSERT (is_matricial (m1) || is_matricial (m2),
	  "non-scalar matrix expected");
  if (!is_matricial (m1))
    return vertical_join (Matrix (m1(0,0), cols (m2), cols (m2)), m2);
  if (!is_matricial (m2))
    return vertical_join (m1, Matrix (m2(0,0), cols (m1), cols (m1)));
  ASSERT (cols (m1) == cols (m2), "unequal number of columns");
  Matrix r (C(0), rows (m1) + rows (m2), cols (m1));
  for (nat j=0; j<cols(m1); j++) {
    for (nat i=0; i<rows(m1); i++)
      r(i,j)= m1(i,j);
    for (nat i=0; i<rows(m2); i++)
      r(i+rows(m1),j)= m2(i,j);
  }
  return r;
}

TMPL Matrix
krylov (const Matrix& m, const Matrix& v) {
  ASSERT (is_matricial (m), "non-scalar matrix expected");
  ASSERT (is_square_matrix (m), "square matrix expected");
  Matrix r= image (v);
  Matrix p= m;
  while (true) {
    nat rk= rows (r);
    r= image (vertical_join (r, r*p));
    if (rows (r) <= rk) return r;
    p= p*p;
  }
}

}//end of namespace
#undef TMPL
#undef Matrix
#undef Matrix_rep
#endif //__MMX_DENSE_PIVOT
