
/******************************************************************************
* MODULE     : dense_polynomial.hpp
* DESCRIPTION: Dense univariate polynomials
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_DENSE_POLYNOMIAL_HPP
#define __MMX_DENSE_POLYNOMIAL_HPP
#include "polynomial.hpp"
#include "dense_vector.hpp"
#include "jet.hpp"
namespace mmx {
#define TMPL template<typename C, typename V>
#define Variant dense_polynomial<V>
#define Polynomial polynomial<C,Variant >
#define Polynomial_rep polynomial_rep<C,Variant >
#define Vector vector<C>
TMPL inline int deg (const Polynomial& P);
TMPL inline nat N (const Polynomial& P);
TMPL inline const C* seg (const Polynomial& P);

/******************************************************************************
* Dense polynomial variants
******************************************************************************/

template<typename V> class dense_polynomial {};

class integer;
class rational;
template<typename V> class floating;
struct kronecker_multiply;

STMPL
struct polynomial_variant_helper<integer> {
  typedef dense_polynomial<kronecker_multiply> PV;
};

STMPL
struct polynomial_variant_helper<rational> {
  typedef dense_polynomial<kronecker_multiply> PV;
};

template<typename V>
struct polynomial_variant_helper<floating<V> > {
  typedef dense_polynomial<kronecker_multiply> PV;
};

/******************************************************************************
* Dense polynomials
******************************************************************************/

template<typename C, typename V>
class Polynomial_rep REP_STRUCT {
  C*  a;
  nat n;
  nat l;  // allocated number of coefficients

public:
  inline void normalize () {
    while ((n != 0) && (a[n-1] == C(0))) n--; }
  inline polynomial_rep (C* a2, nat n2):
    a(a2), n(n2), l(n2) { normalize (); }
  inline polynomial_rep (C* a2, nat n2, nat l2):
    a(a2), n(n2), l(l2) { normalize (); }
  inline ~polynomial_rep () { mmx_delete<C> (a, l); }
  void extend (nat l2) {
    C* b= mmx_new<C> (l2);
    jet_copy (b, a, n);
    jet_clear (b+n, l2-n);
    mmx_delete<C> (a, l);
    a= b;
    n= l2;
    l= l2;
  }

  friend class Polynomial;
  friend int deg LESSGTR (const Polynomial& P);
  friend nat N LESSGTR (const Polynomial& P);
  friend const C* seg LESSGTR (const Polynomial& P);
};

template<typename C, typename V>
class Polynomial {
INDIRECT_PROTO_2 (polynomial, polynomial_rep, C, dense_polynomial<V>)
public:
  inline polynomial () {
    rep= new Polynomial_rep (mmx_new<C> (0), 0);
  }
  template<typename T> inline polynomial (const T& x) {
    C* a= mmx_new<C> (1);
    a[0]= x;
    rep= new Polynomial_rep (a, 1);
    rep->normalize ();
  }
  template<typename T> inline
  polynomial (const polynomial<T,dense_polynomial<V> >& P) {
    C* a= mmx_new<C> (N(P));
    jet_cast (a, seg (P), N(P));
    rep= new Polynomial_rep (a, N(P));
    rep->normalize ();
  }
  template<typename T> polynomial (const T& x, nat n) {
    C* a= mmx_new<C> (n+1);
    jet_clear (a, n);
    a[n]= x;
    rep= new Polynomial_rep (a, n+1);
    rep->normalize ();
  }
  inline polynomial (C* a, nat n) {
    rep= new Polynomial_rep (a, n);
  }
  inline polynomial (C* a, nat n, nat l) {
    rep= new Polynomial_rep (a, n, l);
  }
  polynomial (const iterator<C>& it) {
    nat i, l=0;
    rep= new Polynomial_rep (mmx_new<C> (0), 0);
    for (i=0; busy (it); i++, ++it) {
      if (i >= l) {
	l= max ((nat) 1, l<<1);
	rep->extend (l);
      }
      rep->a[i]= *it;
    }
    rep->normalize ();
  }
  inline C operator[] (nat i) const { return i<rep->n? rep->a[i]: C(0); }
  Polynomial& operator += (const Polynomial& P);
  Polynomial& operator -= (const Polynomial& P);
  Polynomial& operator *= (const C& c);
  Polynomial& operator /= (const C& c);
};
INDIRECT_IMPL_2 (polynomial, polynomial_rep,
		 typename C, C, typename V, dense_polynomial<V>)

TMPL inline nat N (const Polynomial& P) { return P->n; }
TMPL inline int deg (const Polynomial& P) { return ((int) P->n) - 1; }
TMPL inline const C* seg (const Polynomial& P) { return P->a; }

template<typename T, typename F, typename TV, typename FV>
struct cast_helper<polynomial<T,dense_polynomial<TV> >,
		   polynomial<F,dense_polynomial<FV> > >
{
  static inline polynomial<T,dense_polynomial<TV> >
  cv (const polynomial<F,dense_polynomial<FV> >& p) {
    T* c= mmx_new<T> (N(p));
    for (nat i=0; i<N(p); i++) c[i]= as<T> (p[i]);
    return polynomial<T,dense_polynomial<TV> > (c, N(p)); }
};

template<typename Op, typename C, typename V> nat
unary_hash (const Polynomial& p) {
  register nat i, h= 642531, n= N(p);
  for (i=0; i<n; i++)
    h= (h<<1) ^ (h<<5) ^ (h>>27) ^ Op::op (p[i]);
  return h;
}

/******************************************************************************
* In place arithmetic
******************************************************************************/

TMPL Polynomial&
Polynomial::operator += (const Polynomial& P) {
  nat n1= rep->n, n2= N(P);
  if (n1 < n2) rep -> extend (n2);
  jet_add<C> (rep->a, seg (P), n2);
  if (n1 == n2) rep -> normalize ();
  return *this;
}

TMPL Polynomial&
Polynomial::operator -= (const Polynomial& P) {
  nat n1= rep->n, n2= N(P);
  if (n1 < n2) rep->extend (n2);
  jet_sub<C> (rep->a, seg (P), n2);
  if (n1 == n2) rep->normalize ();
  return *this;
}

TMPL Polynomial&
Polynomial::operator *= (const C& c) {
  jet_mul<C> (rep->a, c, rep->n);
  rep->normalize ();
  return *this;
}

TMPL Polynomial&
Polynomial::operator /= (const C& c) {
  jet_div<C> (rep->a, c, rep->n);
  rep->normalize ();
  return *this;
}

/******************************************************************************
* Basic arithmetic operations
******************************************************************************/

TMPL Polynomial
operator - (const Polynomial& P) {
  nat n= N(P);
  C* r= mmx_new<C> (n);
  jet_neg<C> (r, seg (P), n);
  return Polynomial (r, n);
}

TMPL Polynomial
operator + (const Polynomial& P1, const Polynomial& P2) {
  nat m= min (N(P1), N(P2));
  nat n= max (N(P1), N(P2));
  C* r= mmx_new<C> (n);
  jet_add<C> (r, seg (P1), seg (P2), m);
  if (N(P1) > m) jet_copy<C> (r+m, seg(P1)+m, n-m);
  if (N(P2) > m) jet_copy<C> (r+m, seg(P2)+m, n-m);
  return Polynomial (r, n);
}

TMPL Polynomial operator + (const Polynomial& P, const C& c) {
  return P + Polynomial (c); }
TMPL Polynomial operator + (const C& c, const Polynomial& P) {
  return Polynomial (c) + P; }

TMPL Polynomial
big_add (const array<Polynomial >& a) {
  nat i, k= N(a), n=0;
  for (i=0; i<k; i++) n= max (N(a[i]), n);
  C* r= mmx_new<C> (n);
  jet_set<C> (r, 0, n);
  for (i=0; i<k; i++) jet_add<C> (r, seg (a[i]), N(a[i]));
  return Polynomial (r, n);
}

TMPL Polynomial
operator - (const Polynomial& P1, const Polynomial& P2) {
  nat m= min (N(P1), N(P2));
  nat n= max (N(P1), N(P2));
  C* r= mmx_new<C> (n);
  jet_sub<C> (r, seg (P1), seg (P2), m);
  if (N(P1) > m) jet_copy<C> (r+m, seg(P1)+m, n-m);
  if (N(P2) > m) jet_neg<C> (r+m, seg(P2)+m, n-m);
  return Polynomial (r, n);
}

TMPL Polynomial operator - (const Polynomial& P, const C& c) {
  return P - Polynomial (c); }
TMPL Polynomial operator - (const C& c, const Polynomial& P) {
  return Polynomial (c) - P; }

TMPL Polynomial
operator * (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 == 0 || n2 == 0) return Polynomial ();
  C* r= mmx_new<C> (n1+n2-1);
  jet_mul<C,V> (r, seg (P1), seg (P2), n1, n2);
  return Polynomial (r, n1+n2-1);
}

TMPL Polynomial
operator * (const Polynomial& P, const C& c) {
  nat n= N(P);
  C* r= mmx_new<C> (n);
  jet_mul<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
operator * (const C& c, const Polynomial& P) {
  nat n= N(P);
  C* r= mmx_new<C> (n);
  jet_mul<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
big_mul (const array<Polynomial >& a) {
  return segment_dicho_big<mul_op> (seg (a), N(a));
}

TMPL Polynomial
operator / (const Polynomial& P, const C& c) {
  nat n= N(P);
  C* r= mmx_new<C> (n);
  jet_div<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
skew_div (const Polynomial& P, const C& c, bool left) {
  if (left) {
    nat n= N(P);
    C* r= mmx_new<C> (n);
    jet_div<C> (r, seg (P), c, n);
    return Polynomial (r, n);
  }
  else return P / c;
}

TMPL Polynomial
quo (const Polynomial& P, const C& c) {
  nat n= N(P);
  C* r= mmx_new<C> (n);
  jet_quo<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
rem (const Polynomial& P, const C& c) {
  nat n= N(P);
  C* r= mmx_new<C> (n);
  jet_rem<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
operator / (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 < n2 || n2 == 0) return Polynomial ();
  C* q= mmx_new<C> (n1-n2+1);
  C* r= mmx_new<C> (n1);
  jet_copy (r, seg (P1), n1);
  jet_quo_rem<C,V> (q, r, seg (P2), n1, n2);
  ASSERT (jet_equal<C> (r, C(0), n1), "division must be exact");
  mmx_delete<C> (r, n1);
  return Polynomial (q, n1-n2+1);
}

TMPL Polynomial
quo (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 < n2 || n2 == 0) return Polynomial ();
  C* q= mmx_new<C> (n1-n2+1);
  C* r= mmx_new<C> (n1);
  jet_copy (r, seg (P1), n1);
  jet_quo_rem<C,V> (q, r, seg (P2), n1, n2);
  mmx_delete<C> (r, n1);
  return Polynomial (q, n1-n2+1);
}

TMPL Polynomial
rem (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 < n2 || n2 == 0) return P1;
  C* q= mmx_new<C> (n1-n2+1);
  C* r= mmx_new<C> (n1);
  jet_copy (r, seg (P1), n1);
  jet_quo_rem<C,V> (q, r, seg (P2), n1, n2);
  mmx_delete<C> (q, n1-n2+1);
  return Polynomial (r, n1);
}

/******************************************************************************
* Gcd computations
******************************************************************************/

TMPL Polynomial
gcd (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 == 0) return P2;
  if (n2 == 0) return P1;
  C* r= mmx_new<C> (max (n1, n2));
  jet_field_gcd (r, seg (P1), seg (P2), n1, n2);
  //jet_ring_gcd (r, seg (P1), seg (P2), n1, n2);
  return Polynomial (r, max (n1, n2));
}

TMPL Polynomial
lcm (const Polynomial& P1, const Polynomial& P2) {
  return P1 * (P2 / gcd (P1, P2));
}

/******************************************************************************
* Valuation and comparisons
******************************************************************************/

TMPL int
val (const Polynomial& P) {
  for (nat i=0; i<N(P); i++)
    if (P[i] != 0) return (int) i;
  return (int) (((nat) (-1)) >> 1);
}

TMPL int
sign (const Polynomial& P) {
  nat n= N(P);
  if (n == 0) return 0;
  return sign (P[n-1]);
}

TMPL int
compare (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 > n2) return sign (P1[n1-1]);
  if (n2 > n1) return -sign (P2[n2-1]);
  return jet_cmp (seg (P1), seg (P2), n1);
}

template<typename Op,typename C,typename V> inline bool
binary_test (const Polynomial& P1, const Polynomial& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 != n2) return false;
  for (nat i=0; i<n1; i++)
    if (Op::not_op (P1[i], P2[i])) return false;
  return true;
}

/******************************************************************************
* Differentiation, integration and composition
******************************************************************************/

TMPL Polynomial
derive (const Polynomial& P) {
  nat n= N(P);
  if (n <= 1) return 0;
  C* r= mmx_new<C> (n-1);
  jet_derive (r, seg (P), n);
  return Polynomial (r, n-1);
}

TMPL Polynomial
xderive (const Polynomial& P) {
  nat n= N(P);
  C* r= mmx_new<C> (n);
  jet_xderive (r, seg (P), n);
  return Polynomial (r, n);
}

TMPL Polynomial
integrate (const Polynomial& P) {
  nat n= N(P);
  C* r= mmx_new<C> (n+1);
  jet_integrate (r, seg (P), n);
  return Polynomial (r, n+1);
}

template<typename C,typename V,typename K> Polynomial
compose (const Polynomial& P1, const polynomial<K,Variant >& P2) {
  nat n1= N(P1), n2= N(P2);
  if (n1 <= 1) return P1;
  if (n2 == 0) return P1[0];
  nat n= (n1-1) * (n2-1) + 1;
  C* r= mmx_new<C> (n);
  jet_dicho_compose<C,V,K> (r, seg (P1), seg (P2), n1, n2);
  return Polynomial (r, n);
}

TMPL Polynomial
shift (const Polynomial& P, const C& sh) {
  nat n= N(P);
  if (n <= 1 || sh == 0) return P;
  C q[2]; q[0]= sh; q[1]= 1;
  C* r= mmx_new<C> (n);
  jet_dicho_compose<C,V> (r, seg (P), q, n, 2);
  return Polynomial (r, n);
}

TMPL Polynomial
shift (const Polynomial& P, int i) {
  return shift (P, C (i));
}

template<typename C,typename V,typename K> Polynomial
q_difference (const Polynomial& P, const K& q) {
  nat n= N(P);
  if (n <= 1) return P;
  C* r= mmx_new<C> (n);
  jet_q_difference (r, seg (P), q, n);
  return Polynomial (r, n);
}

TMPL Polynomial
dilate (const Polynomial& P, nat p) {
  nat n= N(P);
  if (n <= 1) return P;
  nat k= (n-1)*p + 1;
  C* r= mmx_new<C> (k);
  jet_dilate (r, seg (P), p, n);
  return Polynomial (r, k);
}

/******************************************************************************
* Evaluation
******************************************************************************/

template<typename C> polynomial<C,Polynomial_variant(C) >
annulator (const Vector& v) {
  ASSERT (is_vectorial (v), "non-scalar vector expected");
  C* r= mmx_new<C> (N(v)+1);
  jet_annulator<C,Polynomial_variant(C) > (r, seg(v), N(v));
  return polynomial<C,Polynomial_variant(C) > (r, N(v)+1);
}

TMPL C
eval (const Polynomial& p, const C& c) {
  return jet_eval<C> (seg (p), c, N(p));
}

TMPL Vector
eval (const Polynomial& p, const Vector& v) {
  ASSERT (is_vectorial (v), "non-scalar vector expected");
  C* r= mmx_new<C> (N(v));
  jet_eval<C,Variant > (r, seg(p), seg(v), N(p), N(v));
  return Vector (r, N(v));
}

/******************************************************************************
* Miscellaneous
******************************************************************************/

TMPL Polynomial
range (const Polynomial& P, nat start, nat end) {
  C* r= mmx_new<C> (end-start);
  for (nat i=start; i<end; i++) r[i-start]= P[i];
  return Polynomial (r, end-start);
}

TMPL Polynomial
shift_exp (const Polynomial& P, int shift) {
  if (shift == 0) return P;
  else if (shift <= -((int) N(P))) return 0;
  else if (shift < 0) return range (P, (nat) (-shift), N(P));
  else {
    C* r= mmx_new<C> (N(P) + shift, C(0));
    jet_copy (r+shift, seg (P), N(P));
    return Polynomial (r, N(P) + shift);
  }
}

TMPL Polynomial
reverse (const Polynomial& P) {
  nat n= N(P);
  C* r= mmx_new<C> (n);
  for (nat i=0; i<n; i++) r[i]= P[n-1-i];
  return Polynomial (r, n);
}

TMPL Polynomial
square_roots (const Polynomial& P) {
  nat n= N(P);
  C* r= mmx_new<C> (n);
  jet_square_roots<C,V> (r, seg (P), n);
  return Polynomial (r, n);
}

/******************************************************************************
* Syntactic sugar
******************************************************************************/

TRUE_IDENTITY_OP_SUGAR(TMPL,Polynomial)
SOFT_IDENTITY_OP_SUGAR(TMPL,Polynomial)
ARITH_INT_SUGAR(TMPL,Polynomial,C)
SHIFT_INT_SUGAR(TMPL,Polynomial)
COMPARE_SUGAR(TMPL,Polynomial)
EQUAL_INT_SUGAR(TMPL,Polynomial,Polynomial)
COMPARE_INT_SUGAR(TMPL,Polynomial,Polynomial)

#undef TMPL
#undef Variant
#undef Polynomial
#undef Polynomial_rep
#undef Vector
} // namespace mmx
#endif // __MMX_DENSE_POLYNOMIAL_HPP
