
/******************************************************************************
* MODULE     : jet.hpp
* DESCRIPTION: low level operations on polynomial or serial segments
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__JET__HPP
#define __MMX__JET__HPP
#include "segment.hpp"

namespace mmx {
#define TMPL template<typename C>

/******************************************************************************
* Vectorial operations
******************************************************************************/

#define jet_set segment_set
#define jet_clear segment_clear
#define jet_copy segment_copy
#define jet_cast segment_cast
#define jet_add segment_add
#define jet_sub segment_sub
#define jet_neg segment_neg
#define jet_mul_add segment_mul_add
#define jet_equal segment_equal

TMPL inline void jet_mul (C* dest, const C* s, const C& c, nat n) {
  segment_mul (dest, s, c, n); }
TMPL inline void jet_mul (C* dest, const C& c, nat n) {
  segment_mul (dest, c, n); }
TMPL inline void jet_div (C* dest, const C* s, const C& c, nat n) {
  segment_div (dest, s, c, n); }
TMPL inline void jet_div (C* dest, const C& c, nat n) {
  segment_div (dest, c, n); }
TMPL inline void jet_quo (C* dest, const C* s, const C& c, nat n) {
  segment_quo (dest, s, c, n); }
TMPL inline void jet_quo (C* dest, const C& c, nat n) {
  segment_quo (dest, c, n); }
TMPL inline void jet_rem (C* dest, const C* s, const C& c, nat n) {
  segment_rem (dest, s, c, n); }
TMPL inline void jet_rem (C* dest, const C& c, nat n) {
  segment_rem (dest, c, n); }

TMPL inline int
jet_sign (C* s, nat n) {
  while (n != 0) {
    int sgn= sign (*s);
    if (sgn != 0) return sgn;
    s++; n--;
  }
  return 0;
}

TMPL inline int
jet_cmp (const C* s1, const C* s2, nat n) {
  while (n != 0) {
    int sgn= compare (*s1, *s2);
    if (sgn != 0) return sgn;
    s1++; s2++; n--;
  }
  return 0;
}

/******************************************************************************
* Other linear operators
******************************************************************************/

TMPL inline void
jet_derive (C* dest, nat n) {
  for (nat i=1; i<n; i++)
    dest[i-1]= C(i) * dest[i];
}

TMPL inline void
jet_derive (C* dest, const C* s, nat n) {
  for (nat i=1; i<n; i++)
    dest[i-1]= C(i) * s[i];
}

TMPL inline void
jet_xderive (C* dest, nat n) {
  for (nat i=0; i<n; i++)
    dest[i]= C(i) * dest[i];
}

TMPL inline void
jet_xderive (C* dest, const C* s, nat n) {
  for (nat i=0; i<n; i++)
    dest[i]= C(i) * s[i];
}

TMPL inline void
jet_integrate (C* dest, const C* s, nat n) {
  dest[0]= 0;
  for (nat i=0; i<n; i++)
    dest[i+1]= s[i] / C(i+1);
}

template<typename C,typename K> inline void
jet_q_difference (C* dest, const C* s, const K& q, nat n) {
  K p (1);
  while (n != 0) {
    *dest= p * (*s);
    dest++; s++; n--; p *= q;
  }
}

TMPL inline void
jet_half (C* dest, const C* s, nat l) {
  while (l != 0) {
    *dest= *s;
    dest++; s += 2; l--;
  }
}

TMPL inline void
jet_double (C* dest, const C* s, nat l) {
  while (l != 0) {
    *dest= *s;
    dest += 2; s++; l--;
  }
}

TMPL inline void
jet_double_add (C* dest, const C* s, nat l) {
  while (l != 0) {
    *dest += *s;
    dest += 2; s++; l--;
  }
}

TMPL inline void
jet_triple (C* dest, const C* s, nat l) {
  while (l != 0) {
    *dest= *s;
    dest += 3; s++; l--;
  }
}

TMPL inline void
jet_dilate (C* dest, const C* s, nat p, nat n) {
  jet_clear (dest, (n-1)*p + 1);
  while (n != 0) {
    *dest= *s;
    dest += p; s++; n--;
  }
}

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename C, typename V, typename K>
class multiplier {
public:
  static void mul (C* dest, const C* s1, const K* s2, nat n1, nat n2) {
    jet_clear (dest, n1+n2-1);
    while (n2 != 0) {
      jet_mul_add (dest, s1, *s2, n1);
      dest++; s2++; n2--;
    }
  }
};

template<typename C, typename V, typename K> inline void
jet_mul (C* dest, const C* s1, const K* s2, nat n1, nat n2) {
  multiplier<C,V,K>::mul (dest, s1, s2, n1, n2);
}

template<typename C, typename V> inline void
jet_square (C* dest, const C* s, nat n) {
  multiplier<C,V,C>::mul (dest, s, s, n, n);
}

template<typename C, typename V>
class tmultiplier {
public:
  static void tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    // transposed multiplication (or middle product)
    // s1 has length n1, dest has length n2 and s2 has length n1+n2-1
    s2 += n1;
    jet_clear (dest, n2);
    while (n1 != 0) {
      jet_mul_add (dest, *s1, s2, n2);
      s1++; s2--; n1--;
    }
  }
};

template<typename C, typename V> inline void
jet_tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  tmultiplier<C,V>::tmul (dest, s1, s2, n1, n2);
}

/******************************************************************************
* Division
******************************************************************************/

template<typename C, typename V>
class divisor {
public:
  static void quo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
    while (n1 >= n2 && n1 != 0) {
      int d= n1-n2;
      C q= quo (s1[n1-1], s2[n2-1]);
      dest[d]= q;
      for (nat i=0; i<n2; i++)
	s1[i+d] -= q * s2[i];
      n1--;
    }
  }
};

template<typename C, typename V> inline void
jet_quo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2) {
  // (s1, n1) contains the numerator on input and the remainder on output
  // (s2, n2) contains the denominator. We assume n2>0 and s2[n2-1] != 0
  // (dest, n1-n2+1) contains the quotient
  divisor<C,V>::quo_rem (dest, s1, s2, n1, n2);
}

template<typename C, typename V> void
jet_dicho_quo_rem (C* dest, C* s1, const C* s2, nat n1, nat n2, nat Bound) {
  if (n1 < n2);
  else if (n1 < Bound)
    jet_quo_rem<C,naive_multiply> (dest, s1, s2, n1, n2);
  else {
    nat l= n1-n2+1;
    if (n2 > l) {
      nat d= n2-l;
      jet_dicho_quo_rem<C,V> (dest, s1+d, s2+d, n1-d, l, Bound);
      INLINE_ARRAY (C, buf, l+d-1);
      jet_mul<C,V> (buf, dest, s2, l, d);
      jet_sub (s1, buf, l+d-1);
      DELETE_ARRAY (C, buf, l+d-1);      
    }
    else {
      nat h= (n2+1) >> 1;
      nat d= l-h;
      jet_dicho_quo_rem<C,V> (dest+d, s1+d, s2, n1-d, n2, Bound);
      jet_dicho_quo_rem<C,V> (dest, s1, s2, n1-h, n2, Bound);
    }
  }
}

/******************************************************************************
* Composition
******************************************************************************/

template<typename C, typename V, typename K> void
jet_dicho_compose (C* dest, K* p, const C* s1, const K* s2, nat n1, nat n2) {
  // computes compose ((s1, n1), (s2, n2)) and pow ((s2, n2), n1)
  if (n1 == 1) {
    *dest= *s1;
    jet_copy (p, s2, n2);
  }
  else {
    nat d2= n2 - 1;
    nat h1= n1 >> 1, h2= n1 - h1;
    nat l1= (h1-1) * d2 + 1, l2= (h2-1) * d2 + 1;
    nat k1= l1 + d2, k2= l2 + d2;
    INLINE_ARRAY (C, c1, l1 + l2);
    C* c2= c1 + l1;
    INLINE_ARRAY (K, p1, k1 + k2);
    K* p2= p1 + k1;
    jet_dicho_compose<C,V> (c1, p1, s1, s2, h1, n2);
    jet_dicho_compose<C,V> (c2, p2, s1+h1, s2, h2, n2);
    jet_mul<C,V> (dest, c2, p1, l2, k1);
    jet_add (dest, c1, l1);
    jet_mul<K,V> (p, p1, p2, k1, k2);
    DELETE_ARRAY (K, p1, k1 + k2);
    DELETE_ARRAY (C, c1, l1 + l2);
  }
}

template<typename C, typename V, typename K> void
jet_dicho_compose (C* dest, const C* s1, const K* s2, nat n1, nat n2) {
  ASSERT (n1 > 0 && n2 > 0, "composition of empty segments");
  if (n1 == 1) *dest= *s1;
  else {
    nat d2= n2 - 1;
    nat h1= n1 >> 1, h2= n1 - h1;
    nat l1= (h1-1) * d2 + 1, l2= (h2-1) * d2 + 1;
    nat k1= l1 + d2;
    INLINE_ARRAY (C, c1, l1 + l2);
    C* c2= c1 + l1;
    INLINE_ARRAY (K, p1, k1);
    jet_dicho_compose<C,V> (c1, p1, s1, s2, h1, n2);
    jet_dicho_compose<C,V> (c2, s1+h1, s2, h2, n2);
    jet_mul<C,V> (dest, c2, p1, l2, k1);
    jet_add (dest, c1, l1);
    DELETE_ARRAY (K, p1, k1);
    DELETE_ARRAY (C, c1, l1 + l2);
  }
}

/******************************************************************************
* Root squaring
******************************************************************************/

template<typename C, typename V> void
jet_square_roots (C* dest, const C* s, nat n) {
  nat l1= (n+1) >> 1, l2= n >> 1;
  INLINE_ARRAY (C, c1, n << 1);
  C* c2 = c1+ l1;
  C* aux= c1 + n;
  if (n>0) aux[n-1]= C(0);
  jet_half (c1, s, l1);
  jet_half (c2, s+1, l2);
  jet_square<C,V> (dest, c1, l1);
  jet_square<C,V> (aux , c2, l2);
  jet_sub (dest+1, aux, (l2<<1) - 1);
  DELETE_ARRAY (C, c1, n << 1);
}

/******************************************************************************
* Gcd computations
******************************************************************************/

template<typename C> nat
jet_field_gcd (C* d1, const C* s1, const C* s2, nat n1, nat n2) {
  if (n1<n2) return jet_field_gcd (d1, s2, s1, n2, n1);
  else {
    ASSERT (n1>0 && n2>0 && s1[n1-1] != 0 && s2[n2-1] != 0,
	    "invalid hypothesis for gcd computation");
    nat orig_n2= n2;
    C* d2= mmx_new<C> (orig_n2);
    segment_copy (d1, s1, n1);
    segment_copy (d2, s2, n2);
    while (n2>0) {
      C m= d1[n1-1] / d2[n2-1];
      for (nat i=0; i<n2; i++)
	d1[i+n1-n2] -= d2[i] * m;
      while (n1 > 0 && d1[n1-1] == 0) n1--;
      if (n1 < n2) {
	for (nat i=0; i<n2; i++) swap (d1[i], d2[i]);
	swap (n1, n2);	
      }
    }
    mmx_delete<C> (d2, orig_n2);
    for (nat i=0; i<n1; i++) d1[i]= d1[i] / d1[n1-1];
    return n1;
  }
}

template<typename C> inline C
content (const C* s, nat n) {
  return segment_dicho_big<gcd_op> (s, n);
}

template<typename C> nat
jet_ring_gcd (C* d1, const C* s1, const C* s2, nat n1, nat n2) {
  if (n1<n2) return jet_ring_gcd (d1, s2, s1, n2, n1);
  else {
    ASSERT (n1>0 && n2>0 && s1[n1-1] != 0 && s2[n2-1] != 0,
	    "invalid hypothesis for gcd computation");
    C c1= content (s1, n1);
    C c2= content (s2, n2);

    nat orig_n2= n2;
    C* d2= mmx_new<C> (orig_n2);
    segment_div (d1, s1, c1, n1);
    segment_div (d2, s2, c2, n2);
    while (n2>0) {
      C g = gcd (d1[n1-1], d2[n2-1]);
      C m1= d2[n2-1] / g;
      C m2= d1[n1-1] / g;
      for (nat i=0; i<n2; i++)
	d1[i+n1-n2]= d1[i+n1-n2] * m1 - d2[i] * m2;
      for (nat i=n2; i<n1; i++)
	d1[i-n2]= d1[i-n2] * m1;
      while (n1 > 0 && d1[n1-1] == 0) n1--;
      if (n1 < n2) {
	for (nat i=0; i<n2; i++) swap (d1[i], d2[i]);
	swap (n1, n2);	
      }
      C c= content (d1, n1);
      for (nat i=0; i<n1; i++)
	d1[i]= d1[i] / c;
    }
    mmx_delete<C> (d2, orig_n2);

    C g= gcd (c1, c2);
    for (nat i=0; i<n1; i++)
      d1[i]= d1[i] * g;
    return n1;
  }
}

/******************************************************************************
* Efficient evaluation of polynomials
******************************************************************************/

template<typename C, typename V> void
jet_annulator (C* d, const C* s, nat n) {
  // Given n constants s[0],...s[n-1],
  // construct the polynomial (z-s[0]) ... (z-s[n-1])
  switch (n) {
  case 0:
    d[0]= C(1);
    break;
  case 1:
    d[1]= C(1);
    d[0]= -s[0];
    break;
  case 2:
    d[2]= C(1);
    d[1]= -(s[0]+s[1]);
    d[0]= s[0]*s[1];
    break;
  case 3:
    d[3]= C(1);
    d[2]= -(s[0]+s[1]+s[2]);
    d[1]= s[0]*s[1] + s[1]*s[2]+ s[2]*s[0];
    d[0]= -s[0]*s[1]*s[2];
    break;
  default:
    {
      nat h= n>>1;
      C* aux= mmx_new<C> (n+2);
      jet_annulator<C,V> (aux, s, h);
      jet_annulator<C,V> (aux+h+1, s+h, n-h);
      jet_mul<C,V> (d, aux, aux+h+1, h+1, n+1-h);
      mmx_delete<C> (aux, n+2);
    }
  }
}

template<typename C> C
jet_eval (const C* p, const C& x, nat l) {
  // Evaluate the polynomial (p, n) at x
  if (l == 0) return 0;
  C r= p[l-1];
  for (nat i=l-2; i!=((nat)(-1)); i--)
    r= (x*r) + p[i];
  return r;
}

template<typename C, typename V> void
jet_eval (C* v, const C* p, const C* x, nat l, nat n) {
  // Evaluate the polynomial (p, l) at points x[0], ..., x[n-1]
  if (n < 64 /*4*/) {
    for (nat j=0; j<n; j++)
      if (l == 0) v[j]= 0;
      else {
	C r= p[l-1];
	for (nat i=l-2; i!=((nat)(-1)); i--)
	  r= (x[j]*r) + p[i];
	v[j]= r;
      }
  }
  else if (n > l) {
    nat h= n>>1; 
    jet_eval<C,V> (v, p, x, l, h);
    jet_eval<C,V> (v+h, p, x+h, l, n-h);
  }
  else {
    nat h= n>>1;
    C* a= mmx_new<C> (n+1 /*a*/ + l /*r*/ + l /*q*/);
    C* r= a + n+1;
    C* q= r + l;
    jet_annulator<C,V> (a, x, n);
    jet_copy (r, p, l);
    jet_quo_rem<C,V> (q, r, a, l, n+1);
    jet_eval<C,V> (v, r, x, n, h);
    jet_eval<C,V> (v+h, r, x+h, n, n-h);
    mmx_delete<C> (a, n+1 + l + l);
  }
}

#undef TMPL
} // namespace mmx
#endif //__MMX__JET__HPP
