
/******************************************************************************
* MODULE     : monomial.hpp
* DESCRIPTION: Multivariate power series
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_MONOMIAL_HPP
#define __MMX_MONOMIAL_HPP
#include "table.hpp"
#include "list.hpp"
#include "wrap.hpp"
namespace mmx {
#define TMPL_DEF template<typename C, typename V= generic>
#define TMPL template<typename C, typename V>
#define Monomial monomial<C,V>
#define Stair list<Monomial >
TMPL class monomial;

/******************************************************************************
* Monomials
******************************************************************************/

TMPL_DEF
struct monomial {
  table<C,V> p;
  generic name;
  inline table<C,V> operator * () const { return p; }
  inline monomial (): p (0) {}
  inline monomial (int p2): p (0) { ASSERT (p2==1, "one expected"); }
  inline monomial (const table<C,V>& p2): p (p2) {}
  inline monomial (const V& v, const C& i): p (0) { p[v]= i; }
  inline C operator [] (const V& v) const { return p[v]; }
};

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL table<bool,V>
variables (const Monomial& m) {
  table<bool,V> vs;
  for (iterator<V> it= entries (*m); busy (it); ++it)
    vs[*it]= (m[*it] != 0);
  return vs;
}

TMPL V
as_variable (const Monomial& m) {
  V prod= 1;
  for (iterator<V> it= entries (*m); busy (it); ++it)
    prod= prod * pow (*it, construct (m[*it]));
  return prod;
}

TMPL generic
flatten (const Monomial& m) {
  generic prod= 1;
  for (iterator<V> it= entries (*m); busy (it); ++it)
    prod= prod * pow (flatten (*it), construct (m[*it]));
  return prod;
}

WRAP_INDIRECT_IMPL(TMPL inline,Monomial)
PRINT_SUGAR(TMPL,Monomial)

/******************************************************************************
* Standard arithmetic
******************************************************************************/

TMPL inline Monomial operator * (const Monomial& m1, const Monomial& m2) {
  return Monomial (m1.p + m2.p); }
TMPL inline Monomial operator / (const Monomial& m1, const Monomial& m2) {
  return Monomial (m1.p - m2.p); }
TMPL inline Monomial operator / (int m1, const Monomial& m2) {
  return Monomial (m1) / m2; }
template<typename C, typename V, typename K> inline Monomial
pow (const Monomial& m, const K& p) {
  return Monomial (C(p) * m.p); }
template<typename C, typename V, typename K> inline Monomial
nth_root (const Monomial& m, const K& p) {
  return Monomial (m.p / C(p)); }

TMPL inline Monomial gcd (const Monomial& m1, const Monomial& m2) {
  return Monomial (inf (m1.p, m2.p)); }
TMPL inline Monomial lcm (const Monomial& m1, const Monomial& m2) {
  return Monomial (sup (m1.p, m2.p)); }

TMPL inline bool operator << (const Monomial& m1, const Monomial& m2) {
  return m1.p > m2.p; }
TMPL inline bool operator <<= (const Monomial& m1, const Monomial& m2) {
  return m1.p >= m2.p; }
TMPL inline bool operator >> (const Monomial& m1, const Monomial& m2) {
  return m1.p < m2.p; }
TMPL inline bool operator >>= (const Monomial& m1, const Monomial& m2) {
  return m1.p <= m2.p; }

TMPL inline bool operator == (const Monomial& m1, const int& i) {
  return m1 == Monomial (i); }
TMPL inline bool operator != (const Monomial& m1, const int& i) {
  return m1 != Monomial (i); }
TMPL inline bool operator << (const Monomial& m1, const int& i) {
  return m1 << Monomial (i); }
TMPL inline bool operator <<= (const Monomial& m1, const int& i) {
  return m1 <<= Monomial (i); }
TMPL inline bool operator >> (const Monomial& m1, const int& i) {
  return m1 >> Monomial (i); }
TMPL inline bool operator >>= (const Monomial& m1, const int& i) {
  return m1 >>= Monomial (i); }

/******************************************************************************
* Monomial substitutions
******************************************************************************/

TMPL table<Monomial,V>
identity_substitution (const Monomial& m) {
  // return identity map for variables occurring in m
  table<Monomial,V> r;
  for (iterator<V> it= entries (*m); busy (it); ++it)
    r[*it]= Monomial (*it, 1);
  return r;
}

template<typename C, typename V, typename K, typename W> Monomial
compose (const table<Monomial,W>& t, const monomial<K,W>& m) {
  Monomial r= 1;
  for (iterator<W> it= entries (*m); busy (it); ++it)
    r= r * pow (t[*it], C(m[*it]));
  return r;
}

template<typename C, typename V, typename K, typename W, typename X>
table<monomial<C,V>,X>
compose (const table<monomial<C,V>,W>& t, const table<monomial<K,W>,X>& u) {
  table<monomial<C,V>,X> r;
  for (iterator<W> it= entries (*u); busy (it); ++it)
    r[*it]= compose (t, u[*it]);
  return r;
}

/******************************************************************************
* Widenings
******************************************************************************/

TMPL Monomial
ramify (const Monomial& m, const C& p) {
  Monomial r (1);
  for (iterator<V> it= entries (*m); busy (it); ++it)
    r= r * Monomial (pow (*it, 1/V(p)), p * m[*it]);
  return r;
}

template<typename C, typename V, typename K, typename W>
table<Monomial,V>
make_bounded (const table<monomial<K,W>,V>& t,
	      const Monomial& eps)
{
  // On input: Cartesian coordinates t and a Cartesian monomial eps
  // which represent a bounded monomial.
  // On output: a change of Cartesian coordinates in which eps is bounded
  if (eps <<= 1) return identity_substitution (eps);
  for (iterator<V> it= entries (*eps); busy (it); ++it)
    if (eps[*it] > 0) {
      Monomial m= eps / Monomial (*it, eps[*it]);
      if (compose (t, m) <<= 1) {
	table<Monomial,V> r= make_bounded (t, m);
	r[*it]= Monomial (*it, 1);
	return r;
      }
      else {
	C p= eps[*it];
	table<Monomial,V> r= make_bounded (t, 1/m);
	table<Monomial,V> s;
	for (iterator<V> it2= entries (r); busy (it2); ++it2)
	  s[*it2]= ramify (r[*it2], p);
	V z= pow (as_variable (eps), 1/V(p));
	if (z == 1) s[*it]= nth_root (compose (s, 1/m), p);
	else s[*it]= nth_root (Monomial (z, p) * compose (s, 1/m), p);
	return s;
      }
    }
  ERROR ("unexpected situation in make_bounded");
}

/******************************************************************************
* Stairs
******************************************************************************/

TMPL Stair
stair_cons (const Monomial& m, const Stair& l) {
  if (nil (l)) return cons (m, l);
  else if (m <<= car (l)) return l;
  else if (car (l) <<= m) return stair_cons (m, cdr (l));
  else return cons (car (l), stair_cons (m, cdr (l)));
}

TMPL Stair
stair_append (const Stair& l, const Stair& r) {
  if (nil (l)) return r;
  else return stair_cons (car (l), stair_append (cdr (l), r));
}

TMPL Stair
stair_mul (const Monomial& m, const Stair& l) {
  if (nil (l)) return l;
  else return cons (m * car (l), stair_mul (m, cdr (l)));
}

TMPL C
max_power_in (const Stair& l, const V& v) {
  if (nil (l)) return Minus_infinity(C);
  else return max (car (l) [v], max_power_in (cdr (l), v));
}

template<typename V> int
max_power_in (const list<monomial<int,V> >& l, const V& v) {
  if (nil (l)) return -((int) (((nat) (-1)) >> 1));
  else return max (car (l) [v], max_power_in (cdr (l), v));
}

TMPL Stair
project_up_to (const Stair& l, const V& v, int i) {
  if (nil (l)) return l;
  else {
    int j= car (l) [v];
    Stair r= project_up_to (cdr (l), v, i);
    if (j <= i) return stair_cons (car (l) / Monomial (v, j), r);
    else return r;
  }
}

#undef TMPL_DEF
#undef TMPL
#undef Monomial
#undef Stair
} // namespace mmx
#endif // __MMX_MONOMIAL_HPP
