
/******************************************************************************
* MODULE     : relaxed_series.hpp
* DESCRIPTION: Relaxed univariate power series
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_RELAXED_SERIES_HPP
#define __MMX_RELAXED_SERIES_HPP
#include "series.hpp"
#include "basix_utils.hpp"
namespace mmx {
#define TMPL template<typename C, typename V>
#define Variant relaxed<V>
#define Series series<C,Variant >
#define Series_rep series_rep<C,Variant >

/******************************************************************************
* Relaxed series variants
******************************************************************************/

template<typename V>
class relaxed {
public:
};

class integer;
class rational;
template<typename V> class floating;
struct kronecker_multiply;

STMPL
struct series_variant_helper<integer> {
  typedef relaxed<kronecker_multiply> SV;
};

STMPL
struct series_variant_helper<rational> {
  typedef relaxed<kronecker_multiply> SV;
};

template<typename V>
struct series_variant_helper<floating<V> > {
  typedef relaxed<kronecker_multiply> SV;
};

/******************************************************************************
* Relaxed multiplication
******************************************************************************/

TMPL
class relaxed_mul_series_rep: public Series_rep {
protected:
  Series f, g;

  // NOTE: The following fields keep track of large segments of zeros
  // in the coefficients. This makes the multiplication algorithm linear
  // in the case when we multiply by a polynomial. This is for instance
  // interesting when computing coefficients of holonomic functions.
  nat zeros_f;  // bit 2^p set if f[2^p-1 ... 2^{p+1}-1] = 0
  nat zeros_g;  // bit 2^p set if g[2^p-1 ... 2^{p+1}-1] = 0

public:
  relaxed_mul_series_rep (const Series& f2, const Series& g2):
    f (f2), g (g2), zeros_f (0), zeros_g (0) {}
  ~relaxed_mul_series_rep () {}

  generic expression (const generic& z) const {
    return flatten (f, z) * flatten (g, z); }

  void Set_precision (nat l2) {
    // NOTE: far from optimal, but let us not bother...
    if (l2 <= this->l) return;
    mmx_delete<C> (this->a, this->l);
    this->a= mmx_new<C> (l2);
    this->l= l2;
    jet_clear (this->a, this->l);
    nat k= this->n;
    this->n= 0;
    /* not so nice in case of zero coefficients...
    if (k>0) {
      nat i= 1;
      while (((i<<1) + 1 < k) && ((i<<1) - 1 < this->l))
	i= (i<<1) + 1;
      jet_mul<C,V> (this->a, f->a, g->a, i, i);
      this->n= i;
    }
    */
    while (this->n < k) {
      (void) next ();
      this->n++;
    }
  }

  void Increase_precision (nat l) {
    if (this->l < l) {
      nat k=1;
      while (k <= this->l) k= (k<<1) + 1;
      Set_precision (max (k, l));
    }
    increase_precision (f, l);
    increase_precision (g, l);
  }

  C next () {
    nat k= this->n;
    if (k == 0) this->a[0]= f[0] * g[0];
    else {
      this->a[k] += f[0] * g[k] + f[k] * g[0];
      if ((k&1) == 0) {
	nat p= 0;
	k += 2;
	while ((k&1) == 0 && k != 2) {
	  p++;
	  k= k >> 1;
	  nat len= min (this->l, ((k+2)<<p) - 3) - ((k<<p) - 2);
	  nat ord= min (((nat) 1)<<p, len);
	  if (k == 2) {
	    if (jet_equal (f->a+(1<<p)-1, C(0), 1<<p)) zeros_f |= (1<<p);
	    if (jet_equal (g->a+(1<<p)-1, C(0), 1<<p)) zeros_g |= (1<<p);
	  }
	  C* aux= mmx_new<C> ((ord << 1) - 1);
	  if ((zeros_f & (1<<p)) == 0) {
	    jet_mul<C,V> (aux, f->a + (1<<p)-1, g->a + ((k-1)<<p)-1, ord, ord);
	    jet_add (this->a + (k<<p)-2, aux, len);
	  }
	  if ((zeros_g & (1<<p)) == 0 && k != 2) {
	    jet_mul<C,V> (aux, f->a + ((k-1)<<p)-1, g->a + (1<<p)-1, ord, ord);
	    jet_add (this->a + (k<<p)-2, aux, len);
	  }
	  mmx_delete<C> (aux, (ord << 1) - 1);
	}
      }
    }
    return this->a [this->n];
  }
};

TMPL Series
operator * (const Series& f, const Series& g) {
  return (Series_rep*) new relaxed_mul_series_rep<C,V> (f, g);
}

/******************************************************************************
* Right composition with polynomials
******************************************************************************/

TMPL Series compose (const Series& f, nat i, nat p, nat q, Series* H);

TMPL
struct partial_pol_compose_series_rep: public Series_rep {
  Series f, h;
  nat i, r;
  partial_pol_compose_series_rep<C,V>
    (const Series& f2, nat i2, nat p, nat q, Series* H);
  generic expression (const generic& z) const {
    return mmx::gen ("partial_polynomial_compose"); }
  void Increase_precision (nat l) {
    Series_rep::Increase_precision (mmx::min (l, r));
    increase_precision (h, mmx::min (l, r));
  }
  C next ();
};

TMPL
partial_pol_compose_series_rep<C,V>::partial_pol_compose_series_rep (
  const Series& f2, nat i2, nat p, nat q, Series* H):
    f (f2), i (i2), r ((p-1) * (q-1) + 1)
{
  assert (p > 1);
  Series h1= compose (f, i, p>>1, q, H);
  Series h2= compose (f, i+ (p>>1), p>>1, q, H);
  int d= p>>1;
  h= h1+ shift_exp (h2 * shift_exp (H[p>>1], -d), d);
}

TMPL C
partial_pol_compose_series_rep<C,V>::next () {
  nat n= this->n;
  if (n == r) h= 0;
  if (n >= r) return 0;
  return h[n];
}

TMPL Series
compose (const Series& f, nat i, nat p, nat q, Series* H) {
  if (p == 1) return Series (f[i]);
  return (Series_rep*) new partial_pol_compose_series_rep<C,V> (f, i, p, q, H);
}

TMPL
struct pol_compose_series_rep: public Series_rep {
  Series f, *H, h;
  nat q;
  pol_compose_series_rep<C,V> (const Series& f, const Series& g, nat q);
  ~pol_compose_series_rep<C,V> () { mmx::mmx_classical_delete<Series > (H); }
  generic expression (const generic& z) const {
    return mmx::gen (GEN_COMPOSE, flatten (f, z), flatten (H[1], z)); }
  void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f, l);
    increase_precision (h, l);
  }
  C next ();
};

TMPL
pol_compose_series_rep<C,V>::pol_compose_series_rep
  (const Series& f2, const Series& g, nat q2):
    f (f2), q (q2)
{
  H= mmx::mmx_classical_new<Series > (2);
  H[1]= g;
  h= compose (f, 0, 1, q, H);
}

TMPL C
pol_compose_series_rep<C,V>::next () {
  nat n= this->n;
  nat k= n;
  nat i;
  if (k > 0) while ((k&1) == 0) k= k>>1;
  if (k == 1) {
    if (n > 1) {
      Series* H2= mmx::mmx_classical_new<Series > (n+1);
      for (i=1; i<n; i=i<<1) H2[i]= H[i];
      H2[n]= H[n>>1] * H[n>>1];
      mmx::mmx_classical_delete<Series > (H);
      H= H2;
    }
    h= compose (f, 0, n<<1, q, H);
    increase_precision (h, this->l);
  }
  return h[n];
}

TMPL Series
compose (const Series& f, const Series& g, nat q) {
  ASSERT (g[0] == 0, "constant coefficient must be zero");  
  return (Series_rep*) new pol_compose_series_rep<C,V> (f, g, q);
}

/******************************************************************************
* General relaxed composition
******************************************************************************/

template<typename C> C
multiply_range (nat first, nat end) {
  if (end == first) return C(1);
  else if (end == first+1) return C (first);
  nat half= (first + end) >> 1;
  return multiply_range<C> (first, half) * multiply_range<C> (half, end);
}

TMPL
class relaxed_compose_series_rep: public Series_rep {
protected:
  Series f, g, h;

public:
  relaxed_compose_series_rep (const Series& f2, const Series& g2):
    f (f2), g (g2),
    h ((Series_rep*) new compose_series_rep<C,Variant> (f2, g2)) {}
  ~relaxed_compose_series_rep () {}

  generic expression (const generic& z) const {
    return gen (GEN_COMPOSE, flatten (f, z), flatten (g, z)); }

  void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f, l);
    increase_precision (g, l);
  }

  C next () {
    nat n= this->n;
    if (n==0) return f[0];
    if (n==1) return f[1]*g[1];
    if (n==2) return f[2]*g[1]*g[1]+ f[1]*g[2];
    if (n==3) return (f[3]*g[1]*g[1]+ C(2)*f[2]*g[2])*g[1]+ f[1]*g[3];

    nat p=0;
    while (((p+1) * (1<<(2*p))) <= n) p++;
    if ((n != 4) && ((p * (1<<(2*p-2))) <= (n-1))) return h[n];

    nat nn= (p+1) * (1<<(2*p));
    nat i, q= 2<<p, r= (nn+q-1)/q;
    Series g1 = restrict (g, 0, q);
    Series dg1= derive (g1);
    Series g2 = shift_exp (g - g1, -((int) q));

    Series D= f;
    for (i=1; i<r; i++) D= derive (D);
    D= compose (D, g1, q) / multiply_range<C> (1, r); // divides by (r-1)!
    h= D;

    for (i=r-1; i>0; i--) {
      D= (integrate (C(i) * D * dg1)) + f[i-1];
      h= D + shift_exp (h*g2, q);
    }

    increase_precision (h, this->l);
    return h[n];
  }
};

TMPL Series
bk_compose (const Series& f, const Series& g) {
  // NOTE: not yet default; needs further efficiency and accuracy twiddling
  ASSERT (g[0] == 0, "constant coefficient must be zero");  
  return (Series_rep*) new relaxed_compose_series_rep<C,V> (f, g);
}

#undef TMPL
#undef Variant
#undef Series
#undef Series_rep
} // namespace mmx
#endif // __MMX_SERIES_HPP
