
/******************************************************************************
* MODULE     : holonomic_operators.hpp
* DESCRIPTION: Extra routines for holonomic operators
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_HOLONOMIC_OPERATORS_HPP
#define __MMX_HOLONOMIC_OPERATORS_HPP
#include "skew_polynomial.hpp"
#include "quotient.hpp"
namespace mmx {
#define TMPL template<typename C>
#define Holonomic holonomic<C>
#define Polynomial polynomial<C>
#define Ratfunc quotient<Polynomial,Polynomial >
#define D_multiply skew_multiply<d_operator>
#define D_variant dense_polynomial<D_multiply >
#define D_operator polynomial<Polynomial,D_variant >
#define D_monic_operator polynomial<Ratfunc,D_variant >
#define XD_multiply skew_multiply<xd_operator>
#define XD_variant dense_polynomial<XD_multiply >
#define XD_operator polynomial<Polynomial,XD_variant >
#define XD_monic_operator polynomial<Ratfunc,XD_variant >
#define S_multiply skew_multiply<S_operator<C> >
#define S_variant dense_polynomial<S_multiply >
#define S_operator polynomial<Polynomial,S_variant >
#define S_monic_operator polynomial<Ratfunc,S_variant >
#define T_multiply skew_multiply<T_operator<C> >
#define T_variant dense_polynomial<T_multiply >
#define T_operator polynomial<Polynomial,T_variant >
#define T_monic_operator polynomial<Ratfunc,T_variant >
TMPL class holonomic;

/******************************************************************************
* Normalization
******************************************************************************/

template<typename C, typename V> polynomial<Polynomial,V>
make_reduced (const polynomial<Polynomial,V>& p) {
  if (N(p) == 0) return p;
  Polynomial g= p[0];
  for (nat i=1; i<N(p); i++)
    g= gcd (g, p[i]);
  return skew_div (p, g, true);
}

template<typename C, typename V> polynomial<Ratfunc,V>
make_monic (const polynomial<Ratfunc,V>& p) {
  ASSERT (N(p) != 0, "zero operator cannot be made monic");
  return skew_div (p, p[N(p)-1], true);
}

template<typename C, typename V> polynomial<Ratfunc,V>
make_monic (const polynomial<Polynomial,V>& p) {
  return make_monic (polynomial<Ratfunc,V> (p));
}

template<typename C, typename V> Denominator_type(C)
least_common_denominator (const polynomial<C,V>& p) {
  if (N(p) == 0) return Denominator_type(C) (1);
  Denominator_type(C) l= denominator (p[0]);
  for (nat i=1; i<N(p); i++)
    l= lcm (l, denominator (p[i])); // dichotomic method would be better...
  return l;
}

template<typename C, typename V> polynomial<Polynomial,V>
make_fraction_free (const polynomial<Ratfunc,V>& p) {
  if (N(p) == 0) return 0;
  Polynomial l= least_common_denominator (p);
  Polynomial* c= mmx_new<Polynomial > (N(p));
  for (nat i=0; i<N(p); i++)
    c[i]= numerator (p[i]) * (l / denominator (p[i]));
  return polynomial<Polynomial,V> (c, N(p));
}

template<typename C, typename V> polynomial<Polynomial,V>
make_coefficients_fraction_free (const polynomial<Polynomial,V>& p) {
  if (N(p) == 0) return 0;
  Denominator_type(C) l= least_common_denominator (p[0]);
  for (nat i=1; i<N(p); i++)
    l= lcm (l, least_common_denominator (p[i]));
  return Polynomial (l) * p;
}

/******************************************************************************
* Conversions from D_operator to XD_operator and vice versa
******************************************************************************/

template<typename C> XD_operator
as_XD_operator (const D_operator& L) {
  nat n= N(L);
  if (n == 0) return 0;
  // Build matrix with Stirling numbers
  C* s= mmx_new<C> (n*n, C(0));
  s[0]= 1;
  for (nat i=1; i<n; i++)
    for (nat j=1; j<n; j++) 
      s[n*i + j]= s[n*(i-1) + (j-1)] - (i-1) * s[n*(i-1) + j];
  // Determine power of x by which we have to multiply
  int v= 0;
  for (nat i=0; i<n; i++)
    if (L[i] != 0)
      v= max (v, ((int) i) - val (L[i]));
  // Compute new conversion
  Polynomial* c= mmx_new<Polynomial > (n);
  for (nat j=0; j<n; j++) {
    c[j]= 0;
    for (nat i=j; i<n; i++)
      c[j] += s[n*i + j] * shift_exp (L[i], v-i);
  }
  mmx_delete<C> (s, n*n);
  return XD_operator (c, n);
}

template<typename C> D_operator
as_D_operator (const XD_operator& L) {
  nat n= N(L);
  if (n == 0) return 0;
  // Build matrix with Stirling numbers
  C* s= mmx_new<C> (n*n, C(0));
  s[0]= 1;
  for (nat i=1; i<n; i++)
    for (nat j=1; j<n; j++) 
      s[n*i + j]= s[n*(i-1) + (j-1)] + j * s[n*(i-1) + j];
  // Compute new conversion
  Polynomial* c= mmx_new<Polynomial > (n);
  for (nat j=0; j<n; j++) {
    c[j]= 0;
    for (nat i=j; i<n; i++)
      c[j] += s[n*i + j] * L[i];
  }
  mmx_delete<C> (s, n*n);
  // Multiply by the necessary powers of x and shift back
  int v= (int) (((nat) (-1)) >> 1);
  for (nat i=0; i<n; i++)
    if (c[i] != 0)
      v= min (v, val (c[i]) + ((int) i));
  for (nat i=0; i<n; i++)
    c[i]= shift_exp (c[i], i-v);
  return D_operator (c, n);
}

/******************************************************************************
* Conversions from S_operator to T_operator and vice versa
******************************************************************************/

template<typename C> T_operator
as_T_operator (const S_operator& L) {
  nat n= N(L);
  if (n == 0) return 0;
  Polynomial* c= mmx_new<Polynomial > (n);
  for (nat i=0; i<n; i++)
    c[n-1-i]= shift (L[i], C(1-((int) n)));
  return T_operator (c, n);
}

template<typename C> S_operator
as_S_operator (const T_operator& L) {
  nat n= N(L);
  if (n == 0) return 0;
  Polynomial* c= mmx_new<Polynomial > (n);
  for (nat i=0; i<n; i++)
    c[n-1-i]= shift (L[i], C(((int) n)-1));
  return S_operator (c, n);
}

/******************************************************************************
* Conversions from XD_operator to S_operator and vice versa
******************************************************************************/

template<typename C, typename V> nat
max_N (const polynomial<Polynomial,V>& P) {
  nat i, n= 0;
  for (i=0; i<N(P); i++)
    n= max (n, N(P[i]));
  return n;
}

template<typename C> S_operator
as_S_operator (const XD_operator& L) {
  nat n= N(L), p= max_N (L);
  Polynomial* c= mmx_new<Polynomial > (p);
  for (nat i=0; i<p; i++) {
    C* cc= mmx_new<C> (n);
    for (nat j=0; j<n; j++)
      cc[j]= L[j][i];
    c[p-1-i]= shift (Polynomial (cc, n), C (p-1-i));
  }
  return make_reduced (S_operator (c, p));
}

template<typename C> XD_operator
as_XD_operator (const S_operator& L) {
  nat p= N(L);
  Polynomial* k= mmx_new<Polynomial > (p);
  for (nat i=0; i<p; i++)
    k[i]= shift (L[i], C (-((int) i)));
  S_operator K (k, p);
  nat n= max_N (K);
  Polynomial* c= mmx_new<Polynomial > (n);
  for (nat j=0; j<n; j++) {
    C* cc= mmx_new<C> (p);
    for (nat i=0; i<p; i++)
      cc[i]= K[p-1-i][j];
    c[j]= Polynomial (cc, p);
  }
  return make_reduced (XD_operator (c, n));
}

/******************************************************************************
* Complete conversions using transitivity
******************************************************************************/

template<typename C> T_operator
as_T_operator (const XD_operator& L) {
  return as_T_operator (as_S_operator (L));
}

template<typename C> XD_operator
as_XD_operator (const T_operator& L) {
  return as_XD_operator (as_S_operator (L));
}

template<typename C> S_operator
as_S_operator (const D_operator& L) {
  return as_S_operator (as_XD_operator (L));
}

template<typename C> D_operator
as_D_operator (const S_operator& L) {
  return as_D_operator (as_XD_operator (L));
}

template<typename C> T_operator
as_T_operator (const D_operator& L) {
  return as_T_operator (as_XD_operator (L));
}

template<typename C> D_operator
as_D_operator (const T_operator& L) {
  return as_D_operator (as_XD_operator (L));
}

#undef TMPL
#undef Holonomic
#undef Polynomial
#undef Ratfunc
#undef D_multiply
#undef D_variant
#undef D_operator
#undef D_monic_operator
#undef XD_multiply
#undef XD_variant
#undef XD_operator
#undef XD_monic_operator
#undef S_multiply
#undef S_variant
#undef S_operator
#undef S_monic_operator
#undef T_multiply
#undef T_variant
#undef T_operator
#undef T_monic_operator
} // namespace mmx
#endif // __MMX_HOLONOMIC_OPERATORS_HPP
