
/******************************************************************************
* MODULE     : inst_numeric_polynomial.hpp
* DESCRIPTION: Extra routines for numeric polynomials
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_INST_NUMERIC_POLYNOMIAL_HPP
#define __MMX_INST_NUMERIC_POLYNOMIAL_HPP
#include "numeric_polynomial.hpp"
#include "jet.hpp"
#include "fft.hpp"
namespace mmx {
#define TMPL template<typename C>
#define Real Real_type(C)
#define Center Center_type(C)
#define Real_center Center_type(Real)
#define Radius Radius_type(C)
#define Polynomial polynomial<C>

/******************************************************************************
* Finding the smallest root of a numeric polynomial
******************************************************************************/

TMPL Real
lower_bound_roots (const Polynomial& P) {
  // Fast lower bound computation using Graeffe transforms
  if (P[0] == 0) return 0;
  Polynomial Q= reverse (P);
  int stage, i, d= deg (Q);
  Q /= Q [d];
  Real r= 1, best_bnd;
  Real_center n= 1;
  for (stage=0; true; stage++, n *= 2) {
    /*
    // compute scaling factor
    Real m= 0, M= 0;
    for (i=0; i<d; i++) {
      Real a= abs_D (Q[i]);
      if (a != 0) m= max (m, exp (log (a) / (d-i)));
      a= abs_U (Q[i]);
      if (a != 0) M= max (M, exp (log (a) / (d-i)));
    }
    // nat p= max_prec (Q) + DEFAULT_LOW;
    Real scale= copy (M);
    if (m == Real (0) && stage == 0) return 0;
    */

    // compute scaling factor
    Real scale= 0;
    for (i=0; i<d; i++) {
      Real a= abs (Q[i]);
      if (a != 0) scale= max (scale, exp (log (a) / (d-i)));
    }
    if (scale == Real (0) && stage == 0) return 0;
    // nat p= max_prec (Q) + DEFAULT_LOW;
    Real_center m= lower (scale);
    Real_center M= upper (scale);
    scale= M;

    // update bound
    r *= exp (log (scale) / Real (n));
    Real s= make_interval<Real > (exp (log (m / (Real_center (d) * M)) / n),
				  exp (log (Real_center (2)) / n));
    Real bnd= Real (1) / (r * s);
    if (stage == 0) best_bnd= bnd;
    else best_bnd=
	   make_interval<Real > (max (lower (best_bnd), lower (bnd)),
				 min (upper (best_bnd), upper (bnd)));
    if (radius (s) * Radius (d << 5) < Radius (1)) break;
    if ((m << 1) < M) break;

    // scale Q and square roots
    Real prod= 1;
    C* coeffs= mmx_new<C> (d+1);
    for (i=d; i>=0; i--, prod *= scale)
      coeffs[i]= Q[i] / prod;
    Q= Polynomial (coeffs, d+1);
    Q= square_roots (Q);
  }
  return best_bnd;
}

TMPL Real
lower_bound_roots (const C* P, nat n, const Real& r, const Real& err) {
  //cout << "r[" << n << "]\t\t= " << r << "\n";
  if (abs (P[0]) <= (err << 1)) {
    if (abs_D (P[0]) == 0) return 0;
    Real lambda= Real (upper (Real (2.1) * err / abs (P[0])));
    Real mu    = exp (log (lambda) / n);
    return lower_bound_roots (P, n, r/mu, err/lambda);
  }
  else if (n == 1) return r;
  else {
    Real bnd;
    if (n == 2) {
      bnd= (abs (P[0]) - err) / abs (P[1]);
      //cout << "bnd-1[" << n << "]\t= " << min (r, bnd) << "\n";
    }
    else {
      nat i, h= (n+1) >> 1;
      Real sum= err, prod= pow (r, h);
      for (i=h; i<n; i++, prod *= r)
	sum += abs (P[i]) * prod;
      bnd= lower_bound_roots (P, h, r, sum);
      //cout << "bnd-2[" << n << "]\t= " << min (r, bnd) << "\n";
      C* coeffs= mmx_new<C> (n);
      coeffs[0]= P[0] + make_ball<C> (Center (0), bnd_U (err));
      for (i=1; i<n; i++) coeffs[i]= P[i];
      Polynomial Q (coeffs, n);
      bnd= max (bnd, lower (lower_bound_roots (Q)));
      //cout << "bnd-3[" << n << "]\t= " << min (r, bnd) << "\n";
    }
    while ((bnd << 1) < r) {
      Real new_r  = Real (lower (bnd << 1));
      Real new_err= err * pow (new_r/r, n);
      Real new_bnd= lower_bound_roots (P, n, new_r, new_err);
      bool stop= new_bnd < Real (1.9) * bnd;
      bnd= max (bnd, new_bnd);
      //cout << "bnd-4[" << n << "]\t= " << min (r, bnd) << "\n";
      if (stop) break;
    }
    return min (r, bnd);
  }
}

/******************************************************************************
* Finding extremal values on a disk
******************************************************************************/

#define MAXIMUM_DISK           0
#define MINIMUM_DISK           1
#define MAXIMUM_REAL_PART_DISK 2

TMPL Real
extremum_disk (const C* P, nat n, const Real& r, int type) {
  if (n == 0) return 0;
  else if (n == 1) return abs (P[0]);
  else if (n == 2) {
    if (type == MAXIMUM_DISK) return abs (P[0]) + abs (P[1]) * r;
    else if (type == MINIMUM_DISK) return max (abs (P[0]) - abs (P[1]) * r, 0);
    else return Re (P[0]) + abs (P[1]) * r;
  }
  else {
    nat i;
    /*
    nat max_prec= 0;
    for (i=0; i<n; i++)
      max_prec= max (max_prec, sq_prec (P[i]));
    Real s= Real (upper (r), max_prec);
    */
    Real s= Real (upper (r));
    C* Q= mmx_new<C> (n);
    jet_q_difference (Q, P, C (s), n);
    /*
    mp_exp_t max_expo= mpfr_get_emin (), max_step= mpfr_get_emin ();
    for (i=0; i<n; i++) {
      max_expo= max (max_expo, sq_expo (Q[i]));
      max_step= max (max_step, sq_step (Q[i]));
    }
    */

    //fft_prec  = (nat) max ((mp_exp_t) 0, max_expo - max_step);
    //fft_prec += DEFAULT_LOW;

    nat p=1;
    while (p<3*n) p= p << 1;
    Real* r= mmx_new<Real > (p);
    C* a= mmx_new<C> (p);
    Real delta= abs ((fft_root<C > (2*p)).u - 1);
    Real pow_delta= 1;
    nat stage, dq= n;
    for (stage=0; true; stage++) {
      for (i=0; i<dq; i++) a[i]= Q[i];
      for (; i<p; i++) a[i]= 0;
      fft (a, p, false);
      if (stage == 0) {
	for (i=0; i<p; i++)
	  if (type < MAXIMUM_REAL_PART_DISK) r[i]= abs (a[i]);
	  else r[i]= abs (Re (a[i]));
      }
      else {
	for (i=0; i<p; i++)
	  if (type != MINIMUM_DISK) r[i] += abs (a[i]) * pow_delta;
	  else r[i]= max (0, r[i] - abs (Re (a[i])) * pow_delta);
      }
      /*
      cout << "a";
      for (i=0; i<p; i++)
	cout << ", " << a[i];
      cout << "\n----------\n";
      cout << "r";
      for (i=0; i<p; i++)
	cout << ", " << r[i];
      cout << "\n----------\n";
      */

      dq--;
      for (i=0; i<dq; i++)
	Q[i]= (Real (i+1) / Real (stage+1)) * Q[i+1];
      pow_delta *= delta;

      Real err= 0, pd= pow_delta;
      for (i=0; i<dq; i++, pd *= delta)
	err += abs (Q[i]) * pd;
      //cout << "err= " << err << "\n";

      Real result= r[0];
      for (i=1; i<p; i++)
	if (type != MINIMUM_DISK) result= max (result, r[i]);
	else result= min (result, r[i]);
      if (dq == 0 || p*err < result || result == 0) {
	mmx_delete<C> (Q, n);
	mmx_delete<Real > (r, p);
	mmx_delete<C> (a, p);
	if (type != MINIMUM_DISK) return result + err;
	else return max (0, result - err);
      }
    }
  }
}

TMPL Real
maximum_disk (const C* P, nat n, const Real& r) {
  return extremum_disk (P, n, r, MAXIMUM_DISK);
}

TMPL Real
minimum_disk (const C* P, nat n, const Real& r) {
  return extremum_disk (P, n, r, MINIMUM_DISK);
}

TMPL Real
maximum_real_part_disk (const C* P, nat n, const Real& r) {
  return extremum_disk (P, n, r, MAXIMUM_REAL_PART_DISK);
}

#undef TMPL
#undef Real
#undef Radius
#undef Center
#undef Real_center
#undef Polynomial
} // namespace mmx
#endif // __MMX_INST_NUMERIC_POLYNOMIAL_HPP
