
/******************************************************************************
* MODULE     : polynomial_roots.hpp
* DESCRIPTION: Finding the roots of a complex polynomial
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_POLYNOMIAL_ROOTS_HPP
#define __MMX_POLYNOMIAL_ROOTS_HPP
#include "dense_polynomial.hpp"
#include "complex_ball.hpp"
namespace mmx {
#define TMPL template<typename C>
#define Real Real_type(C)
#define Radius Radius_type(C)
#define Vector vector<C>
#define Polynomial polynomial<C>
#define Polygon vector<nat>

TMPL Vector newton_iterate (const Polynomial& p, const Vector& v);
TMPL Vector adjust_radii (const Polynomial& p, const Vector& v);
TMPL Vector roots (const Polynomial& p);

/******************************************************************************
* Utilities
******************************************************************************/

TMPL inline Vector sharpen (const Vector& v) {
  return unary_map<sharpen_op> (v); }
TMPL inline Polynomial sharpen (const Polynomial& p) {
  C* r= mmx_new<C> (N(p));
  for (nat i=0; i<N(p); i++)
    r[i]= sharpen (p[i]);
  return Polynomial (r, N(p)); }

TMPL C
nth_root (const C& x, nat n) {
  switch (n) {
  case 1: return x;
  case 2: return sqrt (x);
  default: return exp (log (x) / n);
  }
}

STMPL static complex_ball
nth_root (const complex_ball& z, nat n) {
  if (n == 1) return z;
  else if (z == 0)
    return complex_ball (complex_ball_center (0), nth_root (abs_U (z), n));
  else if (n == 2) return sqrt (z);
  else return exp (log (z) / n);
}

TMPL Radius_type(C)
abs_slope (const Polynomial& p, nat i, nat j) {
  return nth_root (abs_U (p[i] / p[j]), j-i);
}

/******************************************************************************
* Newton polygons (=> rough idea about radii of roots)
******************************************************************************/

TMPL nat
newton_polygon (nat* dest, const C* src, nat n) {
  // Put the abscissa of the Newton polygon of [src, n] into [dest, k]
  // with k<=n. We have dest[k-1]= n-1.
  if (n<3) {
    assert (n == 2);
    dest[0]= 0;
    dest[1]= 1;
    return 2;
  }
  else {
    nat cur= 0;
    dest[cur++]= 0;
    dest[cur++]= 1;
    for (nat k=2; k<n; k++) {
      while (cur>1) {
	nat i=dest[cur-2], j=dest[cur-1];
	if (C (k - i) * (src[j] - src[i]) >= C (j - i) * (src[k] - src[i]))
	  break;
	cur--;
      }
      dest[cur++]= k;
    }
    return cur;
  }
}

TMPL vector<double>
magnitudes (const Polynomial& p) {
  nat i, n= N(p);
  double* l= mmx_new<double> (n);
  for (i=0; i<n; i++)
    l[i]= magnitude (abs_U (p[i]));
  return vector<double> (l, n);
}

TMPL Polygon
newton_polygon (const Vector& v) {
  nat n= N(v);
  nat* np= mmx_new<nat> (n);
  nat k= newton_polygon (np, seg (v), n);
  return Polygon (np, k, n);
}

TMPL inline Polygon
newton_polygon (const Polynomial& p) {
  return newton_polygon (magnitudes (p));
}

/******************************************************************************
* Angle frequencies (=> rough idea about arguments of roots)
******************************************************************************/

template<typename C> vector<nat>
angle_frequencies (const Polynomial& p, const Polygon& np, nat n, double ref) {
  nat i, k= N(np);
  nat* freq= mmx_new<nat> (n, 0);
  for (i=0; i<k-1; i++) {
    nat a= np[i], b= np[i+1];
    double phi= as_double (arg (-center (p[a]) / center (p[b]))) / (b-a);
    double twopi= 6.28318530717958647703;
    phi -= ref;
    while (phi < 0.0) phi += twopi;
    for (nat j=0; j<b-a; j++) {
      double psi= phi + (j * twopi) / (b-a);
      int index= (int) (((n*psi) / twopi) + 0.5);
      freq[(nat) (index%n)]++;
      if (j==10) break;
    }
  }
  return vector<nat> (freq, n);
}

template<typename C> nat
main_angle (const Polynomial& p, const Polygon& np, nat n, double ref) {
  const vector<nat> freq= angle_frequencies (p, np, n, ref);
  nat best=0;
  for (nat i=1; i<n; i++)
    if (freq[i] > freq[best]) best= i;
  return best;
}

TMPL C
slope (const Polynomial& p, const Polygon& np, nat i, nat j) {
  if (j-i == 1) return -p[i] / p[j];
  C s= nth_root (-p[i] / p[j], j-i);
  if (s == 0) return s;
  nat a= main_angle (p, np, j-i, as_double (arg (center (s))));
  return exp (2 * a * Pi (C) * Imaginary (C) / (j-i)) * s;
}

/******************************************************************************
* Zoom in on clusters
******************************************************************************/

TMPL C
focus_one (const Polynomial& p, const Polygon& np) {
  for (nat i=0; i<N(np)-1; i++)
    if ((np[i+1] << 1) >= deg(p)) {
      if ((np[i+1] << 1) == deg (p) && i < N(np)-2) {
	C s1= slope (p, np, np[i], np[i+1]);
	C s2= slope (p, np, np[i+1], np[i+2]);
	return s2 * nth_root (s1 / s2, 2);
	// NOTE: sqrt (s1 * s2) may yield a wrong angle
      }
      else return slope (p, np, np[i], np[i+1]);
    }
}

TMPL C
focus (const Polynomial& p) {
  nat i= N(p);
  nat ratio= (N(p) >> 2) + 2;
  Polygon np= newton_polygon (p);
  C last = focus_one (p, np);
  //cout << "  np   = " << np << "\n";
  //cout << "  next = " << last << "\n";
  C cumul= 0;
  while (i != 0) {
    cumul += sharpen (last);
    Polynomial q= shift (p, cumul);
    np= newton_polygon (q);
    C next= focus_one (q, np);
    //cout << "  np   = " << np << "\n";
    //cout << "  cumul= " << cumul << "\n";
    //cout << "  next = " << next << "\n";
    if (ratio * abs_U (next) >= abs_D (last)) break;
    last= next;
    i= i >> 1;
  }
  return cumul;
}

/******************************************************************************
* Root isolation on magnitude
******************************************************************************/

TMPL Polygon
newton_pieces (const Polynomial& p, const Polygon& np,
	       const vector<double>& mag, double threshold)
{
  // Join edges in Newton polygon whose slopes differ by
  // a factor smaller than 2^threshold
  Polygon a= vec<nat> (0);
  for (nat i=0; i<N(np)-2; i++) {
    double s1= (mag[np[i  ]] - mag[np[i+1]]) / (np[i+1] - np[i  ]);
    double s2= (mag[np[i+1]] - mag[np[i+2]]) / (np[i+2] - np[i+1]);
    if (threshold + s1 < s2) a << np[i+1];
  }
  a << deg (p);
  return a;
}

inline nat
largest_edge (const Polygon& np) {
  nat best= np[1]-np[0];
  for (nat i=1; i<N(np)-1; i++)
    best= max (best, np[i+1]-np[i]);
  return best;
}

TMPL bool
first_best (const Polynomial& p1, const Polynomial& p2,
	    const Polygon& ps1, const Polygon& ps2)
{
  nat e1= largest_edge (ps1);
  nat e2= largest_edge (ps2);
  if (e1 < e2) return true;
  if (e2 < e1) return false;
  if (N(ps1) > N(ps2)) return true;
  if (N(ps2) < N(ps1)) return false;
  return
    nth_root (abs_U (p1[ps1[0]] / p1[ps1[1]]), ps1[1]-ps1[0]) <
    nth_root (abs_U (p2[ps2[0]] / p2[ps2[1]]), ps2[1]-ps2[0]);
}

TMPL bool
is_precise (const C& z) {
  return
    (radius (z)<<4) < abs_D (z) &&
    exponent (center (z)) > -1000000;
}

TMPL Vector
piecewise_roots (const Polynomial& p, const Polygon& pieces) {
  Polygon ps= vec<nat> (0);
  for (nat i=1; i<N(pieces)-1; i++)
    if (is_precise (p[pieces[i]])) ps << pieces[i];
  ps << deg(p);

  Vector r;  
  for (nat i=0; i<N(ps)-1; i++)
    r << roots (range (p, ps[i], ps[i+1] + 1));
  r= adjust_radii (p, r);
  // NOTE: this is not really OK, since merely increasing the radii
  // can make several balls to collapse. We really should estimate
  // the error due to the terms from the other slopes.
  // We might actually want to do this a priori by putting the errors
  // in the coefficients p[ps[i]] and p[ps[i+1]].
  ASSERT (N(r) == deg(p), "internal error, bad root count");
  return newton_iterate (p, r);
}

/******************************************************************************
* Root isolation using Graeffe transforms
******************************************************************************/

TMPL Vector
nth_root (const Vector& v, nat p) {
  nat i, n= N(v);
  Vector r (C(0), n);
  for (i=0; i<n; i++)
    r[i]= nth_root (v[i], p);
  return r;
}

TMPL Vector
graeffe_lift (const Polynomial& p, const Vector& sqv) {
  // lift roots of the Graeffe transform of a square-free polynomial p
  // so as to yield the roots of p
  nat i, n= N(sqv);
  const Vector v  = nth_root (sqv, 2);
  const Vector ppv= eval (p, v);
  const Vector pmv= eval (p, -v);
  Vector r (C(0), n);
  for (i=0; i<n; i++)
    if ((is_zero (ppv[i]) && !is_zero (pmv[i])) || is_zero (v[i])) r[i]= v[i];
    else if (is_zero (pmv[i]) && !is_zero (ppv[i])) r[i]= -v[i];
    else if (is_zero (ppv[i]) && is_zero (pmv[i])) {
      // FIXME: this dirty hack is OK for the moment...
      nat j, count=0;
      for (j=0; j<i; j++)
	if (sqv[j] == sqv[i]) count++;
      r[i]= ((count&1 == 0)? v[i]: -v[i]);
      // ERROR ("internal error, multiple Graeffe lift not yet implemented");
    }
    else {
      //cout << "  ppv[" << i << "]= p(" <<  v[i] << ")= " << ppv[i] << "\n";
      //cout << "  pmv[" << i << "]= p(" << -v[i] << ")= " << pmv[i] << "\n";
      ERROR ("internal error, root expected");
    }
  //return newton_iteration (p, r);
  return newton_iterate (p, r);
}

/******************************************************************************
* Newton iteration
******************************************************************************/

inline complex_ball
best (const complex_ball& z1, const complex_ball& z2) {
  if (radius (z1) <= radius (z2)) return z1;
  else return z2;
}

TMPL Vector
newton_iteration (const Polynomial& p, const Vector& v) {
  // Input: a polynomial p and approximations of its roots v.
  // Output: better approximations for the roots using one step of
  // Newton's method. When no better approximation can be found for a root,
  // then we keep the original one.
  const Vector sv = sharpen (v);
  const Vector pv = eval (p, sv);
  const Vector dpv= eval (derive (p), v);
  nat i, n= N(v);
  Vector r (C(0), n);
  for (i=0; i<n; i++)
    if (is_zero (dpv[i])) r[i]= v[i];
    else r[i]= best (v[i], sv[i] - pv[i] / dpv[i]);
  return r;
}

TMPL Vector
newton_iterate (const Polynomial& p, const Vector& v) {
  // Call newton_iteration until approximations do not essentially improve
  Vector w1= v;
  while (true) {
    //cout << "    w1= " << w1 << "\n";
    Vector w2= newton_iteration (p, w1);
    bool stop= true;
    for (nat i=0; i<N(v); i++)
      if ((radius (w2[i]) << 1) < radius (w1[i])) {
	//cout << radius (w1[i]) << " -> " << radius (w2[i]) << "\n";
	stop= false;
	break;
      }
    w1= w2;
    if (stop) break;
  }
  return w1;
}

/******************************************************************************
* Adjust imprecisions
******************************************************************************/

TMPL Radius_type(C)
ensure_zero (const Polynomial& p) {
  // Quicky compute a radius r such that p has a root z with |z|<=r
  vector<double> mag= magnitudes (p);
  const Polygon ps= newton_pieces (p, newton_polygon (mag), mag, 2.01);
  nat j;
  for (j=1; j<N(ps); j++)
    if (is_precise (p[ps[j]])) break;
  ASSERT (j != N(ps), "internal error, fuzzy polynomial");
  C q= p[ps[j-1]] / p[ps[j]];
  return Radius_type (C) (2.5) * nth_root (abs_U (q), ps[j] - ps[j-1]);
  // NOTE: if the ratio of the slopes is E (here 2^2.01),
  // then the constant C by which we have to multiply the smallest slope
  // satisfies (1 - 2 E C)(C - 1) = (1 - E C). In our case C <= 2.5
}

TMPL Vector
adjust_radii (const Polynomial& p, const Vector& v) {
  // v contains approximate roots of p, but the radius of each v[i]
  // might be insufficient to garantee the existence of a root inside v[i].
  // This routines enlarges the radius of v[i] until the existence of
  // a root is guaranteed.
  Vector w (C(0), N(v));
  for (nat i=0; i<N(v); i++) {
    Radius_type(C) r= ensure_zero (shift (p, sharpen (v[i])));
    w[i]= C (center (v[i]), r);
  }
  return w;
  // FIXME: should be replaced by a routine which starts by looking
  // at the first derivatives only (using vectorial evaluation).
}

/******************************************************************************
* Main routine for computing the roots of a complex valued polynomial
******************************************************************************/

TMPL Vector
roots (const Polynomial& p) {
  // cout << "Roots: " << p << "\n";
  switch (N(p)) {
  case 0: return vec<C> ();
  case 1: return vec<C> ();
  case 2: return vec<C> (-p[0]/p[1]);
  case 3:
    {
      C d= nth_root (square (p[1]) - ((p[0]*p[2]) << 2), 2);
      return vec<C> ((-p[1] + d) / (p[2]<<1),
		     (-p[1] - d) / (p[2]<<1));
    }
  default:
    {
      if (!is_precise (p[0])) {
	nat v;
	for (v=1; v<N(p); v++)
	  if (is_precise (p[v])) break;
	ASSERT (v<N(p), "internal error, fuzzy polynomial");
	C z (Center_type(C) (0), ensure_zero (p));
	return append (Vector (z, v), roots (range (p, v, N(p))));
      }
      else {
	C s= focus (p);
	nat i= N(p), iters= 0;
	while (i != 0) { iters++; i= i>>1; }
	Polynomial vp1[iters];
	Polynomial vp2[iters];
	vp1[0]= p;
	vp2[0]= shift (p, s);
	for (i=1; i<iters; i++) {
	  vp1[i]= square_roots (vp1[i-1]);
	  vp2[i]= square_roots (vp2[i-1]);
	}
	Polynomial p1= vp1[iters-1];
	Polynomial p2= vp2[iters-1];
	vector<double> mag1= magnitudes (p1);
	vector<double> mag2= magnitudes (p2);
	Polygon np1= newton_pieces (p1, newton_polygon (mag1), mag1, 5.01);
	Polygon np2= newton_pieces (p2, newton_polygon (mag2), mag2, 5.01);
	//cout << "  np1= " << np1 << "\n";
	//cout << "  np2= " << np2 << "\n";
	if (first_best (p1, p2, np1, np2)) {
	  Vector v= piecewise_roots (p1, np1);
	  for (i=iters-1; i!=0; i--)
	    v= graeffe_lift (vp1[i-1], v);
	  return v;
	}
	else {
	  Vector v= piecewise_roots (p2, np2);
	  for (i=iters-1; i!=0; i--)
	    v= graeffe_lift (vp2[i-1], v);
	  return v + s;
	}
      }
    }
  }
}

#undef TMPL
#undef Real
#undef Radius
#undef Vector
#undef Polynomial
#undef Polygon
} // namespace mmx
#endif // __MMX_POLYNOMIAL_ROOTS_HPP
