
/******************************************************************************
* MODULE     : basix.hpp
* DESCRIPTION: Basic inclusions, definitions and routines
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_BASIX_HPP
#define __MMX_BASIX_HPP
#include "basix-config.hpp"
#include "fast_new.hpp"
#include <iostream>
#include <cassert>

#include <math.h>
#if __GLIBC__ >= 2 && defined(__THROW)
extern "C" double hypot (double, double) __THROW __attribute__ ((const));
#else
extern "C" double hypot (double, double) __attribute__ ((const));
#endif

using std::istream;
using std::ostream;
using std::cin;
using std::cout;
using std::cerr;

namespace mmx {

#define LESSGTR <>
#define STMPL template<>

enum print_format {
  blank, stroke, indent, unindent, lf, hrule
};

ostream& operator << (ostream& out, print_format x);

/******************************************************************************
* Exceptions
******************************************************************************/

#ifdef BASIX_ENABLE_EXCEPTIONS
class generic;
class exception;
class string;
exception as_exception (const generic& x);
exception error_message (const char* msg);
exception error_message (const string& msg);
#define ERROR(msg) throw mmx::error_message (msg);
#define ASSERT(expr,msg) if (!(expr)) ERROR (msg);
#else // disable exceptions
#define ERROR(msg) assert(false)
#define ASSERT(expr,msg) assert(expr)
#endif

#ifdef BASIX_ENABLE_VERIFY
#define VERIFY(expr,msg) ASSERT(expr,msg)
#else // disable verifications
#define VERIFY(expr,msg)
#endif

volatile void fatal_error (char* message, char* routine, char* file);

/******************************************************************************
* Initialization
******************************************************************************/

typedef void (*void_function) ();
void mmx_initialize ();
void mmx_terminate ();

struct global_initializer {
MMX_ALLOCATORS
  global_initializer (void_function initializer);
};

struct global_terminator {
MMX_ALLOCATORS
  global_terminator (void_function terminator);
};

#define INITIALIZER(init) static global_initializer glob_##init (init);
#define TERMINATOR(term) static global_terminator glob_##term (term);

/******************************************************************************
* Indirect data structures prototypes
******************************************************************************/
  
#define INDIRECT_PROTO(T,R)			\
MMX_ALLOCATORS					\
protected:					\
  R* rep;					\
public:						\
  inline T (R* rep2);				\
  inline T (const R* rep2, bool with_inc);	\
  inline T (const T& x);			\
  inline ~T ();					\
  inline T& operator= (const T& x);		\
  inline const R* operator-> () const;		\
  inline void secure ();

#define INDIRECT_PROTO_1(T,R,X)			\
MMX_ALLOCATORS					\
protected:					\
  R<X >* rep;					\
public:						\
  inline T (R<X >* rep2);			\
  inline T (const R<X >* rep2, bool with_inc);	\
  inline T (const T& x);			\
  inline ~T ();					\
  inline T& operator= (const T& x);		\
  inline const R<X >* operator-> () const;	\
  inline R<X >* operator-> ();			\
  inline void secure ();

#define INDIRECT_PROTO_2(T,R,X,Y)		\
MMX_ALLOCATORS					\
protected:					\
  R<X,Y >* rep;					\
public:						\
  inline T (R<X,Y >* rep2);			\
  inline T (const R<X,Y >* rep2, bool inc);	\
  inline T (const T& x);			\
  inline ~T ();					\
  inline T& operator= (const T& x);		\
  inline const R<X,Y >* operator-> () const;	\
  inline void secure ();

#define INDIRECT_PROTO_3(T,R,X,Y,Z)		\
MMX_ALLOCATORS					\
protected:					\
  R<X,Y,Z >* rep;				\
public:						\
  inline T (R<X,Y,Z >* rep2);			\
  inline T (const R<X,Y,Z >* rep2, bool inc);	\
  inline T (const T& x);			\
  inline ~T ();					\
  inline T& operator= (const T& x);		\
  inline const R<X,Y,Z >* operator-> () const;	\
  inline void secure ();

/******************************************************************************
* Indirect data structures with reference counting
******************************************************************************/

struct rep_struct {
MMX_ALLOCATORS
  int ref_count;
  inline rep_struct (): ref_count (1) {}
  virtual inline ~rep_struct () {}
};

#define REP_STRUCT :public rep_struct
#define INC_COUNT(x) { (x)->ref_count++; }
#define DEC_COUNT(x) { if (0==--((x)->ref_count)) delete (x); }
#define INC_NULL_COUNT(x) { if ((x)!=NULL) (x)->ref_count++; }
#define DEC_NULL_COUNT(x) { \
  if ((x)!=NULL && 0==--((x)->ref_count)) delete (x); }

template<typename C> C copy (const C& x) {
  assert (false); // object should really be copied before access
  return x; }
template<typename C> inline C melt (const C& x) {
  return x; }
template<typename C> inline C fuse (const C& x) {
  return x; }
template<typename C> inline nat hard_hash (const C& x) {
  return soft_hash (x); }
template<typename C> inline nat hard_hash_impl (const C* p) {
  return ((nat) ((long unsigned int) ((void*) p))) >> 3; }
template<typename C> inline bool hard_eq (const C& x, const C& y) {
  return eq (x, y); }
template<typename C> inline bool hard_neq (const C& x, const C& y) {
  return neq (x, y); }

#define INDIRECT_IMPL(T,R)						\
  inline T::T (R* rep2): rep(rep2) {}					\
  inline T::T (const R* rep2, bool with_inc):				\
    rep((R*) (void*) rep2) { (void) with_inc; INC_COUNT (rep); }	\
  inline T::T (const T& x): rep(x.rep) { INC_COUNT (rep); }		\
  inline T::~T () { DEC_COUNT (rep); }					\
  inline const R* T::operator -> () const { return rep; }		\
  inline T& T::operator = (const T& x) {				\
    INC_COUNT (x.rep); DEC_COUNT (rep);					\
    rep=x.rep; return *this; }						\
  inline void T::secure () {						\
    if (rep->ref_count>1) *this= copy (*this); }			\
  inline R* inside (const T& x) {					\
    return const_cast<R*> (x.operator -> ()); }				\
  STMPL inline nat hard_hash (const T& x) {				\
    return hard_hash_impl (x.operator -> ()); }				\
  STMPL inline bool hard_eq (const T& x, const T& y) {			\
    return (x.operator -> ()) == (y.operator -> ()); }			\
  STMPL inline bool hard_neq (const T& x, const T& y) {			\
    return (x.operator -> ()) != (y.operator -> ()); }

#define INDIRECT_NULL_IMPL(T,R)						\
  inline T::T (R* rep2): rep(rep2) {}					\
  inline T::T (const R* rep2, bool with_inc):				\
    rep((R*) (void*) rep2) { (void) with_inc; INC_NULL_COUNT (rep); }	\
  inline T::T (const T& x): rep(x.rep) { INC_NULL_COUNT (rep); }	\
  inline T::~T () { DEC_NULL_COUNT (rep); }				\
  inline const R* T::operator -> () const { return rep; }		\
  inline T& T::operator = (const T& x) {				\
    INC_NULL_COUNT (x.rep); DEC_NULL_COUNT (rep);			\
    rep=x.rep; return *this; }						\
  inline void T::secure () {						\
    if (rep->ref_count>1) *this= copy (*this); }			\
  inline R* inside (const T& x) {					\
    return const_cast<R*> (x.operator -> ()); }				\
  STMPL inline nat hard_hash (const T& x) {				\
    return hard_hash_impl (x.operator -> ()); }				\
  STMPL inline bool hard_eq (const T& x, const T& y) {			\
    return (x.operator -> ()) == (y.operator -> ()); }			\
  STMPL inline bool hard_neq (const T& x, const T& y) {			\
    return (x.operator -> ()) != (y.operator -> ()); }

#define INDIRECT_IMPL_1(T,R,XX,X)					\
  template<XX> inline T<X >::T (R<X >* rep2): rep(rep2) {}		\
  template<XX> inline T<X >::T (const R<X >* rep2, bool with_inc):	\
    rep((R<X >*) (void*) rep2) { (void) with_inc; INC_COUNT (rep); }	\
  template<XX> inline T<X >::T (const T<X >& x): rep(x.rep) {		\
    INC_COUNT (rep); }							\
  template<XX> inline T<X >::~T () { DEC_COUNT (rep); }			\
  template<XX> inline const R<X >* T<X >::operator -> () const {	\
    return rep; }							\
  template<XX> inline R<X >* T<X >::operator -> () {			\
    return rep; }							\
  template<XX> inline T<X >&						\
  T<X >::operator = (const T<X >& x) {					\
    INC_COUNT (x.rep); DEC_COUNT (rep);					\
    rep=x.rep; return *this; }						\
  template<XX> inline void T<X >::secure () {				\
    if (rep->ref_count>1) *this= copy (*this); }			\
  template<XX> R<X >* inside (const T<X >& x) {				\
    return const_cast<R<X >*> (x.operator -> ()); }			\
  template<XX> inline nat hard_hash (const T<X >& x) {			\
    return hard_hash_impl (x.operator -> ()); }				\
  template<XX> inline bool hard_eq (const T<X >& x, const T<X >& y) {	\
    return (x.operator -> ()) == (y.operator -> ()); }			\
  template<XX> inline bool hard_neq (const T<X >& x, const T<X >& y) {	\
    return (x.operator -> ()) != (y.operator -> ()); }

#define INDIRECT_NULL_IMPL_1(T,R,XX,X)					  \
  template<XX> inline T<X >::T (R<X >* rep2): rep(rep2) {}		  \
  template<XX> inline T<X >::T (const R<X >* rep2, bool with_inc):	  \
    rep((R<X >*) (void*) rep2) { (void) with_inc; INC_NULL_COUNT (rep); } \
  template<XX> inline T<X >::T (const T<X >& x): rep(x.rep) {		  \
    INC_NULL_COUNT (rep); }						  \
  template<XX> inline T<X >::~T () { DEC_NULL_COUNT (rep); }		  \
  template<XX> inline const R<X >* T<X >::operator -> () const {	  \
    return rep; }							  \
  template<XX> inline R<X >* T<X >::operator -> () {			  \
    return rep; }							  \
  template<XX> inline T<X >&						  \
  T<X >::operator = (const T<X >& x) {					  \
    INC_NULL_COUNT (x.rep); DEC_NULL_COUNT (rep);			  \
    rep=x.rep; return *this; }						  \
  template<XX> inline void T<X >::secure () {				  \
    if (rep->ref_count>1) *this= copy (*this); }			  \
  template<XX> R<X >* inside (const T<X >& x) {				  \
    return const_cast<R<X >*> (x.operator -> ()); }			  \
  template<XX> inline nat hard_hash (const T<X >& x) {			  \
    return hard_hash_impl (x.operator -> ()); }				  \
  template<XX> inline bool hard_eq (const T<X >& x, const T<X >& y) {	  \
    return (x.operator -> ()) == (y.operator -> ()); }			  \
  template<XX> inline bool hard_neq (const T<X >& x, const T<X >& y) {	  \
    return (x.operator -> ()) != (y.operator -> ()); }

#define INDIRECT_IMPL_2(T,R,XX,X,YY,Y)					      \
  template<XX,YY> inline T<X,Y >::T (R<X,Y >* rep2): rep(rep2) {}	      \
  template<XX,YY> inline T<X,Y >::T (const R<X,Y >* rep2, bool inc):	      \
    rep((R<X,Y >*) (void*) rep2) { (void) inc; INC_COUNT (rep); }	      \
  template<XX,YY> inline T<X,Y >::T (const T<X,Y >& x): rep(x.rep) {	      \
    INC_COUNT (rep); }							      \
  template<XX,YY> inline T<X,Y >::~T () { DEC_COUNT (rep); }		      \
  template<XX,YY> inline const R<X,Y >* T<X,Y >::operator -> () const {	      \
    return rep; }							      \
  template<XX,YY> inline T<X,Y >&					      \
  T<X,Y >::operator = (const T<X,Y >& x) {				      \
    INC_COUNT (x.rep); DEC_COUNT (rep);					      \
    rep=x.rep; return *this; }						      \
  template<XX,YY> inline void T<X,Y >::secure () {			      \
    if (rep->ref_count>1) *this= copy (*this); }			      \
  template<XX,YY> R<X,Y >* inside (const T<X,Y >& x) {			      \
    return const_cast<R<X,Y >*> (x.operator -> ()); }			      \
  template<XX,YY> inline nat hard_hash (const T<X,Y >& x) {		      \
    return hard_hash_impl (x.operator -> ()); }				      \
  template<XX,YY> inline bool hard_eq (const T<X,Y >& x, const T<X,Y >& y) {  \
    return (x.operator -> ()) == (y.operator -> ()); }			      \
  template<XX,YY> inline bool hard_neq (const T<X,Y >& x, const T<X,Y >& y) { \
    return (x.operator -> ()) != (y.operator -> ()); }

#define INDIRECT_IMPL_3(T,R,XX,X,YY,Y,ZZ,Z)				\
  template<XX,YY,ZZ> inline T<X,Y,Z >::T (R<X,Y,Z >* rep2):		\
    rep(rep2) {}							\
  template<XX,YY,ZZ> inline						\
  T<X,Y,Z >::T (const R<X,Y,Z >* rep2, bool inc):			\
    rep((R<X,Y,Z >*) (void*) rep2) { (void) inc; INC_COUNT (rep); }	\
  template<XX,YY,ZZ> inline T<X,Y,Z >::T (const T<X,Y,Z >& x):		\
    rep(x.rep) { INC_COUNT (rep); }					\
  template<XX,YY,ZZ> inline T<X,Y,Z >::~T () { DEC_COUNT (rep); }	\
  template<XX,YY,ZZ> inline const R<X,Y,Z >*				\
  T<X,Y,Z >::operator -> () const {					\
    return rep; }							\
  template<XX,YY,ZZ> inline T<X,Y,Z >&					\
  T<X,Y,Z >::operator = (const T<X,Y,Z >& x) {				\
    INC_COUNT (x.rep); DEC_COUNT (rep);					\
    rep=x.rep; return *this; }						\
  template<XX,YY,ZZ> inline void T<X,Y,Z >::secure () {			\
    if (rep->ref_count>1) *this= copy (*this); }			\
  template<XX,YY,ZZ> R<X,Y,Z >* inside (const T<X,Y,Z >& x) {		\
    return const_cast<R<X,Y,Z >*> (x.operator -> ()); }			\
  template<XX,YY,ZZ> inline nat hard_hash (const T<X,Y,Z>& x) {         \
    return hard_hash_impl (x.operator -> ()); }				\
  template<XX,YY,ZZ> inline bool					\
  hard_eq (const T<X,Y,Z >& x, const T<X,Y,Z >& y) {			\
    return (x.operator -> ()) == (y.operator -> ()); }			\
  template<XX,YY,ZZ> inline bool					\
  hard_neq (const T<X,Y,Z >& x, const T<X,Y,Z >& y) {			\
    return (x.operator -> ()) != (y.operator -> ()); }

/******************************************************************************
* Fix to C++ stupidity which prohibits templating over
* parameters which are constant in the scope where the template is used
******************************************************************************/

template<nat n>
class nat_value {
public:
  static const nat val= n;
  //static inline nat val () { return n; }
};

/******************************************************************************
* Forward definitions of some classes
******************************************************************************/

class string;
template<typename V> class floating;
template<typename C> class array;

} // namespace mmx
#include "generic.hpp"
#include "lisp_syntax.hpp"
#include "defaults.hpp"
#include "exception.hpp"
#endif // __MMX_BASIX_HPP
