
/******************************************************************************
* MODULE     : category_glue.hpp
* DESCRIPTION: Routines for defining glue
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_CATEGORY_GLUE_HPP
#define __MMX_CATEGORY_GLUE_HPP
#include "glue.hpp"
namespace mmx {

/******************************************************************************
* Constant functions
******************************************************************************/

template<typename Cond, typename T, typename F, T c>
struct define_constant_function_helper {
  static inline T fun (const F& x) { (void) x; return c; }
  static void define_constant_function (const generic& name) {
    define<Cond,T,F> (name, (T (*) (const F&)) fun); }
};

template<typename Cond, typename T, typename F, T c> inline void
define_constant_function (const generic& name) {
  define_constant_function_helper<Cond,T,F,c>::define_constant_function (name);
}

template<typename T, typename F, T c> inline void
define_constant_function (const generic& name) {
  define_constant_function<always,T,F,c> (name);
}

/******************************************************************************
* Standard operations
******************************************************************************/

template<bool Cond, typename T, typename Op>
struct define_op_helper {
  static inline T unary (const T& x) { return Op::op (x); }
  static inline T binary (const T& x, const T& y) { return Op::op (x, y); }
  static inline bool test (const T& x, const T& y) { return Op::op (x, y); }
  static inline void define_unary () {
    define<always,T,T> (Op::name (), (T(*) (const T&)) unary); }
  static inline void define_binary () {
    define<always,T,T,T> (Op::name (), (T(*) (const T&, const T&)) binary); }
  static inline void define_test () {
    define<always,bool,T,T> (Op::name (), (bool(*)(const T&,const T&)) test); }
};

template<typename T, typename Op>
struct define_op_helper<false,T,Op> {
  static inline void define_unary () {}
  static inline void define_binary () {}
  static inline void define_test () {}
};

template<typename Cond, typename T, typename Op> inline void
define_unary () {
  typedef typename Cond::Cond When;
  define_op_helper<When::value,T,Op>::define_unary ();
}

template<typename Cond, typename T, typename Op> inline void
define_binary () {
  typedef typename Cond::Cond When;
  define_op_helper<When::value,T,Op>::define_binary ();
}

template<typename Cond, typename T, typename Op> inline void
define_test () {
  typedef typename Cond::Cond When;
  define_op_helper<When::value,T,Op>::define_test ();
}

template<bool Cond, typename T, typename X, typename Op>
struct define_extrary_op_helper {
  static inline T left  (const X& y, const T& x) { return Op::op (x, y); }
  static inline T right (const T& x, const X& y) { return Op::op (x, y); }
  static inline void define_left () {
    define<always,T,X,T> (Op::name (), (T(*) (const X&, const T&)) left); }
  static inline void define_right () {
    define<always,T,T,X> (Op::name (), (T(*) (const T&, const X&)) right); }
};

template<typename Cond, typename T, typename X, typename Op> inline void
define_left () {
  typedef typename Cond::Cond When;
  define_extrary_op_helper<When::value,T,X,Op>::define_left ();
}

template<typename Cond, typename T, typename X, typename Op> inline void
define_right () {
  typedef typename Cond::Cond When;
  define_extrary_op_helper<When::value,T,X,Op>::define_right ();
}

/******************************************************************************
* Rings and fields
******************************************************************************/

template<typename Cond, typename T> void
define_ring () {
  define_unary<Cond,T,neg_op> ();
  define_binary<Cond,T,add_op> ();
  define_binary<Cond,T,sub_op> ();
  define_binary<Cond,T,mul_op> ();
}

template<typename Cond, typename T> void
define_euclidean_ring () {
  define_ring<Cond,T> ();
  define_binary<Cond,T,quo_op> ();
  define_binary<Cond,T,rem_op> ();
}

template<typename Cond, typename T> void
define_field () {
  define_euclidean_ring<Cond,T> ();
  define_binary<Cond,T,div_op> ();
}

/******************************************************************************
* Gcd and lcm
******************************************************************************/

template<typename Cond, typename T> void
define_factorial () {
  define_binary<Cond,T,gcd_op> ();
  define_binary<Cond,T,lcm_op> ();
}

/******************************************************************************
* Elementary functions
******************************************************************************/

template<typename Cond, typename T> void
define_elementary () {
  define_unary<Cond,T,sqrt_op> ();
  define_binary<Cond,T,pow_op> ();
  define_unary<Cond,T,exp_op> ();
  define_unary<Cond,T,log_op> ();
  define_unary<Cond,T,cos_op> ();
  define_unary<Cond,T,sin_op> ();
  define_unary<Cond,T,tan_op> ();
  define_unary<Cond,T,acos_op> ();
  define_unary<Cond,T,asin_op> ();
  define_unary<Cond,T,atan_op> ();
}

/******************************************************************************
* Ordered structures
******************************************************************************/

template<typename Cond, typename T> void
define_ordering () {
  define_test<Cond,T,less_op> ();
  define_test<Cond,T,lesseq_op> ();
  define_test<Cond,T,gtr_op> ();
  define_test<Cond,T,gtreq_op> ();
}

} // namespace mmx
#endif // __MMX_CATEGORY_GLUE_HPP
