
/******************************************************************************
* MODULE     : category_lift_glue.hpp
* DESCRIPTION: Routines for lifting operations of a scalar type C to
*              the glue of a compound type like vector<C> or table<C>.
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_CATEGORY_LIFT_GLUE_HPP
#define __MMX_CATEGORY_LIFT_GLUE_HPP
#include "category_glue.hpp"
namespace mmx {

/******************************************************************************
* Rings and fields
******************************************************************************/

template<typename Cond, typename T, typename C> void
lift_ring_ops () {
  define_unary<both<has<C, neg_op>, Cond>, T, neg_op> ();
  define_binary<both<has<C, add_op>, Cond>, T, add_op> ();
  define_binary<both<has<C, sub_op>, Cond>, T, sub_op> ();
  define_binary<both<has<C, mul_op>, Cond>, T, mul_op> ();
}

template<typename Cond, typename T, typename C> void
lift_euclidean_ring_ops () {
  lift_ring_ops<Cond,T,C> ();
  define_binary<both<has<C, quo_op>, Cond>, T, quo_op> ();
  define_binary<both<has<C, rem_op>, Cond>, T, rem_op> ();
}

template<typename Cond, typename T, typename C> void
lift_field_ops () {
  lift_euclidean_ring_ops<Cond,T,C> ();
  define_binary<both<has<C, div_op>, Cond>, T, div_op> ();
}

/******************************************************************************
* Lift ring and field operation between type and scalars
******************************************************************************/

template<typename Cond, typename T, typename C> void
lift_scalar_ring_ops () {
  define_left<both<has<C, add_op>, Cond>, T, C, ladd_op> ();
  define_left<both<has<C, sub_op>, Cond>, T, C, lsub_op> ();
  define_left<both<has<C, mul_op>, Cond>, T, C, lmul_op> ();
  define_right<both<has<C, add_op>, Cond>, T, C, radd_op> ();
  define_right<both<has<C, sub_op>, Cond>, T, C, rsub_op> ();
  define_right<both<has<C, mul_op>, Cond>, T, C, rmul_op> ();
}

template<typename Cond, typename T, typename C> void
lift_scalar_euclidean_ring_ops () {
  lift_scalar_ring_ops<Cond,T,C> ();
  define_right<both<has<C, quo_op>, Cond>, T, C, rquo_op> ();
  define_right<both<has<C, rem_op>, Cond>, T, C, rrem_op> ();
}

template<typename Cond, typename T, typename C> void
lift_scalar_field_ops () {
  lift_scalar_euclidean_ring_ops<Cond,T,C> ();
  define_right<both<has<C, div_op>, Cond>, T, C, rdiv_op> ();
}

template<typename Cond, typename T, typename C> void
lift_scalar_ops () {
  lift_scalar_field_ops<Cond,T,C> ();
  define_left<both<has<C, div_op>, Cond>, T, C, ldiv_op> ();
}

/******************************************************************************
* Gcd and lcm
******************************************************************************/

template<typename Cond, typename T, typename C> void
lift_factorial_ops () {
  define_binary<both<has<C, gcd_op>, Cond>, T, gcd_op> ();
  define_binary<both<has<C, lcm_op>, Cond>, T, lcm_op> ();
}

/******************************************************************************
* Elementary functions
******************************************************************************/

template<typename Cond, typename T, typename C> void
lift_elementary_ops () {
  define_unary<both<has<C, sqrt_op>, Cond>, T, sqrt_op> ();
  define_binary<both<has<C, pow_op>, Cond>, T, pow_op> ();
  define_unary<both<has<C, exp_op>, Cond>, T, exp_op> ();
  define_unary<both<has<C, log_op>, Cond>, T, log_op> ();
  define_unary<both<has<C, cos_op>, Cond>, T, cos_op> ();
  define_unary<both<has<C, sin_op>, Cond>, T, sin_op> ();
  define_unary<both<has<C, tan_op>, Cond>, T, tan_op> ();
  define_unary<both<has<C, acos_op>, Cond>, T, acos_op> ();
  define_unary<both<has<C, asin_op>, Cond>, T, asin_op> ();
  define_unary<both<has<C, atan_op>, Cond>, T, atan_op> ();
}

/******************************************************************************
* Ordered structures
******************************************************************************/

template<typename Cond, typename T, typename C> void
lift_ordering_ops () {
  define_test<both<has<C, less_op>, Cond>, T, less_op> ();
  define_test<both<has<C, lesseq_op>, Cond>, T, lesseq_op> ();
  define_test<both<has<C, gtr_op>, Cond>, T, gtr_op> ();
  define_test<both<has<C, gtreq_op>, Cond>, T, gtreq_op> ();
}

} // namespace mmx
#endif // __MMX_CATEGORY_LIFT_GLUE_HPP
