
/******************************************************************************
* MODULE     : array.hpp
* DESCRIPTION: Arrays (a simplified form of dense vectors)
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_ARRAY_HPP
#define __MMX_ARRAY_HPP
#include "iterator.hpp"
namespace mmx {
#define TMPL template<typename C>
#define Array array<C>
#define Array_rep array_rep<C>
TMPL class array_rep;
TMPL class array;
TMPL inline nat N (const Array& v);
TMPL inline C* seg (Array& v);
TMPL inline const C* seg (const Array& v);

/******************************************************************************
* Dense arrays of arbitrary sizes
******************************************************************************/

TMPL
class array_rep REP_STRUCT {
  C*   a;           // entries of array
  nat  n;           // dimension of array
  nat  l;           // allocated number of entries
public:
  inline array_rep (C* a2, nat n2):
    a (a2), n (n2), l (n2) {}
  inline ~array_rep () { mmx_delete<C> (a, l); }
  void resize (nat n2) {
    if (n2 > l) {
      nat l2= max (n2, l<<1);
      C* b= mmx_new<C> (l2);
      for (nat i=0; i<n; i++) b[i]= a[i];
      mmx_delete<C> (a, l);
      a= b;
      l= l2;
    }
    n= n2;
  }
  friend class Array;
  friend nat N LESSGTR (const Array& v);
  friend C* seg LESSGTR (Array& v);
  friend const C* seg LESSGTR (const Array& v);
};

TMPL
class array {
INDIRECT_PROTO_1 (array, array_rep, C)
public:
  inline array (): rep (new Array_rep (mmx_new<C> (0), 0)) {}
  inline array (C* a, nat n): rep (new Array_rep (a, n)) {}
  inline array (const C& c1):
    rep (new Array_rep (mmx_new<C> (1), 1)) {
      rep->a[0]= c1; }
  inline array (const C& c1, const C& c2):
    rep (new Array_rep (mmx_new<C> (2), 2)) {
      rep->a[0]= c1; rep->a[1]= c2; }
  inline array (const C& c1, const C& c2, const C& c3):
    rep (new Array_rep (mmx_new<C> (3), 3)) {
      rep->a[0]= c1; rep->a[1]= c2; rep->a[2]= c3; }
  inline array (const C& c1, const C& c2, const C& c3, const C& c4):
    rep (new Array_rep (mmx_new<C> (4), 4)) {
      rep->a[0]= c1; rep->a[1]= c2; rep->a[2]= c3; rep->a[3]= c4; }
  inline array (const C& c1, const C& c2, const C& c3,
		const C& c4, const C& c5):
    rep (new Array_rep (mmx_new<C> (5), 5)) {
      rep->a[0]= c1; rep->a[1]= c2; rep->a[2]= c3;
      rep->a[3]= c4; rep->a[4]= c5; }
  inline array (const C& c1, const C& c2, const C& c3,
		const C& c4, const C& c5, const C& c6):
    rep (new Array_rep (mmx_new<C> (6), 6)) {
      rep->a[0]= c1; rep->a[1]= c2; rep->a[2]= c3;
      rep->a[3]= c4; rep->a[4]= c5; rep->a[5]= c6; }

  array (const iterator<C>& it) {
    nat i, l= 1;
    C* a= mmx_new<C> (l);
    for (i=0; busy (it); i++, ++it) {
      if (i >= l) {
	C* b= mmx_new<C> (l<<1);
	for (nat i=0; i<l; i++) b[i]= a[i];
	mmx_delete<C> (a, l);
	a= b;
	l= l<<1;
      }
      a[i]= *it;
    }
    rep= new Array_rep (a, i);
    rep->l= l;
  }

  inline const C& operator [] (nat i) const { return rep->a[i]; }
  inline C& operator [] (nat i) { secure(); return rep->a[i]; }

  Array& operator << (const C& append);
  Array& operator << (const Array& w);
};
INDIRECT_IMPL_1 (array, array_rep, typename C, C)

TMPL inline nat N (const Array& v) { return v->n; }
TMPL inline const C& read (const Array& v, nat i) { return v[i]; }
TMPL inline const C* seg (const Array& v) { return v->a; }
TMPL inline C* seg (Array& v) { v.secure(); return v->a; }

STYPE_TO_TYPE(TMPL,scalar_type,Array,C);

/******************************************************************************
* Array iteration
******************************************************************************/

TMPL
class array_iterator_rep: public iterator_rep<C> {
  const Array v;  // the array to traverse
  nat i;          // current position
public:
  array_iterator_rep (const Array& v2, nat i2= 0): v (v2), i (0) {}
protected:
  bool is_busy () { return i < N(v); }
  void advance () { i++; }
  C current () { return v[i]; }
  iterator_rep<C>* clone () { return new array_iterator_rep (v, i); }
};

TMPL iterator<C>
iterate (const Array& v) {
  return iterator<C> (new array_iterator_rep<C> (v));
}

/******************************************************************************
* Basic operations
******************************************************************************/

template<typename Op, typename C> nat
unary_hash (const Array& v) {
  register nat i, h= 12321, n= N(v);
  for (i=0; i<n; i++)
    h= (h<<3) ^ (h>>29) ^ Op::op (v[i]);
  return h;
}

template<typename Op, typename C> inline Array
unary_map (const Array& v) {
  nat n= N(v);
  C* r= mmx_new<C> (n);
  const C* s= seg (v);
  for (nat i=0; i<n; i++) r[i]= Op::op (s[i]);
  return Array (r, n);
}

template<typename Op, typename C> inline bool
binary_test (const Array& v, const Array& w) {
  nat n= N(v);
  if (N(w) != n) return false;
  const C* a= seg (v);
  const C* b= seg (w);
  for (nat i=0; i<n; i++)
    if (Op::not_op (a[i], b[i])) return false;
  return true;
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Array)
SOFT_IDENTITY_OP_SUGAR(TMPL,Array)
COPY_OP_SUGAR(TMPL,Array)

TMPL generic
flatten (const Array& v) {
  nat i, n= N(v);
  generic args= comma ();
  for (i=0; i<n; i++)
    args= comma (args, flatten (v[i]));
  return xsqtuple (args);
}

PRINT_SUGAR(TMPL,Array);

/******************************************************************************
* Extracting ranges of elements and appending new elements
******************************************************************************/

TMPL inline Array
array_fill (nat n) {
  return Array (mmx_new<C> (n), n);
}

TMPL inline Array
array_fill (const C& c, nat n) {
  C* a= mmx_new<C> (n);
  for (nat i=0; i<n; i++) a[i]= c;
  return Array (a, n);
}

TMPL Array
range (const Array& v, nat start, nat end) {
  nat n= end- start;
  C* a= mmx_new<C> (n);
  const C* b= seg (v) + start;
  for (nat i=0; i<n; i++) a[i]= b[i];
  return Array (a, n);
}

TMPL inline C car (const Array& v) { return v[0]; }
TMPL inline Array cdr (const Array& v) { return range (v, 1, N(v)); }

TMPL Array
cons (const C& c, const Array& v) {
  nat n= N(v);
  C* a= mmx_new<C> (n+1);
  const C* b= seg (v);
  a[0]= c;
  for (nat i=0; i<n; i++) a[i+1]= b[i];
  return Array (a, n+1);
}

TMPL Array
append (const Array& v, const Array& w) {
  nat n= N(v), p= N(w);
  C* a= mmx_new<C> (n+p);
  const C* b= seg (v);
  const C* c= seg (w);
  for (nat i=0; i<n; i++) a[i]= b[i];
  for (nat i=0; i<p; i++) a[i+n]= c[i];
  return Array (a, n+p);
}

TMPL inline Array
operator * (const Array& a, const Array& b) {
  return append (a, b);
}

TMPL Array&
Array::operator << (const C& append) {
  secure ();
  nat n= rep->n;
  rep->resize (n+1);
  rep->a[n]= append;
  return *this;
}

TMPL Array&
Array::operator << (const Array& w) {
  secure ();
  nat n= rep->n, p= N(w);
  rep->resize (n + p);
  C* a= rep->a + n;
  const C* b= seg (w);
  for (nat i=0; i<p; i++) a[i]= b[i];
  return *this;
}

#undef TMPL
#undef Array
#undef Array_rep
} // namespace mmx
#endif // __MMX_ARRAY_HPP
