
/******************************************************************************
* MODULE     : chain.hpp
* DESCRIPTION: Chains (balanced trees considered as lists)
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_CHAIN_HPP
#define __MMX_CHAIN_HPP
#include "iterator.hpp"
namespace mmx {
#define TMPL_DEF template<typename C>
#define TMPL template<typename C>
#define Chain chain<C>
#define Chain_rep chain_rep<C>
TMPL class chain_rep;
TMPL class chain;
TMPL inline Chain left (const Chain& c);
TMPL inline C middle (const Chain& c);
TMPL inline Chain right (const Chain& c);
TMPL inline bool nil (const Chain& c);
TMPL inline nat N (const Chain& c);
TMPL Chain balance_left (const Chain& c);
TMPL Chain balance_right (const Chain& c);
TMPL Chain operator * (const Chain& l, const Chain& r);

/******************************************************************************
* Chain class and its representation class
******************************************************************************/

TMPL_DEF
class chain_rep REP_STRUCT {
  nat   size;
  Chain l;
  C     m;
  Chain r;
  // Invariant: #l < 2 (#r + 1) and #r < 2 (#l + 1)

public:
  inline Chain_rep (const Chain& l2, const C& m2, const Chain& r2);
  friend class Chain;
  friend Chain left LESSGTR (const Chain& c);
  friend C middle LESSGTR (const Chain& c);
  friend Chain right LESSGTR (const Chain& c);
  friend nat N LESSGTR (const Chain& c);
};

TMPL_DEF
class chain {
INDIRECT_PROTO_1 (chain, chain_rep, C)
public:
  inline chain (): rep (NULL) {}
  inline chain (const C& c1):
    rep (new Chain_rep (Chain (), c1, Chain ())) {}
  inline chain (const C& c1, const C& c2):
    rep (new Chain_rep (Chain (c1), c2, Chain ())) {}
  inline chain (const C& c1, const C& c2, const C& c3):
    rep (new Chain_rep (Chain (c1), c2, Chain (c3))) {}
  inline chain (const Chain& l, const C& m, const Chain& r):
    rep (new Chain_rep (l, m, r)) {}
  chain (const iterator<C>& it) {
    rep= NULL;
    for (; busy (it); ++it) *this= (*this) * Chain (*it); }
  C operator [] (nat i) const {
    ASSERT (rep != NULL, "non-empty chain expected");
    if (i < N (rep->l)) return rep->l[i];
    if (i > N (rep->l)) return rep->r[i-N(rep->l)];
    return rep->m; }
  friend Chain left LESSGTR (const Chain& c);
  friend C middle LESSGTR (const Chain& c);
  friend Chain right LESSGTR (const Chain& c);
  friend bool nil LESSGTR (const Chain& c);
  friend nat N LESSGTR (const Chain& c);
};
INDIRECT_NULL_IMPL_1 (chain, chain_rep, typename C, C)

/******************************************************************************
* Basic constructors and accessors
******************************************************************************/

TMPL inline
Chain_rep::chain_rep (const Chain& l2, const C& m2, const Chain& r2):
  size (N (l2) + N (r2) + 1), l (l2), m (m2), r (r2) {}

TMPL inline Chain left (const Chain& c) { return c.rep->l; }
TMPL inline C middle (const Chain& c) { return c.rep->m; }
TMPL inline Chain right (const Chain& c) { return c.rep->r; }
TMPL inline bool nil (const Chain& c) { return c.rep == NULL; }
TMPL inline nat N (const Chain& c) { return (c.rep == NULL? 0: c.rep->size); }

/******************************************************************************
* Chain iteration and output
******************************************************************************/

TMPL iterator<C>
iterate (const Chain& c) {
  if (nil (c)) return iterator<C> ();
  return
    lazy_iterator<C,Chain > (left (c)) *
    iterator<C> (middle (c)) *
    lazy_iterator<C,Chain > (right (c));
}

TMPL generic
flatten (const Chain& c) {
  return flatten (generic ("chain"), iterate (c));
}

PRINT_SUGAR(TMPL,Chain)

/******************************************************************************
* Operations on chains
******************************************************************************/

TMPL Chain
copy (const Chain& c) {
  if (nil (c)) return c;
  return Chain (left (c), middle (c), right (c));
}

TMPL Chain
reverse (const Chain& c) {
  if (nil (c)) return c;
  return Chain (reverse (right (c)), middle (c), reverse (left (c)));
}

TMPL C
car (const Chain& c) {
  ASSERT (!nil (c), "non-empty chain expected");
  if (nil (left (c))) return middle (c);
  return car (left (c));
}

TMPL C
cAr (const Chain& c) {
  ASSERT (!nil (c), "non-empty chain expected");
  if (nil (right (c))) return middle (c);
  return cAr (right (c));
}

template<typename Op, typename C> nat
unary_hash (const Chain& c) {
  if (nil (c)) return 1928;
  nat h1= Op::op (left (c));
  nat h2= Op::op (middle (c));
  nat h3= Op::op (right (c));
  return h1 ^ (h2>>1) ^ (h2<<31) ^ (h3>>3) ^ (h3>>29);
}

template<typename Op, typename C> bool
binary_test (const Chain& c1, const Chain& c2) {
  if (N (c1) != N (c2)) return false;
  return binary_test<Op> (iterate (c1), iterate (c2));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,Chain)
SOFT_IDENTITY_OP_SUGAR(TMPL,Chain)

/******************************************************************************
* Operations that need balancing
******************************************************************************/

TMPL Chain
shift_left (const Chain& c) {
  return Chain (Chain (left (c), middle (c), left (right (c))),
		middle (right (c)),
		right (right (c)));
}

TMPL Chain
shift_right (const Chain& c) {
  return Chain (left (left (c)),
		middle (left (c)),
		Chain (right (left (c)), middle (c), right (c)));
}

TMPL Chain
balance_left (const Chain& c) {
  // Ensure that #left(c) >= #right(c)
  switch (N(c)) {
  case 0:
  case 1:
    return c;
  case 2:
    if (nil (right (c))) return c;
    return Chain (Chain (middle (c)), middle (right (c)), Chain ());
  case 3:
    return c;
  default:
    if (N (left (c)) >= N (right (c))) return c;
    else {
      Chain aux= balance_right (right (c));
      return shift_left (Chain (left (c), middle (c), aux));
    }
  }
}

TMPL Chain
balance_right (const Chain& c) {
  // Ensure that #left(c) <= #right(c)
  switch (N(c)) {
  case 0:
  case 1:
    return c;
  case 2:
    if (nil (left (c))) return c;
    return Chain (Chain (), middle (left (c)), Chain (middle (c)));
  case 3:
    return c;
  default:
    if (N (left (c)) <= N (right (c))) return c;
    else {
      Chain aux= balance_left (left (c));
      return shift_right (Chain (aux, middle (c), right (c)));
    }
  }
}

TMPL Chain
cdr (const Chain& c) {
  ASSERT (!nil (c), "non-empty chain expected");
  if (N (c) == 1) return Chain ();
  if (2 * N (left (c)) <= N (right (c))) return cdr (balance_left (c));
  return Chain (cdr (left (c)), middle (c), right (c));
}

TMPL Chain
cDr (const Chain& c) {
  ASSERT (!nil (c), "non-empty chain expected");
  if (N (c) == 1) return Chain ();
  if (2 * N (right (c)) <= N (left (c))) return cDr (balance_right (c));
  return Chain (left (c), middle (c), cDr (right (c)));
}

TMPL Chain
operator * (const Chain& l, const Chain& r) {
  nat nl= N (l), nr= N (r);
  if (nl < nr) {
    if (nl == 0) return r;
    if (nr < 2*nl + 3) return Chain (l, car (r), cdr (r));
    Chain br= balance_right (r);
    return Chain (l * left (br), middle (br), right (br));
  }
  else {
    if (nr == 0) return l;
    if (nl < 2*nr + 3) return Chain (cDr (l), cAr (l), r);
    Chain bl= balance_left (l);
    return Chain (left (bl), middle (bl), right (bl) * r);
  }
}

TMPL Chain
range (const Chain& c, nat start, nat end) {
  if (start >= end) return Chain ();
  ASSERT (!nil (c), "non-empty chain expected");
  nat n= N (left (c));
  if (end <= n) return range (left (c), start, end);
  if (start > n) return range (right (c), start-n-1, end-n-1);
  Chain l= range (left (c), start, n);
  C     m= middle (c);
  Chain r= range (right (c), 0, end-n-1);
  if (N(l) < 2*(N(r) + 1) && N(r) < 2*(N(l) + 1)) return Chain (l, m, r);
  return l * Chain (m) * r;
}

#undef TMPL_DEF
#undef TMPL
#undef Chain
#undef Chain_rep
} // namespace mmx
#endif // __MMX_CHAIN_HPP
