
/******************************************************************************
* MODULE     : list.hpp
* DESCRIPTION: Linked lists
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_LIST_HPP
#define __MMX_LIST_HPP
#include "iterator.hpp"
namespace mmx {
#define TMPL_DEF template<typename C>
#define TMPL template<typename C>
#define List list<C>
#define List_rep list_rep<C>
TMPL class list_rep;
TMPL class list;
TMPL inline List cons (const C& c, const List& l);
TMPL inline const C& car (const List& l);
TMPL inline C& car (List& l);
TMPL inline const List& cdr (const List& l);
TMPL inline List& cdr (List& l);
TMPL inline bool nil (const List& l);

/******************************************************************************
* List class and its representation class
******************************************************************************/

TMPL_DEF
class list_rep REP_STRUCT {
  C    item;
  List next;

public:
  inline List_rep (const C& item2, const List& next2):
    item(item2), next(next2) {}
  friend class List;
  friend const C& car LESSGTR (const List& l);
  friend C& car LESSGTR (List& l);
  friend const List& cdr LESSGTR (const List& l);
  friend List& cdr LESSGTR (List& l);
};

TMPL_DEF
class list {
INDIRECT_PROTO_1 (list, list_rep, C)
private:
  inline list (const C& item, const List& next):
    rep (new List_rep (item, next)) {}
public:
  inline list (): rep (NULL) {}
  inline list (const C& c1):
    rep (new List_rep (c1, List ())) {}
  inline list (const C& c1, const C& c2):
    rep (new List_rep (c1, List (c2))) {}
  inline list (const C& c1, const C& c2, const C& c3):
    rep (new List_rep (c1, List (c2, c3))) {}
  inline list (const C& c1, const C& c2, const C& c3, const C& c4):
    rep (new List_rep (c1, List (c2, c3, c4))) {}
  list (const iterator<C>& it) {
    if (busy (it)) { C c= *it; ++it; rep= new List_rep (c, it); }
    else rep= NULL; }
  const C& operator [] (nat i) const {
    ASSERT (rep != NULL, "non-empty list expected");
    return i==0? rep->item: read_only (rep->next) [i-1]; }
  C& operator [] (nat i) {
    ASSERT (rep != NULL, "non-empty list expected");
    secure ();
    return i==0? rep->item: rep->next[i-1]; }
  friend List cons LESSGTR (const C& c, const List& l);
  friend const C& car LESSGTR (const List& l);
  friend C& car LESSGTR (List& l);
  friend const List& cdr LESSGTR (const List& l);
  friend List& cdr LESSGTR (List& l);
  friend bool nil LESSGTR (const List& l);
};
INDIRECT_NULL_IMPL_1 (list, list_rep, typename C, C)

/******************************************************************************
* Basic constructors and accessors
******************************************************************************/

TMPL inline List cons (const C& c, const List& l) { return List (c, l); }
TMPL inline const C& car (const List& l) { return l.rep->item; }
TMPL inline const C& read_car (const List& l) { return car (l); }
TMPL inline C& car (List& l) { l.secure(); return l.rep->item; }
TMPL inline const List& cdr (const List& l) { return l.rep->next; }
TMPL inline const List& read_cdr (const List& l) { return cdr (l); }
TMPL inline List& cdr (List& l) { l.secure(); return l.rep->next; }
TMPL inline bool nil (const List& l) { return l.rep == NULL; }
TMPL inline bool atom (const List& l) { return !nil (l) && nil (cdr (l)); }
TMPL nat N (const List& l) { return nil (l)? 0: N (cdr (l)) + 1; }
TMPL inline const C& read (const List& l, nat i) { return l[i]; }

/******************************************************************************
* List iteration
******************************************************************************/

TMPL_DEF
class list_iterator_rep: public iterator_rep<C> {
  List l;  // the list to traverse
public:
  list_iterator_rep (const List& l2): l(l2) {}
protected:
  bool is_busy () { return !nil (l); }
  void advance () { l= read_cdr (l); }
  C current () { return read_car (l); }
  iterator_rep<C>* clone () { return new list_iterator_rep (l); }
};

TMPL iterator<C>
iterate (const List& l) {
  return iterator<C> (new list_iterator_rep<C> (l));
}

/******************************************************************************
* Hashing and equality testing
******************************************************************************/

template<typename Op,typename C> nat
unary_hash (const List& ll) {
  List l= ll;
  register nat h= 253648;
  while (!nil (l)) {
    h= (h<<1) ^ (h<<5) ^ (h>>27) ^ Op::op (read_car (l));
    l= read_cdr (l);
  }
  return h;
}

template<typename Op, typename C> bool
binary_test (const List& l1, const List& l2) {
  if (nil (l1) || nil (l2)) return nil (l1) && nil (l2);
  return Op::op (read_car (l1), read_car (l2)) &&
         binary_test<Op> (read_cdr (l1), read_cdr (l2));
}

TRUE_IDENTITY_OP_SUGAR(TMPL,List)
SOFT_IDENTITY_OP_SUGAR(TMPL,List)

/******************************************************************************
* Printing lists
******************************************************************************/

TMPL generic
flatten (const List& l) {
  generic args= comma ();
  List counter= l;
  while (!nil (counter)) {
    args= comma (args, flatten (read_car (counter)));
    counter= read_cdr (counter);
  }
  return xsqtuple (args);
}

PRINT_SUGAR(TMPL,List)

/******************************************************************************
* Operations on lists
******************************************************************************/

TMPL List
copy (const List& l) {
  if (nil (l)) return l;
  return cons (car (l), cdr (l));
}

TMPL List
operator * (const List& l1, const List& l2) {
  if (nil (l1)) return copy (l2);
  return cons (car (l1), cdr (l1) * l2);
}

TMPL List
reverse (const List& ll) {
  List l= ll, r;
  while (!nil (l)) {
    r= cons (read_car (l), r);
    l= read_cdr (l);
  }
  return r;
}

TMPL List
range (const List& l, nat start, nat end) {
  if (start >= end) return List ();
  ASSERT (!nil (l), "non-empty list expected");
  if (start > 0) return range (cdr (l), start-1, end-1);
  return cons (car (l), range (cdr (l), 0, end-1));
}

#undef TMPL_DEF
#undef TMPL
#undef List
#undef List_rep
} // namespace mmx
#endif // __MMX_LIST_HPP
