
/******************************************************************************
* MODULE     : table.hpp
* DESCRIPTION: Hash tables
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_TABLE_HPP
#define __MMX_TABLE_HPP
#include "list.hpp"
#include "pair.hpp"
namespace mmx {
#define TMPL_DEF template<typename C, typename T, typename V= eq_op>
#define TMPL template<typename C, typename T, typename V>
#define Table table<C,T,V>
#define Table_rep table_rep<C,T,V>
TMPL class table_rep;
TMPL class table;
TMPL class entries_iterator_rep;
TMPL class table_iterator_rep;
TMPL inline C& I (Table& t);
TMPL inline const C& I (const Table& t);
TMPL inline const C& read_I (const Table& t);
TMPL inline nat N (const Table& t);
TMPL inline void reset (Table& t, const T& x);
TMPL inline void simplify (Table& t);
TMPL bool operator == (const Table& t1, const Table& t2);

/******************************************************************************
* Table class and its representation class
******************************************************************************/

TMPL_DEF
class table_rep REP_STRUCT {
  nat size;             // size of table (nr of entries)
  nat n;                // nr of keys (a power of two)
  C*  init;             // default entry
  list<pair<T,C> >* a;  // the array of entries

public:
  Table_rep (nat n2):
    size (0), n (n2),
    init (NULL),
    a (mmx_new<list<pair<T,C> > > (n))
  {
    //cout << "[+init=" << init << "," << this
    //<< "]"; cout.flush ();
  }
  Table_rep (const C& init2, nat n2):
    size (0), n (n2),
    init (mmx_new_one<C> (init2)),
    a (mmx_new<list<pair<T,C> > > (n))
  {
    //cout << "[+init=" << init << "," << this
    //<< ";" << ((int*) ((void*) init))[-1]
    //<< "]"; cout.flush ();
  }
  ~Table_rep () {
    //cout << "[-init=" << init << "," << this;
    //if (init != NULL)
    //cout << ";" << ((int*) ((void*) init))[-1];
    //cout << "]"; cout.flush ();
    if (init != NULL) mmx_delete_one<C> (init);
    mmx_delete<list<pair<T,C> > > (a, n); }
  inline void lazy_initialize () const {
    if (init == NULL) {
      *((C**) ((void*) &init))= mmx_new_one<C> ();
      //cout << "[:init=" << init << "," << this
      //<< ";" << ((int*) ((void*) init))[-1]
      //<< "]"; cout.flush ();
    }
  }
  void resize (nat n);
  void reset (const T& x);
  bool contains (const T& x) const;
  bool get (const T& x, C& y) const;
  const C& get (const T& x) const;
  C& set (const T& x);
  void simplify ();
  double complexity () const;

  friend class Table;
  friend class entries_iterator_rep<C,T,V>;
  friend class table_iterator_rep<C,T,V>;
  friend C& I LESSGTR (Table& t);
  friend const C& I LESSGTR (const Table& t);
  friend const C& read_I LESSGTR (const Table& t);
  friend nat N LESSGTR (const Table& t);
  friend bool operator == LESSGTR (const Table& t1, const Table& t2);
};

TMPL_DEF
class table {
INDIRECT_PROTO_3 (table, table_rep, C, T, V)
  inline table (): rep (new Table_rep (1)) {}
  template<typename K> table (const K& i, nat n=1):
    rep (new Table_rep (C(i), n)) {}
  inline table (const iterator<pair<T,C> >& it): rep (new Table_rep (1)) {
    for (; busy (it); ++it) rep->set ((*it).x1)= (*it).x2; }
  inline const C& operator [] (const T& x) const {
    return rep->get (x); }
  inline C& operator [] (const T& x) {
    secure (); return rep->set (x); }
  friend void reset LESSGTR (Table& t, const T& x);
  friend void simplify LESSGTR (Table& t);
};
INDIRECT_IMPL_3 (table, table_rep, typename C, C, typename T, T, typename V, V)

TMPL inline const C& read (const Table& t, const T& x) { return t[x]; }
TMPL inline void reset (Table& t, const T& x) { t.secure(); t.rep->reset (x); }
TMPL inline void simplify (Table& t) { t.secure(); t.rep->simplify (); }

/******************************************************************************
* Basic access routines for tables
******************************************************************************/

TMPL inline C&
I (Table& t) {
  t->lazy_initialize ();
  return *t->init;
}

TMPL inline const C&
I (const Table& t) {
  t->lazy_initialize ();
  return *t->init;
}

TMPL inline const C&
read_I (const Table& t) {
  t->lazy_initialize ();
  return *t->init;
}

TMPL inline nat
N (const Table& t) {
  return t->size;
}

TMPL void
Table_rep::resize (nat n2) {
  nat i, oldn= n;
  list<pair<T,C> >* olda= a;
  n= n2;
  a= mmx_new<list<pair<T,C> > > (n);
  for (i=0; i<oldn; i++) {
    list<pair<T,C> > l (olda [i]);
    while (!nil (l)) {
      nat code= V::hash_op (read_car(l).x1);
      list<pair<T,C> >& newl= a [code & (n-1)];
      newl= cons (read_car (l), newl);
      l= read_cdr (l);
    }
  }
  mmx_delete<list<pair<T,C> > > (olda, oldn);
  //double acc= 0.0;
  //for (i=0; i<n; i++)
  //acc += N(a[i]) * N(a[i]);
  //cout << "Access ratio: " << (acc/n) << " out of " << n << "\n";
}

TMPL bool
Table_rep::contains (const T& x) const {
  list<pair<T,C> >  l (a [V::hash_op (x) & (n-1)]);
  while (!nil (l)) {
    if (V::op (read_car (l).x1, x)) return true;
    l= read_cdr (l);
  }
  return false;
}

TMPL bool
Table_rep::get (const T& x, C& y) const {
  list<pair<T,C> >  l (a [V::hash_op (x) & (n-1)]);
  while (!nil (l)) {
    if (V::op (read_car (l).x1, x)) {
      y= read_car (l).x2;
      return true;
    }
    l= read_cdr (l);
  }
  return false;
}

TMPL C&
Table_rep::set (const T& x) {
  register nat hv= V::hash_op (x);
  list<pair<T,C> >* l= &(a [hv & (n-1)]);
  while (!nil (*l)) {
    if (V::op (read_car (*l).x1, x))
      return car (*l).x2;
    l= &cdr (*l);
  }
  //cout << "size= " << size << ", n= " << n << "\n";
  if (size>=n) resize (n<<1);
  //cout << "Resized\n";
  lazy_initialize ();
  l= &a [hv & (n-1)];
  *l= cons (pair<T,C> (x, *init), *l);
  size ++;
  return car (*l).x2;
}

TMPL const C&
Table_rep::get (const T& x) const {
  list<pair<T,C> > l (a [V::hash_op (x) & (n-1)]);
  while (!nil (l)) {
    if (V::op (read_car (l).x1, x)) return read_car (l).x2;
    l= read_cdr (l);
  }
  lazy_initialize ();
  return *init;
}

TMPL void
Table_rep::reset (const T& x) {
  list<pair<T,C> > *l= &(a [V::hash_op (x) & (n-1)]);
  while (!nil (*l)) {
    if (V::op (read_car (*l).x1, x)) {
      *l= read_cdr (*l);
      size --;
      if (size < (n>>1)) resize (n>>1);
      return;
    }
    l= &cdr (*l);
  }
}

TMPL void
Table_rep::simplify () {
  lazy_initialize ();
  for (nat i=0; i<n; i++) {
    list<pair<T,C> > *l= &(a[i]);
    while (!nil (*l)) {
      if (V::op (read_car (*l).x2, *init)) {
	*l= read_cdr (*l);
	size --;
      }
      else l= &cdr (*l);
    }
  }
  while (size < (n>>1)) resize (n>>1);
}

TMPL double
Table_rep::complexity () const {
  double sum= 0.0;
  //cout << "{";
  for (nat i=0; i<n; i++) {
    nat l= N(a[i]);
    sum += l*l;
    //cout << " " << l;
  }
  //cout << " }";
  return sum / n;
}

/******************************************************************************
* Iterators
******************************************************************************/

TMPL_DEF
class entries_iterator_rep: public iterator_rep<T> {
  Table t;             // the table to traverse
  nat i;               // current entry in table
  list<pair<T,C> > l;  // current list

protected:
  entries_iterator_rep (const Table& t2, nat i2, const list<pair<T,C> >& l2):
    t(t2), i(i2), l(l2) {}
  void spool () {
    while (nil (l)) {
      i++;
      if (i >= t->n) break;
      l= t->a[i];
    }
  }
  bool is_busy () { return i<t->n; }
  void advance () { l= read_cdr (l); spool (); }
  T current () { return read_car (l).x1; }
  iterator_rep<T>* clone () { return new entries_iterator_rep (t, i, l); }

public:
  entries_iterator_rep (const Table& t2):
    t(t2), i(0), l(t->a[0]) { spool (); }
};

TMPL iterator<T>
entries (const Table& t) {
  return iterator<T> (new entries_iterator_rep<C,T,V> (t));
}

TMPL_DEF
class table_iterator_rep: public iterator_rep<pair<T,C> > {
  Table t;             // the table to traverse
  nat i;               // current entry in table
  list<pair<T,C> > l;  // current list

protected:
  table_iterator_rep (const Table& t2, nat i2, const list<pair<T,C> >& l2):
    t(t2), i(i2), l(l2) {}
  void spool () {
    while (nil (l)) {
      i++;
      if (i >= t->n) break;
      l= t->a[i];
    }
  }
  bool is_busy () { return i<t->n; }
  void advance () { l= read_cdr (l); spool (); }
  pair<T,C> current () { return read_car (l); }
  iterator_rep<pair<T,C> >* clone () {
    return new table_iterator_rep (t, i, l); }

public:
  table_iterator_rep (const Table& t2):
    t(t2), i(0), l(t->a[0]) { spool (); }
};

TMPL iterator<pair<T,C> >
iterate (const Table& t) {
  return iterator<pair<T,C> > (new table_iterator_rep<C,T,V> (t));
}

/******************************************************************************
* Printing tables
******************************************************************************/

TMPL generic
flatten (const Table& t) {
  generic args= comma ();
  for (iterator<pair<T,C> > it= iterate (t); busy (it); ++it)
    args= comma (args, flatten (*it));
  return xapply ("table", args);
}

PRINT_SUGAR(TMPL,Table)

/******************************************************************************
* Abstract operations on tables
******************************************************************************/

template<typename Op, typename C, typename T, typename V> nat
unary_hash (const Table& t) {
  // we add all hash codes of the element pairs; don't "cycle",
  // because the order of the element pairs is not uniquely determined
  nat h=54321;
  for (iterator<pair<T,C> > it= iterate (t); busy (it); ++it) {
    pair<T,C> p= *it;
    nat key_h= V::hash_op (p.x1);
    nat val_h= Op::op (p.x2);
    h += (key_h << 5) ^ (key_h >> 27) ^ val_h;
  }
  return h;
}

TMPL Table
copy (const Table& t) {
  Table r (I(t));
  for (iterator<T> it= entries (t); busy (it); ++it)
    r[*it]= t[*it];
  return r;
}

TMPL Table
fuse (const Table& t) {
  Table r (I(t));
  for (iterator<T> it= entries (t); busy (it); ++it)
    r[fuse (*it)]= fuse (t[*it]);
  return r;
}

template<typename Op, typename C, typename T, typename V> Table
unary_map (const Table& t) {
  Table r (Op::op (I(t)));
  for (iterator<T> it= entries (t); busy (it); ++it)
    r[*it]= Op::op (t[*it]);
  simplify (r);
  return r;
}

template<typename Op, typename C, typename T, typename V> Table
binary_map (const Table& t, const Table& u) {
  Table r (Op::op (I(t), I(u)));
  for (iterator<T> it= entries (t); busy (it); ++it)
    r[*it]= Op::op (t[*it], u[*it]);
  for (iterator<T> it= entries (u); busy (it); ++it)
    if (!(t->contains (*it)))
      r[*it]= Op::op (t[*it], u[*it]);
  simplify (r);
  return r;
}

template<typename Op, typename C, typename T, typename V, typename X> Table
extrary_map (const Table& t, const X& x) {
  Table r (Op::op (I(t), x));
  for (iterator<T> it= entries (t); busy (it); ++it)
    r[*it]= Op::op (t[*it], x);
  simplify (r);
  return r;
}

template<typename Op, typename C, typename T, typename V> Table&
in_place_binary_map (Table& t, const Table& u) {
  Op::in_place (I(t), I(u));
  for (iterator<T> it= entries (u); busy (it); ++it) {
    Op::in_place (t[*it], u[*it]);
    if (V::op (read (t, *it), read_I (t))) reset (t, *it);
  }
  return t;
}

template<typename Op, typename C, typename T, typename V, typename X> Table&
in_place_extrary_map (Table& t, const X& x) {
  Op::in_place (I (t), x);
  for (iterator<T> it= entries (t); busy (it); ++it)
    Op::in_place (t[*it], x);
  simplify (t);
  return t;
}

template<typename Op, typename C, typename T, typename V> bool
binary_test (const Table& t, const Table& u) {
  if (Op::not_op (I(t), I(u))) return false;
  for (iterator<T> it= entries (t); busy (it); ++it)
    if (Op::not_op (t[*it], u[*it])) return false;
  for (iterator<T> it= entries (u); busy (it); ++it)
    if (!t->contains (*it) && Op::not_op (t[*it], u[*it])) return false;
  return true;
}

template<typename Op, typename C, typename T, typename V, typename X> bool
extrary_test (const Table& t, const X& c) {
  if (Op::not_op (I(t), c)) return false;
  for (iterator<T> it= entries (t); busy (it); ++it)
    if (Op::not_op (t[*it], c)) return false;
  return true;
}

template<typename Op, typename C, typename T, typename V> inline C
big (const Table& t) {
  iterator<T> it= entries (t);
  if (done (it)) return Op::template neutral<C> ();
  C r= t[*it];
  for (++it; busy (it); ++it)
    Op::in_place (r, t[*it]);
  return r;
}

/******************************************************************************
* Instantiations of the abstract routines
******************************************************************************/

TMPL inline Table melt (const Table& t) {
  return unary_map<melt_op,C,T,V> (t); }
TMPL inline Table operator - (const Table& t) {
  return unary_map<neg_op,C,T,V> (t); }
TMPL inline Table operator + (const Table& t, const Table& u) {
  return binary_map<add_op,C,T,V> (t, u); }
TMPL inline Table operator + (const Table& t, const C& u) {
  return binary_map<add_op,C,T,V> (t, Table (u)); }
TMPL inline Table operator + (const C& t, const Table& u) {
  return binary_map<add_op,C,T,V> (Table (t), u); }
TMPL inline Table operator - (const Table& t, const Table& u) {
  return binary_map<sub_op,C,T,V> (t, u); }
TMPL inline Table operator - (const Table& t, const C& u) {
  return binary_map<sub_op,C,T,V> (t, Table (u)); }
TMPL inline Table operator - (const C& t, const Table& u) {
  return binary_map<sub_op,C,T,V> (Table (t), u); }
TMPL inline Table operator * (const Table& t, const Table& u) {
  return binary_map<mul_op,C,T,V> (t, u); }
TMPL inline Table operator / (const Table& t, const Table& u) {
  return binary_map<div_op,C,T,V> (t, u); }
TMPL inline Table quo (const Table& t, const Table& u) {
  return binary_map<quo_op,C,T,V> (t, u); }
TMPL inline Table rem (const Table& t, const Table& u) {
  return binary_map<rem_op,C,T,V> (t, u); }
TMPL inline Table operator | (const Table& t, const Table& u) {
  return binary_map<or_op,C,T,V> (t, u); }
TMPL inline Table operator & (const Table& t, const Table& u) {
  return binary_map<and_op,C,T,V> (t, u); }
TMPL inline Table inf (const Table& t, const Table& u) {
  return binary_map<inf_op,C,T,V> (t, u); }
TMPL inline Table sup (const Table& t, const Table& u) {
  return binary_map<sup_op,C,T,V> (t, u); }
template<typename C, typename T, typename V, typename K> inline Table
operator * (const Table& t, const K& sc) {
  return extrary_map<rmul_op> (t, sc); }
template<typename C, typename T, typename V, typename K> inline Table
operator * (const K& sc, const Table& t) {
  return extrary_map<lmul_op> (t, sc); }
template<typename C, typename T, typename V, typename K> inline Table
operator / (const Table& t, const K& sc) {
  return extrary_map<rdiv_op> (t, sc); }
template<typename C, typename T, typename V, typename K> inline Table
operator / (const K& sc, const Table& t) {
  return extrary_map<ldiv_op> (t, sc); }
template<typename C, typename T, typename V, typename K> inline Table
quo (const Table& t, const K& sc) {
  return extrary_map<rquo_op> (t, sc); }
template<typename C, typename T, typename V, typename K> inline Table
rem (const Table& t, const K& sc) {
  return extrary_map<rrem_op> (t, sc); }
TMPL inline Table operator += (Table& t, const Table& u) {
  return in_place_binary_map<add_op,C,T,V> (t, u); }
TMPL inline Table operator -= (Table& t, const Table& u) {
  return in_place_binary_map<sub_op,C,T,V> (t, u); }
TMPL inline Table operator *= (Table& t, const Table& u) {
  return in_place_binary_map<mul_op,C,T,V> (t, u); }
TMPL inline Table operator /= (Table& t, const Table& u) {
  return in_place_binary_map<div_op,C,T,V> (t, u); }
TMPL inline Table operator |= (Table& t, const Table& u) {
  return in_place_binary_map<or_op,C,T,V> (t, u); }
TMPL inline Table operator &= (Table& t, const Table& u) {
  return in_place_binary_map<and_op,C,T,V> (t, u); }
template<typename C, typename T, typename V, typename K> inline Table&
operator *= (Table& t, const K& sc) {
  return in_place_extrary_map<rmul_op> (t, sc); }
template<typename C, typename T, typename V, typename K> inline Table&
operator /= (Table& t, const K& sc) {
  return in_place_extrary_map<rdiv_op> (t, sc); }
TMPL inline bool operator <= (const Table& t, const Table& u) {
  return binary_test<lesseq_op> (t, u); }
TMPL inline bool operator >= (const Table& t, const Table& u) {
  return binary_test<gtreq_op,C,T,V> (t, u); }
template<typename C, typename T, typename V, typename K> inline bool
operator == (const Table& t, const K& c) {
  return extrary_test<equal_op> (t, c); }
template<typename C, typename T, typename V, typename K> inline bool
operator <= (const Table& t, const K& c) {
  return extrary_test<lesseq_op> (t, c); }
template<typename C, typename T, typename V, typename K> inline bool
operator >= (const Table& t, const K& c) {
  return extrary_test<gtreq_op> (t, c); }

TRUE_IDENTITY_OP_SUGAR(TMPL,Table)
SOFT_IDENTITY_OP_SUGAR(TMPL,Table)
VAR_ARITH_INT_SUGAR(TMPL,Table)
STRICT_COMPARE_SUGAR(TMPL,Table)

TMPL inline C big_add (const Table& v) { return big<add_op> (v); }
TMPL inline C big_mul (const Table& v) { return big<mul_op> (v); }
TMPL inline C big_or  (const Table& v) { return big< or_op> (v); }
TMPL inline C big_and (const Table& v) { return big<and_op> (v); }
TMPL inline C big_min (const Table& v) { return big<min_op> (v); }
TMPL inline C big_max (const Table& v) { return big<max_op> (v); }
TMPL inline C big_inf (const Table& v) { return big<inf_op> (v); }
TMPL inline C big_sup (const Table& v) { return big<sup_op> (v); }

#undef TMPL_DEF
#undef TMPL
#undef Table
#undef Table_rep
} // namespace mmx
#endif // __MMX_TABLE_HPP
