
/******************************************************************************
* MODULE     : base_evaluator.cpp
* DESCRIPTION: The base evaluator
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "base_evaluator.hpp"
#include "routine.hpp"
#include "primitive.hpp"
#include "literal.hpp"
#include "generic_internal.hpp"
#include "alias.hpp"
namespace mmx {

/******************************************************************************
* Constructors
******************************************************************************/

evaluator
base_evaluator () {
  return new base_evaluator_rep ();
}

evaluator
base_evaluator (const evaluator& ev) {
  return new base_evaluator_rep (ev);
}

/******************************************************************************
* Environment access
******************************************************************************/

environment
base_evaluator_rep::get_environment () const {
  return env;
}

void
base_evaluator_rep::set (const generic& var, const generic& val) const {
  base_evaluator_rep* me= const_cast<base_evaluator_rep*> (this);
  me->env [var]= val;
}

void
base_evaluator_rep::reset (const generic& var) const {
  base_evaluator_rep* me= const_cast<base_evaluator_rep*> (this);
  mmx::reset (me->env, var);
}

void
base_evaluator_rep::overload (const generic& var, const generic& val,
			      nat pen) const
{
  /*
  cout << "Overload ";
  lisp_print (cout, var);
  if (is<routine> (val)) {
    routine f= as<routine> (val);
    array<nat> ids= f->signature ();
    extern generic type_name (nat id);
    cout << ": ";
    for (nat i=1; i<N(ids); i++) {
      if (i!=1) cout << ", ";
      lisp_print (cout, type_name (ids[i]));
    }
    cout << " -> ";
    lisp_print (cout, type_name (ids[0]));
  }
  cout << "\n";
  */

  base_evaluator_rep* me= const_cast<base_evaluator_rep*> (this);
  if (is<routine> (val)) {
    routine f= as<routine> (val);
    if (eq (var, GEN_NEW)) {
      const array<nat> ids= f->signature ();
      ASSERT (N(ids) == 2, "constructor should take one argument");
      me->env [as<generic> (ids[1])]= val;
    }
    else if (eq (var, GEN_CAST) || eq (var, GEN_UPGRADE) ||
	     eq (var, GEN_DOWNGRADE) || eq (var, GEN_CONVERT))
      {
	const array<nat> ids= f->signature ();
	ASSERT (N(ids) == 2, "converter should take one argument");
	nat src = ids[1], dest= ids[0];
	if (src == dest) return;
	nat trv= 0;
	if (eq (var, GEN_UPGRADE) || eq (var, GEN_CONVERT)) trv += 1;
	if (eq (var, GEN_DOWNGRADE) || eq (var, GEN_CONVERT)) trv += 2;
	inside (env) -> set_converter (src, dest, val, trv, pen);
	//cout << "Converter: " << src << ", " << dest << ", " << val << "\n";
	//cout << "Environment: " << env << "\n";
      }
    else if (N(f->signature ()) == 0)
      me->env [var]= val;
    else {
      if (!env->contains (var))
	me->env [var]= make_abstract (overloaded_routine (var, env));
      as<routine> (read (me->env, var)) -> overload (f);
    }
  }
  else me->env [var]= val;
}

bool
base_evaluator_rep::contains (const generic& var) const {
  return env->contains (var);
}

generic
base_evaluator_rep::get (const generic& var) const {
  return env [var];
}

bool
base_evaluator_rep::get (const generic& var, generic& val) const {
  return env->get (var, val);
}

/******************************************************************************
* Evaluation
******************************************************************************/

static routine
default_routine (const generic& fun, const generic& name) {
  if (is_alias_type (type (fun))) {
    generic ret= get_alias (fun);
    if (is<routine> (ret)) return as<routine> (ret);
  }
  if (is<literal> (fun)) return default_routine (fun);
  return default_routine (name);
}

generic
base_evaluator_rep::eval (const generic& x) const {
#ifdef BASIX_ENABLE_EXCEPTIONS
  try {
#endif
    if (is<literal> (x)) {
      generic r;
      if (env->get (x, r)) return r;
      return x;
    }
    else if (is<compound> (x)) {
      const array<generic> v= compound_to_array (x);
      nat n= N(v);
      /*
      if (n>0 && is<literal> (v[0])) {
        cout << "Evaluate ";
        print_lisp (cout, x);
        cout << "\n";
      }
      */
      switch (n) {
      case 0:
	return x;
      case 1:
	{
	  generic fun= eval (v[0]);
	  if (is<primitive> (fun))
	    return as<primitive> (fun) -> apply (x);
	  else {
	    if (!is<routine> (fun))
	      fun= as<generic> (default_routine (fun, x[0]));
	    return as<routine> (fun) -> apply ();
	  }
	}
      case 2:
	{
	  generic fun= eval (v[0]);
	  if (is<primitive> (fun))
	    return as<primitive> (fun) -> apply (x);
	  else {
	    if (!is<routine> (fun))
	      fun= as<generic> (default_routine (fun, x[0]));
	    generic arg1= eval (v[1]);
	    if (is<exception> (arg1)) return arg1;
	    return as<routine> (fun) -> apply (arg1);
	  }
	}
      case 3:
	{
	  generic fun= eval (v[0]);
	  if (is<primitive> (fun))
	    return as<primitive> (fun) -> apply (x);
	  else {
	    if (!is<routine> (fun))
	      fun= as<generic> (default_routine (fun, x[0]));
	    generic arg1= eval (v[1]);
	    if (is<exception> (arg1)) return arg1;
	    generic arg2= eval (v[2]);
	    if (is<exception> (arg2)) return arg2;
	    return as<routine> (fun) -> apply (arg1, arg2);
	  }
	}
      default:
	{
	  generic fun= eval (v[0]);
	  if (is<primitive> (fun))
	    return as<primitive> (fun) -> apply (x);
	  else {
	    if (!is<routine> (fun))
	      fun= as<generic> (default_routine (fun, x[0]));
	    generic args= xeval (fun, v);
	    if (is<exception> (args)) return args;
	    else return as<routine> (fun) -> xapply (args);
	  }
	}
      }
    }
    else return construct (x);
#ifdef BASIX_ENABLE_EXCEPTIONS
  }
  catch (const exception& err) {
    generic msg= *err;
    generic ret= append (range (msg, 0, N(msg)-1), gen (x));
    return as<generic> (exception (ret));
  }
#endif
}

generic
base_evaluator_rep::xeval (const generic& f, const array<generic>& v) const {
  nat n= N(v);
  array<generic> w= array_fill<generic> (n);
  w[0]= f;
  for (nat i=1; i<n; i++) {
    generic arg= eval (v[i]);
    if (is<exception> (arg)) return arg;
    else w[i]= arg;
  }
  return array_to_compound (w);
}

/******************************************************************************
* Function application
******************************************************************************/

generic
base_evaluator_rep::construct (const generic& x) const {
  generic the_fun= env->get (as<generic> (type (x)));
  if (is<routine> (the_fun))
    return as<routine> (the_fun)->apply (x);
  return x;
}

generic
base_evaluator_rep::xapply (const generic& x) const {
  generic the_fun;
  if (!env->get (x[0], the_fun))
    the_fun= as<generic> (default_routine (x[0]));
  return as<routine> (the_fun)->xapply (x);
}

generic
base_evaluator_rep::apply (const generic& fun) const {
  generic the_fun;
  if (!env->get (fun, the_fun))
    the_fun= as<generic> (default_routine (fun));
  return as<routine> (the_fun)->apply ();
}

generic
base_evaluator_rep::apply (const generic& fun, const generic& x1) const {
  generic the_fun;
  if (!env->get (fun, the_fun))
    the_fun= as<generic> (default_routine (fun));
  return as<routine> (the_fun)->apply (x1);
}

generic
base_evaluator_rep::apply (const generic& fun,
			   const generic& x1, const generic& x2) const
{
  generic the_fun;
  if (!env->get (fun, the_fun))
    the_fun= as<generic> (default_routine (fun));
  return as<routine> (the_fun)->apply (x1, x2);
}

} // namespace mmx
