
/******************************************************************************
* MODULE     : environment.cpp
* DESCRIPTION: Environments for the base evaluator
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "environment.hpp"
#include "literal.hpp"
#include "routine.hpp"
#include "alias.hpp"
namespace mmx {

/******************************************************************************
* Environment access
******************************************************************************/

generic
environment_rep::name () const {
  array<generic> vars (entries (bindings));
  if (nil (next)) return gen (GEN_GLOBAL, vars);
  else return gen (GEN_LOCAL, append (vars, array<generic> (next->name ())));
}

bool
environment_rep::contains (const generic& var) const {
  if (bindings->contains (var)) return true;
  else if (nil (next)) return false;
  else if (next->contains (var)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    me->bindings [var]= next[var];
    return true;
  }
  else return false;
}

bool
environment_rep::get (const generic& var, generic& val) const {
  if (bindings->get (var, val)) return true;
  else if (nil (next)) return false;
  else if (next->get (var, val)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    me->bindings [var]= val;
    return true;
  }
  else return false;
}

generic
environment_rep::get (const generic& var) const {
  generic val;
  if (bindings->get (var, val)) return val;
  else if (nil (next)) return val;
  else if (next->get (var, val)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    me->bindings [var]= val;
    return val;
  }
  else return val;
}

list<string>
environment_rep::strings_for_completion () const {
  list<string> vars;
  iterator<generic> tmp = entries (bindings);
  for ( ; busy (tmp); ++tmp)
    if (is<literal> (*tmp)) 
      vars = cons (literal_to_string (*tmp), vars);
  return vars;
}

/******************************************************************************
* Converters
******************************************************************************/

inline generic tag (const generic& name, nat id) {
  return gen (name, as<generic> (id)); }
inline generic tag (const generic& name, nat id1, nat id2) {
  return gen (name, as<generic> (id1), as<generic> (id2)); }

void
environment_rep::ensure_up_to_date () const {
  if (!nil (next)) {
    next->ensure_up_to_date ();
    if (next_serial != next->serial) {
      environment_rep* me= const_cast<environment_rep*> (this);
      me->serial++;
      me->next_serial= next->serial;
    }
  }
}

void
environment_rep::set_converter (nat src, nat dest, const generic& val,
				nat trv, nat pen)
{
  //cout << "Set converter ";
  //lisp_print (cout, type_name (src));
  //cout << " -> ";
  //lisp_print (cout, type_name (dest));
  //cout << "\n";

  // add the converter
  set (tag (GEN_CONVERTER, src, dest), val);
  set (tag (GEN_TRANSITIVE, src, dest), as<generic> (trv));
  set (tag (GEN_PENALTY, src, dest), as<generic> (pen));

  // add the converter to the list of converters
  generic t= tag (GEN_CONVERTERS, src);
  if (!contains (t)) set (t, as<generic> (array<nat> ()));
  set (t, as<generic> (cons<nat> (dest, as<array<nat> > (get (t)))));

  // time stamp
  environment_rep* me= const_cast<environment_rep*> (this);
  me->serial++;
}

generic
environment_rep::get_converter (nat src, nat dest, nat& pen) const {
  //cout << "Convert: " << src << ", " << dest << "\n";
  if (src == dest) {
    pen= 0;
    return as<generic> (identity_routine (array<nat> (dest, src)));
  }
  if (dest == 0) {
    if (is_alias_type (src)) {
      routine r;
      alias_getter (src, r);
      pen= PENALTY_FALL_BACK;
      return as<generic> (r);
    }
    else {
      pen= PENALTY_FALL_BACK;
      return as<generic> (identity_routine (array<nat> (dest, src)));
    }
  }

  ensure_up_to_date ();
  generic stamp;
  bool ok= get (tag (GEN_CACHE_CONVERTERS, src), stamp);
  if (!ok || !is<nat> (stamp) || serial != as<nat> (stamp)) {
    environment_rep* me= const_cast<environment_rep*> (this);
    table<routine,nat> funt;
    table<nat,nat> trvt;
    table<nat,nat> pent;
    array<nat> todo;

    // setup identity converter
    routine id= identity_routine (array<nat> (src, src));
    me->set (tag (GEN_CACHE_CONVERTER, src, src), as<generic> (id));
    funt [src]= id;
    trvt [src]= 3;
    pent [src]= 0;
    todo << src;

    // setup converter to generic
    if (src != 0) {
      if (is_alias_type (src)) alias_getter (src, id);
      else id= identity_routine (array<nat> ((nat) 0, src));
      me->set (tag (GEN_CACHE_CONVERTER, (nat) 0, src), as<generic> (id));
      funt [0]= id;
      trvt [0]= 3;
      pent [0]= PENALTY_FALL_BACK;
      todo << 0;
    }

    // transitive closure
    for (nat i=0; i<N(todo); i++) {
      nat cur= read (todo, i);
      generic cvs;
      if (get (tag (GEN_CONVERTERS, cur), cvs)) {
	generic all= gen (GEN_CONVERTERS, as<generic> (cur));
	array<nat> succ= as<array<nat> > (cvs);
	for (nat j=0; j<N(succ); j++) {
	  nat     next    = read (succ, j);
	  generic next_fun= get (tag (GEN_CONVERTER, cur, next));
	  generic next_trv= get (tag (GEN_TRANSITIVE, cur, next));
	  generic next_pen= get (tag (GEN_PENALTY, cur, next));
	  ASSERT (is<routine> (next_fun), "routine expected (get_converter)");
	  ASSERT (is<nat> (next_trv), "nat expected (get_converter)");
	  ASSERT (is<nat> (next_pen), "nat expected (get_converter)");
	  nat trv1= read (trvt, cur), trv2= as<nat> (next_trv);
	  nat pen1= read (pent, cur), pen2= as<nat> (next_pen);
	  nat trv= (trv1&1) + (trv2&2);
	  nat pen= max (pen1, pen2);
	  if (((trv1&2) != 0 ||
	       (trv2&1) != 0) &&
	      (!funt->contains (next) ||
	       pen <= read (pent, next)) &&
	      (pen < read (pent, next) ||
	       (trv & (~read (trvt, next))) > 0))
	    {
	      routine fun=
		compose (as<routine> (next_fun), read (funt, cur));
	      me->set (tag (GEN_CACHE_CONVERTER, src, next), as<generic>(fun));
	      me->set (tag (GEN_CACHE_PENALTY, src, next), as<generic>(pen));
	      funt [next]= fun;
	      trvt [next]= trv;
	      pent [next]= pen;
	      todo << next;
	      //cout << "Converter " << type_name (src) << " -> "
	      //<< type_name (next) << ": " << pen << "\n";
	    }
	}
      }
    }

    // time stamp
    me->set (tag (GEN_CACHE_CONVERTERS, src), as<generic> (serial));
  }

  generic fun;
  if (get (tag (GEN_CACHE_CONVERTER, src, dest), fun)) {
    generic penalty;
    get (tag (GEN_CACHE_PENALTY, src, dest), penalty);
    pen= as<nat> (penalty);
    return fun;
  }
  pen= PENALTY_INVALID;
  return as<generic> (routine ());
}

} // namespace mmx
