
/******************************************************************************
* MODULE     : generic.hpp
* DESCRIPTION: Generic expressions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __GENERIC_HPP
#define __GENERIC_HPP
// NOTE: generic.hpp is included by basix.hpp
namespace mmx {
class generic;
class string;
template<typename C> struct constant_helper;
template<typename C> struct internal;
template<typename C> struct inspector;
extern bool booting_done;

/******************************************************************************
* The generic class
******************************************************************************/

class generic_rep: public rep_struct {
protected:
  virtual nat get_type () const = 0;
  virtual nat get_length () const = 0;
  virtual generic get_child (nat i) const = 0;
  virtual nat get_soft_hash_value () const = 0;
  virtual nat get_hash_value () const = 0;
  virtual bool is_eq (const generic& rep) const = 0;
  virtual bool is_equal (const generic& rep) const = 0;
  virtual generic melt_me () const = 0;
  virtual generic fuse_me () const = 0;
  virtual generic expression () const = 0;

public:
  inline generic_rep () {}
  inline virtual ~generic_rep () {}

  friend class generic;
  inline friend nat type (const generic& g);
  inline friend nat N (const generic& g);
  inline friend nat soft_hash (const generic& g);
  inline friend nat hash (const generic& g);
  inline friend bool eq (const generic& g1, const generic& g2);
  inline friend bool neq (const generic& g1, const generic& g2);
  friend bool operator == (const generic& g1, const generic& g2);
  friend bool operator != (const generic& g1, const generic& g2);
  friend generic melt (const generic& g);
  friend generic fuse (const generic& g);
  friend generic flatten (const generic& s);
};

class generic {
INDIRECT_PROTO (generic, generic_rep)
public:
  generic ();
  generic (char* s);
  generic (const string& s);
  generic (int i);
  generic (nat i);
  generic operator () () const;
  generic operator () (const generic& g1) const;
  generic operator () (const generic& g1, const generic& g2) const;
  generic operator () (const generic& g1, const generic& g2,
		       const generic& g3) const;

  inline friend generic_rep* inspect (const generic& g) { return g.rep; }
  inline friend nat type (const generic& g) { return g->get_type (); }
  inline friend nat N (const generic& g) { return g->get_length (); }
  inline generic operator [] (nat i) const { return rep->get_child (i); }
  inline friend nat soft_hash (const generic& g) {
    return g->get_soft_hash_value (); }
  inline friend nat hash (const generic& g) {
    return g->get_hash_value (); }
  inline friend bool eq (const generic& g1, const generic& g2) {
    return g1.rep->is_eq (g2); }
  inline friend bool neq (const generic& g1, const generic& g2) {
    return !g1.rep->is_eq (g2); }
  friend bool operator == (const generic& g1, const generic& g2);
  friend bool operator != (const generic& g1, const generic& g2);
  friend generic melt (const generic& g);
  friend generic fuse (const generic& g);
  friend generic flatten (const generic& g);

  friend class generic_rep;
};
INDIRECT_IMPL (generic, generic_rep)

ostream& operator << (ostream& out, const generic& g);
generic type_name (const generic& g);

/******************************************************************************
* Basic generics formed from literals and compounds
******************************************************************************/

nat new_type_id ();
inline nat binary_id (nat id1, nat id2) { return id1 ^ (0x10115 * id2); }
template<typename T> struct type_information { static nat id; };
template<typename T> nat type_information<T>::id= new_type_id ();
STMPL struct type_information<generic> { static nat id; };
template<typename C> inline bool is (const generic& x) {
  return type (x) == type_information<C>::id; }

/******************************************************************************
* Raw manipulation of generic objects
******************************************************************************/

// the gen functions generate shared generics of type compound,
// while the ugen ones desactivate the expression sharing

generic lit (char* s);
generic ulit (char* s);
generic lit (const string& s);
generic ulit (const string& s);

generic gen (const generic& x1, const array<generic>& v);
generic ugen (const generic& x1, const array<generic>& v);
generic gen ();
generic ugen ();
generic gen (const generic& x1);
generic ugen (const generic& x1);
generic gen (const generic& x1, const string& x2);
generic ugen (const generic& x1, const string& x2);
generic gen (const generic& x1, const generic& x2);
generic ugen (const generic& x1, const generic& x2);
generic gen (const generic& x1, const generic& x2, const generic& x3);
generic ugen (const generic& x1, const generic& x2, const generic& x3);
generic gen (const generic& x1, const generic& x2,
	     const generic& x3, const generic& x4);
generic ugen (const generic& x1, const generic& x2,
	     const generic& x3, const generic& x4);
generic gen (const generic& x1, const generic& x2,
	     const generic& x3, const generic& x4, const generic& x5);
generic ugen (const generic& x1, const generic& x2,
	     const generic& x3, const generic& x4, const generic& x5);
generic gen (const generic& x1, const generic& x2,
	     const generic& x3, const generic& x4,
	     const generic& x5, const generic& x6);
generic ugen (const generic& x1, const generic& x2,
	      const generic& x3, const generic& x4,
	      const generic& x5, const generic& x6);

generic comma ();
generic comma (const generic& x1, const generic& x2);
generic xapply (const generic& f, const generic& args);
generic xaccess (const generic& f, const generic& args);
generic xtuple (const generic& args);
generic xsqtuple (const generic& args);
generic xrow (const generic& args);
generic apply (const generic& f, const generic& x);
generic apply (const generic& f, const generic& x, const generic& y);
generic apply (const generic& f, const generic& x,
	       const generic& y, const generic& z);
generic apply (const generic& f, const array<generic>& x);
generic access (const generic& f, const generic& x);
generic access (const generic& f, const generic& x, const generic& y);
generic access (const generic& f, const array<generic>& x);

bool is_scalar (const generic& g);    // for linear combinations
bool is_central (const generic& g);   // commutes with evrybody
bool is_operator (const generic& g);  // linear operators

bool scalar_expand ();          // a(b+c) = ab+ac for scalar a
bool scalar_power_expand ();    // (ab)^c = a^c b^c for scalar c
bool scalar_power_multiply ();  // (a^b)^c = a^(bc) for scalar c
bool central_expand ();         // a(b+c) = ab+ac for central a
bool central_power_expand ();   // (ab)^c = a^c b^c for central c
bool central_power_multiply (); // (a^b)^c = a^(bc) for central c

/******************************************************************************
* Routines based on the current evaluator
******************************************************************************/

generic construct (const int& i);
generic construct (const nat& i);
generic construct (const double& x);
generic construct (const float& x);
generic construct (const generic& x);
generic eval (const generic& x);

STMPL
struct constant_helper<generic> {
  static generic pi ();
  static generic log2 ();
  static generic euler ();
  static generic catalan ();
  static generic imaginary ();
  static generic nan ();
  static generic fuzz ();
  static generic infinity ();
  static generic plus_infinity ();
  static generic minus_infinity ();
};

generic operator - (const generic& x1);
generic operator + (const generic& x1, const generic& x2);
generic operator - (const generic& x1, const generic& x2);
generic operator * (const generic& x1, const generic& x2);
generic operator / (const generic& x1, const generic& x2);
generic quo (const generic& x1, const generic& x2);
generic rem (const generic& x1, const generic& x2);
generic big_add (const array<generic>& v);
generic big_mul (const array<generic>& v);

generic numerator (const generic& x);
generic denominator (const generic& x);
generic gcd (const generic& x1, const generic& x2);
generic lcm (const generic& x1, const generic& x2);
template<typename C> struct xgcd_matrix;
xgcd_matrix<generic> xgcd (const generic& x1, const generic& x2);

generic pow (const generic& x1, const generic& x2);
generic pow (const generic& x1, const int& x2);
generic pow (const generic& x1, const nat& x2);
generic pow (const int& x1, const generic& x2);
generic pow (const nat& x1, const generic& x2);
generic sqrt (const generic& x1);
generic exp (const generic& x1);
generic log (const generic& x1);
generic cos (const generic& x1);
generic sin (const generic& x1);
generic tan (const generic& x1);
generic acos (const generic& x1);
generic asin (const generic& x1);
generic atan (const generic& x1);
generic trig (const generic& x1);

bool operator == (const generic& x1, const generic& x2);
bool operator != (const generic& x1, const generic& x2);
bool operator <  (const generic& x1, const generic& x2);
bool operator <= (const generic& x1, const generic& x2);
bool operator >  (const generic& x1, const generic& x2);
bool operator >= (const generic& x1, const generic& x2);

generic min (const generic& x1, const generic& x2);
generic max (const generic& x1, const generic& x2);
generic abs (const generic& x1);
generic arg (const generic& x1);
generic Re (const generic& x1);
generic Im (const generic& x1);
generic conj (const generic& x1);
generic gaussian (const generic& x1, const generic& x2);
generic polar (const generic& x1, const generic& x2);

generic sharpen (const generic& x);

generic derive (const generic& x1);
generic derive (const generic& x1, const generic& v);
generic prime (const generic& x1);
generic integrate (const generic& x1);
generic integrate (const generic& x1, const generic& v);
generic substitute (const generic& x1, const generic& x2);
generic compose (const generic& x1, const generic& x2);
generic dilate (const generic& x1, const generic& x2);
generic solve (const generic& x1, const generic& x2);

generic sqrt_init (const generic& x, const generic& c);
generic log_init (const generic& x, const generic& c);
generic acos_init (const generic& x, const generic& c);
generic asin_init (const generic& x, const generic& c);
generic atan_init (const generic& x, const generic& c);
generic integrate_init (const generic& x, const generic& c);
generic solve_lde_init (const generic& x, const generic& c);

generic operator << (const generic& x1, const generic& x2);
generic operator >> (const generic& x1, const generic& x2);
generic shift_exp (const generic& x1, const generic& x2);
generic operator | (const generic& x1, const generic& x2);
generic operator & (const generic& x1, const generic& x2);
generic lres (const generic& x1, const generic& x2);
generic rres (const generic& x1, const generic& x2);
generic uniform (const generic& x1, const generic& x2);
generic specialize (const generic& x1, const generic& x2);

} // namespace mmx
#endif // __GENERIC_HPP
