
/******************************************************************************
* MODULE     : generic_internal.hpp
* DESCRIPTION: Internal generic expression types
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __GENERIC_INTERNAL_HPP
#define __GENERIC_INTERNAL_HPP
#include "type_props.hpp"
#include "mmx_identifiers.hpp"
#include "string.hpp"
#include "symbol.hpp"
#include "array.hpp"
#include "table.hpp"

namespace mmx {

/******************************************************************************
* Internal representation of concrete generics
******************************************************************************/

template<typename C>
struct internal {
  typedef C Repr;
  static inline Repr encode (const C& x) { return x; }
  static inline Repr encode (const C& x, bool flag) { (void) flag; return x; }
  static inline C decode (const Repr& x) { return x; }
};

/******************************************************************************
* Accessing children of the concrete generics
******************************************************************************/

template<typename C>
struct inspector {
  static inline nat length (const C& c) { return 0; }
  static inline generic access (const C& c, nat i) {
    VERIFY (false, "out of range");
    return generic (); }
};

/******************************************************************************
* Concrete instances of generic_rep
******************************************************************************/

template<typename C>
class generic_concrete_rep: public generic_rep {
public:
  typedef typename internal<C>::Repr Repr;
  Repr rep;

protected:
  nat get_type () const { return type_information<C>::id; }
  nat get_length () const {
    return inspector<C>::length (internal<C>::decode (rep)); }
  generic get_child (nat i) const {
    return inspector<C>::access (internal<C>::decode (rep), i); }
  nat get_soft_hash_value () const {
    return soft_hash (rep); }
  nat get_hash_value () const {
    return hash (rep); }
  bool is_eq (const generic& g) const {
    if (type (g) != type_information<C>::id) return false;
    return eq (rep, ((generic_concrete_rep<C>*) inspect (g)) -> rep); }
  bool is_equal (const generic& g) const {
    ASSERT (type (g) == type_information<C>::id, "type mismatch (is_equal)");
    return rep == ((generic_concrete_rep<C>*) inspect (g)) -> rep; }
  generic melt_me () const {
    return as<generic> (internal<C>::decode (melt (rep))); }
  generic fuse_me () const {
    return as<generic> (internal<C>::decode (fuse (rep))); }
  generic expression () const {
    return flatten (internal<C>::decode (rep)); }

public:
  generic_concrete_rep (const C& rep2):
    rep (internal<C>::encode (rep2)) {}
  generic_concrete_rep (const C& rep2, bool unique):
    // if unique is set then *this becomes a unique symbol<C>
    rep (internal<C>::encode (rep2, unique)) {}
};

template<typename C>
struct cast_helper<generic,C> {
  static inline generic cv (const C& x) {
    return new generic_concrete_rep<C> (x); }
};

template<typename C>
struct cast_helper<C,generic> {
  static inline C cv (const generic& x) {
    return
      internal<C>::decode (((generic_concrete_rep<C>*) inspect (x)) -> rep);
  }
};

STMPL
struct cast_helper<generic,generic> {
  static inline generic cv (const generic& x) { return x; }
};

template<typename C> generic
as_generic (const C& c) {
  return construct (new generic_concrete_rep<C> (c));
}

/******************************************************************************
* Utility routines
******************************************************************************/

generic range (const generic& g, nat start, nat end);
generic append (const generic& g1, const generic& g2);
generic cons (const generic& g1, const generic& g2);
generic cons (const generic& g1, const generic& g2, const generic& g3);
generic car (const generic& g);
generic cdr (const generic& g);

bool is_func (const generic& g, const generic& f);
bool is_func (const generic& g, const generic& f, nat n);
nat size (const generic& g);
int big_small_compare (const generic& g1, const generic& g2);
int small_big_compare (const generic& g1, const generic& g2);

} // namespace mmx
#endif // __GENERIC_INTERNAL_HPP
