
/******************************************************************************
* MODULE     : generic.cpp
* DESCRIPTION: Generic expressions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "generic_internal.hpp"
#include "compound.hpp"
#include "evaluator.hpp"
#include "array.hpp"
#include "mmx_syntax.hpp"
#include "memoize.hpp"
namespace mmx {

/******************************************************************************
* Automatic addition of new types
******************************************************************************/

nat
new_type_id () {
  static nat counter= 2;
  ASSERT (counter < 65000, "too many types");
  return counter++;
}

nat type_information<generic>::id= 0;
// type identifier 1 for unknown type

/******************************************************************************
* Printing generic expressions
******************************************************************************/

generic
melt (const generic& x1) {
  return x1.rep->melt_me ();
}

generic
fuse (const generic& x1) {
  MEMOIZE_UNARY (eq_memoizer,generic,generic,fuse,x1,
		 x1.rep->fuse_me ());
}

generic
flatten (const generic& g) {
  MEMOIZE_UNARY (std_memoizer,generic,generic,flatten,g,
		 g.rep->expression());
}

ostream&
operator << (ostream& out, const generic& g) {
  select_output_evaluator ();
  generic f= flatten (g);
  //lisp_print (out, f);
  mmx_print (out, f);
  restore_evaluator ();
  return out;
}

/******************************************************************************
* N-ary operations
******************************************************************************/

void
xapply_sub (array<generic>& v, const generic& g) {
  if (is<compound> (g) && N(g) >= 1 && eq (g[0], GEN_COMMA)) {
    int i, n= N(g);
    for (i=1; i<n; i++)
      xapply_sub (v, g[i]);
  }
  else v << g;
}

generic
xapply (const generic& f, const generic& args) {
  array<generic> v;
  v << f;
  xapply_sub (v, args);
  return array_to_compound (v);
}

generic comma () {
  return gen (generic (GEN_COMMA)); }
generic comma (const generic& g1, const generic& g2) {
  return gen (GEN_COMMA, g1, g2); }
generic xaccess (const generic& f, const generic& args) {
  return xapply (GEN_ACCESS, comma (f, args)); }
generic xtuple (const generic& g) {
  return xapply (GEN_TUPLE, g); }
generic xsqtuple (const generic& g) {
  return xapply (GEN_SQTUPLE, g); }
generic xrow (const generic& g) {
  return xapply (GEN_ROW, g); }
generic apply (const generic& g, const generic& x) {
  return gen (g, x); }
generic apply (const generic& g, const generic& x, const generic& y) {
  return gen (g, x, y); }
generic apply (const generic& g, const generic& x,
	       const generic& y, const generic& z) {
  return gen (g, x, y, z); }
generic apply (const generic& g, const array<generic>& args) {
  return array_to_compound (append (array<generic> (g), args)); }
generic access (const generic& g, const generic& x) {
  return gen (GEN_ACCESS, g, x); }
generic access (const generic& g, const generic& x, const generic& y) {
  return gen (GEN_ACCESS, g, x, y); }
generic access (const generic& g, const array<generic>& args) {
  array<generic> v= array<generic> (generic (GEN_ACCESS), g);
  return array_to_compound (append (v, args)); }

/******************************************************************************
* Predicates
******************************************************************************/

bool
is_scalar (const generic& g) {
  generic r= current_ev->apply ("scalar?", g);
  ASSERT (is<bool> (r), "Bool return value expected");
  return as<bool> (r);
}

bool is_central (const generic& g) { (void) g; return true; }
bool is_operator (const generic& g) { (void) g; return false; }

bool scalar_expand () { return true; }
bool scalar_power_expand () { return true; }
bool scalar_power_multiply () { return true; }
bool central_expand () { return false; }
bool central_power_expand () { return true; }
bool central_power_multiply () { return true; }

/******************************************************************************
* Evaluation
******************************************************************************/

generic
construct (const int& i) {
  return current_ev->construct (as<generic> (i));
}

generic
construct (const nat& i) {
  return current_ev->construct (as<generic> ((int) i));
}

generic
construct (const double& x) {
  return current_ev->construct (as<generic> (x));
}

generic
construct (const float& x) {
  return current_ev->construct (as<generic> ((double) x));
}

generic
construct (const generic& x) {
  return current_ev->construct (x);
}

generic
eval (const generic& x) {
  return current_ev->eval (x);
}

/******************************************************************************
* Constructors
******************************************************************************/

generic::generic (int i):
  rep (new generic_concrete_rep<int> (i))
{
  *this= current_ev->construct (*this);
}

generic::generic (nat i):
  rep (new generic_concrete_rep<int> ((int) i))
{
  ASSERT (((int) i) >= 0, "nat does not fit inside an int");
  *this= current_ev->construct (*this);
}

/******************************************************************************
* Constants
******************************************************************************/

generic constant_helper<generic>::pi () { return GEN_PI; }
generic constant_helper<generic>::log2 () { return log (generic (2)); }
generic constant_helper<generic>::euler () { return GEN_EULER; }
generic constant_helper<generic>::catalan () { return GEN_CATALAN; }
generic constant_helper<generic>::imaginary () { return GEN_I; }
generic constant_helper<generic>::nan () { return GEN_NAN; }
generic constant_helper<generic>::fuzz () { return GEN_FUZZ; }
generic constant_helper<generic>::infinity () { return GEN_INFINITY; }
generic constant_helper<generic>::plus_infinity () { return GEN_INFINITY; }
generic constant_helper<generic>::minus_infinity () { return -infinity (); }

/******************************************************************************
* Basic arithmetic
******************************************************************************/

//bool using_simple ();
//generic type_name (nat id);

generic operator - (const generic& x1) {
  /*
  generic r= current_ev->apply (GEN_MINUS, x1);
  if (!using_simple ())
    cout << "  neg " << x1 << " -> " << r << "\n";
  return r;
  */
  return current_ev->apply (GEN_MINUS, x1); }
generic operator + (const generic& x1, const generic& x2) {
  /*
  generic r= current_ev->apply (GEN_PLUS, x1, x2);
  if (!using_simple ()) {
    cout << "  add " << x1 << ", " << x2 << " -> " << r << "\n";
    cout << "    types: " << type_name (type (x1))
	 << ", " << type_name (type (x2))
	 << " -> " << type_name (type (r)) << "\n";
  }
  return r;
  */
  return current_ev->apply (GEN_PLUS, x1, x2); }
generic operator - (const generic& x1, const generic& x2) {
  /*
  generic r= current_ev->apply (GEN_MINUS, x1, x2);
  if (!using_simple ()) {
    cout << "  sub " << x1 << ", " << x2 << " -> " << r << "\n";
    cout << "    types: " << type_name (type (x1))
	 << ", " << type_name (type (x2))
	 << " -> " << type_name (type (r)) << "\n";
  }
  return r;
  */
  return current_ev->apply (GEN_MINUS, x1, x2); }
generic operator * (const generic& x1, const generic& x2) {
  /*
  generic r= current_ev->apply (GEN_TIMES, x1, x2);
  if (!using_simple ()) {
    cout << "  mul " << x1 << ", " << x2 << " -> " << r << "\n";
    cout << "    types: " << type_name (type (x1))
	 << ", " << type_name (type (x2))
	 << " -> " << type_name (type (r)) << "\n";
  }
  return r;
  */
  return current_ev->apply (GEN_TIMES, x1, x2); }
generic operator / (const generic& x1, const generic& x2) {
  /*
  generic r= current_ev->apply (GEN_OVER, x1, x2);
  if (!using_simple ()) {
    cout << "  div " << x1 << ", " << x2 << " -> " << r << "\n";
    cout << "    types: " << type_name (type (x1))
	 << ", " << type_name (type (x2))
	 << " -> " << type_name (type (r)) << "\n";
  }
  return r;
  */
  return current_ev->apply (GEN_OVER, x1, x2); }
generic quo (const generic& x1, const generic& x2) {
  /*
  generic r= current_ev->apply (GEN_DIV, x1, x2);
  if (!using_simple ()) {
    cout << "  quo " << x1 << ", " << x2 << " -> " << r << "\n";
    cout << "    types: " << type_name (type (x1))
	 << ", " << type_name (type (x2))
	 << " -> " << type_name (type (r)) << "\n";
  }
  return r;
  */
  return current_ev->apply (GEN_DIV, x1, x2); }
generic rem (const generic& x1, const generic& x2) {
  /*
  generic r= current_ev->apply (GEN_MOD, x1, x2);
  if (!using_simple ()) {
    cout << "  rem " << x1 << ", " << x2 << " -> " << r << "\n";
    cout << "    types: " << type_name (type (x1))
	 << ", " << type_name (type (x2))
	 << " -> " << type_name (type (r)) << "\n";
  }
  return r;
  */
  return current_ev->apply (GEN_MOD, x1, x2); }
generic big_add (const array<generic>& v) {
  /*
  generic r= current_ev->xapply (apply (GEN_PLUS, v));
  if (!using_simple ())
    cout << "  big add " << v << " -> " << r << "\n";
  return r;
  */
  switch (N(v)) {
  case 0: return 0;
  case 1: return v[0];
  case 2: return v[0] + v[1];
  default: return current_ev->xapply (apply (GEN_PLUS, v));
  } }
generic big_mul (const array<generic>& v) {
  /*
  generic r= current_ev->xapply (apply (GEN_TIMES, v));
  if (!using_simple ())
    cout << "  big mul " << v << " -> " << r << "\n";
  return r;
  */
  switch (N(v)) {
  case 0: return 1;
  case 1: return v[0];
  case 2: return v[0] * v[1];
  default: return current_ev->xapply (apply (GEN_TIMES, v));
  } }

/******************************************************************************
* Gcd and lcm
******************************************************************************/

generic numerator (const generic& x) {
  return current_ev->apply (GEN_NUMERATOR, x); }
generic denominator (const generic& x) {
  return current_ev->apply (GEN_DENOMINATOR, x); }
generic gcd (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_GCD, x1, x2); }
generic lcm (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_LCM, x1, x2); }

/******************************************************************************
* Elementary functions
******************************************************************************/

generic pow (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_POWER, x1, x2); }
generic pow (const generic& x1, const int& x2) {
  return current_ev->apply (GEN_POWER, x1, generic (x2)); }
generic pow (const generic& x1, const nat& x2) {
  return current_ev->apply (GEN_POWER, x1, generic (x2)); }
generic pow (const int& x1, const generic& x2) {
  return current_ev->apply (GEN_POWER, generic (x1), x2); }
generic pow (const nat& x1, const generic& x2) {
  return current_ev->apply (GEN_POWER, generic (x1), x2); }
generic sqrt (const generic& x1) {
  return current_ev->apply (GEN_SQRT, x1); }
generic exp (const generic& x1) {
  return current_ev->apply (GEN_EXP, x1); }
generic log (const generic& x1) {
  return current_ev->apply (GEN_LOG, x1); }
generic cos (const generic& x1) {
  return current_ev->apply (GEN_COS, x1); }
generic sin (const generic& x1) {
  return current_ev->apply (GEN_SIN, x1); }
generic tan (const generic& x1) {
  return current_ev->apply (GEN_TAN, x1); }
generic acos (const generic& x1) {
  return current_ev->apply (GEN_ARCCOS, x1); }
generic asin (const generic& x1) {
  return current_ev->apply (GEN_ARCSIN, x1); }
generic atan (const generic& x1) {
  return current_ev->apply (GEN_ARCTAN, x1); }
generic trig (const generic& x1) {
  return xsqtuple (comma (cos (x1), sin (x1))); }

/******************************************************************************
* Predicates
******************************************************************************/

bool
operator == (const generic& x1, const generic& x2) {
  if (type (x1) == type (x2)) return x1.rep->is_equal (x2);
  generic r= current_ev->apply (GEN_EQUAL, x1, x2);
  if (!is<bool> (r)) return false;
  return as<bool> (r);
}

bool
operator != (const generic& x1, const generic& x2) {
  if (type (x1) == type (x2)) return !x1.rep->is_equal (x2);
  generic r= current_ev->apply (GEN_UNEQUAL, x1, x2);
  if (!is<bool> (r)) return true;
  return as<bool> (r);
}

bool
operator < (const generic& x1, const generic& x2) {
  generic r= current_ev->apply (GEN_LESS, x1, x2);
  ASSERT (is<bool> (r), "Boolean return value expected");
  return as<bool> (r);
}

bool
operator <= (const generic& x1, const generic& x2) {
  generic r= current_ev->apply (GEN_LESSEQ, x1, x2);
  ASSERT (is<bool> (r), "Boolean return value expected");
  return as<bool> (r);
}

bool
operator > (const generic& x1, const generic& x2) {
  generic r= current_ev->apply (GEN_GTR, x1, x2);
  ASSERT (is<bool> (r), "Boolean return value expected");
  return as<bool> (r);
}

bool
operator >= (const generic& x1, const generic& x2) {
  generic r= current_ev->apply (GEN_GTREQ, x1, x2);
  ASSERT (is<bool> (r), "Boolean return value expected");
  return as<bool> (r);
}

/******************************************************************************
* Real and complex structures
******************************************************************************/

generic min (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_MIN, x1, x2); }
generic max (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_MAX, x1, x2); }
generic abs (const generic& x1) {
  return current_ev->apply (GEN_ABS, x1); }
generic arg (const generic& x1) {
  return current_ev->apply (GEN_ARG, x1); }
generic Re (const generic& x1) {
  return current_ev->apply (GEN_RE, x1); }
generic Im (const generic& x1) {
  return current_ev->apply (GEN_IM, x1); }
generic conj (const generic& x1) {
  return current_ev->apply (GEN_CONJ, x1); }
generic gaussian (const generic& x1, const generic& x2) {
  return x1 + x2 * Imaginary (generic); }
generic polar (const generic& x1, const generic& x2) {
  return x1 * exp (x2 * Imaginary (generic)); }

/******************************************************************************
* Routines for approximate arithmetic and reliable computations
******************************************************************************/

generic sharpen (const generic& x);

/******************************************************************************
* Calculus
******************************************************************************/

generic derive (const generic& x1) {
  return current_ev->apply (GEN_DERIVE, x1); }
generic derive (const generic& x1, const generic& v) {
  MEMOIZE_BINARY (std_memoizer,generic,generic,generic,derive,x1,v,
		  current_ev->apply (GEN_DERIVE, x1, v)); }
generic prime (const generic& x1) {
  return current_ev->apply ("'", x1); }
generic integrate (const generic& x1) {
  return current_ev->apply (GEN_INTEGRATE, x1); }
generic integrate (const generic& x1, const generic& v) {
  return current_ev->apply (GEN_INTEGRATE, x1, v); }
generic substitute (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_SUBSTITUTE, x1, x2); }
generic compose (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_COMPOSE, x1, x2); }
generic dilate (const generic& x1, const generic& x2) {
  return current_ev->apply ("dilate", x1, x2); }
generic solve (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_SOLVE, x1, x2); }

/******************************************************************************
* Analytic continuation
******************************************************************************/

generic sqrt_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("sqrt_init", x1, x2); }
generic log_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("log_init", x1, x2); }
generic acos_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("acos_init", x1, x2); }
generic asin_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("asin_init", x1, x2); }
generic atan_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("atan_init", x1, x2); }
generic integrate_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("integrate_init", x1, x2); }
generic solve_lde_init (const generic& x1, const generic& x2) {
  return current_ev->apply ("solve_lde_init", x1, x2); }

/******************************************************************************
* Miscellaneous
******************************************************************************/

generic operator << (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_LESSLESS, x1, x2); }
generic operator >> (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_GTRGTR, x1, x2); }
generic shift_exp (const generic& x1, const generic& x2) {
  return current_ev->apply ("shift_exp", x1, x2); }
generic operator | (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_OR, x1, x2); }
generic operator & (const generic& x1, const generic& x2) {
  return current_ev->apply (GEN_AND, x1, x2); }
generic lres (const generic& x1, const generic& x2) {
  return current_ev->apply ("lres", x1, x2); }
generic rres (const generic& x1, const generic& x2) {
  return current_ev->apply ("rres", x1, x2); }
generic uniform (const generic& x1, const generic& x2) {
  return current_ev->apply ("uniform", x1, x2); }
generic specialize (const generic& x1, const generic& x2) {
  return current_ev->apply ("specialize", x1, x2); }

} // namespace mmx
