/******************************************************************************
* MODULE     : parse_tools.hpp
* DESCRIPTION: Tools for parsing
* COPYRIGHT  : (C) 2007  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __PARSE_TOOLS_HPP
#define __PARSE_TOOLS_HPP
#include "string.hpp"
#include "pair.hpp"
#include "table.hpp"
#include "generic.hpp"
#include "generic_internal.hpp"
#include "compound.hpp"

namespace mmx {

  /*
    IMPORTANT: all positions are counted from 0 internally.
  */

class source_position {
MMX_ALLOCATORS
public:
  // This storage is redundant but helpful.
  nat position;
  nat line;
  nat column;
  source_position (void) : position (0), line (0), column (0) {}
  source_position (nat p, nat l, nat c)
    : position (p), line (l), column (c) {}
  source_position (const source_position& p)
    : position (p.position), line (p.line), column (p.column) {}
  inline source_position& operator= (const source_position& p) {
    position = p.position;
    line = p.line;
    column = p.column;
    return *this;
  }
};

inline source_position
min (const source_position& p1, const source_position& p2) {
  if (p1.position < p2.position)
    return p1;
  return p2;
};
  
inline source_position
max (const source_position& p1, const source_position& p2) {
  if (p1.position > p2.position)
    return p1;
  return p2;
};
  
class syntax_error {
MMX_ALLOCATORS
public:
  string type;
  string message;
  string file_name;
  nat input_number;
  source_position begin;
  source_position end;
  
  inline syntax_error () {}
  inline syntax_error(const string& t, const string& m, const string& f,
		      nat i, const source_position& b,
		      const source_position& e)
    : type (t), message (m), file_name (f),
      input_number (i), begin (b), end (e) {};
  inline syntax_error(const syntax_error& e)
    : type (e.type), message (e.message), file_name (e.file_name),
      input_number (e.input_number), begin (e.begin), end (e.end) {};
  inline syntax_error& operator=(const syntax_error& e) {
    type = e.type;
    message = e.message;
    file_name = e.file_name;
    input_number = e.input_number;
    begin = e.begin;
    end = e.end;
    return *this;
  }
};

struct parse_instance {
MMX_ALLOCATORS
  // internal data for the parser
  string        lex_file_name; // name of the stream
                               // "" is reserved for stdin
  nat           lex_input_number; // number of the input in interactive mode
  
  char*         lex_string; // part of the stream to be parsed and
			    // starting at the above specified offset

  nat           lex_length; // length of lex_string

  nat           lex_start; // source_positions in lex_string used in the lexer
  nat           lex_start_line;
  nat           lex_start_begin_line;
  
  nat           lex_prev; 
  nat           lex_prev_line; 
  nat           lex_prev_begin_line; 

  nat           lex_pos; // current source_position in lex_string in the lexer
  nat           lex_line;
  nat           lex_begin_line;

  generic       result; // output of the parser

  list<syntax_error> errors; // list of syntax error and warning

  parse_instance (const string& f, nat i, const string& s);
  // Setup with lex_file_name f, lex_string s and lex_input_number i
};

class source_location {
MMX_ALLOCATORS
public:
  generic obj; // located object
  string file_name; // file_name of the source_location
  nat input_number; // input number if file is stdin
  source_position begin;
  source_position end;

  inline source_location () : obj (), file_name (""),
			      input_number (0), begin (), end () {}
  inline source_location (generic g, string f, nat i,
			  const source_position& b,
			  const source_position e) : 
    obj (g), file_name (f),
    input_number (i), begin (b), end(e) {}
  inline source_location (const source_location& l) : 
    obj (l.obj), file_name (l.file_name),
    input_number (l.input_number),
    begin (l.begin), end (l.end) {}
  inline source_location& operator= (const source_location& l) {
    obj = l.obj;
    file_name = l.file_name;
    input_number = l.input_number;
    begin = l.begin;
    end = l.end;
    return *this;
  }
};

inline bool
nil (const source_location& l) {
  return (l.end.line == 0 && l.end.position == 0);
};

inline generic
flatten (const source_location& l) {
  return flatten(gen (GEN_TUPLE, l.obj, l.file_name, l.input_number,
		      gen (GEN_TUPLE, l.begin.line, l.begin.column),
		      gen (GEN_TUPLE, l.end.line, l.end.column)));
}
  
void source_insert (const generic& g, const source_location& l);
  // Insert g into source_source_location with source_location l.

void source_delete (const generic& g);
  // Delete g from source_source_location.

void source_locate (const generic& g, source_location& l);
  // Return in l the source_location of g. If l is already valid
  // source_location in input, then the returned value for l is the
  // minimal interval containing both g and l.
  // WARNING: g must not share subexpressions in different files or inputs

generic source_extend (const generic& g1, const generic& g2);
  // Set the source_location of g1 in order to contain g1 and g2.
  // Delete g2 in source_source_location and return g1.

inline generic
source_extend (const generic& g1, const generic& g2, const generic& g3) {
  return source_extend (source_extend (g1, g2), g3);
}

void print_source_location_table (ostream& out);

} // namespace mmx
#endif // __PARSE_TOOLS_HPP
