
/******************************************************************************
* MODULE     : cpp_printer.cpp
* DESCRIPTION: C++ tree printer
* COPYRIGHT  : (C) 2000  Joris van der Hoeven, 2007  Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <string.hpp>
#include <list.hpp>
#include <table.hpp>
#include <generic_internal.hpp>
#include <cpp_syntax.hpp>

namespace mml {

string CPP_ACCESS (".[]");
string CPP_ALIAS ("alias");
string CPP_AND ("&&");
string CPP_ARROW ("->");
string CPP_ASSIGN ("=");
string CPP_BEGIN ("begin");
string CPP_BITWISE_AND ("&");
string CPP_BITWISE_OR ("|");
string CPP_BITWISE_XOR ("^");
string CPP_BOOL ("bool");
string CPP_BRACKETS (".()");
string CPP_BREAK ("break");
string CPP_CAST ("cast");
string CPP_CLASS ("class");
string CPP_COMMA (",");
string CPP_COMPLEMENT ("~");
string CPP_CONST ("const");
string CPP_CONSTRUCT ("construct");
string CPP_CONSTRUCTOR ("constructor");
string CPP_CONTINUE ("continue");
string CPP_DELETE ("delete");
string CPP_DESTRUCTOR ("destructor");
string CPP_DO ("do");
string CPP_DOT (".");
string CPP_ELSE ("else");
string CPP_EQUAL ("==");
string CPP_EXIT ("exit");
string CPP_EXTERN ("extern");
string CPP_FOR ("for");
string CPP_GTR (">");
string CPP_GTREQ (">=");
string CPP_GTRGTR (">>");
string CPP_IF ("if");
string CPP_INHERIT ("inherit");
string CPP_INITIALIZE (":");
string CPP_INLINE ("inline");
string CPP_INT ("int");
string CPP_LESS ("<");
string CPP_LESSEQ ("<=");
string CPP_LESSLESS ("<<");
string CPP_MINUS ("-");
string CPP_MINUS_ASSIGN ("-=");
string CPP_MOD ("%");
string CPP_MOD_ASSIGN ("%=");
string CPP_NAMESPACE ("namespace");
string CPP_NEW ("new");
string CPP_NOT ("!");
string CPP_OR ("||");
string CPP_OSTREAM ("ostream");
string CPP_OVER ("/");
string CPP_OVER_ASSIGN ("/=");
string CPP_PLUS ("+");
string CPP_PLUS_ASSIGN ("+=");
string CPP_POINTER ("pointer");
string CPP_POSTDEC (".--");
string CPP_POSTINC (".++");
string CPP_PREDEC ("--.");
string CPP_PREINC ("++.");
string CPP_PUBLIC ("public");
string CPP_RETURN ("return");
string CPP_SCOPE ("::");
string CPP_SEMICOLON (";");
string CPP_SPECIALIZE (".<>");
string CPP_STATIC ("static");
string CPP_STRING ("string");
string CPP_STRUCT ("struct");
string CPP_SWITCH ("switch");
string CPP_TEMPLATE ("template");
string CPP_TIMES ("*");
string CPP_TIMES_ASSIGN ("*=");
string CPP_TYPE ("type");
string CPP_TYPEDEF ("typedef");
string CPP_UNALIAS ("unalias");
string CPP_UNEQUAL ("!=");
string CPP_UNFUNCTION ("unfunction");
string CPP_UNPOINTER ("unpointer");
string CPP_USING ("using");
string CPP_VERBATIM ("verbatim");
string CPP_VIRTUAL ("virtual");
string CPP_VOID ("void");
string CPP_WHILE ("while");
string CPP_XOR ("^^");

static table<generic, generic> op_name;

/******************************************************************************
* Useful symbols and subroutines
******************************************************************************/

inline generic
cAr (const array<generic>& a) {
  if (N(a) == 0)
    return generic ();
  return a[N(a)-1];
}

inline array<generic>
cDr (const array<generic>& a) {
  if (N(a) == 0)
    return a;
  return range (a,0,N(a)-1);
}

static void
add (const generic& var, const generic& val) {
  op_name (var)= val;
  op_name (val)= val;
  op_name (string ("operator ") * val)= val;
}

static void
cpp_init () {
  if (N (op_name) == 0) {
    add (CPP_PLUS, "+");
    add (CPP_MINUS, "-");
    add (CPP_TIMES, "*");
    add (CPP_OVER, "/");
    add (CPP_MOD, "%");
    add (CPP_XOR, "^");
    add (CPP_AND, "&");
    add (CPP_OR, "|");
    add (CPP_NOT, "!");
    add (CPP_ASSIGN, "=");
    add (CPP_LESS, "<");
    add (CPP_GTR, ">");
    add (CPP_PLUS_ASSIGN, "+=");
    add (CPP_MINUS_ASSIGN, "-=");
    add (CPP_TIMES_ASSIGN, "*=");
    add (CPP_OVER_ASSIGN, "/=");
    add (CPP_MOD_ASSIGN, "%=");
    add (CPP_LESSLESS, "<<");
    add (CPP_GTRGTR, ">>");
    add (CPP_EQUAL, "==");
    add (CPP_UNEQUAL, "!=");
    add (CPP_LESSEQ, "<=");
    add (CPP_GTREQ, ">=");
    add (CPP_PREINC, "++");
    add (CPP_PREDEC, "--");
    add (CPP_ACCESS, "[]");
    add (CPP_COMPLEMENT, "~");
    add (CPP_ARROW, "->");
  }
}

inline bool
is_opname (const generic& s) {
  return op_name->contains (s);
}

inline generic
cpp_name (const generic& s) {
  cpp_init ();
  if (op_name->contains (s)) return (op_name[s]);
  return s;
}

inline generic
cpp_opname (const generic& s) {
  cpp_init ();
  if (op_name->contains (s))
    return generic (string ("operator ") *
		   as<string> (op_name[s]));
  return s;
}

/******************************************************************************
* The cpp printer class
******************************************************************************/

class cpp_printer {
private:
  ostream& out;
  list<generic> sl;
  bool     lf_flag;
  bool     lf_done;
  nat      indent;

private:
  void pp_begin (const generic& g);
  void pp_instr (const generic& g);
  void pp_init (const generic& g);


  void pp_decl (const generic& var, bool large= true);
  void pp_decl (const generic& var, const generic& type, bool large);
  void pp_type (const generic& g);
  void pp_funarg (const generic& g);
  void pp_funargs (const generic& g);

  void pp_expr (const generic& g);
  void pp_exprs (const generic& g);
  void pp_assign (const generic& g);

  void pp_cond (const generic& g);
  void pp_seqor (const generic& g);
  void pp_seqand (const generic& g);
  void pp_or (const generic& g);
  void pp_xor (const generic& g);
  void pp_and (const generic& g);
  void pp_eq (const generic& g);
  void pp_rel (const generic& g);

  void pp_shift (const generic& g);
  void pp_lshift (const generic& g);
  void pp_rshift (const generic& g);
  void pp_add (const generic& g);
  void pp_mult (const generic& g);

  void pp_unary (const generic& g);
  void pp_postfix (const generic& g);
  void pp_primary (const generic& g);

  inline string lf (void) {
    string x = "\n";
    for (nat i = 0; i < indent; ++i)
      x << "  ";
    return x;
  }
    
public:
  inline cpp_printer (ostream& out2): out (out2), indent (0) {}
  inline ~cpp_printer () {}

  void pp (const generic& g);
};

/******************************************************************************
* Required routines for generic structures
******************************************************************************/

inline generic
flatten (const cpp_printer& p) {
  return "#cpp_printer<>";
  
}

inline nat
hash (const cpp_printer& p) {
  return 0;
}

/******************************************************************************
* Instructions
******************************************************************************/

void
cpp_printer::pp_instr (const generic& g) {
  if (is_func (g, CPP_VERBATIM, 1) && is<string> (g[1])) {
    string s = as<string> (g[1]);
    int i, n = N(s);
    for (i=0; i<n; i++)
      if (s[i] == '\n') out << lf ();
      else out << s[i];
  }
  else if (is_func (g, CPP_NAMESPACE, 2)) {
    out << "namespace " << g[1] << " {" << lf (); 
    indent ++;
    pp_begin (g[2]);
    indent --;
    out << lf () << "}" << lf ();
  }
  else if (is_func (g, CPP_INITIALIZE)) {
    pp_init (g);
    generic last= cAr (compound_to_array (g));
    if (is_func (last, CPP_BEGIN)) pp_instr (last);
    else pp_instr (gen (CPP_BEGIN, last));
  }
  else if (is_func (g, CPP_BEGIN)) {
    out << "{" << lf ();
    indent ++;
    pp_begin (g);
    indent --;
    out << lf () << "}" << lf ();
  }
  else if (is_func (g, CPP_IF, 2)) {
    out << "if (";
    pp_expr (g[1]);
    out << ") ";
    if (is_func (g[2], CPP_IF)) pp_instr (gen (CPP_BEGIN, g[2]));
    else pp_instr (g[2]);
  }
  else if (is_func (g, CPP_IF, 3)) {
    out << "if (";
    pp_expr (g[1]);
    out << ") ";
    if (is_func (g[2], CPP_IF, 2)) pp_instr (gen (CPP_BEGIN, g[2]));
    else pp_instr (g[2]);
    out << "else ";
    pp_instr (g[3]);
  }
  else if (g == CPP_BREAK)
    out << "break;" << lf ();
  else if (g == CPP_CONTINUE)
    out << "continue;" << lf ();
  else if (g == CPP_RETURN)
    out << "return;" << lf ();
  else if (is_func (g, CPP_RETURN, 1)) {
    out << "return ";
    pp_expr (g[1]);
    out << ";" << lf ();
  }
  else pp_decl (g);
}

void
cpp_printer::pp_init (const generic& g) {
  array<generic> l= cDr (cdr (compound_to_array (g)));
  iterator<generic> i = iterate (l);
  bool first= true;
  for ( ; !busy (i); ++i) {
    if (first) out << ": ";
    else out << ", ";
    pp_expr (*i);
    first= false;
  }
}

void
cpp_printer::pp_begin (const generic& g) {
  if (is_func (g, CPP_BEGIN)) {
    lf_flag= false;
    lf_done= true;
    array<generic> l= cdr (compound_to_array (g));
    iterator<generic> i = iterate (l);
    for ( ; !busy (i); ++i) {
      if (lf_flag) { out << lf (); lf_flag= false; lf_done= true; }
      pp_instr (*i);
      lf_done= false;
    }
  }
  else pp_instr (g);
}

/******************************************************************************
* Declarations
******************************************************************************/

void
cpp_printer::pp_decl (const generic& var, const generic& type, bool large) {
  // prefixes
  assert (is<string> (var));
  if (is_func (type, CPP_CONST, 1)) {
    out << "const ";
    pp_decl (var, type[1], large);
  }
  else if (is_func (type, CPP_EXTERN, 1)) {
    out << "extern ";
    pp_decl (var, type[1], large);
  }
  else if (is_func (type, CPP_EXTERN, 2)) {
    out << "extern " << type[1] << " ";
    pp_decl (var, type[2], large);
  }
  else if (is_func (type, CPP_INLINE, 1)) {
    out << "inline ";
    pp_decl (var, type[1], large);
  }
  else if (is_func (type, CPP_STATIC, 1)) {
    out << "static ";
    pp_decl (var, type[1], large);
  }
  else if (is_func (type, CPP_VIRTUAL, 1)) {
    out << "virtual ";
    pp_decl (var, type[1], large);
  }
  else if (is_func (type, CPP_TEMPLATE)) {
    array<generic> x= cDr (cdr (compound_to_array (type)));
    out << "template<";
    bool first = true;
    iterator<generic> i;
    for (i = iterate (x); busy(i); ++i) {
      if (!first) out << ", ";
      pp_funarg (*i);
      first = false;
    }
    out << "> ";
    pp_decl (var, cAr (compound_to_array (type)), large);
  }
  else if (is_func (type, CPP_CONSTRUCTOR)) {
    array<generic> args = cdr (type);
    pp_decl (array_to_compound (array<generic> ((CPP_UNFUNCTION) << var) * args),
	     generic (""), large);
  }
  else if (is_func (type, CPP_DESTRUCTOR, 0))
    pp_decl (gen (CPP_UNFUNCTION, var), string ("~"), large);
  else if (is_func (type, CPP_INHERIT)) {
    out << "struct " << var << ": ";
    iterator<generic> i;
    bool first = true;
    for ( i = iterate (cdr (compound_to_array (type))); busy(i); ++i) {
      if (!first) out << ", ";
      pp_type (*i);
      first = false;
    }
  }
  else {
    pp_type (type);
    if (type == string (""));
    else if (type == string ("~"));
    else if (large && (type != CPP_CLASS)) out << lf ();
    else out << " ";
    pp_expr (var);
  }
}

void
cpp_printer::pp_decl (const generic& g, bool large) {
  if (is_func (g, CPP_TYPEDEF, 2)) {
    out << "typedef ";
    pp_decl (gen (CPP_TYPE, g[1], g[2]));
  }
  else if (is_func (g, CPP_ASSIGN, 2)) {
    bool large2= false;
    bool emptyb= false;
    if (is_func (g[1], CPP_TYPE, 2)) {
      large2= is_func (g[2], CPP_INITIALIZE);
      large2 |= is_func (g[2], CPP_BEGIN);
      large2 |= is_func (g[2], CPP_STRUCT);
      emptyb |= is_func (g[2], CPP_BEGIN, 0);
      emptyb |= is_func (g[2], CPP_INITIALIZE) &&
	is_func (g[2][N(g[2])-1], CPP_BEGIN, 0);
      if (large && large2 && (!lf_done) && (!emptyb)) out << lf ();
      if (is_func (g[2], CPP_CONSTRUCT)) {
	array<generic> var= cons (g[1][1], cdr (compound_to_array (g[2])));
	if (is_func (g, CPP_ASSIGN, 2))
	  pp_decl (array_to_compound (var), gen (CPP_CONST, g[1][2]), large2);
	else pp_decl (array_to_compound (var), g[1][2], large2);
      }
      else {
	if (is_func (g, CPP_ASSIGN, 2) && (!large2))
	  pp_decl (g[1][1], gen (CPP_CONST, g[1][2]), large2);
	else pp_decl (g[1][1], g[1][2], large2);
      }
    }
    else pp_cond (g[1]);
    if (large && large2) {
      generic h= g[2];
      if (is_func (h, CPP_INITIALIZE)) {
	pp_init (h);
	generic last= cAr (compound_to_array (h));
	if (is_func (last, CPP_BEGIN)) h= last;
	else h= gen (CPP_BEGIN, last);
      }
      if (emptyb) {
	out << " {";
	pp_begin (h);
	out << "}";
	out << lf ();
	lf_flag= false;
      }
      else {
	out << " {" << lf ();
	indent ++;
	pp_begin (h);
	indent --;
	out << "}";
	if (is_func (h, CPP_STRUCT)) out << ";";
	out << lf ();
	lf_flag= true;
      }
    }
    else {
      if (!is_func (g[2], CPP_CONSTRUCT)) {
	out << " = ";
	pp_expr (g[2]);
      }
      if (large) out << ";" << lf ();
    }
  }
  else if (is_func (g, CPP_TYPE, 2)) {
    pp_decl (g[1], g[2], false);
    if (large) out << ";" << lf ();
  }
  else if (is_func (g, CPP_USING, 1)) {
    out << "using ";
    pp_expr (g[1]);
    out << ";" << lf ();
  }
  else {
    pp_expr (g);
    out << ";" << lf ();
  }
}

/******************************************************************************
* Function arguments in function declarations
******************************************************************************/

void
cpp_printer::pp_type (const generic& g) {
  if (is_func (g, CPP_ALIAS, 1)) {
    pp_type (g[1]);
    out << "&";
  }
  else if (is_func (g, CPP_POINTER, 1)) {
    pp_type (g[1]);
    out << "*";
  }
  else if (is_func (g, CPP_SCOPE, 2)) {
    pp_type (g[1]);
    out << "::";
    pp_type (g[2]);
  }
  else if (is_func (g, CPP_CONST, 1)) {
    out << "const ";
    pp_type (g[1]);
  }
  else if (is_func (g, CPP_PUBLIC, 1)) {
    out << "public ";
    pp_type (g[1]);
  }
  else if (is<compound> (g)) {
    array<generic> l= compound_to_array (g);
    generic fun= car (l);
    pp_type (fun);
    l= cdr (l);
    out << "(";
    while (N (l) != 0) {
      pp_type (car (l));
      if (N(cdr (l)) !=0 && is<compound> (car (l)))
	out << " ";
      l= cdr (l);
      if (N (l) !=0) out << ", ";
    }
    out << ")";
  }
  else if (g == CPP_CLASS) out << "struct";
  else if (g == CPP_VOID) out << "void";
  else if (g == CPP_BOOL) out << "bool";
  else if (g == CPP_INT) out << "int";
  else if (g == CPP_STRING) out << "string";
  else if (g == CPP_OSTREAM) out << "ostream";
  else out << g;
}

void
cpp_printer::pp_funarg (const generic& g) {
  if (is_func (g, CPP_TYPE, 2) && (g[2] == CPP_CLASS))
    pp_decl (g[1], string ("class"), false);
  else if (is_func (g, CPP_TYPE, 2))
    pp_decl (g[1], g[2], false);
  else if (is_func (g, CPP_ASSIGN, 2))
    pp_decl (g, false);
  else pp_type (g);
}

void
cpp_printer::pp_funargs (const generic& g) {
  if (N(g) == 1) pp_funarg (car (g));
  if (N(g) <= 1) return;
  pp_funarg (car (g));
  out << ", ";
  pp_funargs (cdr (g));
}

/******************************************************************************
* Expressions
******************************************************************************/

void
cpp_printer::pp_expr (const generic& g) {
  pp_assign (g);
}

void
cpp_printer::pp_exprs (const generic& g) {
  if (N(g) == 1) pp_expr (car (g));
  if (N(g) <= 1) return;
  pp_expr (car (g));
  out << ", ";
  pp_exprs (cdr (g));
}

/******************************************************************************
* Assignments
******************************************************************************/

void
cpp_printer::pp_assign (const generic& g) {
  if (is_func (g, CPP_ASSIGN, 2)) {
    pp_cond (g[1]);
    out << " = ";
    pp_assign (g[2]);
  }
  else if (is_func (g, CPP_PLUS_ASSIGN, 2)) {
    pp_cond (g[1]);
    out << " += ";
    pp_assign (g[2]);
  }
  else if (is_func (g, CPP_MINUS_ASSIGN, 2)) {
    pp_cond (g[1]);
    out << " -= ";
    pp_assign (g[2]);
  }
  else if (is_func (g, CPP_TIMES_ASSIGN, 2)) {
    pp_cond (g[1]);
    out << " *= ";
    pp_assign (g[2]);
  }
  else if (is_func (g, CPP_OVER_ASSIGN, 2)) {
    pp_cond (g[1]);
    out << " /= ";
    pp_assign (g[2]);
  }
  else if (is_func (g, CPP_MOD_ASSIGN, 2)) {
    pp_cond (g[1]);
    out << " %= ";
    pp_assign (g[2]);
  }
  else pp_cond (g);
}

/******************************************************************************
* Conditional expressions
******************************************************************************/

void
cpp_printer::pp_cond (const generic& g) {
  if (is_func (g, CPP_IF, 3)) {
    pp_expr (g[1]);
    out << " ? ";
    pp_expr (g[2]);
    out << " : ";
    pp_expr (g[3]);
  }
  else pp_seqor (g);
}

/******************************************************************************
* Logical expressions
******************************************************************************/

void
cpp_printer::pp_seqor (const generic& g) {
  if (is_func (g, CPP_OR, 2)) {
    pp_seqor (g[1]);
    out << " || ";
    pp_seqor (g[2]);
  }
  else pp_seqand (g);
}

void
cpp_printer::pp_seqand (const generic& g) {
  if (is_func (g, CPP_AND, 2)) {
    pp_seqand (g[1]);
    out << " && ";
    pp_seqand (g[2]);
  }
  else pp_or (g);
}

void
cpp_printer::pp_or (const generic& g) {
  if (is_func (g, CPP_BITWISE_OR, 2)) {
    pp_or (g[1]);
    out << " | ";
    pp_or (g[2]);
  }
  else pp_xor (g);
}

void
cpp_printer::pp_xor (const generic& g) {
  if (is_func (g, CPP_BITWISE_XOR, 2)) {
    pp_xor (g[1]);
    out << " ^ ";
    pp_xor (g[2]);
  }
  else pp_and (g);
}

void
cpp_printer::pp_and (const generic& g) {
  if (is_func (g, CPP_BITWISE_AND, 2)) {
    pp_and (g[1]);
    out << " & ";
    pp_and (g[2]);
  }
  else pp_eq (g);
}

/******************************************************************************
* Relational expressions
******************************************************************************/

void
cpp_printer::pp_eq (const generic& g) {
  if (is_func (g, CPP_EQUAL, 2)) {
    pp_rel (g[1]);
    out << " == ";
    pp_rel (g[2]);
  }
  else if (is_func (g, CPP_UNEQUAL, 2)) {
    pp_rel (g[1]);
    out << " != ";
    pp_rel (g[2]);
  }
  else pp_rel (g);
}

void
cpp_printer::pp_rel (const generic& g) {
  if (is_func (g, CPP_LESS, 2)) {
    pp_shift (g[1]);
    out << " < ";
    pp_shift (g[2]);
  }
  else if (is_func (g, CPP_LESSEQ, 2)) {
    pp_shift (g[1]);
    out << " <= ";
    pp_shift (g[2]);
  }
  else if (is_func (g, CPP_GTR, 2)) {
    pp_shift (g[1]);
    out << " > ";
    pp_shift (g[2]);
  }
  else if (is_func (g, CPP_GTREQ, 2)) {
    pp_shift (g[1]);
    out << " >= ";
    pp_shift (g[2]);
  }
  else pp_shift (g);
}

/******************************************************************************
* Arithmetic expressions
******************************************************************************/

void
cpp_printer::pp_shift (const generic& g) {
  if (is_func (g, CPP_LESSLESS, 2)) {
    pp_lshift (g[1]);
    out << " << ";
    pp_add (g[2]);
  }
  else if (is_func (g, CPP_GTRGTR, 2)) {
    pp_add (g[1]);
    out << " >> ";
    pp_rshift (g[2]);
  }
  else pp_add (g);
}

void
cpp_printer::pp_lshift (const generic& g) {
  if (is_func (g, CPP_LESSLESS, 2)) {
    pp_lshift (g[1]);
    out << " << ";
    pp_add (g[2]);
  }
  else pp_add (g);
}

void
cpp_printer::pp_rshift (const generic& g) {
  if (is_func (g, CPP_GTRGTR, 2)) {
    pp_add (g[1]);
    out << " >> ";
    pp_rshift (g[2]);
  }
  else pp_add (g);
}

void
cpp_printer::pp_add (const generic& g) {
  if (is_func (g, CPP_PLUS, 2)) {
    pp_add (g[1]);
    out << " + ";
    pp_add (g[2]);
  }
  else if (is_func (g, CPP_MINUS, 2)) {
    pp_add (g[1]);
    out << " - ";
    pp_mult (g[2]);
  }
  else pp_mult (g);
}

void
cpp_printer::pp_mult (const generic& g) {
  if (is_func (g, CPP_TIMES, 2)) {
    pp_mult (g[1]);
    out << " * ";
    pp_mult (g[2]);
  }
  else if (is_func (g, CPP_OVER, 2)) {
    pp_mult (g[1]);
    out << " / ";
    pp_unary (g[2]);
  }
  else if (is_func (g, CPP_MOD, 2)) {
    pp_mult (g[1]);
    out << " % ";
    pp_unary (g[2]);
  }
  else pp_unary (g);
}

/******************************************************************************
* Unary, postfix and primary expressions
******************************************************************************/

void
cpp_printer::pp_unary (const generic& g) {
  if (is_func (g, CPP_PREINC, 1)) {
    out << "++";
    pp_unary (g[1]);
  }
  else if (is_func (g, CPP_PREDEC, 1)) {
    out << "--";
    pp_unary (g[1]);
  }
  else if (is_func (g, CPP_UNPOINTER, 1)) {
    out << "*";
    pp_unary (g[1]);
  }
  else if (is_func (g, CPP_UNALIAS, 1)) {
    out << "&";
    pp_unary (g[1]);
  }
  else if (is_func (g, CPP_PLUS, 1)) {
    out << "+";
    pp_unary (g[1]);
  }
  else if (is_func (g, CPP_MINUS, 1)) {
    out << "-";
    pp_unary (g[1]);
  }
  else if (is_func (g, CPP_NOT, 1)) {
    out << "!";
    pp_unary (g[1]);
  }
  else if (is_func (g, CPP_COMPLEMENT, 1)) {
    out << "~";
    pp_unary (g[1]);
  }
  else if (is_func (g, CPP_DELETE, 1)) {
    out << "delete ";
    pp_expr (g[1]);
  }
  else if (is_func (g, CPP_NEW)) {
    out << "new ";
    pp_type (g[1]);
    out << " (";
    pp_exprs (cdr (cdr (g)));
    out << ")";
  }
  else pp_postfix (g);
}

/******************************************************************************
* Postfix expressions
******************************************************************************/

static bool
is_applicable (const generic& g) {
  if (is<string> (g)) {
    string s= as<string> (g);
    if (N(s)==0) return false;
    nat i;
    for (i=0; i<N(s); i++)
      if (((s[i]<'0') || (s[i]>'9')) &&
          ((s[i]<'a') || (s[i]>'z')) &&
          ((s[i]<'A') || (s[i]>'Z')) &&
          ((s[i]<'_') || (s[i]>'_')))
        return false;
    return true;
  }
  return false;
}

void
cpp_printer::pp_postfix (const generic& g) {
  if (is_func (g, CPP_SCOPE, 2)) {
    pp_type (g[1]);
    out << "::";
    pp_postfix (g[2]);
  }
  else if (is_func (g, CPP_DOT, 2)) {
    pp_postfix (g[1]);
    out << ".";
    pp_primary (g[2]);
  }
  else if (is_func (g, CPP_ARROW, 2)) {
    pp_postfix (g[1]);
    out << "->";
    pp_primary (g[2]);
  }
  else if (is_func (g, CPP_POSTINC, 1)) {
    pp_postfix (g[1]);
    out << "++";
  }
  else if (is_func (g, CPP_POSTDEC, 1)) {
    pp_postfix (g[1]);
    out << "--";
  }
  else if (is_func (g, CPP_UNFUNCTION)) {
    pp_postfix (g[1]);
    out << " (";
    pp_funargs (cdr (cdr (g)));
    out << ")";
  }
  else if (is_func (g, CPP_SPECIALIZE)) {
    generic fun= g[1];
    pp_postfix (fun);
    out << " < ";
    pp_funargs (cdr (cdr (g)));
    out << " > ";
  }
  else if (is_func (g, CPP_ACCESS)) {
    pp_postfix (g[1]);
    out << "[";
    pp_exprs (cdr (cdr (g)));
    out << "]";
  }
  else if (((N(g) > 0) && is_applicable (g[0])) ||
	   ((!nil (sl)) && (car (sl) == g))) {
    pp_postfix  (car (g));
    out << " (";
    pp_exprs (cdr (g));
    out << ")";
  }
  else pp_primary (g);
}

/******************************************************************************
* Primary expressions
******************************************************************************/

void
cpp_printer::pp_primary (const generic& g) {
  if (is_opname (g)) out << cpp_opname (g);
  else if (is<string> (g)) out << "\"" << (as<string> (g)) << "\"";
  else if (N(g) == 0) out << g;
  else if (is_func (g, CPP_BRACKETS, 1)) {
    out << "(";
    pp_expr (g[1]);
    out << ")";
  }
  else {
    out << "(";
    sl= cons (g, sl);
    pp_expr (g);
    sl= cdr (sl);
    out << ")";
  }
}

/******************************************************************************
* Interface
******************************************************************************/

void
cpp_printer::pp (const generic& g) {
  pp_begin (g);
}

void
cpp_print (ostream& out, const generic& g) {
  cpp_printer pp (out);
  pp.pp (g);
}

} // namespace mmx
