
/******************************************************************************
* MODULE     : mmx_printer.cpp
* DESCRIPTION: Printing as mathemagix expressions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "generic_internal.hpp"
#include "list.hpp"
#include "literal.hpp"
#include "compound.hpp"
namespace mmx {

/******************************************************************************
* The mmx_printer class
******************************************************************************/

class mmx_printer {
MMX_ALLOCATORS
private:
  ostream&      out;
  list<generic> gl;

private:
  void pp_verb (const string& s);
  void pp_op (const generic& op, bool infix= true);
  void pp_keyw (bool before, const string& s, bool after);

  void pp_where (const generic& g);
  void pp_E0s (const generic& g, nat start);

  void pp_E1 (const generic& g);
  void pp_Lflux (const generic& g);
  void pp_Rflux (const generic& g);

  void pp_F0 (const generic& g);
  void pp_Seqors (const generic& g);
  void pp_Ors (const generic& g);
  void pp_Xors (const generic& g);
  void pp_F1 (const generic& g);
  void pp_Seqands (const generic& g);
  void pp_Ands (const generic& g);
  void pp_F2 (const generic& g);
  void pp_F3 (const generic& g);

  void pp_T0 (const generic& g);
  void pp_T1 (const generic& g);
  void pp_T2 (const generic& g);
  void pp_T3 (const generic& g);
  void pp_T4 (const generic& g);
  void pp_T5 (const generic& g);

  void pp_R0 (const generic& g);
  void pp_R1 (const generic& g);
  void pp_R2 (const generic& g);
  void pp_R3 (const generic& g);

public:
  inline mmx_printer (ostream& out2): out (out2), gl () {}
  inline ~mmx_printer () {}

  void pp_E0 (const generic& g);
};

/******************************************************************************
* Helper routines
******************************************************************************/

void
mmx_printer::pp_verb (const string& s) {
  out << s;
}

void
mmx_printer::pp_op (const generic& op, bool infix) {
  if (infix) out << " ";
  out << literal_to_string (op);
  if (infix) out << " ";
}

void
mmx_printer::pp_keyw (bool before, const string& s, bool after) {
  if (before) out << " ";
  out << s;
  if (after) out << " ";
}

/******************************************************************************
* Tuples
******************************************************************************/

void
mmx_printer::pp_where (const generic& g) {
  pp_E0 (g[1]);
  pp_op (g[0]);
  pp_E0s (g, 2);
}

void
mmx_printer::pp_E0s (const generic& g, nat start) {
  nat n= N (g);
  if (start >= n);
  else if (is_func (g[start], GEN_WHERE) && start==n-1) pp_where (g[start]);
  else if (is_func (g[start], GEN_VWHERE) && start==n-1) pp_where (g[start]);
  else {
    bool vertical= false;
    for (nat i=start; i<n; i++)
      vertical= vertical || is_func (g[i], GEN_ROW);
    for (nat i=start; i<n; i++) {
      if (i != start) {
	if (vertical) pp_verb ("; ");
	else pp_verb (", ");
      }
      pp_E0 (g[i]);
    }
  }
}

/******************************************************************************
* Expressions
******************************************************************************/

void
mmx_printer::pp_E0 (const generic& g) {
  pp_E1 (g);
}

void
mmx_printer::pp_E1 (const generic& g) {
  if (is_func (g, GEN_LESSLESS, 2)) {
    pp_Lflux (g[1]);
    pp_op (g[0]);
    pp_F0 (g[2]);
  }
  else if (is_func (g, GEN_GTRGTR, 2)) {
    pp_F0 (g[1]);
    pp_op (g[0]);
    pp_Rflux (g[2]);
  }
  else pp_F0 (g);
}

void
mmx_printer::pp_Lflux (const generic& g) {
  if (is_func (g, GEN_LESSLESS, 2)) {
    pp_Lflux (g[1]);
    pp_op (g[0]);
    pp_F0 (g[2]);
  }
  else pp_F0 (g);
}

void
mmx_printer::pp_Rflux (const generic& g) {
  if (is_func (g, GEN_GTRGTR, 2)) {
    pp_F0 (g[1]);
    pp_op (g[0]);
    pp_Rflux (g[2]);
  }
  else pp_F0 (g);
}

/******************************************************************************
* Formulas
******************************************************************************/

void
mmx_printer::pp_F0 (const generic& g) {
  if (is_func (g, GEN_IMPLIES, 2) ||
      is_func (g, GEN_EQUIV, 2)) {
    pp_F1 (g[1]);
    pp_op (g[0]);
    pp_F1 (g[2]);
  }
  else pp_F1 (g);
}

void
mmx_printer::pp_F1 (const generic& g) {
  if (is_func (g, GEN_SEQOR, 2)) {
    pp_Seqors (g[1]);
    pp_op (g[0]);
    pp_Seqors (g[2]);
  }
  else if (is_func (g, GEN_OR, 2)) {
    pp_Ors (g[1]);
    pp_op (g[0]);
    pp_Ors (g[2]);
  }
  else if (is_func (g, GEN_XOR, 2)) {
    pp_Xors (g[1]);
    pp_op (g[0]);
    pp_Xors (g[2]);
  }
  else pp_F2 (g);
}

void
mmx_printer::pp_Seqors (const generic& g) {
  if (is_func (g, GEN_SEQOR, 2)) {
    pp_Seqors (g[1]);
    pp_op (g[0]);
    pp_Seqors (g[2]);
  }
  else pp_F2 (g);
}

void
mmx_printer::pp_Ors (const generic& g) {
  if (is_func (g, GEN_OR, 2)) {
    pp_Ors (g[1]);
    pp_op (g[0]);
    pp_Ors (g[2]);
  }
  else pp_F2 (g);
}

void
mmx_printer::pp_Xors (const generic& g) {
  if (is_func (g, GEN_XOR, 2)) {
    pp_Xors (g[1]);
    pp_op (g[0]);
    pp_Xors (g[2]);
  }
  else pp_F2 (g);
}

void
mmx_printer::pp_F2 (const generic& g) {
  if (is_func (g, GEN_AND, 2)) {
    pp_Ands (g[1]);
    pp_op (g[0]);
    pp_Ands (g[2]);
  }
  else if (is_func (g, GEN_SEQAND, 2)) {
    pp_Seqands (g[1]);
    pp_op (g[0]);
    pp_Seqands (g[2]);
  }
  else pp_F3 (g);
}

void
mmx_printer::pp_Seqands (const generic& g) {
  if (is_func (g, GEN_SEQAND, 2)) {
    pp_Seqands (g[1]);
    pp_op (g[0]);
    pp_Seqands (g[2]);
  }
  else pp_F3 (g);
}

void
mmx_printer::pp_Ands (const generic& g) {
  if (is_func (g, GEN_AND, 2) || is_func (g, GEN_SEQAND, 2)) {
    pp_Ands (g[1]);
    pp_op (g[0]);
    pp_Ands (g[2]);
  }
  else pp_F3 (g);
}

void
mmx_printer::pp_F3 (const generic& g) {
  if (is_func (g, GEN_EQUAL, 2) ||
      is_func (g, GEN_UNEQUAL, 2) ||
      is_func (g, GEN_LESS, 2) ||
      is_func (g, GEN_LESSEQ, 2) ||
      is_func (g, GEN_GTR, 2) ||
      is_func (g, GEN_GTREQ, 2)) {
    pp_T0 (g[1]);
    pp_op (g[0]);
    pp_T0 (g[2]);
  }
  else if (is_func (g, GEN_TYPE, 2)) {
    pp_T0 (g[1]);
    pp_verb (": ");
    pp_T0 (g[2]);
  }
  else if (is_func (g, GEN_IN, 2)) {
    if (is_func (g[1], GEN_TYPE, 2)) pp_F3 (g[1]);
    else pp_T0 (g[1]);
    pp_op (g[0]);
    pp_T0 (g[2]);
  }
  else pp_T0 (g);
}

/******************************************************************************
* Terms
******************************************************************************/

void
mmx_printer::pp_T0 (const generic& g) {
  pp_T1 (g);
}

void
mmx_printer::pp_T1 (const generic& g) {
  if (is_func (g, GEN_INTO, 2) ||
      is_func (g, GEN_MAPSTO, 2)) {
    pp_T2 (g[1]);
    pp_op (g[0]);
    pp_T2 (g[2]);
  }
  else pp_T2 (g);
}

void
mmx_printer::pp_T2 (const generic& g) {
  if (is_func (g, GEN_RANGE, 2) ||
      is_func (g, GEN_RANGEEQ, 2)) {
    pp_T3 (g[1]);
    pp_op (g[0]);
    pp_T3 (g[2]);
  }
  else pp_T3 (g);
}

void
mmx_printer::pp_T3 (const generic& g) {
  if (is_func (g, GEN_PLUS, 2)) {
    pp_T3 (g[1]);
    pp_op (g[0]);
    pp_T3 (g[2]);
  }
  else if (is_func (g, GEN_MINUS, 2)) {
    pp_T3 (g[1]);
    pp_op (g[0]);
    pp_T4 (g[2]);
  }
  else pp_T4 (g);
}

void
mmx_printer::pp_T4 (const generic& g) {
  if (is_func (g, GEN_TIMES, 2) ||
      is_func (g, GEN_COMPOSE, 2)) {
    pp_T4 (g[1]);
    pp_op (g[0]);
    pp_T4 (g[2]);
  }
  else if (is_func (g, GEN_OVER, 2)) {
    pp_T4 (g[1]);
    pp_op (g[0]);
    pp_T5 (g[2]);
  }
  else if (is_func (g, GEN_DIV, 2)) {
    pp_T4 (g[1]);
    pp_keyw (true, "div", true);
    pp_T5 (g[2]);
  }
  else if (is_func (g, GEN_MOD, 2)) {
    pp_T4 (g[1]);
    pp_keyw (true, "mod", true);
    pp_T5 (g[2]);
  }
  else pp_T5 (g);
}

void
mmx_printer::pp_T5 (const generic& g) {
  if (is_func (g, GEN_SIZE, 1) ||
      is_func (g, GEN_NOT, 1) ||
      is_func (g, GEN_MINUS, 1)) {
    pp_op (g[0], false);
    pp_R0 (g[1]);
  }
  else if (is<literal> (g) && starts (literal_to_string (g), "-"))
    pp_verb (literal_to_string (g));
  else pp_R0 (g);
}

/******************************************************************************
* Radicals
******************************************************************************/

void
mmx_printer::pp_R0 (const generic& g) {
  pp_R1 (g);
}

void
mmx_printer::pp_R1 (const generic& g) {
  if (is_func (g, GEN_POWER, 2)) {
    pp_R2 (g[1]);
    pp_op (g[0], false); // suppress spaces
    pp_R2 (g[2]);
  }
  else pp_R2 (g);
}

static bool
is_applicable (const generic& g) {
  if (is<literal> (g)) {
    const string s= literal_to_string (g);
    if (N(s)==0) return false;
    for (nat i=0; i<N(s); i++)
      if (((s[i]<'0') || (s[i]>'9')) &&
          ((s[i]<'a') || (s[i]>'z')) &&
          ((s[i]<'A') || (s[i]>'Z')) &&
          ((s[i]!='_') && (s[i]!='?')))
        return false;
  }
  return true;
}

void
mmx_printer::pp_R2 (const generic& g) {
  if (is_func (g, GEN_FACTORIAL, 1)) {
    pp_R2 (g[1]);
    pp_verb ("!");
  }
  else if (is_func (g, GEN_PRIME, 1)) {
    pp_R2 (g[1]);
    pp_op ("'", false);
  }
  else if (is_func (g, GEN_BACKPRIME, 1)) {
    pp_R2 (g[1]);
    pp_op ("`", false);
  }
  else if (is_func (g, GEN_STAR, 1)) {
    pp_R2 (g[1]);
    pp_op ("*", false);
  }
  else if (is_func (g, GEN_HAT, 1)) {
    pp_R2 (g[1]);
    pp_op ("^", false);
  }
  else if (is_func (g, GEN_TILDA, 1)) {
    pp_R2 (g[1]);
    pp_op ("~", false);
  }
  else if (is_func (g, GEN_QUOTE, 1) || is_func (g, GEN_BACKQUOTE, 1)) {
    pp_op (g[0], false);
    if (is<literal> (g[1])) // this test should be further improved
      pp_verb (literal_to_string (g[1]));
    else {
      pp_verb ("(");
      pp_E0 (g[1]);
      pp_verb (")");
    }
  }
  else if (is_func (g, GEN_DOT, 2)) {
    pp_R2 (g[1]);
    pp_verb (".");
    pp_R3 (g[2]);
  }
  else if (is_func (g, GEN_ACCESS)) {
    pp_R2 (g[1]);
    pp_verb ("[");
    pp_E0s (g, 2);
    pp_verb ("]");
  }
  else if ((!is<literal> (g) && is_applicable (g[0])) ||
	   (!nil (gl)) && eq (read_car (gl), g)) {
    pp_R2  (g[0]);
    pp_verb (" (");
    pp_E0s (g, 1);
    pp_verb (")");
  }
  else pp_R3 (g);
}

void
mmx_printer::pp_R3 (const generic& g) {
  if (is<literal> (g) && !starts (literal_to_string (g), "-")) {
    if (eq (g, GEN_CATALAN)) pp_verb ("K");
    else if (eq (g, GEN_DELTA)) pp_verb ("d");
    else if (eq (g, GEN_E)) pp_verb ("e");
    else if (eq (g, GEN_EULER)) pp_verb ("gamma");
    else if (eq (g, GEN_I)) pp_verb ("i");
    else if (eq (g, GEN_PARTIAL)) pp_verb ("d");
    else if (eq (g, GEN_PI)) pp_verb ("pi");
    else if (eq (g, GEN_DERIVATIVE)) pp_verb ("D");
    else pp_verb (literal_to_string (g));
  }
  else if (is_func (g, GEN_SQTUPLE)) {
    pp_verb ("[");
    pp_E0s (g, 1);
    pp_verb ("]");
  }
  else if (is_func (g, GEN_TUPLE)) {
    pp_verb ("(");
    pp_E0s (g, 1);
    pp_verb (")");
  }
  else if (is_func (g, GEN_ROW))
    pp_E0s (g, 1);
  else if (is_func (g, GEN_WHERE) || is_func (g, GEN_VWHERE)) {
    pp_verb ("(");
    pp_where (g);
    pp_verb (")");
  }
  else {
    pp_verb ("(");
    gl= cons (g, gl);
    pp_E0 (g);
    gl= read_cdr (gl);
    pp_verb (")");
  }
}

/******************************************************************************
* Interface
******************************************************************************/

void
mmx_print (ostream& out, const generic& g) {
  mmx_printer pp (out);
  pp.pp_E0 (g);
}

} // namespace mmx
