
/******************************************************************************
* MODULE     : Fast memory allocation
* DESCRIPTION: Fast allocations is realized by using a linked list
*              of allocations for each fixed size divisible by
*              a word legth up to MAX_FAST. Otherwise,
*              usual memory allocation is used.
* ASSUMPTIONS: The word size of the computer is 4.
*              Otherwise, change WORD_LENGTH.
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "basix.hpp"
#include <cstring>
namespace mmx {

#ifdef ENABLE_FAST_ALLOCATORS

#define WORD_LENGTH sizeof(void*)
#define WORD_LENGTH_INC (sizeof(void*)-1)
#define WORD_MASK (-sizeof(void*))
#define LAST_FAST  256    // should be a power of two
#define MAX_FAST   (LAST_FAST+WORD_LENGTH)
#define BLOCK_SIZE 65536  // should be >>> MAX_FAST

char*  alloc_table[MAX_FAST]; // Static declaration initializes with NULL's
char*  alloc_mem=NULL;
size_t alloc_remains=0;
nat    allocated=0;
nat    fast_chunks=0;
nat    large_uses=0;
size_t mmx_used_bytes ();

#define ALLOC_PTR(i) (*((void **) (alloc_table+i)))
#define IND(ptr) (*((void **) ptr))

/******************************************************************************
* General purpose fast allocation routines
******************************************************************************/

void*
safe_malloc (register size_t sz) {
  void* ptr= malloc (sz);
  if (ptr==NULL) {
    cerr << "Fatal error: out of memory\n";
    //cout << 1/0 << "\n";
    exit (1);
  }
  return ptr;
}

void*
enlarge_malloc (register size_t sz) {
  if (alloc_remains<sz) {
    alloc_mem    = (char *) safe_malloc (BLOCK_SIZE);
    alloc_remains= BLOCK_SIZE;
    fast_chunks++;
  }
  register void* ptr= alloc_mem;
  alloc_mem    += sz;
  alloc_remains-= sz;
  return ptr;
}

void*
mmx_malloc (register size_t sz) {
  if (sz == 0) return NULL;
  sz= (sz+WORD_LENGTH_INC)&WORD_MASK;
  if (sz<MAX_FAST) {
    register void *ptr= ALLOC_PTR (sz);
    if (ptr==NULL) return enlarge_malloc (sz);
    ALLOC_PTR (sz)= IND (ptr);
    return ptr;
  }
  else {
    //cout << "Big alloc of " << sz << " bytes\n";
    //cout << "Memory used: " << mmx_used_bytes () << " bytes\n";
    large_uses += sz;
    return safe_malloc (sz);
  }
}

void*
mmx_realloc (register void* old_ptr, register size_t old_sz,
	     register size_t new_sz)
{
  void* new_ptr= mmx_malloc (new_sz);
  memcpy (new_ptr, old_ptr, min (new_sz, old_sz));
  mmx_free (old_ptr, old_sz);
  return new_ptr;
}

void
mmx_free (register void* ptr, register size_t sz) {
  if (sz == 0) return;
  sz=(sz+WORD_LENGTH_INC)&WORD_MASK;
  if (sz<MAX_FAST) {
    IND (ptr)     = ALLOC_PTR (sz);
    ALLOC_PTR (sz)= ptr;
  }
  else {
    large_uses -= sz;    
    free (ptr);
    //cout << "Big free of " << sz << " bytes\n";
    //cout << "Memory used: " << mmx_used_bytes () << " bytes\n";
  }
}

/******************************************************************************
* Statistics
******************************************************************************/

size_t
mmx_compute_free (void* ptr) {
  nat i= (nat) -1;
  while (ptr!=NULL) {
    i++;
    ptr= IND (ptr);
  }
  return i;
}

size_t
mmx_used_bytes () {
  size_t free_bytes= alloc_remains;
  size_t chunks_use= BLOCK_SIZE*fast_chunks;
  for (nat i=WORD_LENGTH; i<MAX_FAST; i+=WORD_LENGTH)
    free_bytes += i*mmx_compute_free (alloc_table+i);
  size_t small_uses= chunks_use- free_bytes;
  return small_uses+ large_uses;
}

void
mmx_mem_info () {
  cout << "\n---------------- memory statistics ----------------\n";
  size_t free_bytes= alloc_remains;
  size_t chunks_use= BLOCK_SIZE*fast_chunks;
  for (nat i=WORD_LENGTH; i<MAX_FAST; i+=WORD_LENGTH)
    free_bytes += i*mmx_compute_free (alloc_table+i);
  size_t small_uses= chunks_use- free_bytes;
  size_t total_uses= small_uses+ large_uses;
  // cout << "Fast chunks   : " << chunks_use << " bytes\n";
  // cout << "Free on chunks: " << alloc_remains << " bytes\n";
  cout << "User          : " << total_uses << " bytes\n";
  cout << "Allocator     : " << chunks_use+ large_uses << " bytes\n";
  cout << "Small mallocs : "
       << ((100*((float) small_uses))/((float) total_uses)) << "%\n";
}

/******************************************************************************
* Memory management
******************************************************************************/

#else // not ENABLE_FAST_ALLOCATORS

static size_t mmx_bytes_in_use= 0;

void*
mmx_malloc (size_t new_size) {
  //if (new_size > 32)
  //cout << "[+" << new_size << "]"; cout.flush ();
  mmx_bytes_in_use += new_size;
  return malloc (new_size);
}

void*
mmx_realloc (void* ptr, size_t old_size, size_t new_size) {
  //if (new_size > 32)
  //cout << "[" << old_size << "->" << new_size << "]"; cout.flush ();
  mmx_bytes_in_use += new_size - old_size;
  return realloc (ptr, new_size);
}

void
mmx_free (void* ptr, size_t old_size) {
  //if (old_size > 32)
  //cout << "[-" << old_size << "]"; cout.flush ();
  mmx_bytes_in_use -= old_size;
  free (ptr);
}

size_t
mmx_used_bytes () {
  return mmx_bytes_in_use;
}

#endif // ENABLE_FAST_ALLOCATORS

} // namespace mmx
