<TeXmacs|1.0.6.9>

<style|<tuple|mmxdoc|mathemagix>>

<\body>
  <tmdoc-title|The <name|Mathemagix> glue mechanism>

  <section|Introduction>

  New functionality can be added to <name|Mathemagix> using the ``glue''
  mechanism. There are two typical types of glue:

  <\enumerate>
    <item>The user has written a new C++ class or a new C++ template class,
    and would like to use this class from withing the interpreter.

    <item>The user has written an interface to an external software which
    comes with its own language and typesystem. In this case, one would like
    to have an automatic mechanism for making the functionality of the
    external software available from within <name|Mathemagix>.
  </enumerate>

  In both cases, the user has to write a ``glue definition routine'', which
  declares the functionality provided by the glue, using a standard API. This
  routine can then be called from some central point (at startup or when
  loading a dynamically linked glue library) in order to make the glue
  functionality available from within the current evaluator.

  <section|Gluing C++ classes>

  <subsection|A simple example>

  Let us first consider writing a glue definition routine for a simple C++
  class named <verbatim|color>. Assume also that the <verbatim|color> class
  comes with two routines which we wish to export to <name|Mathemagix>:

  <\cpp-fragment>
    color named_color (const string& name);

    color invert_color (const color& c);
  </cpp-fragment>

  Then the glue definition routine for <verbatim|color> typically looks as
  follows:

  <\cpp-fragment>
    void

    define_color () {

    \ \ define_type\<less\>always,color\<gtr\> ("Color");

    \ \ define\<less\>always\<gtr\> ("named_color", named_color);

    \ \ define\<less\>always\<gtr\> ("invert", invert_color);

    }
  </cpp-fragment>

  The <verbatim|define_type> instruction exports the type <verbatim|color> to
  <name|Mathemagix> under the name <verbatim|Color>. The two next lines
  export the routines <verbatim|named_color> and <verbatim|invert_color>
  under the same names. The type of the routines <verbatim|named_color> and
  <verbatim|invert_color> is automatically inferred by the gluing mechanism.
  The <verbatim|always> template parameter is not really important for our
  simple example; below, we will see how to use it for the conditional
  definition of glue in the case of template types.

  Whenever a new type is exported to <name|Mathemagix>, a few standard
  routines are required to be implemented for this type. In the case of our
  class <verbatim|color>, the following routines must be provided:

  <\cpp-fragment>
    nat hash (const color& c);

    generic flatten (const color& c);

    bool operator == (const color& c1, const color& c2);

    bool operator != (const color& c1, const color& c2);
  </cpp-fragment>

  The first routine computes a hash value for <verbatim|c>, where
  <verbatim|nat> stands for <verbatim|unsigned int>. The second flattening
  routine converts <verbatim|c> to a generic expression; this routine will be
  used when printing a color. The last two routines implement equality
  testing. Notice that there are several types of equality in
  <name|Mathemagix>. In addition to the <verbatim|==> and <verbatim|!=>
  operators, you may wish to define

  <\cpp-fragment>
    bool hard_eq (const color& c1, const color& c2);

    bool hard_neq (const color& c1, const color& c2);

    bool eq (const color& c1, const color& c2);

    bool neq (const color& c1, const color& c2);
  </cpp-fragment>

  The routine <verbatim|hard_eq> is a fast test whether <verbatim|c1> and
  <verbatim|c2> are represented in the same way in memory. In the case of
  pointer objects like vectors, one typically tests whether the pointers
  match; in particular, two identical vectors which are stored at different
  locations would not be ``hard equal''. The routine <verbatim|eq> tests
  whether <verbatim|c1> and <verbatim|c2> are syntactically equal. Typically,
  the rational number <math|2/1> might be equal to the integer <math|1>,
  without being syntactically equal.

  <subsection|Adding a template class>

  Let us now consider gluing a template class, such as
  <verbatim|list\<less\>C\<gtr\>>. In this case, one typically has to write
  two glue definition routines. The first one is a template which might look
  as follows:

  <\cpp-fragment>
    template\<less\>typename Cond, typename C\<gtr\> void

    define_list () {

    \ \ define_type\<less\>Cond,List \<gtr\> (gen ("List",
    type_name\<less\>C\<gtr\> ()));

    \ \ define\<less\>Cond\<gtr\> ("list", empty_list\<less\>C\<gtr\>);

    \ \ define\<less\>Cond\<gtr\> ("cons", cons\<less\>C\<gtr\>);

    \ \ define\<less\>Cond\<gtr\> ("car", car\<less\>C\<gtr\>);

    \ \ define\<less\>Cond\<gtr\> ("cdr", cdr\<less\>C\<gtr\>);

    \ \ define\<less\>both\<less\>has\<less\>C,add_op\<gtr\>,Cond\<gtr\>
    \<gtr\> ("sum", sum\<less\>C\<gtr\>);

    }
  </cpp-fragment>

  The first template parameter <verbatim|Cond> is used for defining glue
  under certain conditions only. A typical use of this parameter is the last
  routine, which defines the sum of the elements of a list if and only if the
  type <verbatim|C> admits an addition (more explanations on operator classes
  <verbatim|add_op> will be given below). In absence of the condition,
  instanstation of <verbatim|define_list\<less\>C\<gtr\>> for a type
  <verbatim|C> without an addition, would try to instantiate
  <verbatim|sum\<less\>C\<gtr\>>, leading to a compilation error.

  Since C++ only allows us to instantiate the template <verbatim|define_list>
  at compile time, we also need a second glue definition routine
  <verbatim|define_standard_list> which performs the instantiation for a
  small number of frequently used types. This routine might look as follows:

  <\cpp-fragment>
    void

    define_standard_list () {

    \ \ define_list\<less\>always,generic\<gtr\> ();

    \ \ define_list\<less\>always,int\<gtr\> ();

    }
  </cpp-fragment>

  Most template glue definition routines are in particular instantiated for
  the <verbatim|generic> class of symbolic expressions. In addition, it can
  be useful to instantiate for a few other classes for which performance
  might be an issue. Notice that the <verbatim|define_list> template should
  be distributed with the header files of your package: another library might
  which to instantiate the template for additional classes.

  <subsection|A short reference guide>

  The main routines for defining new types and routines are specified in
  <verbatim|glue.hpp>. Here follows a short description:

  <\explain|<\cpp-code>
    void <with|color|blue|define_type>\<less\>Cond,C\<gtr\> (const generic&
    name)
  </cpp-code>>
    This routine conditionally declares a new C++ type <verbatim|C> and
    exports as <verbatim|name>. Whenever a new type is defined, a few other
    derived types are defined as well (see section <reference|derived-types>
    below).
  </explain>

  <\explain|<\cpp-code>
    void <with|color|blue|define_constant>\<less\>Cond,C\<gtr\> (const
    generic& name, const C& x)
  </cpp-code>>
    This routine conditionally exports the constant <cpp-code|C x> as
    <verbatim|name>.
  </explain>

  <\explain|<\cpp-code>
    void <with|color|blue|define_constructor>\<less\>Cond,C\<gtr\> (generic
    (*fun) (const C&))
  </cpp-code>>
    This routine conditionally exports a constructor <verbatim|fun> for
    instances of type <verbatim|C>.
  </explain>

  <\explain|<\cpp-code>
    void <with|color|blue|define>\<less\>Cond,D\<gtr\> (const generic& name,
    D (*fun) ())

    void <with|color|blue|define>\<less\>Cond,D,S1\<gtr\> (const generic&
    name, D (*fun) (const S1&))

    void <with|color|blue|define>\<less\>Cond,D,S1,S2\<gtr\> (const generic&
    name, D (*fun) (const S1&, const S2&))

    ...
  </cpp-code>>
    This routine conditionally exports a function <verbatim|fun> as
    <verbatim|name>.
  </explain>

  <\explain|<\cpp-code>
    void <with|color|blue|define_from_iterator>\<less\>Cond,C\<gtr\> (const
    generic& name)
  </cpp-code>>
    In the case when you implemented a constructor from
    <verbatim|tuple\<less\>C\<gtr\>>, this routine allows you to define
    another constructor with a given <verbatim|name> from
    <verbatim|iterator\<less\>C\<gtr\>>.
  </explain>

  <\explain|<\cpp-code>
    void <with|color|blue|define_caster>\<less\>Cond,D,S\<gtr\>
    (<with|color|brown|[>D (*fun) (const S&) <with|color|brown|[>, nat
    penalty <with|color|brown|]><with|color|brown|]>)

    void <with|color|blue|define_upgrader>\<less\>Cond,D,S\<gtr\>
    (<with|color|brown|[>D (*fun) (const S&) <with|color|brown|[>, nat
    penalty <with|color|brown|]><with|color|brown|]>)

    void <with|color|blue|define_downgrader>\<less\>Cond,D,S\<gtr\>
    (<with|color|brown|[>D (*fun) (const S&) <with|color|brown|[>, nat
    penalty <with|color|brown|]><with|color|brown|]>)

    void <with|color|blue|define_converter>\<less\>Cond,D,S\<gtr\>
    (<with|color|brown|[>D (*fun) (const S&) <with|color|brown|[>, nat
    penalty <with|color|brown|]><with|color|brown|]>)
  </cpp-code>>
    This routine conditionally exports a type conversion routine
    <verbatim|fun: S-\<gtr\>D> (which defaults to the default converter from
    <verbatim|S> to <verbatim|D>). An optional <verbatim|penalty> may be
    specified, which has the effect that conversion chains with the least
    penalty are preferred. The four different forms correspond to whether the
    converters are left and/or right transitive. Upgraders are used for
    automatic constructors (such as <verbatim|integer-\<gtr\>rational>) and
    downgraders for type inheritance (such as
    <verbatim|line_shape-\<gtr\>shape>). Casters cannot be composed with
    other converters and general converters can be composed with anyone else
    (except casters).
  </explain>

  <\explain|<\cpp-code>
    void <with|color|blue|define_primitive>\<less\>Cond\<gtr\> (const
    generic& name, generic (*f) (const generic&))
  </cpp-code>>
    This routine conditionally exports a language primitive <verbatim|f>. The
    argument to the primitive is a tuple which is not evaluated before
    <verbatim|f> is called. The primitive should take care of the possible
    evaluation of its arguments itself.
  </explain>

  <subsection|Implicitly glued classes><label|derived-types>

  Whenever a new type <verbatim|C> is defined by the user, a few other
  related types are added automatically. More precisely, <name|Mathemagix>
  will automatically define the following types:

  <\explain|<\cpp-code>
    alias\<less\>C\<gtr\>
  </cpp-code>>
    This type is used for aliases to instances of type <verbatim|C> (see
    <verbatim|alias.hpp>). The <verbatim|alias\<less\>C\<gtr\>> type plays a
    similar role as the C++ reference type <verbatim|C&>, but there are some
    subtle differences. In <name|Mathemagix>, aliases are much slower, but
    more functional and powerful in nature. More precisely,
    <verbatim|alias\<less\>C\<gtr\>> is an abstract class with two promises
    for read-access and write-access. In particular, an alias need not
    correspond to a physical location in memory. To understand one major
    difference, consider the following session:

    <with|prog-language|mathemagix|prog-session|default|<\session>
      <\input|<with|color|red|Mmx <with|mode|math|\<gg\>>> >
        v: Vector (Generic) := vector (1, 2, 3);
      </input>

      <\output>
        <with|mode|math|<left|[>1,2,3<right|]>>
      </output>

      <\input|<with|color|red|Mmx <with|mode|math|\<gg\>>> >
        x == v[4]; ()
      </input>

      <\input|<with|color|red|Mmx <with|mode|math|\<gg\>>> >
        v := vector (1, 2, 3, 4, 5)
      </input>

      <\output>
        <with|mode|math|<left|[>1,2,3,4,5<right|]>>
      </output>

      <\input|<with|color|red|Mmx <with|mode|math|\<gg\>>> >
        x+1
      </input>

      <\output>
        <with|mode|math|6>
      </output>
    </session>>

    In <name|C++>, the second line would be incorrect. However, in
    <name|Mathemagix>, a read-access for <verbatim|v[4]> is only performed at
    the last line, when computing <verbatim|x+1>. Notice also that
    <name|Mathemagix> performs only a read-access for <verbatim|v[4]>,
    whereas <name|C++> would typically perform a write-access (see
    remark<nbsp><reference|alias-rem> below).
  </explain>

  <\explain|<\cpp-code>
    tuple\<less\>C\<gtr\>
  </cpp-code>>
    This type stands for a tuple of elements of type <verbatim|C>. By
    defining a vector constructor using

    <\cpp-fragment>
      define\<less\>Cond\<gtr\> ("vector", make_vector\<less\>C\<gtr\>);
    </cpp-fragment>

    where

    <\cpp-fragment>
      template\<less\>typename C\<gtr\> vector\<less\>C\<gtr\> make_vector
      (const tuple\<less\>C\<gtr\>& t);
    </cpp-fragment>

    this will allow you to enter vectors in the <name|Mathemagix> interpreter
    using

    <\mmx-fragment>
      vector(a,b,c,d,e,f)
    </mmx-fragment>
  </explain>

  <\explain|<\cpp-code>
    alias\<less\>tuple\<less\>C\<gtr\> \<gtr\>
  </cpp-code>>
    This type is also added, for coherence.
  </explain>

  <\remark>
    <label|alias-rem>Concerning <name|C++> references types, one should
    notice another subtlety when writing accessors for compound data types.
    For instance, consider the methods

    <\cpp-fragment>
      template\<less\>typename C,typename T\<gtr\> C

      table\<less\>C,T\<gtr\>::operator [] (const T& key) const;

      template\<less\>typename C,typename T\<gtr\> C&

      table\<less\>C,T\<gtr\>::operator [] (const T& key);
    </cpp-fragment>

    The overloading will allow for both read-access and write-access using
    the same syntax. However, it sometimes happens that you have a (non
    constant) table which corresponds to a global environment. In that case,
    any access to this table will be a write-access independently if you
    really perform some modifications of the table. This may lead to subtle
    errors if you really wanted to perform a read-access, since a
    write-access might actually modify the table (allocating a non existant
    key-value pair, for instance). This subtlety does not occur for the
    <name|Mathemagix> <verbatim|Alias\<less\>C\<gtr\>> type.
  </remark>

  <subsection|Implicit conversions>

  The routines <verbatim|define_caster>, <verbatim|define_upgrader>,
  <verbatim|define_downgrader> and <verbatim|define_converter> allow the user
  to define automatic converters between different types. This facility is
  quite powerful, but has to be used with care: since automatic converters
  are applied in a very systematic way, they may even be applied in sitations
  which the user did not foresee.

  First of all, the user has to carefully select between the four different
  types of converters: casters, upgraders, downgraders and converters. These
  types differ in the way they may be composed: casters are neither left nor
  right composable, upgraders are right composable, downgraders are left
  composable and converters are both left and right composable. Given a left
  composable converter <verbatim|B-\<gtr\>C> and an arbitrary converter
  <verbatim|A-\<gtr\>B>, <name|Mathemagix> automatically adds a converter
  <verbatim|A-\<gtr\>C> (which is left composable if <verbatim|A-\<gtr\>B> is
  left composable and right composable if <verbatim|B-\<gtr\>C> is right
  composable). Similarly, if <verbatim|A-\<gtr\>B> is right composable and
  <verbatim|B-\<gtr\>C> is arbitrary, then we add a converter
  <verbatim|A-\<gtr\>C>.

  Typically, upgraders correspond to type constructors, such as
  <verbatim|Integer -\<gtr\> Rational> or <verbatim|C <group|-\<gtr\>>
  Matrix(C)>. Similarly, downgraders correspond to type inheritance, such as
  <verbatim|Rectangle -\<gtr\> Shape>, <verbatim|Sum_series(C) -\<gtr\>
  Series(C)>, etc. Casters are often used for converters which may involve
  some loss of data, such as <verbatim|Integer-\<gtr\>Double>. Left and right
  composable converters are rarely needed.

  A second important property of a converter is the correponding penalty:
  when several conversion schemes can be used in order to apply a function to
  some arguments, the scheme with the lowest penalty will be preferred (here
  we notice that the penalty of a conversion <verbatim|(A,B)-\<gtr\>(C,D)> is
  the maximum of the penalties of the conversions <verbatim|A-\<gtr\>C> and
  <verbatim|B-\<gtr\>D>). Among all possible schemes with the lowest penalty,
  the most specialized function will be chosen (a type <verbatim|T> is
  strictly more specialized than <verbatim|U> if there exists a converter
  <group|<verbatim|T-\<gtr\>U>> but no converter <verbatim|U-\<gtr\>T>). If
  no conversion schemes can be found to apply the function, then it will be
  applied symbolically.

  Currently, the following penalties are provided:

  <\description>
    <item*|<verbatim|PENALTY_NONE>>This corresponds to an exact match.

    <item*|<verbatim|PENALTY_AUTOMATIC>>This corresponds to the penalty for
    automatic language-related conversions, such as
    <verbatim|Alias(C)-\<gtr\>C>.

    <item*|<verbatim|PENALTY_INCLUSION>>This penalty should be used for
    conversions <verbatim|T-\<gtr\>U>, where <verbatim|T> may be viewed as a
    mathematical subset of <verbatim|U>. Example:
    <verbatim|Integer-\<gtr\>Rational>. This penalty is the default one for
    conversions <verbatim|T-\<gtr\>U> when <verbatim|T> is different from
    <verbatim|Generic>.

    <item*|<verbatim|PENALTY_HOMOMORPHISM>>This penalty should be used for
    conversions <verbatim|T-\<gtr\>U> which can be viewed as mathematical
    homomorphisms, but not as inclusions. Examples:
    <verbatim|Integer-\<gtr\>Int> or, more generally,
    <verbatim|Integer-\<gtr\>Modular(p)>.

    <item*|<verbatim|PENALTY_VARIANT>>Sometimes, two distinct libraries
    implement the same or a similar type. In that case, it may be interesting
    to provide automatic converters between these types (in both ways). Using
    the higher penalty <verbatim|PENALTY_VARIANT> for this kind of
    conversions will ensure that operations are performed in the library of
    the types of the arguments, unless an implementation is only available in
    the other library.

    <item*|<verbatim|PENALTY_CAST>>This penalty should be used for all
    conversions which, even though convenient for the user, entail some loss
    of information. Example: <verbatim|Integer-\<gtr\>Double>.

    <item*|<verbatim|PENALTY_FALL_BACK>>For any type <verbatim|T>, this is
    the penalty of the conversion <verbatim|T-\<gtr\>Generic>. Hence, if the
    user provides a generic fall back implementation of an operation, then
    this will be the penalty for the application of the fall back method.

    <item*|<verbatim|PENALTY_PROMOTE_GENERIC>>Many composite types, such as
    <verbatim|Complexify(C)>, come with an inclusion
    <verbatim|C-\<gtr\>Complexify(C)>. Although it is generally correct to
    use <verbatim|PENALTY_INCLUSION> for the corresponding penalty, many
    unwanted conversions may arise when <verbatim|C=Generic>. For this
    reason, the default penaly for conversions of the kind
    <verbatim|Generic-\<gtr\>T> is the maximal penalty
    <verbatim|PENALTY_PROMOTE_GENERIC>.
  </description>

  Some common sources of bugs when using overloading and automatic
  conversions are the following:

  <\enumerate>
    <item>You specified an upgrader <verbatim|Generic-\<gtr\>Polynomial(Generic)>,
    but addition on generic polynomials can not be applied so as to add one
    to a polynomial. The point here is that the penalty of the conversion
    <verbatim|Generic-\<gtr\>Polynomial(Generic)> should be the maximal
    penalty <verbatim|PENALTY_PROMOTE_GENERIC>, which is larger than the
    penalty <verbatim|PENALTY_FALL_BACK> for the symbolic addition of
    expressions. The solution to this problem is to implement the following
    two specialized additions:

    <\verbatim>
      \ \ \ \ +: (Polynomial (Generic), Generic) -\<gtr\> Polynomial
      (Generic)

      \ \ \ \ +: (Generic, Polynomial (Generic)) -\<gtr\> Polynomial
      (Generic)
    </verbatim>

    Actually, these operations may be useful for other coefficient types than
    <verbatim|Generic>, since they can usually be implemented in a
    particularly efficient way.

    <item>You forgot to define a generic fall back method for some operation.
    Sometimes, your implementation may rely on the assumption that a given
    operation <verbatim|foo> has no implementation, so that it will be
    applied symbolically. When providing an implementation of <verbatim|foo>
    for some type, this assumption may suddenly be violated and provoke
    infinite loops or incorrect results. In that case, you should provide a
    default symbolic implementation for <verbatim|foo>.
  </enumerate>

  <section|Categorial properties>

  <subsection|Introduction>

  Several common operations, such as addition, multiplication, <abbr|etc.>,
  are exported over and over again by many different types. <name|Mathemagix>
  provides two abstractions to work more efficiently with such common
  operations: operator classes and categories. Categories reflect the usual
  mathematical categories to which different classes belong. For instance,
  <verbatim|integer> both matches the <verbatim|ring_cat> and
  <verbatim|ordering_cat>. These abstractions have the following advantages:

  <\enumerate>
    <item>Similar operations on compound types can be factored in a single
    template which takes an operator class as additional parameter. For
    instance, one might define a template

    <\cpp-fragment>
      template\<less\>typename Op, typename C\<gtr\> vector\<less\>C\<gtr\>

      binary_operation (const vector\<less\>C\<gtr\>& v1, const
      vector\<less\>C\<gtr\>& v2);
    </cpp-fragment>

    for a vectorial binary operation which acts componentwise. This allows us
    to define the addition on vectors as <verbatim|binary_operation\<less\>add_op,C\<gtr\>>.
    Notice that the operator classes may come with additional static methods
    which easy the implementation of templates such as
    <verbatim|binary_operation>.

    <item>When exporting a standard operator to the glue, one may use
    simplified instructions, such as

    <\cpp-fragment>
      define_binary\<less\>always,mul_op,string\<gtr\> ();
    </cpp-fragment>

    Categories allow for similar simplifications:

    <\cpp-fragment>
      define_ring\<less\>always,integer\<gtr\> ();

      define_ordering\<less\>always,integer\<gtr\> ();
    </cpp-fragment>

    <item>Categorial properties can be used for conditional glue definitions:

    <\cpp-fragment>
      define_ordering\<less\> has\<less\>C,ordering_cat\<gtr\>,
      vector\<less\>C\<gtr\> \<gtr\> ();
    </cpp-fragment>

    Notice that a class which matches <verbatim|ring_cat> is assumed to
    correspond to a mathematical ring. For instance, the addition should be
    commutative, etc. This rule is not strict, however, since it is
    convenient to regard the class <verbatim|double> as a ring. Notice also
    that a class which matches both <verbatim|ring_cat> and
    <verbatim|ordering_cat> is not necessarily an ordered ring. In the
    future, we might added categories such as <verbatim|ordered_ring_cat> if
    needed.
  </enumerate>

  <subsection|Operator and category based glue definitions>

  Operator and category based routines for glue definition can be found in
  <verbatim|category_glue.hpp>. Here follows a short description of the
  provided functionality.

  <\explain|<\cpp-code>
    void <with|color|blue|define_unary>\<less\>Cond,C,Op\<gtr\> ()
  </cpp-code>>
    This routine conditionally exports the unary operator <verbatim|Op>,
    considered as a function from <verbatim|C> to <verbatim|C>.
  </explain>

  <\explain|<\cpp-code>
    void <with|color|blue|define_binary>\<less\>Cond,C,Op\<gtr\> ()
  </cpp-code>>
    This routine conditionally exports the binary operator <verbatim|Op>,
    considered as a function from <verbatim|(C,C)> to <verbatim|C>.
  </explain>

  <\explain|<\cpp-code>
    void <with|color|blue|define_test>\<less\>Cond,C,Op\<gtr\> ()
  </cpp-code>>
    This routine conditionally exports the binary test <verbatim|Op>,
    considered as a function from <verbatim|(C,C)> to<nbsp><verbatim|bool>.
  </explain>

  The standard operator types are defined in <verbatim|operator.hpp>. Each
  operator in particular knows about its name; therefore, the above routines
  need not provide an additional name argument. Glue definitions for standard
  categories are all of the form

  <\explain|<\cpp-code>
    void <with|color|blue|define_<with|color|magenta|<em|name>>>\<less\>Cond,C\<gtr\>
    ()
  </cpp-code>>
    This routine conditionally exports all routines from the category
    <verbatim|<with|color|magenta|<em|name>>>.
  </explain>

  The names of standard categories can be found in
  <verbatim|category_glue.hpp>. Some frequently used categories are the
  following:

  <\description>
    <item*|ring>The category of rings, with operations
    <math|<op|+>,<op|\<um\>>,<op|\<times\>>> as well as unary <math|<op|->>.

    <item*|euclidean_ring>The category of euclidean rings, with additional
    operations <math|quo> and <math|rem>.

    <item*|field>An euclidean ring which is actual a field and provides the
    additional operations <math|/>.

    <item*|elementary>A domain on which the elementary functions
    <math|sqrt,pow,exp,log,cos,sin,tan,acos,asin> and <math|atan> are
    defined.

    <item*|ordering>An ordered domain (but not necessarily totally ordered)
    with operations <math|<op|\<leqslant\>,<op|\<less\>>>,<op|\<geqslant\>>,<op|\<gtr\>>>.
  </description>

  <subsection|Definition of categorial meta properties>

  Besides defining conditional glue routines, it is necessary to the define
  the categorial meta-properties of classes separately (it should be possible
  to avoid this double bookkeeping, but this is not easy to program reliably
  in <verbatim|C++>). Categorial meta properties for a type such as
  <verbatim|integer> should be defined in corresponding meta files
  <verbatim|integer_meta.hpp>.

  The file <verbatim|category_meta.hpp> provides several macros which should
  be used for the definition of new meta properties. Any sequence of
  definitions should be enclosed in a block of the following form:

  <\cpp-fragment>
    #define META_TYPE <with|prog-font-shape|italic|<with|color|magenta|current_type_name>>

    #define META_TMPL <with|prog-font-shape|italic|<with|color|magenta|current_template_prefix>>

    #define META_TPNM <with|prog-font-shape|italic|<with|color|magenta|opt_typename_prefix>>

    #define META_COND <with|prog-font-shape|italic|<with|color|magenta|current_condition>>

    <with|prog-font-shape|italic|<with|color|magenta|your_meta_definitions>>

    #undef META_COND

    #undef META_TPNM

    #undef META_TMPL

    #undef META_TYPE
  </cpp-fragment>

  The <verbatim|META_TYPE> macro should refer to the current type to which
  your definitions will apply, <abbr|e.g.> <verbatim|vector\<less\>C\<gtr\>>
  or <verbatim|integer>. The <verbatim|META_TMPL> macro stands for the
  corresponding template prefix, <abbr|e.g.>
  <verbatim|template\<less\>typename C\<gtr\>> in the case of
  <verbatim|vector\<less\>C\<gtr\>> and nothing in the case of
  <verbatim|integer>. The optional typename prefix should be
  <verbatim|typename\<less\>\<gtr\>> in the case of template types such as
  and left empty otherwise. The <verbatim|META_COND> stands for the current
  conditions under which the meta definitions will be valid. In the case when
  different conditions are needed for different parts of the meta
  definitions, <verbatim|META_COND> should be undefined and redefined.

  The following meta definitions are available inside
  <verbatim|#define>-<verbatim|#undef> blocks of the above kind:

  <\explain>
    <\cpp-code>
      <with|color|blue|HAS>(<em|<with|color|magenta|my_op>>)

      <with|color|blue|HAS>(<em|<with|color|magenta|my_cat>>)
    </cpp-code>
  <|explain>
    Declare that the current type has a given operation
    <verbatim|<em|<with|color|magenta|my_op>>> or that it matches a given
    category<nbsp><verbatim|<em|<with|color|magenta|my_cat>>>.
  </explain>

  <\explain>
    <\cpp-code>
      <with|color|blue|HAS_<verbatim|<em|<with|color|magenta|CATEGORY>>>>()
    </cpp-code>
  <|explain>
    Declare that the current type matches a given category <em|and> declare
    the corresponding operations. For instance the command
    <verbatim|HAS_RING()> is equivalent to

    <\cpp-fragment>
      HAS(neg_op)

      HAS(add_op)

      HAS(sub_op)

      HAS(mul_op)

      HAS(ring_cat)
    </cpp-fragment>
  </explain>

  <\explain>
    <\cpp-code>
      <with|color|blue|HAS_<verbatim|<em|<with|color|magenta|CATEGORY>>>_ABOVE_<em|<with|color|magenta|BASE>>>()
    </cpp-code>
  <|explain>
    Similar to <verbatim|HAS_<em|<with|color|magenta|CATEGORY>>>, except that
    the declarations corresponding to <verbatim|<em|<with|color|magenta|BASE>>>
    are omitted.
  </explain>

  <section|An alternative scheme for glue definition>

  So far, we have described how to define glue for new C++ classes or C++
  template classes. When gluing other languages, it may be necessary to have
  a more direct control over the gluing process, by directly calling the glue
  definition methods of the abstract evaluator.

  <subsection|Evaluators>

  Let us start by giving more details on how expressions are evaluated in
  <name|Mathemagix>. The global variable <verbatim|current_ev> contains the
  current evaluator which is used for the evaluation of expressions. All
  semantic operations (such as <math|<op|+>>, <math|<op|->>, <math|exp>,
  <math|derive>, <abbr|etc.>) on instances of the <verbatim|generic> type are
  performed using this current evaluator (some syntactic operations, such as
  <verbatim|eq>, <verbatim|gen>, <abbr|etc.> do not depend on the current
  evaluator). In addition, a <verbatim|generic> expression can be explicitly
  evaluated using <verbatim|eval>.

  The current evaluator itself is an instance of the abstract
  <verbatim|evaluator> class, whose interface is specified in
  <verbatim|evaluator.hpp>. <name|Mathemagix> provides different evaluators:

  <\itemize>
    <item>The trivial evaluator evaluates any expression
    <math|f(x<rsub|1>,\<ldots\>,x<rsub|n>)> to itself.

    <item>The output evaluator acts like the trivial evaluator but performs a
    few additional simplifications. For instance, <math|x+0>, <math|0+x>,
    <math|x\<times\>1>, <math|1\<times\>x> and <math|x<rsup|1>> are all
    evaluated to <math|x>.

    <item>The standard evaluator corresponds to the <verbatim|mmx-light>
    interpreter.
  </itemize>

  New evaluators can be defined by the user, by proving a new concrete
  implementation of the virtual methods of the abstract evaluator.

  Notice that it can be useful to locally change the current evaluator. This
  can be done using the following mechanism:

  <\cpp-fragment>
    select_evaluator (my_evaluator);

    <\with|prog-font-shape|italic>
      <with|color|magenta|instructions_for_my_evaluator>
    </with>

    restore_evaluator ();
  </cpp-fragment>

  For instance, in order to convert an arbitrary <verbatim|generic> to an
  expression <verbatim|generic> which can be printed out, we call the routine
  <verbatim|flatten> using the output evaluator. Notice that the same
  mechaniusm can be used for other expression-based conversions. For
  instance, in order to transform a sparse polynomial into a dense one, one
  might construct a ``dense polynomial evaluator'' which evaluates scalars
  and the indetermine to dense polynomials and implements the ring operations
  on dense polynomials. It then suffices to flatten the sparse polynomial
  using the dense polynomial evaluator.

  <subsection|Glue definition methods of the abstract evaluator>

  <section|Dynamically linked libraries>

  A package with a collection of types, routines and glue definition routines
  should be compiled into a dynamic library, which can then be loaded on the
  fly into the interpreter. Names of <name|Mathemagix> glue libraries should
  be of the form <verbatim|libmmx<with|color|brown|<em|name>>.la> or
  <verbatim|libmmx<with|color|brown|<em|name>>.so> and the principal glue
  definition routine for the library should carry the name
  <verbatim|define_<with|color|brown|<em|name>>>. Notice that <name|C++>
  names are mangled, so you may need to declare
  <verbatim|define_<with|color|brown|<em|name>>> as a <verbatim|void (*) ()>.

  In the subdirectories <verbatim|examples/fibonacci> and
  <verbatim|examples/gaussian>, you may find two simple examples on how to
  glue new code to <name|Mathemagix>. For larger projects, we refer to the
  documentation on coding conventions and how to add new packages.

  <subsection|Computing fibonacci numbers>

  Assume that we want to add a routine <verbatim|fibonacci> for computing
  Fibonacci numbers to the glue. The file <verbatim|fibonnacci.cpp> with the
  routine <verbatim|fibonacci> and the corresponding glue definition routine
  would typically look at follows:

  <\cpp-fragment>
    #include "glue.hpp"

    using namespace mmx;

    \;

    int

    fibonacci (const int& n) {

    \ \ if (n \<less\>= 1) return 1;

    \ \ return fibonacci (n-1) + fibonacci (n-2);

    }

    \;

    void

    define_standard_fibonacci () {

    \ \ define\<less\>always\<gtr\> ("fibonacci", fibonacci);

    }

    \;

    void (*define_fibonacci) () = define_standard_fibonacci;
  </cpp-fragment>

  The very last line is added to prevent name mangling of
  <verbatim|define_fibonacci>. We may now compile the file
  <verbatim|fibonnacci.cpp> into a shared library
  <verbatim|libmmxfibonacci.so>:

  <\shell-fragment>
    g++ --shared `basix-config --cppflags --libs`

    \ \ \ \ fibonacci.cpp -o libmmxfibonacci.so
  </shell-fragment>

  After putting the directory which contains <verbatim|libmmxfibonacci.so> in
  your <verbatim|LD_LIBRARY_PATH>, you may now use the library from within
  <name|Mathemagix>:

  <with|prog-language|mathemagix|prog-session|default|<\session>
    <\input|<with|color|red|Mmx <with|mode|math|\<gg\>>> >
      use "fibonacci"
    </input>

    <\input|<with|color|red|Mmx <with|mode|math|\<gg\>>> >
      fibonacci(37)
    </input>

    <\output>
      <with|mode|math|39088169>
    </output>
  </session>>

  <subsection|Complexified numbers>

  To be completed.

  <section|Exception handling>

  It is possible to catch exceptions occurring in glued C++ routines within
  the <name|Mathemagix> interpreter. In order to make this work, you should
  first configure <name|Mathemagix> using the (default) option
  <verbatim|--enable-exceptions>. Next, your glue code may raise exceptions
  of the type <verbatim|mmx::exception>.

  In fact, it is better not to directly raise exceptions by yourself, but
  rather use the convenience macros defined in <verbatim|basix.hpp>. Indeed,
  <name|Mathemagix> distinguishes two main kind of exceptions, whose default
  behaviour can be configured:

  <\itemize>
    <item>Normal exceptions are the typical exceptions that you want to make
    visible within the interpreter. Since the interpreter is much slower than
    the <name|C++> code anyway, it is a good practice to heavily test for
    erroneous exceptional cases in the glue routines. For instance, when
    making an array access, one should typically test the bounds. Normal
    exceptions are enabled using the default <verbatim|--enable-exceptions>
    configuration option. When disabled, they will be replaced by assertions
    inside the code.

    <item>Low level exceptions are additional checks that you may wish to add
    in critical parts of the<nbsp><name|C++> code. For instance, low level
    <name|C++> array access is performance-critical, since it might be used
    intensively by other <name|C++> routines. When using the
    <verbatim|--enable-verify> configuration option, you may make add
    additional checks for such low level routines. Since this may slow down
    the global performance of the system, this option is disabled by default,
    and should mainly be used for debugging purposes. Instead of directly
    gluing low level routines to the editor, we rather recommend to write a
    small wrapper with the necessary checks for the routine.
  </itemize>

  The two above types of exceptions both come with their corresponding
  macros. The following macros should be used for raising exceptions:

  <\explain|<\cpp-code>
    <with|color|blue|ERROR>(message)
  </cpp-code>>
    Raises the error message <verbatim|message>.
  </explain>

  <\explain|<\cpp-code>
    <with|color|blue|ASSERT>(condition,message)
  </cpp-code>>
    Raises the error message <verbatim|message> if the <verbatim|condition>
    is not satisfied.
  </explain>

  <\explain|<\cpp-code>
    <with|color|blue|VERIFY>(condition,message)
  </cpp-code>>
    When compiling using <verbatim|--enable-verify>, this macro raises the
    error message <verbatim|message> if the <verbatim|condition> is not
    satisfied.
  </explain>

  <section|The <name|Mathemagix> module system>

  Still to be written and documented.
</body>