
/******************************************************************************
* MODULE     : integer.hpp
* DESCRIPTION: Integers
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_INTEGER_HPP
#define __MMX_INTEGER_HPP
#include "mmx_gmp.hpp"
#include "string.hpp"
namespace mmx {

class integer;
class integer_rep;
class rational;
integer copy (const integer& i);

class integer_rep REP_STRUCT {
private:
  mpz_t x;
public:
  inline integer_rep () { mpz_init (x); }
  inline ~integer_rep () { mpz_clear (x); }
  friend class integer;
  friend class rational;
};

class integer {
INDIRECT_PROTO (integer, integer_rep)
public: //private:
  inline const mpz_t& operator * () const { return rep->x; }
  inline mpz_t& operator * () { secure(); return rep->x; }

public:
  // constructors
  inline integer (): rep (new integer_rep ()) {}
  inline integer (signed int i):
    rep (new integer_rep ()) { mpz_set_si (rep->x, i); }
  inline integer (unsigned int i):
    rep (new integer_rep ()) { mpz_set_ui (rep->x, i); }
  inline integer (signed short int i):
    rep (new integer_rep ()) { mpz_set_si (rep->x, i); }
  inline integer (unsigned short int i):
    rep (new integer_rep ()) { mpz_set_ui (rep->x, i); }
  inline integer (signed long int i):
    rep (new integer_rep ()) { mpz_set_si (rep->x, i); }
  inline integer (unsigned long int i):
    rep (new integer_rep ()) { mpz_set_ui (rep->x, i); }
  inline integer (char* s, int base= 0):
    rep (new integer_rep ()) { mpz_set_str (rep->x, s, base); }
  inline integer (const string& s, int base= 0):
    rep (new integer_rep ()) {
      char* r= as_charp (s);
      mpz_set_str (rep->x, r, base);
      free_charp (r); }

  // conversion and input/output
  inline friend int as_int (const integer& x) {
    return (int) mpz_get_si (*x); }
  friend string as_string (const integer& x);

  // basic arithmetic
  inline friend integer copy (const integer& x1) {
    integer r; mpz_set (*r, *x1); return r; }
  inline friend integer operator - (const integer& x1) {
    integer r; mpz_neg (*r, *x1); return r; }
  inline friend integer operator + (const integer& x1, const integer& x2) {
    integer r; mpz_add (*r, *x1, *x2); return r; }
  inline friend integer operator - (const integer& x1, const integer& x2) {
    integer r; mpz_sub (*r, *x1, *x2); return r; }
  inline friend integer operator * (const integer& x1, const integer& x2) {
    integer r; mpz_mul (*r, *x1, *x2); return r; }
  inline friend integer
  operator / (const integer& x1, const integer& x2) {
    ASSERT (mpz_sgn (*x2) != 0, "division by zero");
#ifdef BASIX_ENABLE_VERIFY
    integer r; mpz_fdiv_r (*r, *x1, *x2);
    ASSERT (mpz_sgn (*r) == 0, "unexact division");
    mpz_divexact (*r, *x1, *x2); return r;
#else
    integer r; mpz_divexact (*r, *x1, *x2); return r;
#endif
  }
  inline friend integer quo (const integer& x1, const integer& x2) {
    if (mpz_sgn (*x2) == 0) return x2;
    integer r; mpz_fdiv_q (*r, *x1, *x2); return r; }
  inline friend integer rem (const integer& x1, const integer& x2) {
    if (mpz_sgn (*x2) == 0) return x1;
    integer r; mpz_fdiv_r (*r, *x1, *x2); return r; }
  inline friend integer operator << (const integer& x1, long int shift) {
    integer r; mpz_mul_2si (*r, *x1, shift); return r; }
  inline friend integer operator >> (const integer& x1, long int shift) {
    integer r; mpz_mul_2si (*r, *x1, -shift); return r; }

  // in-place arithmetic
  inline friend integer& operator += (integer& x1, const integer& x2) {
    mpz_add (*x1, *x1, *x2); return x1; }
  inline friend integer& operator -= (integer& x1, const integer& x2) {
    mpz_sub (*x1, *x1, *x2); return x1; }
  inline friend integer& operator *= (integer& x1, const integer& x2) {
    mpz_mul (*x1, *x1, *x2); return x1; }
  inline friend integer& operator /= (integer& x1, const integer& x2) {
    return x1= x1 / x2; }
  inline friend integer& operator <<= (integer& x1, long int shift) {
    return x1= x1 << shift; }
  inline friend integer& operator >>= (integer& x1, long int shift) {
    return x1= x1 >> shift; }

  // accelerated arithmetic for int operands
  inline friend integer operator + (const integer& x1, const int& x2) {
    integer r; mpz_add_si (*r, *x1, x2); return r; }
  inline friend integer operator + (const int& x1, const integer& x2) {
    integer r; mpz_add_si (*r, *x2, x1); return r; }
  inline friend integer operator - (const integer& x1, const int& x2) {
    integer r; mpz_sub_si (*r, *x1, x2); return r; }
  inline friend integer operator - (const int& x1, const integer& x2) {
    integer r; mpz_si_sub (*r, x1, *x2); return r; }
  inline friend integer operator * (const integer& x1, const int& x2) {
    integer r; mpz_mul_si (*r, *x1, x2); return r; }
  inline friend integer operator * (const int& x1, const integer& x2) {
    integer r; mpz_mul_si (*r, *x2, x1); return r; }
  inline friend integer operator / (const integer& x1, const int& x2) {
    return x1 / integer (x2); }
  inline friend integer operator / (const int& x1, const integer& x2) {
    return integer (x1) / x2; }

  // comparing integer numbers
  inline friend bool operator == (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) == 0; }
  inline friend bool operator != (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) != 0; }
  inline friend bool operator < (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) < 0; }
  inline friend bool operator > (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) > 0; }
  inline friend bool operator <= (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) <= 0; }
  inline friend bool operator >= (const integer& x1, const integer& x2) {
    return mpz_cmp (*x1, *x2) >= 0; }
  inline friend int sign (const integer& x1) {
    int r= mpz_sgn (*x1); return (r<0? -1: (r==0? 0: 1)); }
  inline friend integer abs (const integer& x1) {
    integer r; mpz_abs (*r, *x1); return r; }
  inline friend integer min (const integer& x1, const integer& x2) {
    return x1 <= x2? x1: x2; }
  inline friend integer max (const integer& x1, const integer& x2) {
    return x1 >= x2? x1: x2; }

  // bit operations
  inline friend integer operator & (const integer& x1, const integer& x2) {
    integer r; mpz_and (*r, *x1, *x2); return r; }
  inline friend integer operator | (const integer& x1, const integer& x2) {
    integer r; mpz_ior (*r, *x1, *x2); return r; }
  inline friend integer operator ^ (const integer& x1, const integer& x2) {
    integer r; mpz_xor (*r, *x1, *x2); return r; }
  inline friend integer operator ~ (const integer& x1) {
    integer r; mpz_com (*r, *x1); return r; }
  inline bool operator [] (nat index) const {
    return (bool) mpz_tstbit (rep->x, index); }
  inline friend nat hamming_norm (const integer& x1) {
    return mpz_popcount (*x1); }
  inline friend nat hamming_distance (const integer& x1, const integer& x2) {
    return mpz_hamdist (*x1, *x2); }

  // other functions on integers
  inline friend bool is_square (const integer& a) {
    return mpz_perfect_square_p (*a); }
  inline friend integer sqrt (const integer& a) {
    VERIFY (is_square (a), "not a perfect square");
    integer r; mpz_sqrt (*r, *a); return r; }
  inline friend integer pow (const integer& x1, const int& i) {
    ASSERT (i >= 0, "positive int expected");
    integer r; mpz_pow_ui (*r, *x1, (nat) i); return r; }
  inline friend integer pow (const integer& x1, const nat& i) {
    integer r; mpz_pow_ui (*r, *x1, i); return r; }
  inline friend integer gcd (const integer& a, const integer& b) {
    integer r; mpz_gcd (*r, *a, *b); return r; }
  inline friend integer lcm (const integer& a, const integer& b) {
    integer r; mpz_lcm (*r, *a, *b); return r; }
  inline friend xgcd_matrix<integer> xgcd(const integer& a, const integer& b) {
    integer g; xgcd_matrix<integer> r;
    mpz_gcdext (*g, *(r.a), *(r.b), *a, *b);
    r.c= -b/g; r.d= a/g; return r; }
  inline friend integer invert_modulo (const integer& a, const integer& b) {
    integer r; mpz_invert (*r, *a, *b); return r; }
  inline friend int jacobi (const integer& a, const integer& b) {
    ASSERT ((b&1) != 0, "second argument must be odd");
    return mpz_jacobi (*a, *b); }
  inline friend int legendre (const integer& a, const integer& b) {
    ASSERT ((b&1) != 0, "second argument must be odd");
    return mpz_legendre (*a, *b); }
  inline friend int kronecker (const integer& a, const integer& b) {
    return mpz_kronecker (*a, *b); }
  inline friend integer factorial (nat i) {
    integer r; mpz_fac_ui (*r, i); return r; }
  inline friend integer binomial (nat i, nat j) {
    integer r; mpz_bin_uiui (*r, i, j); return r; }

  // miscellaneous functions
  inline friend nat hash (const integer& x) {
    const __mpz_struct& rep= (*x)[0];
    if (rep._mp_size == 0) return 0;
    return ((nat) (rep._mp_d[0])) ^ ((nat) (rep._mp_size << 11));
  }
  inline friend nat size (const integer& x, nat base) {
    return mpz_sizeinbase (*x, base); }
  inline friend nat bit_size (const integer& x) {
    return mpz_sizeinbase (*x, 2); }
  inline friend nat limb_size (const integer& x) {
    return mpz_size (*x); }
  
  friend class rational;
  friend integer numerator (const rational& x1);
  friend integer denominator (const rational& x1);
  template<typename V> friend class floating;
  template<typename V> friend integer as_integer (const floating<V>& x);
};
INDIRECT_IMPL (integer, integer_rep)

generic construct (const integer& i);
generic flatten (const integer& i);

TRUE_TO_SOFT_IDENTITY_SUGAR(,integer)
PRINT_SUGAR(,integer)
POOR_MAN_ELEMENTARY_SUGAR(,integer)

}
#endif // __MMX_INTEGER_HPP
