
/******************************************************************************
* MODULE     : gmp_kronecker.cpp
* DESCRIPTION: Breaking big integers into parts and recomposing them
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "mmx_gmp.hpp"
namespace mmx {

#define ASSERT_RANGE(cond) assert(cond)
//#define ASSERT_RANGE(cond) VERIFY(cond,"limb out of range")

inline void
mpn_set (mp_limb_t* dest, const mp_limb_t* src, mp_size_t size) {
  mp_size_t i;
  for (i=0; i<size; i++) dest[i]= src[i];
}

/******************************************************************************
* Encoding a polynomial of integers using Kronecker's method
******************************************************************************/

inline void
mpn_lshift_bis (mp_limb_t* dest, mp_size_t dest_size,
		const mp_limb_t* src , mp_size_t src_size,
		unsigned int count)
{
  assert (dest_size >= src_size && dest_size > 0);
  if (dest_size == src_size) {
    if (count == 0) mpn_set (dest, src, src_size);
    else (void) mpn_lshift (dest, src, src_size, count);
  }
  else {
    mp_size_t i;
    if (src_size == 0) dest[src_size]= 0;
    else {
      if (count == 0) {
	mpn_set (dest, src, src_size);
	dest[src_size]= 0;
      }
      else dest[src_size]= mpn_lshift (dest, src, src_size, count);
    }
    for (i=src_size+1; i<dest_size; i++) dest[i]= 0;
  }
}

void
mpz_kronecker_encode (mpz_t dest,
		      mpz_t* src, unsigned long n, unsigned long bits)
{
  // src is an array of length n of integers which fit into bits-1 bits.
  // We set dest := src[0]+ src[1] 2^{bits} + ... + src[n-1] 2^{(n-1) bits}.

  unsigned long i;
  mp_size_t dest_size;
  mp_size_t cleaned;
  mpz_t aux;
  int dest_sgn;

  while (n>0 && mpz_sgn (src[n-1]) == 0) n--;
  mpz_set_si (dest, 0);
  if (n == 0) return;
  dest_sgn= mpz_sgn (src[n-1]);

  mpz_realloc2 (dest, n * bits);
  dest_size  = dest->_mp_alloc;
  cleaned= dest_size;
  //cout << "dest_size= " << dest_size << "\n";

  mpz_init2 (aux, bits + BITS_PER_LIMB);
  for (i=n-1; i != ((unsigned long) -1); i--) {
    mp_size_t    j;
    mp_size_t    bit     = i * bits;
    unsigned int shift   = bit & (BITS_PER_LIMB-1);
    mp_size_t    offset  = bit / BITS_PER_LIMB;
    mp_size_t    extra   = cleaned - offset;
    mp_size_t    aux_size= aux->_mp_alloc;
    mp_size_t    src_ssz = src[i]->_mp_size;
    mp_size_t    src_size= (src_ssz>0? src_ssz: -src_ssz);
    int          src_sgn = (src_ssz>=0? 1: -1);
    if (aux_size + offset > dest_size) aux_size--;
    /*
    cout << "Step " << i << "\n";
    cout << "  bit     = " << bit << "\n";
    cout << "  shift   = " << shift << "\n";
    cout << "  offset  = " << offset << "\n";
    cout << "  extra   = " << extra << "\n";
    cout << "  aux_size= " << aux_size << "\n";
    cout << "  src_size= " << src_size << "\n";
    cout << "  dest_sgn= " << dest_sgn << "\n";
    cout << "  src_sgn = " << src_sgn << "\n";
    */

    ASSERT_RANGE (aux->_mp_alloc >= aux_size);
    mpn_lshift_bis (aux->_mp_d, aux_size, src[i]->_mp_d, src_size, shift);
    if (src_sgn == dest_sgn) {
      for (j=0; j<extra; j++)
	dest->_mp_d[j+offset]= aux->_mp_d[j];
      if (aux_size > extra) {
	ASSERT_RANGE (dest->_mp_alloc >= cleaned + aux_size - extra);
	mpn_add_n (dest->_mp_d + cleaned, dest->_mp_d + cleaned,
		   aux->_mp_d + extra, aux_size - extra);
      }
    }
    else {
      ASSERT_RANGE (dest->_mp_alloc >= offset + extra);
      for (j=0; j<extra; j++)
	dest->_mp_d[j+offset]= 0;
      ASSERT_RANGE (dest->_mp_alloc >= dest_size);
      mpn_sub (dest->_mp_d + offset,
	       dest->_mp_d + offset, dest_size - offset,
	       aux->_mp_d, aux_size);
    }
    cleaned= offset;
  }
  mpz_clear (aux);

  while (dest->_mp_d[dest_size-1] == 0) dest_size--;
  dest->_mp_size= (dest_sgn>0? dest_size: -dest_size);
}

/******************************************************************************
* Decoding a polynomial of integers using Kronecker's method
******************************************************************************/

inline void
mpn_rshift_bis (mp_limb_t* dest, mp_size_t dest_size,
		const mp_limb_t* src , mp_size_t src_size,
		unsigned int count)
{
  assert (dest_size > 0);
  if (src_size >= dest_size) {
    if (count == 0) mpn_set (dest, src, dest_size);
    else (void) mpn_rshift (dest, src, dest_size, count);
  }
  else {
    mp_size_t i;
    if (src_size != 0) {
      if (count == 0) mpn_set (dest, src, src_size);
      else mpn_rshift (dest, src, src_size, count);
    }
    for (i=src_size; i<dest_size; i++) dest[i]= 0;
  }
}

void
mpz_kronecker_decode (mpz_t* dest, unsigned long n, unsigned long bits,
                      const mpz_t src_bis)
{
  // let src be an integer of the form a_0 + ... + a_{n-1} 2^{(n-1) bits},
  // where the a_i are positive or negative integers of at most bits-1 bits.
  // On exit, we have dest[i] := a_i for each i.

  unsigned long i, src_bit_size;
  mp_size_t src_size, aux_size;
  int src_sgn;
  mpz_t src, aux, tmp;
  mp_limb_t mask= (((mp_limb_t) 1) << (bits & (BITS_PER_LIMB-1))) - 1;
  if (mask == 0) mask= (mp_limb_t) -1;
  mpz_init2 (src, (mpz_size (src_bis) + 1) * BITS_PER_LIMB);
  mpz_set (src, src_bis);
  ASSERT_RANGE (((long int) src->_mp_alloc) > ((long int) mpz_size (src)));
  src->_mp_d[mpz_size (src)]= 0;

  /* handle trailing zeros */
  if (n == 0) return;
  src_bit_size= mpz_sizeinbase (src, 2);
  if (src_bit_size < (n-1) * bits) {
    unsigned long new_n= (src_bit_size / bits) + 1;
    for (i= new_n; i<n; i++) mpz_set_si (dest[i], 0);
    n= new_n;
  }
  ASSERT_RANGE (src_bit_size < n*bits);
  ASSERT_RANGE (src_bit_size >= (n-1) * bits);

  src_size= mpz_size (src) + 1;
  src_sgn= mpz_sgn (src);
  mpz_init2 (aux, bits + BITS_PER_LIMB);
  aux_size= aux->_mp_alloc;
  mpz_init2 (tmp, bits + BITS_PER_LIMB);
  mpz_set_si (tmp, 1);
  mpz_mul_2exp (tmp, tmp, bits);

  /*
  cout << "mask\t= " << mask << "\n";
  cout << "n\t= " << n << "\n";
  cout << "src_sz\t= " << src_size << "\n";
  cout << "src_sgn\t= " << src_sgn << "\n";
  cout << "aux_sz\t= " << aux_size << "\n";
  */

  for (i=0; i<n; i++) {
    int          dest_sgn = 1;
    mp_size_t    bit      = i * bits;
    unsigned int shift    = bit & (BITS_PER_LIMB-1);
    mp_size_t    offset   = bit / BITS_PER_LIMB;

    /*
    cout << "Stage " << i << "\n";
    cout << "  bit\t= " << bit << "\n";
    cout << "  total\t= " << src_bit_size << "\n";
    cout << "  shift\t= " << shift << "\n";
    cout << "  off\t= " << offset << "\n";
    */

    /* put the raw next coefficient in aux */
    ASSERT_RANGE (aux->_mp_alloc >= aux_size);
    mpn_rshift_bis (aux->_mp_d, aux_size,
		    src->_mp_d+offset, src_size-offset, shift);
    ASSERT_RANGE (aux->_mp_alloc >= aux_size);
    aux->_mp_d[aux_size-2] &= mask;
    aux->_mp_size= aux_size - 1;
    while (aux->_mp_size > 0 && aux->_mp_d[aux->_mp_size-1] == 0)
      aux->_mp_size--;

    /* check whether the next coefficient has a negative sign */
    if (i != n-1) {
      unsigned int k= (bits-1) & (BITS_PER_LIMB-1);
      mp_size_t    l= (bits-1) / BITS_PER_LIMB;
      dest_sgn= ((aux->_mp_d[l] & (1<<k)) == 0? 1: -1);
      if (dest_sgn < 0) {
	/* in-place update of src in case of a negative sign */
	mp_size_t    next_bit   = (i+1) * bits;
	unsigned int next_shift = next_bit & (BITS_PER_LIMB-1);
	mp_size_t    next_offset= next_bit / BITS_PER_LIMB;
	mp_limb_t    plus       = ((mp_limb_t) 1) << next_shift;
	ASSERT_RANGE (src->_mp_alloc >= src_size);
	if (mpn_add (src->_mp_d + next_offset,
		     src->_mp_d + next_offset, src_size - next_offset,
		     &plus, 1) != 0)
	  ASSERT (false, "unexpected carry (mpz_kronecker_decode)");
      }
      //cout << "  sgn\t= " << dest_sgn << "\n";
    }

    /* put the sign-corrected next coefficient in dest[i] */
    if (dest_sgn >= 0) mpz_set (dest[i], aux);
    else mpz_sub (dest[i], tmp, aux);
    if (dest_sgn != src_sgn) dest[i]->_mp_size= -dest[i]->_mp_size;
  }

  mpz_clear (src);
  mpz_clear (aux);
  mpz_clear (tmp);
}

} // namespace mmx
