
/******************************************************************************
* MODULE     : integer_kronecker.cpp
* DESCRIPTION: Breaking big integers into parts and recomposing them
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "jet_integer.hpp"
//#include "basix_utils.hpp"
namespace mmx {

/******************************************************************************
* Naive encoding and decoding of integer polynomials using Kronecker's method
******************************************************************************/

void
kronecker_encode_naive (integer& dest, const integer* src, nat n, nat bits) {
  if (n == 0) dest= 0;
  else if (n == 1) dest= src[0];
  else {
    nat h= n>>1;
    integer aux;
    kronecker_encode_naive (aux , src+h, n-h, bits);
    kronecker_encode_naive (dest, src  , h  , bits);
    dest += (aux << (h*bits));
  }
}

void
kronecker_decode_naive (integer* dest, nat n, nat bits, const integer& src) {
  if (n == 0);
  else if (n == 1) dest[0]= src;
  else if (src > 0) {
    nat h= n>>1;
    integer aux= src >> (h*bits);
    if (src[h*bits-1]) aux += 1;
    kronecker_decode_naive (dest+h, n-h, bits, aux);
    aux= src - (aux << (h*bits));
    kronecker_decode_naive (dest, h, bits, aux);
  }
  else {
    integer bis= -src;
    nat h= n>>1;
    integer aux= bis >> (h*bits);
    if (bis[h*bits-1]) aux += 1;
    kronecker_decode_naive (dest+h, n-h, bits, -aux);
    aux= bis - (aux << (h*bits));
    kronecker_decode_naive (dest, h, bits, -aux);
  }
}

/******************************************************************************
* Integer wrappers for Kronecker's encoding of integer polynomials
******************************************************************************/

void
kronecker_encode (integer& dest, const integer* src, nat n, nat bits) {
  nat i;
  mpz_t* Src= mmx_new<mpz_t> (n);
  for (i=0; i<n; i++) {
    const mpz_t& aux= *(src[i]);
    Src[i]->_mp_alloc= aux->_mp_alloc;
    Src[i]->_mp_size = aux->_mp_size;
    Src[i]->_mp_d    = aux->_mp_d;
  }
  mpz_kronecker_encode (*dest, Src, n, bits);
  mmx_delete<mpz_t> (Src, n);
}

void
kronecker_decode (integer* dest, nat n, nat bits, const integer& src) {
  nat i;
  mpz_t* Dest= mmx_new<mpz_t> (n);
  for (i=0; i<n; i++) mpz_init (Dest[i]);
  mpz_kronecker_decode (Dest, n, bits, *src);
  for (i=0; i<n; i++) {
    mpz_t& aux= *(dest[i]);
    aux->_mp_alloc= Dest[i]->_mp_alloc;
    aux->_mp_size = Dest[i]->_mp_size;
    aux->_mp_d    = Dest[i]->_mp_d;
  }
  mmx_delete<mpz_t> (Dest, n);
}

/******************************************************************************
* Multiply two integer polynomials stored in segments using Kronecker's method
******************************************************************************/

nat
max_bit_size (const integer* src, nat n) {
  nat m= 0;
  for (nat i=0; i<n; i++)
    m= max (m, bit_size (src[i]));
  return m;
}

void
kronecker_mul (integer* dest,
	       const integer* src1, nat n1,
	       const integer* src2, nat n2)
{
  /*
  cout << "n1= " << n1 << ", " << "n2= " << n2 << "\n";
  for (nat i=0; i<n1; i++)
    cout << "  s" << i << "\t" << src1[i] << "\n";
  for (nat i=0; i<n2; i++)
    cout << "  t" << i << "\t" << src2[i] << "\n";
  */

  while (n1 > 0 && src1[n1-1] == 0) { if (n2>0) dest[n1+n2-2]= 0; n1--; }
  while (n2 > 0 && src2[n2-1] == 0) { if (n1>0) dest[n1+n2-2]= 0; n2--; }
  if (n1 == 0 || n2 == 0) {
    if (n1 + n2 > 1)
      for (nat i=0; i<n1+n2-1; i++)
	dest[i]= 0;
    return;
  }

  nat bits1= max_bit_size (src1, n1);
  nat bits2= max_bit_size (src2, n2);
  nat bits= bits1 + bits2 + bit_size (integer (min (n1, n2))) + 1;

  integer aux1, aux2;
  //cout << "Encoding\n";
  //nat start= mmx_user_time ();
  kronecker_encode (aux1, src1, n1, bits);
  kronecker_encode (aux2, src2, n2, bits);
  //cout << "Done in " << mmx_user_time () - start << "ms\n";
  //cout << "Multiplying\n";
  //start= mmx_user_time ();
  integer aux= aux1*aux2;
  //cout << "Done in " << mmx_user_time () - start << "ms\n";
  //cout << "Decoding\n";
  //start= mmx_user_time ();
  kronecker_decode (dest, n1+n2-1, bits, aux);
  //cout << "Done in " << mmx_user_time () - start << "ms\n";

  /*
  integer* dest1= mmx_new<integer> (n1+n2-1);
  kronecker_decode_naive (dest1, n1+n2-1, bits, aux);
  kronecker_decode (dest, n1+n2-1, bits, aux);
  for (nat i=0; i<n1+n2-1; i++) {
    cout << "naive\t" << i << "\t" << dest1[i] << "\n";
    cout << "fast\t" << i << "\t" << dest[i] << "\n";
  }
  mmx_delete<integer> (dest1, n1+n2-1);
  */
}

} // namespace mmx
