
/******************************************************************************
* MODULE     : rational_kronecker.cpp
* DESCRIPTION: Fast multiplication of polynomials with rational coefficients
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "jet_rational.hpp"
namespace mmx {

/******************************************************************************
* Putting an array of rationals under a common denominator
******************************************************************************/

integer
least_common_denominator (const rational* src, nat n) {
  switch (n) {
  case 0: return 1;
  case 1: return denominator (src[0]);
  case 2: return lcm (denominator (src[0]), denominator (src[1]));
  case 3: return lcm (denominator (src[0]), lcm (denominator (src[1]),
						 denominator (src[2])));
  default: return lcm (least_common_denominator (src, n>>1),
		       least_common_denominator (src + (n>>1), n - (n>>1)));
  }
}

void
denominator_mul (integer* dest, const rational* src, nat n, const integer& d) {
  for (nat i=0; i<n; i++)
    dest[i]= numerator (src[i]) * (d / denominator (src[i]));
}

void
denominator_div (rational* dest, const integer* src, nat n, const integer& d) {
  for (nat i=0; i<n; i++)
    dest[i]= rational (src[i]) / d;
}

/******************************************************************************
* Multiply two polynomials with rational coefficients using Kronecker's method
******************************************************************************/

void
kronecker_mul (rational* dest,
	       const rational* src1, nat n1,
	       const rational* src2, nat n2)
{
  integer  d1  = least_common_denominator (src1, n1);
  integer  d2  = least_common_denominator (src2, n2);
  /*
  cout << "d1= " << d1 << ", " << "d2= " << d2 << "\n";
  for (nat i=0; i<n1; i++)
    cout << "  s" << i << "\t" << src1[i] << "\n";
  for (nat i=0; i<n2; i++)
    cout << "  t" << i << "\t" << src2[i] << "\n";
  */
  integer* Src1= mmx_new<integer> (n1);
  integer* Src2= mmx_new<integer> (n2);
  integer* Dest= mmx_new<integer> (n1+n2-1);
  denominator_mul (Src1, src1, n1, d1);
  denominator_mul (Src2, src2, n2, d2);
  kronecker_mul (Dest, Src1, n1, Src2, n2);
  denominator_div (dest, Dest, n1+n2-1, d1*d2);
  mmx_delete<integer> (Src1, n1);
  mmx_delete<integer> (Src2, n2);
  mmx_delete<integer> (Dest, n1+n2-1);
}

} // namespace mmx
