
/******************************************************************************
* MODULE     : jet_floating.hpp
* DESCRIPTION: segments of floating point numbers
* COPYRIGHT  : (C) 2007  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_JET_FLOATING_HPP
#define __MMX_JET_FLOATING_HPP
#include "floating.hpp"
#include "jet_integer.hpp"
namespace mmx {
#define TMPL template<typename V>
#define Floating floating<V>
#define Rnd V::rnd()

template<typename C> void
newton_edge (const C* src, nat n, nat pos, nat& i, nat& j) {
  // Find the edge of the Newton polygon of [src, n] at position pos.
  if (n<3) {
    if (n == 2) { i=0; j=1; }
    else i=j=0;
  }
  else {
    i=0; j=1;
    for (nat k=2; k<n; k++) {
      if (C (k - j) * (src[j] - src[i]) < C (j - i) * (src[k] - src[j])) j=k;
      else if (j<pos) { i=j; j=k; }
    }
  }
}

template<typename C> void
newton_slope (const C* src, nat n, nat pos) {
  // Find the slope of the the Newton polygon of [src, n] at position pos.
  if (n<2) return C(0);
  else {
    nat i, j;
    newton_edge (src, n, pos, i, j);
    return (src[j] - src[i]) / C (j - i);
  }
}

TMPL bool
is_number (const Floating* src, nat n) {
  // Test whether all elements of [src, n] are numbers.
  for (nat i=0; i<n; i++)
    if (!is_number (src[i])) return false;
  return true;
}

TMPL void
scale_edge (const Floating* src, nat n, nat pos, nat& i, nat& j) {
  // Find the edge of the Newton polygon of the exponents of
  // the segment [src, n] at position pos.
  double* expo_src= mmx_new<double> (n);
  for (nat k=0; k<n; k++)
    expo_src[k]= (double) exponent (src[k]);
  newton_edge<double> (expo_src, n, pos, i, j);
  mmx_delete<double> (expo_src, n);
}

TMPL Floating
scale_slope (const Floating* src, nat n, nat pos) {
  // Find the scaling factor for [src, n] at position pos.
  if (n<2) return Floating (1);
  else {
    nat i, j;
    scale_edge (src, n, pos, i, j);
    if (src[i] != 0 && src[j] != 0)
      return exp (log (abs (src[j]) / abs (src[i])) / Floating (j-i));
    return Floating (1);
  }
}

TMPL void
scale (Floating* dest, const Floating* src, nat n,
       const Floating& c, const Floating& a)
{
  // Set dest[i] = c src[i] a^i for all i
  if (n < 16) {
    if (n>0) dest[0]= c * src[0];
    Floating b= c;
    for (nat i=1; i<n; i++) {
      b *= a;
      dest[i]= b * src[i];
    }
  }
  else {
    nat i=1;
    Floating p= a;
    while ((i<<1) < n) {
      p *= p;
      i <<= 1;
    }
    scale (dest, src, i, c, a);
    scale (dest+i, src+i, n-i, p*c, a);
  }
}

TMPL void
decompose (integer* dest, mp_exp_t& e, const Floating* src, nat n, nat bits) {
  // Given a (scaled) segment [src, n], compute a segment [dest, i] and
  // a minimal exponent e, such that each dest[i] fits into bits bits and
  // [dest, n] 2^e is an approximation of [src, n].
  if (n == 0) { e=0; return; }
  e= exponent (src[0]);
  for (nat i=1; i<n; i++)
    e= max (e, exponent (src[i]));
  e -= (bits-1);
  for (nat i=0; i<n; i++)
    dest[i]= as_integer (src[i] >> e);
}

TMPL void
recompose (Floating* dest, const integer* src, nat n, mp_exp_t e) {
  // Set [dest, n] := [src, n] 2^e
  for (nat i=0; i<n; i++)
    dest[i]= Floating (src[i]) << e;
}

TMPL void
kronecker_mul (Floating* dest,
	       const Floating* src1, nat n1,
	       const Floating* src2, nat n2)
{
  Floating* Src1= mmx_new<Floating > (n1);
  Floating* Src2= mmx_new<Floating > (n2);
  Floating* Dest= mmx_new<Floating > (n1+n2-1);
  Floating sc1= scale_slope (src1, n1, n1>>1);
  Floating sc2= scale_slope (src2, n2, n2>>1);
  Floating sc = max (sc1, sc2);
  //for (nat i=0; i<n1; i++)
  //cout << "  " << i << "\t -> " << src1[i] << "\n";
  scale (Src1, src1, n1, Floating (1), 1/sc);
  scale (Src2, src2, n2, Floating (1), 1/sc);
  //for (nat i=0; i<n1; i++)
  //cout << "  " << i << "\t -> " << Src1[i] << "\n";

  integer* mant1= mmx_new<integer> (n1);
  integer* mant2= mmx_new<integer> (n2);
  integer* dmant= mmx_new<integer> (n1+n2-1);
  mp_exp_t e1, e2;
  nat bits= V::prec() + (BITS_PER_LIMB >> 1);
  decompose (mant1, e1, Src1, n1, bits);
  decompose (mant2, e2, Src2, n2, bits);
  //for (nat i=0; i<n1; i++)
  //cout << "  " << i << "\t -> " << mant1[i] << "\n";
  kronecker_mul (dmant, mant1, n1, mant2, n2);
  //for (nat i=0; i<n1+n2-1; i++)
  //cout << "  " << i << "\t -> " << dmant[i] << "\n";
  recompose (Dest, dmant, n1+n2-1, e1+e2);
  mmx_delete<integer> (mant1, n1);
  mmx_delete<integer> (mant2, n2);
  mmx_delete<integer> (dmant, n1+n2-1);

  //for (nat i=0; i<n1+n2-1; i++)
  //cout << "  " << i << "\t -> " << Dest[i] << "\n";
  scale (dest, Dest, n1+n2-1, Floating (1), sc);
  //for (nat i=0; i<n1+n2-1; i++)
  //cout << "  " << i << "\t -> " << dest[i] << "\n";
  mmx_delete<Floating > (Src1, n1);
  mmx_delete<Floating > (Src2, n2);
  mmx_delete<Floating > (Dest, n1+n2-1);
}

TMPL
class multiplier<Floating,kronecker_multiply,Floating> {
public:
  static inline void
  mul (Floating* d, const Floating* s1, const Floating* s2, nat n1, nat n2) {
    return kronecker_mul (d, s1, n1, s2, n2); }
};

#undef TMPL
#undef Floating
#undef Rnd
}
#endif // __MMX_JET_FLOATING_HPP
