;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; MODULE     : mmx-mode.el
;; DESCRIPTION: A Mathemagix mode for Emacs
;; COPYRIGHT  : (C) 2007  Gregoire Lecerf
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Usage: put the following code in your .emacs file after setting the path
;; (setq auto-mode-alist
;;      (append auto-mode-alist
;;       '(("\\.mmx" . mmx-mode))))
;; (autoload 'mmx-mode
;;   "[the correct path]/mmx-mode.el"
;;   "Major mode for editing Mathemagix files" t)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; This software falls under the GNU general public license and comes
;; WITHOUT ANY WARRANTY WHATSOEVER. See the file LICENSE for more details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Support for highlighting and indentation
;

(defconst mmx-version "@version@"
  "mmx-mode version number")

(defun mmx-version ()
  "Echo the current version of mmmx-mode in the minibuffer."
  (interactive)
  (message "mmx-mode version %s" mmx-version))

;; Hook for this major-mode
(defvar mmx-mode-hook nil)

;; Syntax table for comments //... and /{...}/
;; Strings are handled via font locking
(defvar mmx-mode-syntax-table
  (let ((mmx-mode-syntax-table (make-syntax-table)))
    (modify-syntax-entry ?_ "w" mmx-mode-syntax-table)
    (modify-syntax-entry ?? "w" mmx-mode-syntax-table)
    (modify-syntax-entry ?/ ". 124b" mmx-mode-syntax-table)
    (modify-syntax-entry ?\n "> b" mmx-mode-syntax-table)
    (modify-syntax-entry ?{ "(} 2n" mmx-mode-syntax-table)
    (modify-syntax-entry ?} ")} 3n" mmx-mode-syntax-table)
    (modify-syntax-entry ?[ "(]" mmx-mode-syntax-table)
    (modify-syntax-entry ?] ")[" mmx-mode-syntax-table)
    (modify-syntax-entry ?( "()" mmx-mode-syntax-table)
    (modify-syntax-entry ?) ")(" mmx-mode-syntax-table)
    (modify-syntax-entry ?\" "." mmx-mode-syntax-table)
    mmx-mode-syntax-table)
  "Syntax table for mmx-mode")

;; Common functions for moving in Mathemagix expressions

(defun mmx-search-forward-quotes (limit)
  "Move point after the next double quotes and
   return its type: \" has type 1; /\" has type 2.
   If no quotes are found then 0 is returned and point is moved to limit."
  (setq found 0)
  (while (and (= found 0) (< (point) limit))
    (if (and
	 (re-search-forward "\\(/\"\\|[\\]*\"\\)" limit 'noerror 1)
	 (not (eq 'font-lock-comment-face (get-char-property (point) 'face))))
	(progn
	  (setq b (match-beginning 0))
	  (setq e (match-end 0))
	  (goto-char b)
	  (cond
	   ((looking-at "/\"") (setq found 2))
	   ((and (looking-at "[\\]*\"") (= (mod (- e b) 2) 1)) (setq found 1)))
	  (goto-char e))))
  found)

(defun mmx-end-of-quotes-1 (limit)
  "Move point after the next quotes of type 1. Go to limit otherwise."
  (setq found nil)
  (while (and (not found) (< (point) limit))
    (if (and
	 (re-search-forward "[\\]*\"" limit 'noerror 1)
	 (not (eq 'font-lock-comment-face (get-char-property (point) 'face))))
	(progn
	  (setq b (match-beginning 0))
	  (setq e (match-end 0))
	  (goto-char e)
	  (if (= (mod (- e b) 2) 1)
	      (setq found t)))))
  found)

(defun foo ()
  (interactive)
  (mmx-search-forward-quotes (point-max)))

(defun mmx-end-of-quotes-2 (limit)
  "Move point after closing quotes of types 2, ie \"/.
   Otherwise go to limit."
  (setq found nil)
  (while (and (not found) (< (point) limit))
    (if (and
	 (re-search-forward "\"/" limit 'noerror 1)
	 (not (eq 'font-lock-comment-face (get-char-property (point) 'face))))
	(progn
	  (setq found t)
	  (goto-char (match-end 0))
	  t)))
  found)

(defun mmx-skip-chars-forward (e limit)
  "A wrapper of skip-chars-forward that skips comments."
  (while (progn
	   (skip-chars-forward e limit)
	   (if (< (point) limit)
	     (progn
	       (setq f (get-char-property (point) 'face))
	       (if (eq f 'font-lock-comment-face)
		   (progn
		     (forward-char)
		     t)
		 nil))))))

(defun mmx-skip-chars-backward (e)
  "A wrapper of skip-chars-backward that skips comments and strings."
  (while (progn
	   (skip-chars-backward e)
	   (if (bobp)
	       nil
	     (progn
	       (setq f (get-char-property (- (point) 1) 'face))
	       (if (or (eq f 'font-lock-comment-face)
		       (eq f 'font-lock-string-face))
		   (progn
		     (backward-char)
		     t)
		 nil))))))

(defun mmx-re-search-forward (e a b)
  "A wrapper of re-search-forward that skips comments and strings"
  (while (progn
	   (setq found (re-search-forward e a b))
	   (if found
	       (progn
		 (setq f (get-char-property (point) 'face))
		 (or (eq f 'font-lock-comment-face)
		     (eq f 'font-lock-string-face)))
	     nil)))
  found)

(defun mmx-re-search-backward (e a b)
  "A wrapper of re-search-backward that skips comments and strings"
  (while (progn
	   (setq found (re-search-backward e a b))
	   (if found
	       (progn
		 (setq f (get-char-property (point) 'face))
		 (or (eq f 'font-lock-comment-face)
		     (eq f 'font-lock-string-face)))
	     nil)))
  found)

(defun mmx-skip-forward-blanks (limit)
  "Skip spaces and comments."
  (interactive)
  (while
      (progn
	(skip-chars-forward " \t\n" limit)
	(if (= (point) limit)
	    nil
	  (progn
	    (setq f (get-char-property (point) 'face))
	    (if (eq f 'font-lock-comment-face)
		(progn
		  (forward-char)
		  t)
	      nil))))))

(defun mmx-skip-forward-tuple (limit)
  "Skip a Mathemagix tuple."
  (interactive)
  (mmx-skip-forward-blanks limit)
  (if (looking-at "[(]")
      (progn
	(parse-partial-sexp (point) limit 0)
	t)
    nil))

(defun mmx-skip-forward-quantifier (limit)
  "Skip a Mathemagix quantifier declaration."
  (interactive)
  (mmx-skip-forward-blanks limit)
  (if (looking-at "\\(\\<forall\\>\\|\\<exists\\>\\)")
      (progn
	(goto-char (match-end 0))
	(mmx-skip-forward-tuple limit))))

;; Matchers for font-lock-mode

(defun mmx-string-matcher (limit)
  "Matcher for strings braced into \"...\", /\".../\" and /\"...\"/"
  (setq x (mmx-search-forward-quotes limit))
  (cond	((eq x 1)
	   (progn
	     (setq bm (- (point) 1))
	     (setq e (mmx-end-of-quotes-1 limit))
	     (setq em (point))
	     (set-match-data (list bm em))
	     e))
	((eq x 2)
	 (progn
	   (setq bm (- (match-end 0) 2))
	   (setq e (mmx-end-of-quotes-2 limit))
	   (setq em (point))
	   (set-match-data (list bm em))
	   e))
	(t nil)))

(defun mmx-function-declaration-matcher (limit)
  (if (mmx-re-search-forward "==\\|:=" limit 'noerror)
      (progn
	(setq e (match-end 0))
	(goto-char (match-beginning 0))
	(mmx-beginning-of-instruction)
	(mmx-skip-forward-quantifier e)
	(mmx-skip-forward-blanks e)
	(if (save-excursion (mmx-re-search-forward "[(]" e 'noerror))
	    (progn
	      (looking-at "[a-zA-Z0-9_?]+")
	      (set-match-data (list (match-beginning 0) (match-end 0)))
	      (goto-char e)
	      t)
	  (progn
	    (goto-char e)
	    (mmx-function-declaration-matcher limit))))
    nil))

(defconst mmx-font-lock-keywords
  (list 
   '(mmx-string-matcher . (0 font-lock-string-face nil t))
   '(mmx-function-declaration-matcher . (0 font-lock-function-name-face t nil))
   '("\\(^\\|[^:]\\):[>]?[ \t]*\\([a-zA-Z0-9_?]+\\)" 2 font-lock-type-face)
   '("\\<\\([0-9]+\\(\\|[.][0-9]+\\)\\|true\\|false\\|mmout\\|mmin\\|mmerr\\|nil\\)\\>" . font-lock-constant-face)
   '("\\<\\(alias\\|and\\|begin\\|break\\|case\\|continue\\|error\\|explode\\|fuse\\|generate\\|lambda\\|loop\\|melt\\|operator\\|or\\|postfix\\|prefix\\|quit\\|return\\|step\\|supports[?]\\|type\\|until\\|use\\|while\\|for\\|if\\|then\\|else\\|in\\|do\\|forall\\|exists\\|while\\)\\>" . font-lock-keyword-face))
  "Expressions highlighting for the Mathemagix mode")

;; Indentation

(defvar mmx-indent 2 "Indentation size")

(defun mmx-backward-block-delimiter ()
  "Move before a block or instruction delimiter."
  (interactive)
  (save-excursion
    (mmx-skip-chars-backward "^()[]{},;")
    (setq xe (point))
    (setq xb (- (point) 1)))
  (save-excursion
    (if (mmx-re-search-backward "\\(do\\|loop\\|==\\|:=\\|+=\\|-=\\|/=\\|\*=\\|\\<if\\>\\|\\<else\\>\\|\\<then\\>\\)[ \t\n]*[{]?" 0 'noerror)
	(progn
	  (setq ye (match-end 0))
	  (setq yb (match-beginning 0))
	  (if (looking-at "if")
	      (progn
		(skip-chars-backward " \t")
		(if (> (point) 4)
		    (progn
		      (forward-char -4)
		      (if (looking-at "else") (setq yb (point))))))))
      (progn
	(setq yb 1)
	(setq ye 1))))
  (if (> xe ye)
      (progn
	(goto-char xb)
	xe)
    (progn
      (goto-char yb)
      ye)))

(defun cdre (x l)
  "Remove x if it is (car l). Return () otherwise."
  (if (null l)
      ()
    (if (= x (car l))
	(cdr l)
      ())))

(defun itcdr (x l)
  "Remove all the x on the top of l."
  (if (null l)
      ()
    (if (= x (car l))
	(itcdr x (cdr l))
      l)))

(defun conse (x l)
  "Cons x to l unless l is null."
  (if (null l)
      ()
    (cons x l)))
    
(defun mmx-parse-backward-block-delimiters (state)
  "Mathemagix block and instruction parser."
  (interactive)
  (mmx-backward-block-delimiter)
  (cond
   ((looking-at "}") (cons 1 state))
   ((looking-at "{") (cdre 1 (itcdr 4 state)))
   ((looking-at ")") (cons 2 state))
   ((looking-at "(") (cdre 2 (itcdr 4 state)))
   ((looking-at "[]]") (cons 3 state))
   ((looking-at "[[]") (cdre 3 (itcdr 4 state)))
   ((looking-at ",") (cons 4 state))
   ((looking-at ";") (cons 4 state))
   ((looking-at "\\(do\\|loop\\|==\\|:=\\|+=\\|-=\\|/=\\|\*=\\)[ \t\n]*{") (cdre 1 (itcdr 4 state)))
   ((looking-at "\\(do\\|loop\\|==\\|:=\\|+=\\|-=\\|/=\\|\*=\\)")
      (if (eq 0 (car state)) () state))
   ((looking-at "if[ \t\n]*{") (cdre 5 (cdre 1 (itcdr 4 state))))
   ((looking-at "if") (cdre 5 (itcdr 4 state)))
   ((looking-at "then[ \t\n]*{")
    (progn
      (setq tmp (cdre 1 (itcdr 4 state)))
      (if (eq (car tmp) 5) tmp (conse 5 tmp))))
   ((looking-at "else[ \t\n]*{") (conse 5 (cdre 1 (itcdr 4 state))))
   ((looking-at "\\(then\\|else[ \t]+if\\)")
    (if (eq (car state) 0)
	()
      (progn
	(setq tmp (itcdr 4 state))
	(if (eq 5 (car tmp)) tmp (conse 5 (conse 4 tmp))))))
   ((looking-at "else") 
    (if (eq (car state) 0)
	()
      (conse 5 (conse 4 (itcdr 4 state)))))))

(defun mmx-beginning-of-block ()
  "Go to beginning of block of instructions."
  (interactive)
  (setq state (list 0))
  (while (progn
	   (setq state (mmx-parse-backward-block-delimiters state))
	   (not (null state)))))

(defun mmx-beginning-of-block-with-init (i)
  "Go to beginning of block of instructions."
  (interactive)
  (setq state i)
  (while (progn
	   (setq state (mmx-parse-backward-block-delimiters state))
	   (not (null state)))))

(defun mmx-beginning-of-instruction ()
  "Go to beginning of instruction."
  (interactive)
  (setq level 0)
  (setq cont t)
  (setq z (+ 1 (point)))
  (while (progn
	   (setq z (mmx-backward-block-delimiter))
	   (setq cont nil)
	   (cond
	    ((looking-at ")")
	     (progn (setq level (1+ level)) (setq cont t)))
	    ((looking-at "(") 
	     (progn (setq level (1- level)) (setq cont t)))
	    ((looking-at "[]]")
	     (progn (setq level (1+ level)) (setq cont t)))
	    ((looking-at "[[]")
	     (progn (setq level (1- level)) (setq cont t))))
	   (or cont (not (or (= level 0) (bobp))))))
  (if (not (bobp))
      (goto-char z)))

(defun mmx-skip-forward-consecutive-closing-blocks-on-current-line ()
  "Skip closing delimiters until the end of the current line."
  (interactive)
  (mmx-skip-chars-forward "])},; \t" (save-excursion (end-of-line) (point)))
  (if (looking-at "\\(\\<then\\>\\|\\<else\\>\\)")
      (progn
	(goto-char (match-beginning 0))
	(list 5 0))
    (list 0)))

(defun mmx-indent-line ()
  "Compute indentation of the current line."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (setq i (mmx-skip-forward-consecutive-closing-blocks-on-current-line))
    (mmx-beginning-of-block-with-init i)
    (if (looking-at "[(]")
	(progn
	  (goto-char (match-end 0))
	  (setq x (current-column)))
      (if (= (point) 1)
	  (setq x 0) ; hack for beginning of buffer
	(setq x (+ (current-indentation) mmx-indent)))))
  (save-excursion
    (setq e)
    (beginning-of-line)
    (if (looking-at "[ \t]*<<")
	(progn
	  (mmx-beginning-of-instruction)
	  (mmx-re-search-forward "<<" e 'noerror)
	  (setq x (- (current-column) 2))))
    (if (looking-at "[ \t]*>>")
	(progn
	  (mmx-beginning-of-instruction)
	  (mmx-re-search-forward ">>" e 'noerror)
	  (setq x (- (current-column) 2)))))
  (save-excursion
    (beginning-of-line)
    (delete-region (point) (+ (point) (current-indentation)))
    (insert-char 32 x))
  (if (< (current-column) x)
      (forward-char (- x (current-column)))))

;; Keymap

(defvar mmx-mode-map nil
  "Keymap used in Mathemagix mode.")

;; Main function.
(defun mmx-mode ()
  (interactive)
  (kill-all-local-variables)
  (set-syntax-table mmx-mode-syntax-table)
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(mmx-font-lock-keywords nil nil ((?_ . "w"))))
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  (make-local-variable 'comment-multi-line)
  (setq comment-multi-line t)
  (setq font-lock-mode-maximum-decoration t)
  (setq indent-tabs-mode nil)
  (setq indent-line-function 'mmx-indent-line)
  (setq mmx-mode-map (make-sparse-keymap))
  (define-key mmx-mode-map [(return)] 'reindent-then-newline-and-indent)
  (use-local-map mmx-mode-map)
  (setq major-mode 'mmx-mode)
  (setq mode-name "Mathemagix")
  (turn-on-font-lock)
  (run-hooks 'mmx-mode-hook))

(provide 'mmx-mode)
