
/******************************************************************************
* MODULE     : mmx-light.cpp
* DESCRIPTION: Light command line evaluator of expressions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven and Gregoire Lecerf
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <unistd.h>
#include "tuple.hpp"
#include "list.hpp"
#include "glue.hpp"
#include "mmx_syntax.hpp"
#include "shell.hpp"
#include "basix_utils.hpp"
#include "generic_internal.hpp"

namespace mmx {
  evaluator mmxlight_evaluator ();
}

using namespace mmx;

/*
 * Glue new functions for the interface
 */

static bool
toggle_quiet_mode (void) {
  quiet_mode = !quiet_mode;
  return quiet_mode;
}

static bool
toggle_debug_mode (void) {
  debug_mode = !debug_mode;
  return debug_mode;
}

static bool
toggle_type_mode (void) {
  type_mode = !type_mode;
  return type_mode;
}

static bool
toggle_time_mode (void) {
  time_mode = !time_mode;
  return time_mode;
}

static bool
toggle_math_mode (void) {
  math_mode = !math_mode;
  return math_mode;
}

static bool
is_batch_mode (void) {
  return batch_mode;
}

static bool
is_texmacs_mode (void) {
  return texmacs_mode;
}

static nat
set_history_size (const nat& s) {
  history_size = 0;
  return history_size;
}

static generic
mmx_quit () {
  cerr.flush ();
  cout.flush ();
  if (!batch_mode) {
    shell_save_history ();
    shell_save_session ();
  }
  restore_evaluator ();
  mmx_terminate ();
  exit (0);
}

static void
define_interface (void) {
  define<always> ("quiet_mode", toggle_quiet_mode);
  define<always> ("debug_mode", toggle_debug_mode);
  define<always> ("type_mode", toggle_type_mode);
  define<always> ("time_mode", toggle_time_mode);
  define<always> ("math_mode", toggle_math_mode);
  define<always> ("batch_mode?", is_batch_mode);
  define<always> ("texmacs_mode?", is_texmacs_mode);
  define<always> ("set_history_size", set_history_size);
  define<always> ("history", get_interactive_source);
  define<always> ("quit", mmx_quit);
}

static generic
replace_literals (const generic& in) {
  if (is_func (in, "literal_floating", 1))
    return current_ev->eval (in);
  else if (is_func (in, "literal_integer", 1))
    return current_ev->eval (in);
  else if (is_func (in, "literal_string"))
    return current_ev->eval (in);
  else if (is<compound> (in)) {
    const array<generic> a= compound_to_array (in);
    array<generic> b= array_fill<generic> (N(a));
    for (nat i=0; i<N(a); i++) b[i]= replace_literals (a[i]);
    return array_to_compound (b);
  }
  else return in;
}

static generic
fail_proof_pre_evaluate (const generic& in) {
#ifdef BASIX_ENABLE_EXCEPTIONS
  try {
#endif
    return replace_literals (in);
#ifdef BASIX_ENABLE_EXCEPTIONS
  }
  catch (const exception& err) {
    return as<generic> (err);
  }
#endif
}

static generic
fail_proof_evaluate (const generic& in) {
#ifdef BASIX_ENABLE_EXCEPTIONS
  try {
#endif
    return current_ev->eval (in);
#ifdef BASIX_ENABLE_EXCEPTIONS
  }
  catch (const exception& err) {
    return as<generic> (err);
  }
#endif
}

static void
handle_evaluation (const mmx::string& s, nat n) {
  nat start, total;
  list<syntax_error> errors;
  generic in, out;
  static generic nil_generic = generic();

  if (N(s)==0) return;
  in = mmx_parse ("", n, s, errors);
  if (shell_terminal_print_syntax_errors (errors)) return;
  if (eq (in, nil_generic)) return;

  out= fail_proof_pre_evaluate (in);
  start= mmx_user_time ();
  if (!is<exception> (out))
    out= fail_proof_evaluate (out);
  total= mmx_user_time () - start;

  if (!is<tuple<generic> > (out) || N (as<tuple<generic> > (out)) > 0) {
    if (is<exception> (out))
      shell_terminal_print_exception (out);
    else {
      generic t= "None";
      if (type_mode) t= type_name (type (out));
      select_output_evaluator ();
      //cerr << "Flattening... "; cerr.flush ();
      generic flout= flatten (out);
      //cerr << "Done\n";
      restore_evaluator ();
      if (texmacs_mode) shell_texmacs_output (flout, t);
      else shell_terminal_output (flout, t);
    }
  }
  if (time_mode && total >= 1) cout << "Computed in " << total << " ms\n";
  cout.flush();
};

int
main (int argc, char** argv) {
  mmx::string in;
  list<mmx::string> to_load;
  iterator<mmx::string> it;
  bool neof;

  mmx_initialize ();
  select_evaluator (mmxlight_evaluator ());
  if (dl_exists ("basix")) dl_link ("basix");
  else cerr << "basix did not open\n";
  if (dl_exists ("gnumerix")) dl_link ("gnumerix");
  else cerr << "gnumerix did not open\n";
  if (dl_exists ("algebramix")) dl_link ("algebramix");
  else cerr << "algebramix did not open\n";
  if (dl_exists ("symbolix")) dl_link ("symbolix");
  else cerr << "symbolix did not open\n";
  if (dl_exists ("analyzit")) dl_link ("analyzit");
  else cerr << "analyzit did not open\n";

  define_interface ();
  to_load = shell_initialize (argc, argv);
  batch_mode = !isatty(STDIN_FILENO) && !texmacs_mode;

  if (!nil (to_load)) {
    batch_mode = true;
    for (it = iterate (to_load); busy(it); ++it)
      handle_evaluation("load \"" * *it * "\"", 0);
    mmx_quit();
  }

  shell_load_history ();
  
  if (replay_mode) {
    const array<string> v = shell_restore_session ();
    string in;
    nat i,j,k;
    for (i = 0; i < N(v); ++i) {
      in = v[i];
      cout << as_string (i + 1) * "] ";
      for (j = 0; j < N(in); ++j) {
	cout << in[j];
	if (in[j] == '\n' && j+1 < N(in)) {
	  for (k=0; k < N(as_string (i + 1)); ++k)
	    cout << " ";
	  cout << "] ";
	}
      }

      if (in == mmx::string ("quit\n") || in == mmx::string ("exit\n") ||
	  in == mmx::string ("bye\n"))
	mmx_quit();
      handle_evaluation (in, i);
    }
    mmx_quit();
  }

  if (batch_mode) {
    // run in batch mode
    shell_terminal_batch_input (in);
    handle_evaluation (in, 0);
    mmx_quit();
  }

  while (true) {
    if (texmacs_mode)
      neof = shell_texmacs_input (in);
    else
      neof = shell_terminal_input (in);
    if (in == mmx::string ("quit\n") || in == mmx::string ("exit\n") ||
	in == mmx::string ("bye\n"))
      mmx_quit();
    handle_evaluation (in, input_number ());
    if (!neof) {
      cout << "\n";
      mmx_quit();
    }
  }
}
