
/******************************************************************************
* MODULE     : calculus_glue.cpp
* DESCRIPTION: Standard calculus
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "evaluator.hpp"
#include "glue.hpp"
#include "rational.hpp"
#include "literal.hpp"
#include "compound.hpp"
#include "big_sum.hpp"
#include "big_product.hpp"
#include "generic_internal.hpp"
namespace mmx {
#define Table table<generic,generic>
#define Array array<generic>
#define Iterator iterator<generic>
#define PRIME(f) gen (GEN_PRIME, f)
#define GRAD(f) gen (GEN_GRAD, f)

/******************************************************************************
* Additional array arithmetic
******************************************************************************/

Array
derive (const Array& v, const generic& y) {
  nat n= N(v);
  generic* r= mmx_new<generic> (n);
  for (nat i=0; i<n; i++)
    r[i]= derive (v[i], y);
  return Array (r, n);
}

Array
operator * (const Array& v, const generic& x) {
  nat n= N(v);
  generic* r= mmx_new<generic> (n);
  for (nat i=0; i<n; i++)
    r[i]= v[i] * x;
  return Array (r, n);
}

generic
dot (const Array& v, const Array& w) {
  nat n= N(v);
  ASSERT (N(w) == n, "lengths don't match");
  generic sum= 0;
  for (nat i=0; i<n; i++)
    sum += v[i] * w[i];
  return sum;
}

/******************************************************************************
* Differentiation
******************************************************************************/

template<typename T> generic
scalar_derive (const T& x, const generic& y) {
  (void) x; (void) y; return 0;
}

generic
literal_derive (const literal& x, const generic& y) {
  return as<generic> (x) == y ? generic (1): generic (0);
}

generic
compound_derive (const compound& x, const generic& y) {
  const Array v= *x;
  if (N(v) <= 1) return 0;
  generic fun;
  if (N(v) == 2 && current_ev->get (PRIME (v[0]), fun))
    return as<routine> (fun) -> apply (v[1]) * derive (v[1], y);
  Array args = range (v, 1, N(v));
  Array dargs= derive (args, y);
  if (current_ev->get (GRAD (v[0]), fun)) {
    generic r= as<routine> (fun) -> apply (args);
    if (is<Array > (r))
      return dot (as<Array > (r), dargs);
  }
  return gen (gen (gen (GEN_DERIVATIVE, v[0]), args), dargs);
}

generic
big_sum_derive (const big_sum& x, const generic& y) {
  const Table t= *x;
  Array v;
  for (Iterator it= entries (t); busy (it); ++it)
    v << t[*it] * derive (*it, y);
  return big_add (v);
}

generic
big_product_derive (const big_product& x, const generic& y) {
  const Table t= *x;
  Array v;
  for (Iterator it= entries (t); busy (it); ++it) {
    generic h= *it, sc= t[h];
    v << derive (sc, y) * log (h) + sc * derive (h, y) / h;
  }
  //return big_add (v) * as<generic> (x);
  return big_add (v * as<generic> (x));
}

generic log_prime (const generic& x) { return 1 / x; }
generic exp_prime (const generic& x) { return exp (x); }
generic cos_prime (const generic& x) { return -sin (x); }
generic sin_prime (const generic& x) { return cos (x); }
generic tan_prime (const generic& x) { return 1 / square (cos (x)); }
generic acos_prime (const generic& x) { return -1 / sqrt (1 - square (x)); }
generic asin_prime (const generic& x) { return 1 / sqrt (1 - square (x)); }
generic atan_prime (const generic& x) { return 1 / (1 + square (x)); }

//generic test_grad (const generic& x, const generic& y) {
//return as<generic> (array<generic> (x + sin (y), cos (x))); }

generic
fused_derive (const generic& x, const generic& v) {
  return derive (fuse (x), fuse (v));
}

/******************************************************************************
* Substitution
******************************************************************************/

template<typename T> generic
scalar_substitute (const T& x, const generic& y) {
  (void) y; return as_generic (x);
}

generic
literal_substitute (const literal& x, const generic& y) {
  generic g= as<generic> (x);
  if (is<Table > (y)) {
    const Table t= as<Table > (y);
    if (t->contains (g)) return t[g];
  }
  return g;
}

generic
compound_substitute (const compound& x, const generic& y) {
  generic g= as<generic> (x);
  if (is<Table > (y)) {
    const Table t= as<Table > (y);
    if (t->contains (g)) return t[g];
  }

  const Array v= *x;
  nat i, n= N(v);
  Array w= array_fill<generic> (n);
  for (i=0; i<n; i++)
    w[i]= substitute (v[i], y);
  return array_to_compound (w);
}

generic
big_sum_substitute (const big_sum& x, const generic& y) {
  const Table sum= *x;
  Array v;
  for (Iterator it= entries (sum); busy (it); ++it)
    v << substitute (sum[*it], y) * substitute (*it, y);
  return big_add (v);
}

generic
big_product_substitute (const big_product& x, const generic& y) {
  const Table prod= *x;
  Array v;
  for (Iterator it= entries (prod); busy (it); ++it)
    v << pow (substitute (*it, y), substitute (prod[*it], y));
  return big_mul (v);
}

/******************************************************************************
* Composition
******************************************************************************/

generic
default_compose (const generic& f, const generic& g) {
  return gen (GEN_COMPOSE, f, g);
}

/******************************************************************************
* Interface
******************************************************************************/

void
define_calculus () {
  define<always> (GEN_DERIVE, scalar_derive<rational>);
  define<always> (GEN_DERIVE, literal_derive);
  define<always> (GEN_DERIVE, compound_derive);
  define<always> (GEN_DERIVE, big_sum_derive);
  define<always> (GEN_DERIVE, big_product_derive);
  define<always> ("derive", fused_derive);

  define<always> (PRIME (GEN_LOG), log_prime);
  define<always> (PRIME (GEN_EXP), exp_prime);
  define<always> (PRIME (GEN_COS), cos_prime);
  define<always> (PRIME (GEN_SIN), sin_prime);
  define<always> (PRIME (GEN_TAN), tan_prime);
  define<always> (PRIME (GEN_ARCCOS), acos_prime);
  define<always> (PRIME (GEN_ARCSIN), asin_prime);
  define<always> (PRIME (GEN_ARCTAN), atan_prime);

  define<always> (GEN_SUBSTITUTE, scalar_substitute<rational>);
  define<always> (GEN_SUBSTITUTE, literal_substitute);
  define<always> (GEN_SUBSTITUTE, compound_substitute);
  define<always> (GEN_SUBSTITUTE, big_sum_substitute);
  define<always> (GEN_SUBSTITUTE, big_product_substitute);

  define<always> (GEN_COMPOSE, default_compose);
}

#undef Table
#undef Array
#undef Iterator
#undef PRIME
#undef GRAD
} // namespace mmx
