
/******************************************************************************
* MODULE     : elementary_glue.cpp
* DESCRIPTION: Evaluation of elementary functions on generic expressions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "evaluator.hpp"
#include "glue.hpp"
#include "rational.hpp"
#include "literal.hpp"
#include "compound.hpp"
#include "big_sum.hpp"
#include "big_product.hpp"
#include "generic_internal.hpp"
namespace mmx {
#define Table table<generic,generic>
#define Array array<generic>
#define Iterator iterator<generic>

/******************************************************************************
* Exp-log functions
******************************************************************************/

generic
calculus_sqrt (const generic& x) {
  if (is<integer> (x) && is_square (as<integer> (x)))
    return as<generic> (sqrt (as<integer> (x)));
  if (is<rational> (x) && is_square (as<rational> (x)))
    return as<generic> (sqrt (as<rational> (x)));
  return pow (x, generic (1) / generic (2));
}

generic
calculus_exp (const generic& x) {
  if (x == 0) return 1;
  return pow (generic (GEN_E), x);
}

generic
calculus_log (const generic& x) {
  if (x == 1) return 0;
  if (x == generic (GEN_E)) return 1;
  if (is<big_sum> (x)) {
    const Table t= sum_to_table (x);
    if (N(t) == 1) {
      generic y = *entries (t);
      generic sc= t[y];
      return log (sc) + log (y);
    }
    else return gen (GEN_LOG, x);
  }
  if (is<big_product> (x)) {
    const Table t= product_to_table (x);
    Array v;
    for (Iterator it= entries (t); busy (it); ++it)
      v << t[*it] * log (*it);
    return big_add (v);
  }
  return gen (GEN_LOG, x);
}

/******************************************************************************
* Trigonometric functions
******************************************************************************/

generic
calculus_cos (const generic& x) {
  if (x == 0) return 1;
  if (is_func (x, GEN_ARCCOS, 1)) return x[1];
  return gen (GEN_COS, x);
}

generic
calculus_sin (const generic& x) {
  if (x == 0) return 0;
  if (is_func (x, GEN_ARCSIN, 1)) return x[1];
  return gen (GEN_SIN, x);
}

generic
calculus_tan (const generic& x) {
  if (x == 0) return 0;
  if (is_func (x, GEN_ARCTAN, 1)) return x[1];
  return gen (GEN_TAN, x);
}

generic
calculus_acos (const generic& x) {
  if (x == 1) return 0;
  if (is_func (x, GEN_COS, 1)) return x[1];
  return gen (GEN_ARCCOS, x);
}

generic
calculus_asin (const generic& x) {
  if (x == 0) return 0;
  if (is_func (x, GEN_SIN, 1)) return x[1];
  return gen (GEN_ARCSIN, x);
}

generic
calculus_atan (const generic& x) {
  if (x == 0) return 0;
  if (is_func (x, GEN_TAN, 1)) return x[1];
  return gen (GEN_ARCTAN, x);
}

/******************************************************************************
* Same functions on rational numbers, so as to prevent bad casts
******************************************************************************/

generic rational_sqrt (const rational& x) {
  return calculus_sqrt (as<generic> (x)); }
generic rational_exp (const rational& x) {
  return calculus_exp (as<generic> (x)); }
generic rational_log (const rational& x) {
  return calculus_log (as<generic> (x)); }
generic rational_cos (const rational& x) {
  return calculus_cos (as<generic> (x)); }
generic rational_sin (const rational& x) {
  return calculus_sin (as<generic> (x)); }
generic rational_tan (const rational& x) {
  return calculus_tan (as<generic> (x)); }
generic rational_acos (const rational& x) {
  return calculus_acos (as<generic> (x)); }
generic rational_asin (const rational& x) {
  return calculus_asin (as<generic> (x)); }
generic rational_atan (const rational& x) {
  return calculus_atan (as<generic> (x)); }

/******************************************************************************
* Interface
******************************************************************************/

void
define_elementary () {
  define<always> (GEN_SQRT, calculus_sqrt);
  define<always> (GEN_EXP, calculus_exp);
  define<always> (GEN_LOG, calculus_log);
  define<always> (GEN_COS, calculus_cos);
  define<always> (GEN_SIN, calculus_sin);
  define<always> (GEN_TAN, calculus_tan);
  define<always> (GEN_ARCCOS, calculus_acos);
  define<always> (GEN_ARCSIN, calculus_asin);
  define<always> (GEN_ARCTAN, calculus_atan);

  define<always> (GEN_SQRT, rational_sqrt);
  define<always> (GEN_EXP, rational_exp);
  define<always> (GEN_LOG, rational_log);
  define<always> (GEN_COS, rational_cos);
  define<always> (GEN_SIN, rational_sin);
  define<always> (GEN_TAN, rational_tan);
  define<always> (GEN_ARCCOS, rational_acos);
  define<always> (GEN_ARCSIN, rational_asin);
  define<always> (GEN_ARCTAN, rational_atan);
}

#undef Table
#undef Array
#undef Iterator
} // namespace mmx
