
/******************************************************************************
* MODULE     : solve_glue.cpp
* DESCRIPTION: Solving equations
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "evaluator.hpp"
#include "glue.hpp"
#include "literal.hpp"
#include "compound.hpp"
#include "big_sum.hpp"
#include "big_product.hpp"
#include "generic_internal.hpp"
#include "list_sort.hpp"
namespace mmx {
#define Table table<generic,generic>
#define Vector vector<generic>
#define List list<generic>
#define Iterator iterator<generic>

string MMX_FAIL= "fail";
string MMX_UNINIT= "uninit";

/******************************************************************************
* Dependencies
******************************************************************************/

void
count_occurrences (const generic& x, table<nat,generic>& r) {
  r[x]++;
  if (is_scalar (x) || is<literal> (x));
  else if (is<compound> (x)) {
    for (nat i=0; i<N(x); i++)
      count_occurrences (x[i], r);
  }
  else if (is<big_sum> (x)) {
    const Table t= sum_to_table (x);
    for (Iterator it= entries (t); busy (it); ++it) {
      count_occurrences (*it, r);
      count_occurrences (t[*it], r);
    }
  }
  else if (is<big_product> (x)) {
    const Table t= product_to_table (x);
    for (Iterator it= entries (t); busy (it); ++it) {
      count_occurrences (*it, r);
      count_occurrences (t[*it], r);
    }
  }
}

table<nat,generic>
count_occurrences (const generic& x) {
  table<nat,generic> t (0);
  count_occurrences (x, t);
  return t;
}

nat
occurrences (const generic& x, const generic& y) {
  return count_occurrences (x) [y];
}

bool
depends (const generic& x, const generic& y) {
  return count_occurrences (x) [y] != 0;
}

/******************************************************************************
* Solving single equations
******************************************************************************/

nat
find_unique (const Table& t, const generic& x, generic& sc, generic& rad) {
  nat counter= 0;
  for (Iterator it= entries (t); busy (it); ++it) {
    if (depends (*it, x) || depends (t[*it], x)) {
      sc = t[*it];
      rad= *it;
      if (depends (*it, x)) counter++;
      if (depends (t[*it], x)) counter++;
    }
  }
  return counter;
}

generic
simple_solve (const generic& l, const generic& r, const generic& x) {
  //cout << "Solve " << l << " = " << r << " in " << x << "\n";
  if (l == r) return x;
  if (is<big_sum> (l) || (is<big_sum> (r) && depends (r, x))) {
    generic y= r - l;
    if (!is<big_sum> (y)) return simple_solve (y, 0, x);
    generic sc, rad;
    if (find_unique (sum_to_table (y), x, sc, rad) != 1) return MMX_FAIL;
    generic z= sc * rad - y;
    if (depends (sc, x)) return simple_solve (sc, z / rad, x);
    return simple_solve (rad, z / sc, x);
  }
  if (is<big_product> (l) || (is<big_product> (r) && depends (r, x))) {
    generic y= r / l;
    if (!is<big_product> (y)) return simple_solve (y, 1, x);
    generic p, rad;
    if (find_unique (product_to_table (y), x, p, rad) != 1) return MMX_FAIL;
    generic z= pow (rad, p) / y;
    if (depends (p, x)) return simple_solve (p, log (z) / log (rad), x);
    return simple_solve (rad, pow (z, 1/p), x);
  }

  nat nl= occurrences (l, x), nr= occurrences (r, x);
  if (nl + nr != 1) return MMX_FAIL;
  if (nl == 0) return simple_solve (r, l, x);
  if (l == x) return r;

  if (is_func (l, GEN_EXP, 1)) return simple_solve (l[1], log (r), x);
  if (is_func (l, GEN_LOG, 1)) return simple_solve (l[1], exp (r), x);
  if (is_func (l, GEN_COS, 1)) return simple_solve (l[1], acos (r), x);
  if (is_func (l, GEN_SIN, 1)) return simple_solve (l[1], asin (r), x);
  if (is_func (l, GEN_TAN, 1)) return simple_solve (l[1], atan (r), x);
  if (is_func (l, GEN_ARCCOS, 1)) return simple_solve (l[1], cos (r), x);
  if (is_func (l, GEN_ARCSIN, 1)) return simple_solve (l[1], sin (r), x);
  if (is_func (l, GEN_ARCTAN, 1)) return simple_solve (l[1], tan (r), x);

  return MMX_FAIL;
}

generic
simple_solve_equation (const generic& z, const generic& x) {
  //cout << "Solving " << z << " in " << x << "\n";
  if (is_func (z, GEN_EQUAL, 2)) return simple_solve (z[1], z[2], x);
  return simple_solve (z, 0, x);
}

/******************************************************************************
* Solving systems of equations
******************************************************************************/

List
substitute (const List& x, const generic& y) {
  if (nil (x)) return x;
  return cons (substitute (car (x), y), substitute (cdr (x), y));
}

Table
substitute (const Table& t, const generic& y) {
  Table r (0);
  for (Iterator it= entries (t); busy (it); ++it)
    r [*it]= substitute (t[*it], y);
  return r;
}

List
simple_solve_system (const List& l, const List& r) {
  //cout << "Solve " << l << " in " << r << "\n";
  Table t (0);
  List eqs= sort (l, small_big_compare);
  List vars= r;
  while (!nil (eqs)) {
    //cout << "Solving " << eqs << ", " << vars << "\n";
    generic var= MMX_UNINIT, eq, sol;
    List remaining_eqs;
    for (; !nil (eqs); eqs= cdr (eqs)) {
      List remaining_vars;
      for (; !nil (vars); vars= cdr (vars)) {
	if (occurrences (car (eqs), car (vars)) == 1) break;
	else remaining_vars= cons (car (vars), remaining_vars);
      }
      if (nil (vars)) {
	vars= reverse (remaining_vars);
	remaining_eqs = cons (car (eqs), remaining_eqs);
      }
      else {
	var = car (vars);
	vars= reverse (remaining_vars) * cdr (vars);
	break;
      }
    }
    if (nil (eqs)) {
      eqs= reverse (remaining_eqs);
      eq = car (eqs);
      eqs= cdr (eqs);
    }
    else {
      eq = car (eqs);
      eqs= reverse (remaining_eqs) * cdr (eqs);
    }
    sol= simple_solve_equation (eq, var);
    //cout << "Solved " << var << " := " << sol << "\n";
    if (sol == MMX_FAIL) {
      List ret;
      for (vars= r; !nil (vars); vars= cdr (vars))
	ret= cons (generic (MMX_FAIL), ret);
      return ret;
    }
    if (var != MMX_UNINIT) {
      Table u (0);
      u [var]= sol;
      generic subst= as<generic> (u);
      eqs= substitute (eqs, subst);
      t  = substitute (t, subst);
      t [var]= sol;
    }
  }

  List ret;
  for (vars= r; !nil (vars); vars= cdr (vars)) {
    generic var= car (vars);
    ret= cons (t->contains (var)? t[var]: var, ret);
  }
  return reverse (ret);
}

/******************************************************************************
* Interface
******************************************************************************/

void
define_solve () {
  define<always> (GEN_SOLVE, simple_solve_equation);
  define<always> (GEN_SOLVE, simple_solve_system);
}

#undef Table
#undef Vector
#undef List
#undef Iterator
} // namespace mmx
